#include "mali_cv/Image.h"

#include "mali_cv/Error.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/Context.h"
#include "mali_cv/Helper.h"
#include "arm_compute/core/Utils.h"
#include <cstring>

#include <utility>

using namespace mali_cv;

#ifdef TEST_TENSORS
CLImageTensor::CLImageTensor():_buffer(nullptr),_info()
{
}

void CLImageTensor::init( arm_compute::ICLTensor& parent )
{
	const arm_compute::ITensorInfo* parent_info = parent.info();

	ERROR_ON( parent_info->num_dimensions() < 4 );
	ERROR_ON( dim2_to_use >= parent_info->dimension(2) );
	ERROR_ON( dim3_to_use >= parent_info->dimension(3) );

	_buffer = parent.cl_buffer();

	arm_compute::TensorInfo info(
		arm_compute::TensorShape(parent_info->dimension(0), parent_info->dimension(1)),
		parent_info->format());

	info.auto_padding();

	size_t offset = parent_info->offset_first_element_in_bytes() +
					dim2_to_use * parent_info->strides_in_bytes()[2] +
					dim3_to_use * parent_info->strides_in_bytes()[3];

	info.init(
		info.tensor_shape(),
		info.format(),
		arm_compute::Strides(parent_info->strides_in_bytes()[0], parent_info->strides_in_bytes()[1]),
		offset,
		offset + info.total_size() - info.offset_first_element_in_bytes());

	info.set_is_resizable(parent_info->is_resizable());

	_info = std::move(info);
}

arm_compute::ITensorInfo* CLImageTensor::info() const
{
	return &_info;
}

arm_compute::ITensorInfo* CLImageTensor::info()
{
	return &_info;
}

const cl::Buffer &CLImageTensor::cl_buffer() const
{
	return _buffer;
}

uint8_t *CLImageTensor::do_map(cl::CommandQueue &q, bool blocking)
{
	ARM_COMPUTE_ERROR_ON(_buffer.get() == nullptr);
	return static_cast<uint8_t *>(q.enqueueMapBuffer(_buffer, blocking ? CL_TRUE : CL_FALSE, CL_MAP_READ | CL_MAP_WRITE, 0, info()->total_size()));
}

void CLImageTensor::do_unmap(cl::CommandQueue &q)
{
	ARM_COMPUTE_ERROR_ON(_buffer.get() == nullptr);
	q.enqueueUnmapMemObject(_buffer, buffer());
}

#endif /* TEST_TENSORS */

CLImage::CLImage(): IImage(),
#ifdef TEST_TENSORS
	tensor(),image()
#else /* TEST_TENSORS */
	multi_img(),
	tensor(*multi_img.cl_plane(0)),
	image(*multi_img.cl_plane(0))
#endif /* TEST_TENSORS */
{
}

void CLImage::internal_create(Context &, size_t w, size_t h, Format format)
{
	arm_compute::Format f = ARM_COMPUTEHelper::convert_to_arm_compute_format(format);

#ifdef TEST_TENSORS
	arm_compute::TensorInfo info(arm_compute::TensorShape(w, h, 3U, 2U), f);
	info.auto_padding();
	tensor.allocator()->init(info);
	tensor.allocator()->allocate();
	image.init(tensor);
#else /* TEST_TENSORS */
	/* The allocator takes care of the padding */
	multi_img.init_auto_padding(w, h, f);
	multi_img.allocate();
#endif /* TEST_TENSORS */
}

void CLImage::store(RawImage &out)
{
	auto q = cl::CommandQueue::getDefault();
	out.create(width, height, format);
#ifdef TEST_TENSORS
	image.map(q);
	copy_plane(image, out.data());
	image.unmap(q);
#else /* TEST_TENSORS */
	unsigned char* tmp_ptr = out.data();
	for(int i = 0; i < 3 && (nullptr != multi_img.cl_plane(i)->cl_buffer().get()); ++i)
	{
		multi_img.cl_plane(i)->map(q);
		tmp_ptr = copy_plane(*multi_img.plane(i), tmp_ptr);
		multi_img.cl_plane(i)->unmap(q);
	}
#endif /* TEST_TENSORS */
	q.finish();
}

void CLImage::load(const RawImage &in)
{
	ERROR_ON(in.width() != width);
	ERROR_ON(in.height() != height);
	ERROR_ON(in.format() != format);

	auto q = cl::CommandQueue::getDefault();

#ifdef TEST_TENSORS
	image.map(q);
	fill_plane(in.data(),image);
	image.unmap(q);
#else /* TEST_TENSORS */
	unsigned char* tmp_ptr = in.data();
	for(int i = 0; i < 3 && (nullptr != multi_img.cl_plane(i)->cl_buffer().get()); ++i)
	{
		multi_img.cl_plane(i)->map(q);
		tmp_ptr = fill_plane(tmp_ptr , *multi_img.plane(i));
		multi_img.cl_plane(i)->unmap(q);
	}
#endif /* TEST_TENSORS */
	q.finish();
}

arm_compute::ValidRegion CLImage::valid_region() const
{
    return image.info()->valid_region();
}
