#include "mali_cv/Target.h"

#include "arm_compute/runtime/CL/CLFunctions.h"
#include "arm_compute/core/CL/CLKernels.h" //Just to check this header is well formed
#include "arm_compute/core/CL/CLHelpers.h"
#include "arm_compute/runtime/CL/CLArray.h"
#include "arm_compute/runtime/CL/CLMultiHOG.h"
#include "mali_cv/Context.h"
#include "mali_cv/Image.h"
#include "mali_cv/Lut.h"
#include "mali_cv/Convolution.h"
#include "mali_cv/Pyramid.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/RemapTable.h"
#include "mali_cv/Timer.h"
#include "TestResults.h"

#include <cmath>
#include <cstring>
#include <iostream>
#include <sstream>

using namespace mali_cv;

CLTarget::CLTarget(Context &ctx): ITarget(ctx)
{
	_timer = Timer(name());
}

void CLTarget::absolute_difference(Image &a, Image &b, Image &c)
{
	arm_compute::CLAbsoluteDifference abs_diff;

	abs_diff.configure( &a.cl.tensor, &b.cl.tensor, &c.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		_timer.start();
		abs_diff.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::bitwise_or( Image &a, Image &b, Image &c)
{
	arm_compute::CLBitwiseOr bit_or;
	bit_or.configure( &a.cl.tensor, &b.cl.tensor, &c.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		_timer.start();
		bit_or.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::bitwise_and( Image &a, Image &b, Image &c)
{
	arm_compute::CLBitwiseAnd bit_and;

	bit_and.configure( &a.cl.tensor, &b.cl.tensor, &c.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		_timer.start();
		bit_and.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::bitwise_not(Image &a, Image &c)
{
	arm_compute::CLBitwiseNot bit_not;

	bit_not.configure( &a.cl.tensor, &c.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		_timer.start();
		bit_not.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::bitwise_xor(Image &a, Image &b, Image &c)
{
	arm_compute::CLBitwiseXor bit_xor;

	bit_xor.configure( &a.cl.tensor, &b.cl.tensor, &c.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		_timer.start();
		bit_xor.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::accumulate(Image &a, Image &c)
{
	arm_compute::CLAccumulate accu;

	accu.configure(&a.cl.tensor, &c.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		_timer.start();
		accu.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::accumulate_squared(Image &a, Image &c, uint32_t shift)
{
	arm_compute::CLAccumulateSquared accu_sqr;

	accu_sqr.configure(&a.cl.tensor, shift, &c.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		_timer.start();
		accu_sqr.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}
void CLTarget::accumulate_weighted(Image &a, Image &c, float alpha)
{
	arm_compute::CLAccumulateWeighted accu_weight;

	accu_weight.configure(&a.cl.tensor, alpha, &c.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		_timer.start();
		accu_weight.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::arithmetic_addition(Image &a, Image &b, Image &c, uint8_t policy)
{
	arm_compute::CLArithmeticAddition arith_add;

	arith_add.configure(&a.cl.tensor, &b.cl.tensor, &c.cl.tensor, (arm_compute::ConvertPolicy)policy);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		_timer.start();
		arith_add.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::arithmetic_subtraction(Image &a, Image &b, Image &c, uint8_t policy)
{
	arm_compute::CLArithmeticSubtraction arith_sub;

	arith_sub.configure(&a.cl.tensor, &b.cl.tensor, &c.cl.tensor, (arm_compute::ConvertPolicy)policy);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		_timer.start();
		arith_sub.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::harris_corners(Image &a, float threshold, float min_dist, float sensitivity, uint8_t gradient, uint8_t block_size, IKeyPointArray &corners, size_t &num_corners)
{
	arm_compute::CLHarrisCorners harris_corners;

	harris_corners.configure(&a.cl.image, threshold, min_dist, sensitivity, gradient, block_size, &corners.cl.array, _ctx.border_mode(),_ctx.constant_border_value() );
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		corners.cl.array.clear();

		_timer.start();
		harris_corners.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}

	num_corners = corners.cl.array.num_values();
}

void CLTarget::fast_corners(Image &a, float threshold, bool nonmax_suppression, IKeyPointArray &corners, size_t &num_corners)
{
	arm_compute::CLFastCorners fast_corners;


	fast_corners.configure(&a.cl.image, threshold, nonmax_suppression, &corners.cl.array, nullptr, _ctx.border_mode(), _ctx.constant_border_value());

	cl::CommandQueue q = cl::CommandQueue::getDefault();

	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		_timer.start();
		fast_corners.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CLTarget::warp_affine(Image &a, Image &c, const float matrix[3][2], InterpolationMode interpolation)
{
	arm_compute::CLWarpAffine warp;
	warp.configure( &a.cl.tensor, &c.cl.tensor, (const float*)matrix, interpolation, _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		warp.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CLTarget::warp_perspective(Image &a, Image &c, const float matrix[3][3], InterpolationMode interpolation)
{
	arm_compute::CLWarpPerspective warp;
	warp.configure( &a.cl.tensor, &c.cl.tensor, (const float*)matrix, interpolation, _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		warp.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CLTarget::fill_image_borders( CLImage &a, uint num_pixels )
{
#if 0
	if( _ctx.border_mode() == BorderMode::UNDEFINED )
		return;
	NDRange gws = { DIV_CEIL(a.width + num_pixels * 2, 16) + a.height, 2};
	mali::CLKernelPtr kernel;

	switch(_ctx.border_mode())
	{
		case BorderMode::CONSTANT:
			switch(num_pixels)
			{
				case 1:
					kernel = _ctx.cl.get_kernel( "fill_image_borders_1px_U8_constant");
					break;
				case 2:
					kernel = _ctx.cl.get_kernel( "fill_image_borders_2px_U8_constant");
					break;
				default:
					ERROR("Number of pixels not supported");
			}
			kernel->add_arg((cl_uint)_ctx.constant_border_value());
			break;
		case BorderMode::REPLICATE:
			switch(num_pixels)
			{
				case 1:
					kernel = _ctx.cl.get_kernel( "fill_image_borders_1px_U8_replicate");
					break;
				case 2:
					kernel = _ctx.cl.get_kernel( "fill_image_borders_2px_U8_replicate");
					break;
				default:
					ERROR("Number of pixels not supported");
			}
			break;
		default:
			ERROR("Impossible");
	}
	add_kernel_arg(kernel, a);
	queue()->enqueue_kernel( kernel, 0, gws, 0 );
#else
	ERROR("Not implemented");
#endif
}

void CLTarget::erode(Image &a, Image &c)
{
	arm_compute::CLErode erode;

	erode.configure(&a.cl.tensor, &c.cl.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		erode.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::dilate(Image &a, Image &c)
{
	arm_compute::CLDilate dilate;

	dilate.configure(&a.cl.tensor, &c.cl.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		dilate.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::gaussian3x3(Image &a, Image &c)
{
	arm_compute::CLGaussian3x3 gaussian3x3;

	gaussian3x3.configure(&a.cl.tensor, &c.cl.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		gaussian3x3.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::gaussian5x5( Image &a, Image &c)
{
	arm_compute::CLGaussian5x5 gaussian5x5;

	gaussian5x5.configure(&a.cl.tensor, &c.cl.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		gaussian5x5.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::box3x3(Image &a, Image &c)
{
	arm_compute::CLBox3x3 box3x3;

	box3x3.configure(&a.cl.tensor, &c.cl.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		box3x3.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::median3x3(Image &a, Image &c)
{
	arm_compute::CLMedian3x3 median3x3;

	median3x3.configure(&a.cl.tensor, &c.cl.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		median3x3.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::scharr(Image &a, Image *c, Image *d)
{
	arm_compute::CLScharr3x3 scharr3x3;

	scharr3x3.configure(&a.cl.tensor, c ? &c->cl.tensor : nullptr, d ? &d->cl.tensor : nullptr,
	                    _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		scharr3x3.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::sobel(Image &a, Image *c, Image *d)
{
	arm_compute::CLSobel3x3 sobel3x3;

	sobel3x3.configure(&a.cl.tensor, c ? &c->cl.tensor : nullptr, d ? &d->cl.tensor : nullptr,
	                    _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		sobel3x3.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::sobel5x5(Image &a, Image *c, Image *d)
{
	arm_compute::CLSobel5x5 sobel5x5;

	sobel5x5.configure(&a.cl.tensor, c ? &c->cl.tensor : nullptr, d ? &d->cl.tensor : nullptr,
	                    _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		sobel5x5.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::sobel7x7(Image &a, Image *c, Image *d)
{
	arm_compute::CLSobel7x7 sobel7x7;

	sobel7x7.configure(&a.cl.tensor, c ? &c->cl.tensor : nullptr, d ? &d->cl.tensor : nullptr,
	                    _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		sobel7x7.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::convolution( Image &a, Image &c, Convolution &matrix)
{
	ERROR_ON(matrix.cl.width != 3 && matrix.cl.width != 5 && matrix.cl.width != 7 && matrix.cl.width != 9);
	ERROR_ON(matrix.cl.height != 3 && matrix.cl.height != 5 && matrix.cl.height != 7 && matrix.cl.height != 9);

	// Run test
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		if(matrix.cl.width == matrix.cl.height)
		{
		    switch(matrix.cl.width)
		    {
			case 3:
				{
				arm_compute::CLConvolution3x3 conv;
				conv.configure( &a.cl.tensor, &c.cl.tensor, matrix.cl.convolution.get(), matrix.cl._scale , _ctx.border_mode(), _ctx.constant_border_value());
				conv.run();
				break;
				}
			case 5:
				{
				arm_compute::CLConvolution5x5 conv;
				conv.configure( &a.cl.tensor, &c.cl.tensor, matrix.cl.convolution.get(), matrix.cl._scale , _ctx.border_mode(), _ctx.constant_border_value());
				conv.run();
				break;
				}
			case 7:
				{
				arm_compute::CLConvolution7x7 conv;
				conv.configure( &a.cl.tensor, &c.cl.tensor, matrix.cl.convolution.get(), matrix.cl._scale , _ctx.border_mode(), _ctx.constant_border_value());
				conv.run();
				break;
				}
			case 9:
				{
				arm_compute::CLConvolution9x9 conv;
				conv.configure( &a.cl.tensor, &c.cl.tensor, matrix.cl.convolution.get(), matrix.cl._scale , _ctx.border_mode(), _ctx.constant_border_value());
				conv.run();
				break;
				}
			default:
				ERROR("Not supported");
				break;
		    }
		}
		else
		{
			arm_compute::CLConvolutionRectangle conv;
			conv.configure(&a.cl.tensor, &c.cl.tensor, matrix.cl.convolution.get(), matrix.cl.width, matrix.cl.height,
			matrix.cl._scale, _ctx.border_mode(), _ctx.constant_border_value());
			conv.run();
			break;
		}
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::scale( Image &a, Image &c, InterpolationMode interpolation)
{
	arm_compute::CLScale scale;
	scale.configure(&a.cl.tensor, &c.cl.tensor, interpolation, _ctx.border_mode(), _ctx.constant_border_value());

	cl::CommandQueue q = cl::CommandQueue::getDefault();

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		scale.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CLTarget::laplacian_reconstruct(Pyramid &pyramid, Image &a, Image &c)
{
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	arm_compute::CLLaplacianReconstruct lr;
	lr.configure(&pyramid.cl.pyramid, &a.cl.image, &c.cl.image,_ctx.border_mode(),_ctx.constant_border_value());

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		lr.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::laplacian_pyramid(Image &a, Pyramid &pyramid, Image &b, bool timers_enabled)
{
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	arm_compute::CLLaplacianPyramid lp;
	lp.configure(&a.cl.image, &pyramid.cl.pyramid, &b.cl.image, _ctx.border_mode(),_ctx.constant_border_value());

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		lp.run();
		q.finish();
		if(timers_enabled)
		{
			_timer.stop();
			TestResults::report().add(name(), _timer.get_elapsed_ms());
			_timer.print_elapsed_ms();
		}
	}
}

void CLTarget::gaussian_pyramid( Image &a, Pyramid &pyramid)
{
	ERROR_ON(Format::LUMINANCE != a.cl.format );

	cl::CommandQueue q = cl::CommandQueue::getDefault();
	std::unique_ptr<arm_compute::CLGaussianPyramid> gaussian_pyramid;

	if( arm_compute::SCALE_PYRAMID_HALF == pyramid.scale )
	{
		gaussian_pyramid.reset(new arm_compute::CLGaussianPyramidHalf());
	}
	else if( arm_compute::SCALE_PYRAMID_ORB == pyramid.scale )
	{
		gaussian_pyramid.reset(new arm_compute::CLGaussianPyramidOrb());
	}
	else
	{
		ERROR("Scale factor not supported for gaussian pyramid");
	}

	gaussian_pyramid->configure(&a.cl.image, &pyramid.cl.pyramid, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		gaussian_pyramid->run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::color_convert( Image &a, Image &c )
{
	arm_compute::CLColorConvert color_convert;

	cl::CommandQueue q = cl::CommandQueue::getDefault();

	switch(a.cl.format)
	{
		case Format::RGB888:
			switch(c.cl.format)
			{
				case Format::RGBA8888:
					color_convert.configure(&a.cl.tensor, &c.cl.tensor);
					break;
#ifndef TEST_TENSORS
				case Format::IYUV:
				case Format::YUV444:
				case Format::NV12:
					color_convert.configure(&a.cl.image, &c.cl.multi_img);
					break;
#endif /* TEST_TENSORS */
				default:
					ERROR("Not supported");
					break;

			}
			break;
		case Format::RGBA8888:
			switch(c.cl.format)
			{
				case Format::RGB888:
					color_convert.configure(&a.cl.tensor, &c.cl.tensor);
					break;
#ifndef TEST_TENSORS
				case Format::NV12:
				case Format::IYUV:
				case Format::YUV444:
					color_convert.configure(&a.cl.image, &c.cl.multi_img);
					break;
#endif /* TEST_TENSORS */
				default:
					ERROR("Not supported");
					break;
			}
			break;
		case Format::NV12:
		case Format::NV21:
			switch(c.cl.format)
			{
#ifndef TEST_TENSORS
				case Format::RGB888:
				case Format::RGBA8888:
					color_convert.configure(&a.cl.multi_img, &c.cl.image);
					break;
				case Format::IYUV:
				case Format::YUV444:
					color_convert.configure(&a.cl.multi_img, &c.cl.multi_img);
					break;
#endif /* TEST_TENSORS */
				default:
					ERROR("Not supported");
					break;
			}
			break;
		case Format::UYVY:
		case Format::YUYV422:
			switch(c.cl.format)
			{
				case Format::RGB888:
				case Format::RGBA8888:
					color_convert.configure(&a.cl.tensor, &c.cl.tensor);
					break;
#ifndef TEST_TENSORS
				case Format::IYUV:
				case Format::NV12:
					color_convert.configure(&a.cl.image, &c.cl.multi_img);
					break;
#endif /* TEST_TENSORS */
				default:
					ERROR("Not supported");
					break;
			}
			break;
		case Format::IYUV:
			switch(c.cl.format)
			{
#ifndef TEST_TENSORS
				case Format::RGB888:
				case Format::RGBA8888:
					color_convert.configure(&a.cl.multi_img, &c.cl.image);
					break;
				case Format::YUV444:
				case Format::NV12:
					color_convert.configure(&a.cl.multi_img, &c.cl.multi_img);
					break;
#endif /* TEST_TENSORS */
				default:
					ERROR("Not supported");
					break;
			}
			break;
		default:
			ERROR("Not supported");
			break;

	}

	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		_timer.start();
		color_convert.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::canny(Image &a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image &c)
{
	arm_compute::CLCannyEdge canny;

	canny.configure(&a.cl.image, &c.cl.image, upper, lower, gradient_size, norm_type, _ctx.border_mode(), _ctx.constant_border_value());
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		canny.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::mean_stddev(Image &a, Scalar<float> &mean, Scalar<float> &stddev)
{
	arm_compute::CLMeanStdDev mean_stddev;

	mean_stddev.configure(&a.cl.image, &mean.cl.scalar, &stddev.cl.scalar);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j = 0; j < _ctx.num_iterations(); j++ )
	{
		_timer.start();
		mean_stddev.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::convert_depth(Image &a, Image &c, ConvertPolicy policy , uint32_t shift)
{
	arm_compute::CLDepthConvert depth_conv;

	depth_conv.configure(&a.cl.tensor, &c.cl.tensor, policy, shift);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		depth_conv.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::threshold_binary(Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value)
{
	arm_compute::CLThreshold thresh;

	thresh.configure(&a.cl.tensor, &b.cl.tensor, threshold, false_value, true_value, arm_compute::ThresholdType::BINARY);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		thresh.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::threshold_range(Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value, uint8_t upper)
{
	arm_compute::CLThreshold thresh;

	thresh.configure(&a.cl.tensor, &b.cl.tensor, threshold, false_value, true_value, arm_compute::ThresholdType::RANGE, upper);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		thresh.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::optical_flow(Image &a, Image &b, IKeyPointArray *old_points, IKeyPointArray *new_points_estimates,
                            IKeyPointArray *new_points, Termination termination, float epsilon, uint32_t num_iterations,
                            bool use_initial_estimate, size_t window_dimension, size_t num_levels)
{
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	arm_compute::CLGaussianPyramidHalf gp1;
	arm_compute::CLGaussianPyramidHalf gp2;
	arm_compute::CLOpticalFlow of;
	arm_compute::CLPyramid old_pyramid;
	arm_compute::CLPyramid new_pyramid;
	arm_compute::PyramidInfo pyramid_info;

	pyramid_info.init(num_levels, 0.5f, a.cl.width, a.cl.height, arm_compute::Format::U8);

	old_pyramid.init_auto_padding(pyramid_info);
	old_pyramid.allocate();
	new_pyramid.init_auto_padding(pyramid_info);
	new_pyramid.allocate();

	gp1.configure(&a.cl.image, &old_pyramid, _ctx.border_mode(), _ctx.constant_border_value());
	gp2.configure(&b.cl.image, &new_pyramid, _ctx.border_mode(), _ctx.constant_border_value());

	gp1.run();
	gp2.run();

	of.configure(&old_pyramid, &new_pyramid, &old_points->cl.array, &new_points_estimates->cl.array, &new_points->cl.array, termination, epsilon,
	             num_iterations, window_dimension, use_initial_estimate, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j = 0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		of.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::integral(Image &a, Image &b)
{
	arm_compute::CLIntegralImage integral;

	integral.configure(&a.cl.tensor, &b.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		integral.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::pixelwise_mul(Image &a, Image &b, Image &c, float scale, uint8_t overflow_policy, uint8_t round_policy)
{
	arm_compute::CLPixelWiseMultiplication pixelwise_mul;
	arm_compute::ConvertPolicy op = (overflow_policy) ? arm_compute::ConvertPolicy::SATURATE : arm_compute::ConvertPolicy::WRAP;
	arm_compute::RoundingPolicy rp = (round_policy) ? arm_compute::RoundingPolicy::TO_NEAREST_EVEN : arm_compute::RoundingPolicy::TO_ZERO;

	pixelwise_mul.configure(&a.cl.tensor, &b.cl.tensor, &c.cl.tensor, scale, op, rp);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for( unsigned int j= 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		pixelwise_mul.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::histogram(Image& a, Distribution& b)

{
	arm_compute::CLHistogram hist;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	hist.configure(&a.cl.image, b.cl.dist.get());
	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		b.cl.dist->map(q);
		b.cl.dist->clear();
		b.cl.dist->unmap(q);

		_timer.start();
		hist.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::magnitude(Image &a, Image &b, Image &c)
{
	arm_compute::CLMagnitude mag;

	mag.configure(&a.cl.tensor, &b.cl.tensor, &c.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		mag.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::phase(Image &a, Image &b, Image &c)
{
	arm_compute::CLPhase phase;

	phase.configure(&a.cl.tensor, &b.cl.tensor, &c.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		phase.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::equalize_histogram(Image &a, Image &b)
{
	arm_compute::CLEqualizeHistogram hist_equalize;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	hist_equalize.configure(&a.cl.image, &b.cl.image);
	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		hist_equalize.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::tablelookup( Image &a , Lut &b, Image &c )
{
	arm_compute::CLTableLookup lut;
	lut.configure( &a.cl.tensor, &b.cl.lut, &c.cl.tensor );
	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		lut.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::channel_combine(Image &plane0, Image &plane1, Image &plane2, Image &plane3, Image &out)
{
	arm_compute::CLChannelCombine ch_combine;

	// Initialize function appropriately.
	switch( out.cl.format )
	{
		case Format::NV12:
		case Format::NV21:
		case Format::IYUV:
		case Format::YUV444:
#ifdef TEST_TENSORS
			ERROR("Not supported");
#else /* TEST_TENSORS */
			ch_combine.configure(&plane0.cl.image, &plane1.cl.image, &plane2.cl.image, &out.cl.multi_img);
#endif /* TEST_TENSORS */
			break;
		case Format::RGBA8888:
			ch_combine.configure(&plane0.cl.tensor, &plane1.cl.tensor, &plane2.cl.tensor, &plane3.cl.tensor, &out.cl.tensor);
			break;
		default:
			ch_combine.configure(&plane0.cl.tensor, &plane1.cl.tensor, &plane2.cl.tensor, nullptr, &out.cl.tensor);
			break;
	}

	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		ch_combine.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::channel_extract( Image &a, Channel channel, Image &c )
{
	arm_compute::CLChannelExtract ch_extract;

	// Initialize function appropriately.
	switch( a.cl.format )
	{
		case Format::NV12:
		case Format::NV21:
		case Format::IYUV:
		case Format::YUV444:
#ifdef TEST_TENSORS
			ERROR("Not supported");
#else /* TEST_TENSORS */
			ch_extract.configure(&a.cl.multi_img, channel, &c.cl.image);
#endif /* TEST_TENSORS */
			break;
		default:
			ch_extract.configure(&a.cl.tensor, channel, &c.cl.tensor);
			break;
	}

	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		ch_extract.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::derivative(Image &a, Image *c, Image *d)
{
	arm_compute::CLDerivative der;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	der.configure( &a.cl.tensor, c ? &c->cl.tensor : nullptr, d ? &d->cl.tensor : nullptr, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		der.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::minmaxlocation( Image &a, Scalar<int32_t> &min, Scalar<int32_t> &max, ICoord2DArray *min_loc, ICoord2DArray *max_loc,
                               Scalar<uint32_t> *min_count, Scalar<uint32_t> *max_count)
{
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	arm_compute::CLMinMaxLocation min_max_loc;
	uint32_t* min_count_ptr = (min_count) ? &min_count->cl.scalar : nullptr;
	uint32_t* max_count_ptr = (max_count) ? &max_count->cl.scalar : nullptr;
	arm_compute::CLCoordinates2DArray* min_loc_ptr = (min_loc) ? &min_loc->cl.array : nullptr;
	arm_compute::CLCoordinates2DArray* max_loc_ptr = (max_loc) ? &max_loc->cl.array : nullptr;
	min_max_loc.configure(&a.cl.image, &min.cl.scalar, &max.cl.scalar, min_loc_ptr, max_loc_ptr, min_count_ptr, max_count_ptr);
	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		min_max_loc.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::non_linear_filter(Image &a, Image &b, NonLinearFilterFunction function, MatrixPattern pattern, uint mask_size, uint8_t* mask)
{
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	arm_compute::CLNonLinearFilter nl_filter;

	nl_filter.configure(&a.cl.tensor, &b.cl.tensor, function, mask_size, pattern, mask, _ctx.border_mode(), _ctx.constant_border_value());
	for( unsigned int i = 0; i < _ctx.num_iterations(); ++i )
	{
		_timer.start();
		nl_filter.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::remap(Image &a, RemapTable &table, Image &d, InterpolationMode interpolation)
{
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	arm_compute::CLRemap remap;
	remap.configure(&a.cl.tensor, &table.cl.table[0]->tensor, &table.cl.table[1]->tensor, &d.cl.tensor, interpolation, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		remap.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::activation_layer(Tensor &in, Tensor &out, ActivationLayerInfo act_info)
{
	arm_compute::CLActivationLayer actl;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	actl.configure(&in.cl.tensor, &out.cl.tensor, act_info);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		actl.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::pooling_layer(Tensor &in, Tensor &out, PoolingLayerInfo pool_info)
{
	arm_compute::CLPoolingLayer pool;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	pool.configure(&in.cl.tensor, &out.cl.tensor, pool_info);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		pool.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::tensor_ops(Tensor &in1, Tensor &in2, Tensor &out, TensorOp op)
{
	arm_compute::IFunction* func = nullptr;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	arm_compute::CLArithmeticAddition add;
	arm_compute::CLArithmeticSubtraction sub;
	arm_compute::CLPixelWiseMultiplication mul;

	switch(op)
	{
		case TensorOp::ADD:
			add.configure(&in1.cl.tensor, &in2.cl.tensor, &out.cl.tensor, ConvertPolicy::WRAP);
			func = &add;
			break;
		case TensorOp::SUB:
			sub.configure(&in1.cl.tensor, &in2.cl.tensor, &out.cl.tensor, ConvertPolicy::WRAP);
			func = &sub;
			break;
		case TensorOp::MUL:
			mul.configure(&in1.cl.tensor, &in2.cl.tensor, &out.cl.tensor, 1.f, ConvertPolicy::WRAP, RoundingPolicy::TO_ZERO);
			func = &mul;
			break;
		case TensorOp::LUT:
		case TensorOp::TRANSPOSE:
		default:
			ERROR("Unsupported tensor operation");
	}

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		func->run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::softmax_layer(Tensor &in, Tensor &out)
{
	arm_compute::CLSoftmaxLayer sml;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	sml.configure(&in.cl.tensor, &out.cl.tensor);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		sml.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::normalization_layer(Tensor &a, Tensor &b, NormalizationLayerInfo norm_info)
{
	arm_compute::CLNormalizationLayer norm;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	norm.configure(&a.cl.tensor, &b.cl.tensor, norm_info);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		norm.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::batchnormalization_layer(Tensor &a, Tensor &b, Tensor &mean, Tensor &var, Tensor &beta, Tensor &gamma, const float epsilon)
{
	arm_compute::CLBatchNormalizationLayer norm;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	norm.configure(&a.cl.tensor, &b.cl.tensor, &mean.cl.tensor, &var.cl.tensor, &beta.cl.tensor, &gamma.cl.tensor, epsilon);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		norm.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::convolution_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info, Tensor *rw)
{
	cl::CommandQueue q = cl::CommandQueue::getDefault();

    if(rw)
    {

        arm_compute::CLConvolutionLayerReshapeWeights convlrshape;
        convlrshape.configure(&w.cl.tensor, b ? &b->cl.tensor : nullptr, & rw->cl.tensor, true);
        convlrshape.run();
    }

	arm_compute::CLConvolutionLayer convl;

	convl.configure(&in.cl.tensor, rw ? & rw->cl.tensor : &w.cl.tensor, b ? &b->cl.tensor : nullptr, &out.cl.tensor, conv_info,
            rw ? arm_compute::WeightsInfo(true, w.cl.tensor.info()->dimension(0), w.cl.tensor.info()->dimension(1), w.cl.tensor.info()->dimension(3)) : arm_compute::WeightsInfo());

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		convl.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::locally_connected_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info)
{
	arm_compute::CLLocallyConnectedLayer lc;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	lc.configure(&in.cl.tensor, &w.cl.tensor, b ? &b->cl.tensor : nullptr, &out.cl.tensor, conv_info);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		lc.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::fully_connected_layer(Tensor &input, Tensor &weight, Tensor *bias, Tensor &output, bool are_weights_reshaped)
{
	arm_compute::CLTensor                            weight_reshaped;
	arm_compute::CLFullyConnectedLayer               fc;
	arm_compute::CLFullyConnectedLayerReshapeWeights reshape_weights;
	cl::CommandQueue                                 q = cl::CommandQueue::getDefault();

	const arm_compute::CLTensor *weight_to_use = &weight.cl.tensor;
	if(are_weights_reshaped)
	{
		weight_to_use = &weight_reshaped;

		// Note: the tensor weight_reshaped will be initialized within the configuration of CLFullyConnectedLayerReshapeWeights
		reshape_weights.configure(&weight.cl.tensor, &weight_reshaped);
	}

	fc.configure(&input.cl.tensor, weight_to_use, bias ? &bias->cl.tensor : nullptr, &output.cl.tensor, true, are_weights_reshaped);

	if(are_weights_reshaped)
	{
		// Reshape the weights before running convolution layer
		weight_reshaped.allocator()->allocate();
		reshape_weights.run();
		q.finish();
	}

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		fc.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::transpose( Image &a, Image &b)
{
	arm_compute::CLTranspose transpose;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	transpose.configure(&a.cl.tensor, &b.cl.tensor);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		transpose.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::gemm_fp32( Tensor &a, Tensor &b, Tensor &c, Tensor &d, float alpha, float beta)
{
	arm_compute::CLGEMM gemm;

	gemm.configure(&a.cl.tensor,  &b.cl.tensor, &c.cl.tensor, &d.cl.tensor, alpha, beta);

	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		_timer.start();
		gemm.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::gemm_fp16( Image &a, Image &b, Image &c, Image &d, float alpha, float beta)
{
	arm_compute::CLGEMM gemm;

	gemm.configure(&a.cl.tensor,  &b.cl.tensor, &c.cl.tensor, &d.cl.tensor, alpha, beta);

	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		_timer.start();
		gemm.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::gemm_lowp(Image &a, Image &b, Image &d, int32_t a_offset, int32_t b_offset, int32_t c_offset, int32_t c_mult_int, int32_t shift)
{
	arm_compute::CLGEMMLowp gemm;
	gemm.configure(&a.cl.tensor, &b.cl.tensor, &d.cl.tensor, a_offset, b_offset, c_offset, c_mult_int, shift);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		gemm.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::concatenate_depth(std::vector<std::unique_ptr<Tensor>>& a, Tensor &b)
{
	arm_compute::CLDepthConcatenate cd;
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	std::vector<arm_compute::ICLTensor*> input;

	for(auto& temp : a)
	{
		input.push_back(&(temp->cl.tensor));
	}

	cd.configure(input, &b.cl.tensor);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		cd.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::interleave_4x4(Image &in, Image &out)
{
	arm_compute::CLGEMMInterleave4x4 inter4x4;
	inter4x4.configure(&in.cl.tensor,  &out.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		inter4x4.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::hog_descriptor(Image &a, HOG &b, Tensor &c)
{
	arm_compute::CLHOGDescriptor hog_descriptor;

	hog_descriptor.configure( &a.cl.tensor, &c.cl.tensor, &b.cl.hog, _ctx.border_mode(), _ctx.constant_border_value() );

	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		_timer.start();
		hog_descriptor.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CLTarget::hog_detector(Tensor &a, HOG &b, IKeyPointArray &keypoint_detection_windows)
{
	arm_compute::CLHOGDetector hog_detector;
	arm_compute::CLDetectionWindowArray detection_windows(10000);

	hog_detector.configure(&a.cl.tensor, &b.cl.hog, &detection_windows, b.cl.hog.info()->block_stride());

	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		// Reset detection windows
		detection_windows.clear();

		_timer.start();
		hog_detector.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}

	// Copy the detection  windows
	std::vector<IKeyPoint> keypoints;

	detection_windows.map(q, true);

	for(uint i = 0; i < detection_windows.num_values(); i++)
	{
		IKeyPoint val;
		val.x = detection_windows.at(i).x;
		val.y = detection_windows.at(i).y;
		val.tracking_status = 1;
		val.strength = detection_windows.at(i).score;

		keypoints.push_back(val);
	}

	detection_windows.unmap(q);

	keypoint_detection_windows.cl.fill(keypoints);
}

void CLTarget::hog_multi_detection(Image &a, HOG &b, IKeyPointArray &keypoint_detection_windows)
{
	arm_compute::CLHOGMultiDetection    hog_multi_detection;
	arm_compute::CLMultiHOG             multi_hog(1);
	arm_compute::CLDetectionWindowArray detection_windows(10000);
	arm_compute::CLSize2DArray          detection_window_strides(1);
	cl::CommandQueue                    q = cl::CommandQueue::getDefault();

	// Resize detection_window_strides
	detection_window_strides.resize(1);

	// Map detection window strides array
	detection_window_strides.map(q, true);

	for(uint i = 0; i < 1; i++)
	{
		auto hog_model = reinterpret_cast<arm_compute::CLHOG*>(multi_hog.cl_model(i));

		// Initialize HOG model
		hog_model->init(*(b.cl.hog.info()));

		hog_model->map(q, true);
		b.cl.hog.map(q, true);

		// Initialize HOG descriptor
		std::memcpy(hog_model->descriptor(), b.cl.hog.descriptor(), b.descriptor_size * sizeof(float));

		hog_model->unmap(q);
		b.cl.hog.unmap(q);

		// Initialize detection window stride
		detection_window_strides.at(i) = multi_hog.cl_model(i)->info()->block_stride();
	}

	// Unmap detection window strides array
	detection_window_strides.unmap(q);

	hog_multi_detection.configure( &a.cl.tensor, &multi_hog, &detection_windows, &detection_window_strides, arm_compute::BorderMode::REPLICATE, 0, 0.0f, false, 30.0f);

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		// Reset detection windows
		detection_windows.clear();

		_timer.start();
		hog_multi_detection.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}

	// Copy the detection  windows
	std::vector<IKeyPoint> keypoints;

	detection_windows.map(q, true);

	for(uint i = 0; i < detection_windows.num_values(); i++)
	{
		IKeyPoint val;
		val.x = detection_windows.at(i).x;
		val.y = detection_windows.at(i).y;
		val.tracking_status = 1;
		val.strength = detection_windows.at(i).score;

		keypoints.push_back(val);
	}

	detection_windows.unmap(q);

	keypoint_detection_windows.cl.fill(keypoints);
}

void CLTarget::transpose_1xW(Image &a, Image &b)
{
	arm_compute::CLGEMMTranspose1xW transpose1xW;

	transpose1xW.configure(&a.cl.tensor, &b.cl.tensor);
	cl::CommandQueue q = cl::CommandQueue::getDefault();
	for(uint i = 0; i < _ctx.num_iterations() ; i++)
	{
		_timer.start();
		transpose1xW.run();
		q.finish();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}
