#include "mali_cv/Image.h"

#include "mali_cv/Error.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/Helper.h"

using namespace mali_cv;

CVImage::CVImage():IImage(),img()
{
}

static cv::Size compute_matrix_size(size_t width, size_t height, Format format)
{
	cv::Size size = cv::Size();
	switch( format )
	{
		case Format::NV12:
		case Format::NV21:
		case Format::IYUV:
			size = cv::Size(width, height / 2 * 3);
			break;
		case Format::RGB888:
		case Format::RGBA8888:
		case Format::UYVY:
		case Format::YUYV422:
			size = cv::Size(width, height);
			break;
		case Format::YUV444:
			size = cv::Size(width, height * 3);
			break;
		case Format::LUMINANCE:
		case Format::LUMINANCE16:
		case Format::FP32C1:
		default:
			size = cv::Size(width, height);
	}
	return size;
}

void CVImage::internal_create( Context &ctx, size_t w, size_t h, Format f)
{
	width = w;
	height = h;
	format = f;
	img = cv::Mat( compute_matrix_size(width, height, f), CVHelper::convert_to_cv_format( format ));
}

void CVImage::store( RawImage &out )
{
	out.create( width, height, format);

	memcpy( out.data(), img.data, out.num_pixels()*out.pixel_size());
}

void CVImage::load(const RawImage &in )
{
	ERROR_ON( in.width() != width );
	ERROR_ON( in.height() != height );

	switch(img.type())
	{
		case CV_8U:
		{
			size_t size = 0;

			if(in.pixel_size() == 1)
			{
				size = in.num_pixels();
			}
			else
			{
				auto s = compute_matrix_size(width, height, format);
				size = s.area();
			}

			memcpy( img.data, in.data(), size);
			break;
		}
		case CV_16U:
		case CV_16S:
			if(in.pixel_size() == 2)
				memcpy( img.data, in.data(), in.num_pixels() * in.pixel_size());
			else
				ERROR("TODO");
			break;
		case CV_32S:
			if(in.pixel_size() == 4)
				memcpy( img.data, in.data(), in.num_pixels() * in.pixel_size());
			else
				ERROR("TODO");
			break;
		case CV_8UC2:
			if(in.pixel_size() == 2)
				memcpy( img.data, in.data(), in.num_pixels() * in.pixel_size());
			else
				ERROR("TODO");
			break;
		case CV_8UC3:
			if(in.pixel_size() == 3)
				memcpy( img.data, in.data(), in.num_pixels() * in.pixel_size());
			else
				ERROR("TODO");
			break;
		case CV_8UC4:
			if(in.pixel_size() == 4)
				memcpy( img.data, in.data(), in.num_pixels() * in.pixel_size());
			else
				ERROR("TODO");
			break;
		case CV_32F:
			if(in.pixel_size() == 4)
				memcpy( img.data, in.data(), in.num_pixels() * in.pixel_size());
			else
				ERROR("TODO");
			break;
		default:
		{
			ERROR_ON( in.pixel_size() != 1);
			int16_t *out = (int16_t *) img.data;
			for(uint i = 0; i < in.num_pixels(); i++)
			{
				out[i] = in.data()[i];
			}
		}
	}
}
