#include "mali_cv/Target.h"

#include "arm_compute/runtime/Scheduler.h"
#include "mali_cv/Error.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/Convolution.h"
#include "mali_cv/Pyramid.h"
#include "mali_cv/RemapTable.h"
#include "mali_cv/Timer.h"
#include "mali_cv/Types.h"
#include "TestResults.h"

using namespace mali_cv;

namespace
{
int cv_border_type( BorderMode mode )
{
	switch( mode )
	{
		case BorderMode::CONSTANT:
			return cv::BORDER_CONSTANT;
		case BorderMode::UNDEFINED: /* Note: There is no undefined in OpenCV */
		case BorderMode::REPLICATE:
			return cv::BORDER_REPLICATE;
		default:
			ERROR("Border mode not supported");
			return cv::BORDER_DEFAULT;
	}
}
}

CVTarget::CVTarget(Context &ctx): ITarget(ctx),_eigen(ctx, name())
{
	_timer = Timer(name());
	cv::setNumThreads(arm_compute::Scheduler::get().num_threads()); //make sure OpenCV uses the same number of threads as the runtime's schedulers
}

void CVTarget::absolute_difference( Image &a, Image &b, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::absdiff( a.cv.img, b.cv.img, c.cv.img);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::bitwise_or( Image &a, Image &b, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::bitwise_or( a.cv.img, b.cv.img, c.cv.img);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::bitwise_not( Image &a, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::bitwise_not( a.cv.img, c.cv.img);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::accumulate( Image &a, Image &c)
{
	const cv::Size s(a.cv.img.cols, a.cv.img.rows);
	cv::Mat out(s, CV_32F);
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		c.cv.img.convertTo(out, CV_32F);
		_timer.start();
		cv::accumulate(a.cv.img , out);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
	out.convertTo(c.cv.img, CV_16U);
}

void CVTarget::accumulate_squared( Image &a, Image &c, uint32_t shift)
{
	ERROR("cv::accumulate only handles floating point input/output");
}

void CVTarget::accumulate_weighted( Image &a, Image &c, float alpha)
{
	ERROR("cv::accumulate only handles floating point input/output");
}

void CVTarget::arithmetic_addition( Image &a, Image &b, Image &c, uint8_t policy)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		if( Format::LUMINANCE16 == c.cv.format )
		{
			cv::add( a.cv.img, b.cv.img, c.cv.img,  cv::noArray(),  CV_16S);
		}
		else
		{
			cv::add( a.cv.img, b.cv.img, c.cv.img);
		}

		_timer.stop();

		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::arithmetic_subtraction( Image &a, Image &b, Image &c, uint8_t policy)
{
	if( Format::LUMINANCE16 == c.cv.format )
	{
		for(uint i = 0; i < _ctx.num_iterations(); i++)
		{
			_timer.start();
			cv::subtract( a.cv.img, b.cv.img, c.cv.img,  cv::noArray(),  CV_16S);
			_timer.stop();
			TestResults::report().add(name(), _timer.get_elapsed_ms() );
			_timer.print_elapsed_ms();
		}
	}
	else
	{
		for(uint i = 0; i < _ctx.num_iterations(); i++)
		{
			_timer.start();
			cv::subtract( a.cv.img, b.cv.img, c.cv.img);
			_timer.stop();
			TestResults::report().add(name(), _timer.get_elapsed_ms() );
			_timer.print_elapsed_ms();
		}
	}
}

void CVTarget::harris_corners(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners, size_t& num_corners)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
        cv::Mat dst;

		_timer.start();
		cv::goodFeaturesToTrack( a.cv.img, dst, 10000, 0.3, min_dist, cv::noArray(), block_size, true, sensitivity );
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();

        for(int i = 0; i < dst.rows; ++i)
        {
            cv::Vec2f corner = dst.at<cv::Vec2f>(i);
            cv::KeyPoint kp;
            kp.pt.x = corner[0];
            kp.pt.y = corner[1];
            corners.cv.array.push_back(kp);
        }

        num_corners = dst.rows;
	}
}

void CVTarget::fast_corners(Image &a,float threshold, bool nonmax_suppression, IKeyPointArray &corners, size_t &num_corners)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		corners.cv.array.clear();

		_timer.start();
		cv::FAST(a.cv.img, corners.cv.array , threshold, false, cv::FastFeatureDetector::TYPE_9_16);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();

        num_corners = corners.cv.array.size();
	}
}

void CVTarget::warp_perspective(Image &a, Image &c, const float matrix[3][3], InterpolationMode mode)
{
	cv::Mat warp_mat( 3, 3, CV_32FC1 );
	warp_mat.at<float>(0,0) = matrix[0][0];
	warp_mat.at<float>(0,1) = matrix[1][0];
	warp_mat.at<float>(0,2) = matrix[2][0];
	warp_mat.at<float>(1,0) = matrix[0][1];
	warp_mat.at<float>(1,1) = matrix[1][1];
	warp_mat.at<float>(1,2) = matrix[2][1];
	warp_mat.at<float>(2,0) = matrix[0][2];
	warp_mat.at<float>(2,1) = matrix[1][2];
	warp_mat.at<float>(2,2) = matrix[2][2];

	const int interpolation = CVHelper::convert_to_cv_interpolation(mode);

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::warpPerspective( a.cv.img, c.cv.img, warp_mat, c.cv.img.size(), interpolation | cv::WARP_INVERSE_MAP, cv_border_type(_ctx.border_mode()) );
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::warp_affine(Image &a, Image &c, const float matrix[3][2], InterpolationMode mode)
{
	cv::Mat warp_mat( 2, 3, CV_32FC1 );
	warp_mat.at<float>(0,0) = matrix[0][0];
	warp_mat.at<float>(0,1) = matrix[1][0];
	warp_mat.at<float>(0,2) = matrix[2][0];
	warp_mat.at<float>(1,0) = matrix[0][1];
	warp_mat.at<float>(1,1) = matrix[1][1];
	warp_mat.at<float>(1,2) = matrix[2][1];

	const int interpolation = CVHelper::convert_to_cv_interpolation(mode);

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::warpAffine( a.cv.img, c.cv.img, warp_mat, c.cv.img.size(), interpolation | cv::WARP_INVERSE_MAP, cv_border_type(_ctx.border_mode()) );
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::bitwise_and( Image &a, Image &b, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::bitwise_and( a.cv.img, b.cv.img, c.cv.img);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::bitwise_xor( Image &a, Image &b, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::bitwise_xor( a.cv.img, b.cv.img, c.cv.img);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::pixelwise_mul( Image &a, Image &b, Image &c, float scale, uint8_t overflow_policy, uint8_t round_policy )
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		if( Format::LUMINANCE16 == c.cv.format )
		{
			cv::multiply( a.cv.img, b.cv.img, c.cv.img, scale,  CV_16S);
		}
		else
		{
			cv::multiply( a.cv.img, b.cv.img, c.cv.img, scale );
		}

		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::erode( Image &a, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::erode( a.cv.img, c.cv.img, cv::getStructuringElement(cv::MORPH_RECT, cv::Size(3, 3)), cv::Point(-1,-1), 1, cv_border_type(_ctx.border_mode()), _ctx.constant_border_value());
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::dilate( Image &a, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::dilate( a.cv.img, c.cv.img, cv::getStructuringElement(cv::MORPH_RECT, cv::Size(3, 3)), cv::Point(-1,-1), 1, cv_border_type(_ctx.border_mode()), _ctx.constant_border_value());
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::gaussian3x3( Image &a, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::GaussianBlur(a.cv.img, c.cv.img, cv::Size(3, 3), 0, 0, cv_border_type(_ctx.border_mode()));
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::gaussian5x5( Image &a, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::GaussianBlur(a.cv.img, c.cv.img, cv::Size(5, 5), 0, 0, cv_border_type(_ctx.border_mode()));
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::box3x3( Image &a, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::boxFilter(a.cv.img, c.cv.img, -1, cv::Size(3, 3), cv::Point(-1,-1),true,cv_border_type(_ctx.border_mode()));
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::median3x3( Image &a, Image &c)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::medianBlur(a.cv.img, c.cv.img, 3); // Can't set borders
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::convolution( Image &a, Image &c, Convolution &matrix)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::filter2D(a.cv.img, c.cv.img, c.cv.img.type(), matrix.cv.convolution, cv::Point(-1, -1), 0, cv_border_type(_ctx.border_mode()) ); // Can't set border constant value
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::scale( Image &a, Image &c, InterpolationMode mode)
{
	const int interpolation = CVHelper::convert_to_cv_interpolation(mode);

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::resize(a.cv.img, c.cv.img, c.cv.img.size(), 0, 0, interpolation);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::laplacian_pyramid( Image &a, Pyramid &pyramid, Image &b, bool timers_enabled)
{
	ERROR("Not implemented");
}

void CVTarget::laplacian_reconstruct( Pyramid &pyramid, Image &a, Image &c)
{
	ERROR("Not implemented");
}

void CVTarget::gaussian_pyramid( Image &a, Pyramid &pyramid)
{
	pyramid.cv.pyramid[0]->img = a.cv.img; // Copy the input data to the first level of the pyramid

	if( SCALE_PYRAMID_HALF == pyramid.scale )
	{
		for(uint i = 0; i < _ctx.num_iterations() ; i++)
		{
			_timer.start();

			for(uint l = 1; l < pyramid.n_levels; ++l)
			{
				CVImage &in  = *pyramid.cv.pyramid[l - 1];
				CVImage &out = *pyramid.cv.pyramid[l];
				cv::pyrDown( in.img, out.img, cv::Size(), cv_border_type(_ctx.border_mode()));
			}
			_timer.stop();
			TestResults::report().add(name(), _timer.get_elapsed_ms() );
			_timer.print_elapsed_ms();
		}
	}
	else if( SCALE_PYRAMID_ORB == pyramid.scale )
	{
		for(uint i = 0; i < _ctx.num_iterations() ; i++)
		{
			_timer.start();

			for(uint l = 1; l < pyramid.n_levels; ++l)
			{
				CVImage &in  = *pyramid.cv.pyramid[l - 1];
				CVImage &out = *pyramid.cv.pyramid[l];
				cv::Mat tmp;
				cv::GaussianBlur(in.img, tmp, cv::Size(5, 5), 0, 0, cv_border_type(_ctx.border_mode()));
				cv::resize(tmp, out.img, out.img.size(), 0, 0, cv::INTER_NEAREST);
			}

			_timer.stop();
			TestResults::report().add(name(), _timer.get_elapsed_ms() );
			_timer.print_elapsed_ms();
		}
	}
	else
	{
		ERROR("Scale factor not supported for gaussian pyramid ");
	}
}

void CVTarget::canny(Image &a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image &c)
{
	ERROR_ON( norm_type != 1 && norm_type != 2);
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::Canny(a.cv.img,c.cv.img, lower, upper, gradient_size, norm_type == 2 );
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::color_convert( Image &a, Image &c )
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		switch(a.cv.format)
		{
			case Format::RGBA8888:
			{
				switch(c.cv.format)
				{
					case Format::RGB888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_RGBA2RGB);
						break;
					case Format::IYUV:
						cv::cvtColor(a.cv.img , c.cv.img, CV_RGBA2YUV_IYUV);
						break;
					case Format::UNKNOWN:
					default:
						ERROR("Not supported");
						break;

				}
				break;
			}
			case Format::UYVY:
			{
				switch(c.cv.format)
				{
					case Format::RGBA8888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_YUV2RGBA_UYVY);
						break;
					case Format::RGB888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_YUV2RGB_UYVY);
						break;
					case Format::UNKNOWN:
					default:
						ERROR("Not supported");
						break;
				}
				break;
			}
			case Format::YUYV422:
			{
				switch(c.cv.format)
				{
					case Format::RGBA8888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_YUV2RGBA_YUYV);
						break;
					case Format::RGB888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_YUV2RGB_YUYV);
						break;
					case Format::UNKNOWN:
					default:
						ERROR("Not supported");
						break;
				}
				break;
			}

			case Format::RGB888:
			{
				switch(c.cv.format)
				{
					case Format::RGBA8888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_RGB2RGBA);
						break;
					case Format::IYUV:
						cv::cvtColor(a.cv.img , c.cv.img, CV_RGB2YUV_IYUV);
						break;
					case Format::UNKNOWN:
					default:
						ERROR("Not supported");
						break;

				}
				break;
			}
			case Format::NV12:
			{
				switch(c.cv.format)
				{
					case Format::RGBA8888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_YUV2RGBA_NV12);
						break;
					case Format::RGB888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_YUV2RGB_NV12);
						break;
					default:
						ERROR("Not supported");
						break;
				}
				break;
			}
			case Format::NV21:
			{
				switch(c.cv.format)
				{
					case Format::RGBA8888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_YUV2RGBA_NV21);
						break;
					case Format::RGB888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_YUV2RGB_NV21);
						break;
					default:
						ERROR("Not supported");
						break;
				}
				break;
			}
			case Format::IYUV:
			{
				switch(c.cv.format)
				{
					case Format::RGBA8888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_YUV2RGBA_IYUV);
						break;
					case Format::RGB888:
						cv::cvtColor(a.cv.img , c.cv.img, CV_YUV2RGB_IYUV);
						break;
					default:
						ERROR("Not supported");
						break;
				}
				break;
			}
			case Format::UNKNOWN:
			default:
			{
				ERROR("Not supported");
				break;
			}
		}

		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::sobel( Image &a, Image *c, Image *d )
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		if(c)
		{
			cv::Sobel(a.cv.img, c->cv.img, CV_16S, 1, 0, 3, 1, 0, cv_border_type(_ctx.border_mode()));
		}

		if(d)
		{
			cv::Sobel(a.cv.img, d->cv.img, CV_16S, 0, 1, 3, 1, 0, cv_border_type(_ctx.border_mode()));
		}

		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::sobel5x5( Image &a, Image *c, Image *d )
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		if(c)
		{
			cv::Sobel(a.cv.img, c->cv.img, CV_16S, 1, 0, 5, 1, 0, cv_border_type(_ctx.border_mode()));
		}

		if(d)
		{
			cv::Sobel(a.cv.img, d->cv.img, CV_16S, 0, 1, 5, 1, 0, cv_border_type(_ctx.border_mode()));
		}

		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::sobel7x7( Image &a, Image *c, Image *d )
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		if(c)
		{
			cv::Sobel(a.cv.img, c->cv.img, CV_16S, 1, 0, 7, 1, 0, cv_border_type(_ctx.border_mode()));
		}
		if(d)
		{
			cv::Sobel(a.cv.img, d->cv.img, CV_16S, 0, 1, 7, 1, 0, cv_border_type(_ctx.border_mode()));
		}
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CVTarget::scharr( Image &a, Image *c, Image *d )
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		if(c)
		{
			cv::Scharr(a.cv.img, c->cv.img, CV_16S, 1, 0, 1, 0, cv_border_type(_ctx.border_mode()));
		}

		if(d)
		{
			cv::Scharr(a.cv.img, d->cv.img, CV_16S, 0, 1, 1, 0, cv_border_type(_ctx.border_mode()));
		}

		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::derivative( Image &a, Image *c, Image *d )
{
	float dkernel[] = {-1, 0, 1};

	cv::Mat kernel_x = cv::Mat(1, 3, CV_32FC1, dkernel);
	cv::Mat kernel_y = cv::Mat(3, 1, CV_32FC1, dkernel);

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		if(c)
		{
			cv::filter2D(a.cv.img, c->cv.img, c->cv.img.type(), kernel_x, cv::Point(-1,-1), 0, cv_border_type(_ctx.border_mode()) ); // Can't set border constant value
		}

		if(d)
		{
			cv::filter2D(a.cv.img, d->cv.img, d->cv.img.type(), kernel_y, cv::Point(-1,-1), 0, cv_border_type(_ctx.border_mode()) ); // Can't set border constant value		}
		}

		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::magnitude( Image &a, Image &b, Image &c)
{
	a.cv.img.convertTo(a.cv.img, CV_32F);
	b.cv.img.convertTo(b.cv.img, CV_32F);
	c.cv.img.convertTo(c.cv.img, CV_32F);

	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		_timer.start();
		cv::magnitude( a.cv.img, b.cv.img, c.cv.img);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}

	a.cv.img.convertTo(a.cv.img, CV_16S);
	b.cv.img.convertTo(b.cv.img, CV_16S);
	c.cv.img.convertTo(c.cv.img, CV_16S);
}

void CVTarget::phase( Image &a, Image &b, Image &c)
{
	ERROR("OpenCV only supports float phase");
}

void CVTarget::remap(Image &a, RemapTable &table, Image &d, InterpolationMode policy )
{
	const int interpolation = CVHelper::convert_to_cv_interpolation(policy);

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::remap(a.cv.img, d.cv.img, table.cv.table[0]->img, table.cv.table[1]->img, interpolation, cv_border_type(_ctx.border_mode()),_ctx.constant_border_value());
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

static unsigned int channel_index(Channel channel)
{
	unsigned int index = 0;
	switch( channel )
	{
		case Channel::C0:
		case Channel::R:
		case Channel::Y:
			index = 0;
			break;
		case Channel::C1:
		case Channel::G:
		case Channel::U:
			index = 1;
			break;
		case Channel::C2:
		case Channel::B:
		case Channel::V:
			index = 2;
			break;
		case Channel::C3:
		case Channel::A:
			index = 3;
			break;
		default:
			ERROR("Unsupported channel.");
			break;
	}
	return index;
}

void CVTarget::channel_extract( Image &a, Channel channel, Image &c )
{
	cv::Mat channels[4];
	unsigned int index = channel_index(channel);
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::split(a.cv.img, channels);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
		c.cv.img = channels[index];
	}
}

void CVTarget::mean_stddev(Image& a, Scalar<float>& mean, Scalar<float>& stddev)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::meanStdDev( a.cv.img, mean.cv.scalar, stddev.cv.scalar );
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::tablelookup( Image &a , Lut &b, Image &c )
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::LUT(a.cv.img, b.cv.lut, c.cv.img);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::convert_depth( Image &a, Image &c, ConvertPolicy policy , uint32_t shift)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		switch(a.cv.format)
		{
			case Format::LUMINANCE:
			{
				switch(c.cv.format)
				{
					case Format::LUMINANCE16:
					{
						a.cv.img.convertTo( c.cv.img , CV_16S, pow(2,shift));
						break;
					}
					case Format::LUMINANCE32:
					{
						a.cv.img.convertTo( c.cv.img , CV_32S, pow(2,shift));
						break;
					}
					default:
					{
						ERROR("Format not supported");
						break;
					}
				}
				break;
			}
			case Format::LUMINANCE16:
			{
				switch(c.cv.format)
				{
					case Format::LUMINANCE32:
					{
						a.cv.img.convertTo( c.cv.img , CV_32S, pow(2,shift));
						break;
					}
					default:
					{
						ERROR("Format not supported");
						break;
					}
				}
				break;
			}
			default:
			{
				ERROR("Format not supported");
				break;
			}
		}

		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CVTarget::minmaxlocation( Image &a, Scalar<int32_t> &min, Scalar<int32_t> &max, ICoord2DArray *min_loc, ICoord2DArray *max_loc,
                               Scalar<uint32_t> *min_count, Scalar<uint32_t> *max_count)
{
	double cv_min = 0;
	double cv_max = 0;

	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		if ( (nullptr == min_loc) || (nullptr == max_loc) || (nullptr == min_count) || (nullptr == max_count) )
		{
			_timer.start();
			cv::minMaxLoc( a.cv.img, &cv_min, &cv_max );
			_timer.stop();

			min.cv.scalar = cv_min;
			max.cv.scalar = cv_max;

			TestResults::report().add(name(), _timer.get_elapsed_ms() );
			_timer.print_elapsed_ms();
		}
	}
}

void CVTarget::histogram(Image& a, Distribution& b)
{
	float ranges[] = { static_cast<float>(b.cv.h_offset), static_cast<float>(b.cv.h_offset) + b.cv.h_range };
	const float* hist_ranges = { ranges };
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::calcHist(&a.cv.img, 1, 0, cv::Mat(), b.cv.hist, 1, &b.cv.h_size, &hist_ranges);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::channel_combine(Image &plane0, Image &plane1, Image &plane2, Image &plane3, Image &out)
{
	std::vector<cv::Mat> channels;
	channels.push_back(plane0.cv.img);
	channels.push_back(plane1.cv.img);
	channels.push_back(plane2.cv.img);
	if(Format::RGBA8888 == out.cv.format)
	    channels.push_back(plane3.cv.img);


	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::merge(channels, out.cv.img);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::optical_flow( Image &a, Image &b, IKeyPointArray *old_points, IKeyPointArray *new_points_estimates,
                             IKeyPointArray *new_points, Termination termination, float epsilon, uint32_t num_iterations,
                             bool use_initial_estimate, size_t window_dimension, size_t num_levels)
{

	cv::TermCriteria termcrit(cv::TermCriteria::COUNT | cv::TermCriteria::EPS, num_iterations, epsilon);
	cv::Size WinSize(window_dimension, window_dimension);

	std::vector<uchar> status;
	std::vector<float> err;

	std::vector<cv::Mat> old_pyramid;
	std::vector<cv::Mat> new_pyramid;

	std::vector<cv::Point2f> prev_points;
	std::vector<cv::Point2f> next_points;

	new_points->cv.array.resize(old_points->cv.array.size());

	for(unsigned int i = 0; i < old_points->cv.array.size(); i++)
	{
		const cv::Point2f point(old_points->cv.array[i].pt.x, old_points->cv.array[i].pt.y);
		prev_points.push_back(point);
		next_points.push_back(point);
	}

	int flag = 0;
	if(use_initial_estimate)
	{
		flag = cv::OPTFLOW_USE_INITIAL_FLOW;
	}

	buildOpticalFlowPyramid(a.cv.img, old_pyramid, WinSize, num_levels, 0, cv_border_type(_ctx.border_mode()),
	                        cv_border_type(_ctx.border_mode()), use_initial_estimate);
	buildOpticalFlowPyramid(b.cv.img, new_pyramid, WinSize, num_levels, 0, cv_border_type(_ctx.border_mode()),
	                        cv_border_type(_ctx.border_mode()), use_initial_estimate);

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		calcOpticalFlowPyrLK(old_pyramid, new_pyramid, prev_points, next_points, status, err,
		                     WinSize, num_levels, termcrit, flag | cv::OPTFLOW_LK_GET_MIN_EIGENVALS, 0.0001);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}

	for(unsigned int i = 0; i < new_points->cv.array.size(); i++)
	{
		new_points->cv.array[i].pt.x = roundf(next_points[i].x);
		new_points->cv.array[i].pt.y = roundf(next_points[i].y);
	}
}

void CVTarget::integral(Image &a, Image &b)
{
	cv::Mat tmp;

	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		_timer.start();
		cv::integral(a.cv.img, tmp);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}

	cv::Rect roi(1, 1, a.cv.width, a.cv.height);
	b.cv.img = tmp(roi).clone();
}

void CVTarget::equalize_histogram(Image& a, Image& b)
{
	for( uint i = 0; i < _ctx.num_iterations(); ++i )
	{
		_timer.start();
		cv::equalizeHist(a.cv.img, b.cv.img);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CVTarget::non_linear_filter(Image &a, Image &b, NonLinearFilterFunction function, MatrixPattern pattern, uint mask_size, uint8_t* mask)
{
	if(MatrixPattern::BOX == pattern)
	{
		for(uint i = 0; i < _ctx.num_iterations(); i++)
		{
			_timer.start();
			switch(function)
			{
			case NonLinearFilterFunction::MEDIAN:
				cv::medianBlur(a.cv.img, b.cv.img, mask_size); // Can't set borders
				break;
			case NonLinearFilterFunction::MIN:
				cv::erode(a.cv.img, b.cv.img, cv::getStructuringElement(cv::MORPH_RECT, cv::Size(mask_size, mask_size)), cv::Point(-1,-1), 1, cv_border_type(_ctx.border_mode()), _ctx.constant_border_value());
				break;
			case NonLinearFilterFunction::MAX:
				cv::dilate(a.cv.img, b.cv.img, cv::getStructuringElement(cv::MORPH_RECT, cv::Size(mask_size, mask_size)), cv::Point(-1,-1), 1, cv_border_type(_ctx.border_mode()), _ctx.constant_border_value());
				break;
			default:
				ERROR_ON("Operation not supported");
				break;
			}
			_timer.stop();
			TestResults::report().add(name(), _timer.get_elapsed_ms() );
			_timer.print_elapsed_ms();
		}
	}
	else
	{
		ERROR_ON("NOT IMPLEMENTED");
	}
}

void CVTarget::hog_descriptor(Image &a, HOG &b, Tensor &c)
{
	std::vector<cv::Point> locations;
	std::vector<float> descriptors;

	cv::HOGDescriptor hog( cv::Size(a.cv.width, a.cv.height), b.cv.hog.blockSize, b.cv.hog.blockStride, b.cv.hog.cellSize, b.cv.hog.nbins );

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		hog.compute( a.cv.img, descriptors, cv::Size(0,0), cv::Size(0,0), locations );
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
}

void CVTarget::activation_layer(Tensor &in, Tensor &out, ActivationLayerInfo act_info)
{
	_eigen.activation_layer(in.cv, out.cv, act_info);
}

void CVTarget::normalization_layer(Tensor &a, Tensor &b, NormalizationLayerInfo norm_info)
{
	ERROR_ON("NOT IMPLEMENTED");
}

void CVTarget::batchnormalization_layer(Tensor &a, Tensor &b, Tensor &mean, Tensor &var, Tensor &beta, Tensor &gamma, float epsilon)
{
	_eigen.batchnormalization_layer( a.cv, b.cv, mean.cv, var.cv, beta.cv, gamma.cv, epsilon);
}

void CVTarget::pooling_layer(Tensor &in, Tensor &out, PoolingLayerInfo pool_info)
{
	_eigen.pooling_layer( in.cv, out.cv, pool_info);
}

void CVTarget::softmax_layer(Tensor &in, Tensor &out)
{
	_eigen.softmax_layer( in.cv, out.cv );
}

void CVTarget::tensor_ops(Tensor &in1, Tensor &in2, Tensor &out, TensorOp op)
{
	_eigen.tensor_ops( in1.cv, in2.cv, out.cv, op);
}

void CVTarget::convolution_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info, Tensor *rw)
{
    UNUSED(rw);
	_eigen.convolution_layer( in.cv, w.cv, b ? &b->cv : nullptr, out.cv, conv_info);
}

void CVTarget::locally_connected_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info)
{
	_eigen.locally_connected_layer( in.cv, w.cv, b ? &b->cv : nullptr, out.cv, conv_info);
}

void CVTarget::fully_connected_layer(Tensor &input, Tensor &weight, Tensor* bias, Tensor &output, bool are_weights_reshaped)
{
	_eigen.fully_connected_layer( input.cv, weight.cv, bias ? &bias->cv : nullptr, output.cv);
}

void CVTarget::transpose( Image &a, Image &b)
{
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		cv::transpose(a.cv.img, b.cv.img);
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void CVTarget::gemm_fp32( Tensor &a, Tensor &b, Tensor &c, Tensor &d, float alpha, float beta)
{
	_eigen.gemm(a.cv, b.cv, c.cv, d.cv, alpha, beta);
}
