#include "mali_cv/EigenFunctions.h"

#include "mali_cv/Error.h"
#include "TestResults.h"

using namespace mali_cv;

EigenFunctions::EigenFunctions( Context &ctx, const std::string& target_name)
#ifdef USE_EIGEN
	:_ctx(ctx),_name(target_name),_timer(target_name)
#endif /* USE_EIGEN */
{
}

void EigenFunctions::activation_layer(EigenTensor &in, EigenTensor &out, ActivationLayerInfo act_info)
{
#ifdef USE_EIGEN
	ActivationFunction func = act_info.activation();
	uint32_t a = act_info.a();
	uint32_t b = act_info.b();

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		switch(func)
		{
			case ActivationFunction::LINEAR:
				out.tensor = static_cast<float>(a) * in.tensor + static_cast<float>(b);
				break;
			case ActivationFunction::LOGISTIC:
				out.tensor = 1.f / (1.f + (-in.tensor).exp());
				break;
			case ActivationFunction::RELU:
				out.tensor = in.tensor.cwiseMax(0.f);
				break;
			case ActivationFunction::BOUNDED_RELU:
				out.tensor = in.tensor.cwiseMax(0.f).cwiseMin(static_cast<float>(a));
				break;
			case ActivationFunction::SOFT_RELU:
				out.tensor = (1.f + in.tensor.exp()).log();
				break;
			case ActivationFunction::ABS:
				out.tensor = in.tensor.abs();
				break;
			case ActivationFunction::SQRT:
				out.tensor = in.tensor.sqrt();
				break;
			case ActivationFunction::SQUARE:
				out.tensor = in.tensor.square();
				break;
			case ActivationFunction::TANH:
				out.tensor = static_cast<float>(a) * (static_cast<float>(b) * in.tensor).tanh();
				break;
			default:
				ERROR("Unsupported activation function");
		}
		_timer.stop();
		TestResults::report().add(_name, _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
#else
	ERROR("Not implemented");
#endif
}

void EigenFunctions::batchnormalization_layer(EigenTensor &a, EigenTensor &b, EigenTensor &mean, EigenTensor &var, EigenTensor &beta, EigenTensor &gamma, float epsilon)
{
#ifdef USE_EIGEN
	auto dims_a = a.tensor.dimensions();
	int fm = a.tensor.dimension(dims_a.size() - 3);
	std::array<int, 6> shuffle_in{ { 0, 1, 2, 5, 4, 3} };
	Eigen::Tensor<float, 6, Eigen::RowMajor> tmp_a = a.tensor.shuffle(shuffle_in);
	auto dims_tmp = tmp_a.dimensions();

	const int depth = mean.tensor.dimension(5);
	const int rest_size = std::accumulate(dims_a.begin(), dims_a.end(), 1, std::multiplies<int>()) / fm;
	Eigen::DSizes<int, 2> rest_by_depth(rest_size, depth);
	Eigen::DSizes<int, 2> rest_by_one(rest_size, 1);
	Eigen::DSizes<int, 2> one_by_depth(1, depth);
	Eigen::DSizes<int, 2> depth_by_one(depth, 1);
	Eigen::Tensor<float, 2, Eigen::RowMajor> tmp_b;

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		tmp_b = (tmp_a.reshape(rest_by_depth) -
				mean.tensor.reshape(one_by_depth).broadcast(rest_by_one)) *
				((var.tensor + var.tensor.constant(epsilon)).rsqrt() * gamma.tensor)
				.eval()
				.reshape(one_by_depth)
				.broadcast(rest_by_one) +
				beta.tensor.reshape(one_by_depth).broadcast(rest_by_one);

		_timer.stop();
		TestResults::report().add(_name, _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}

	// Reshape back output tensor
	b.tensor = tmp_b.reshape(dims_tmp).shuffle(shuffle_in);
#else
	ERROR("Not implemented");
#endif
}
void EigenFunctions::pooling_layer(EigenTensor &in, EigenTensor &out, PoolingLayerInfo pool_info)
{
#ifdef USE_EIGEN
	unsigned int pool_pad_x, pool_pad_y, pool_stride_x, pool_stride_y = 0;
	std::tie(pool_pad_x, pool_pad_y) = pool_info.pad_stride_info().pad();
	std::tie(pool_stride_x, pool_stride_y) = pool_info.pad_stride_info().stride();
	unsigned int pool_size = pool_info.pool_size();
	PoolingType pool_type = pool_info.pool_type();

	// Define pad_type
	Eigen::PaddingType pad_type = Eigen::PADDING_VALID;

	// Pad input
	Eigen::array<std::pair<int, int>, 6> paddings;
	paddings[0] = paddings[1] = paddings[2] = paddings[3] = std::make_pair(0, 0);
	if(pool_pad_x > 0 || pool_pad_y > 0)
	{
	    paddings[4] = std::make_pair(pool_pad_x, pool_pad_x);
	    paddings[5] = std::make_pair(pool_pad_y, pool_pad_y);
	}
	Eigen::Tensor<float, 6, Eigen::RowMajor> padded_in = in.tensor.pad(paddings);

	// Clash upper dimensions and reshape Tensor to 3D Tensor
	auto dims_in = padded_in.dimensions();
	int cols_in = padded_in.dimension(dims_in.size() - 1);
	int rows_in = padded_in.dimension(dims_in.size() - 2);
	int upper_dims_in = std::accumulate(dims_in.begin(), dims_in.end(), 1, std::multiplies<int>()) / (rows_in * cols_in);
	std::array<int, 4> dim4d_in{ { upper_dims_in, rows_in, cols_in, 1 } };

	// Reshape input and output tensor
	Eigen::Tensor<float, 4, Eigen::RowMajor> tmp_in = padded_in.reshape(dim4d_in);

	Eigen::Tensor<float, 4, Eigen::RowMajor> tmp_out;
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		// Perform pooling
		if(PoolingType::MAX == pool_type)
		{
			tmp_out = Eigen::SpatialMaxPooling(tmp_in, pool_size, pool_size, pool_stride_x, pool_stride_y, pad_type);
		}
		else if(PoolingType::AVG == pool_type)
		{
			tmp_out = Eigen::SpatialAvgPooling(tmp_in, pool_size, pool_size, pool_stride_x, pool_stride_y, pad_type);
		}
		else
		{
			ERROR("Not implemented");
		}
		_timer.stop();
		TestResults::report().add(_name, _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}

	// Reshape back (Get rid of channels as they are fixed to 1)
	auto dims_out = dims_in;
	dims_out[dims_out.size() - 1] = tmp_out.dimensions()[2];
	dims_out[dims_out.size() - 2] = tmp_out.dimensions()[1];
	out.tensor = tmp_out.reshape(dims_out);
#else
	ERROR("Not implemented");
#endif
}

void EigenFunctions::softmax_layer(EigenTensor &in, EigenTensor &out)
{
#ifdef USE_EIGEN
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		// Clash upper dimensions and reshape Tensor to 2D Tensor
		auto dims = in.tensor.dimensions();
		int upper_dims = std::accumulate(dims.begin(), dims.end(), 1, std::multiplies<int>()) /
			in.tensor.dimension(in.tensor.dimensions().size() - 1);
		Eigen::DSizes<int, 2> dim2d(upper_dims, (in.tensor.dimension(in.tensor.dimensions().size() - 1)));
		Eigen::Tensor<float, 2, Eigen::RowMajor> tmp = in.tensor.reshape(dim2d);

		// Set up broadcasting and reduction dimensions
		Eigen::DSizes<int, 1> along_class(1);
		Eigen::DSizes<int, 2> batch_by_one(tmp.dimensions()[0], 1);
		Eigen::DSizes<int, 2> one_by_class(1, tmp.dimensions()[1]);

		// Perform Softmax
		// shifted_logits = logits - max(logits along classes);
		tmp = (tmp - tmp.maximum(along_class).eval().reshape(batch_by_one).broadcast(one_by_class)).exp();
		// softmax = softmax * (1 / sum(softmax along classes));
		tmp = (tmp * tmp.sum(along_class).inverse().eval().reshape(batch_by_one).broadcast(one_by_class));

		// Reshape back
		out.tensor = tmp.reshape(in.tensor.dimensions());
		_timer.stop();
		TestResults::report().add(_name, _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
#else
	ERROR("Not implemented");
#endif
}

void EigenFunctions::tensor_ops(EigenTensor &in1, EigenTensor &in2, EigenTensor &out, TensorOp op)
{
#ifdef USE_EIGEN
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		switch(op)
		{
			case TensorOp::ADD:
				out.tensor = in1.tensor + in2.tensor;
				break;
			case TensorOp::SUB:
				out.tensor = in1.tensor - in2.tensor;
				break;
			case TensorOp::MUL:
				out.tensor = in1.tensor * in2.tensor;
				break;
			case TensorOp::LUT:
			case TensorOp::TRANSPOSE:
			default:
				ERROR("Unsupported tensor operation");
		}
		_timer.stop();
		TestResults::report().add(_name, _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
#else
	ERROR("Not implemented");
#endif
}

void EigenFunctions::convolution_layer(EigenTensor &in, EigenTensor &w, EigenTensor *b, EigenTensor &out, PadStrideInfo conv_info)
{
#ifdef USE_EIGEN
	unsigned int pad_x, pad_y, stride_x, stride_y = 0;
	std::tie(pad_x, pad_y) = conv_info.pad();
	std::tie(stride_x, stride_y) = conv_info.stride();

	// Define pad_type
	Eigen::PaddingType pad_type = Eigen::PADDING_VALID;

	// Pad input
	Eigen::array<std::pair<int, int>, 6> paddings;
	paddings[0] = paddings[1] = paddings[2] = paddings[3] = std::make_pair(0, 0);
	if(pad_x || pad_y)
	{
		paddings[4] = std::make_pair(pad_x, pad_x);
		paddings[5] = std::make_pair(pad_y, pad_y);
	}
	Eigen::Tensor<float, 6, Eigen::RowMajor> padded_in = in.tensor.pad(paddings);

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		// Reshape kernels tensor to 4D (rows, cols, fmi, fmo)
		auto dims_k = w.tensor.dimensions();
		int cols_k = w.tensor.dimension(dims_k.size() - 1);
		int rows_k = w.tensor.dimension(dims_k.size() - 2);
		int fmi_k = w.tensor.dimension(dims_k.size() - 3);
		int fmo_k = w.tensor.dimension(dims_k.size() - 4);
		std::array<int, 4> dim4d_k{ { fmo_k, fmi_k, rows_k, cols_k } };
		std::array<int, 4> shuffle_k{ { 3, 2, 1, 0 } };
		Eigen::Tensor<float, 4, Eigen::RowMajor> tmp_k = w.tensor.reshape(dim4d_k).shuffle(shuffle_k);

		// Reshape input tensor match the expected spatial convolution shape(rows, cols, fmi, batch)
		auto dims_in = padded_in.dimensions();
		int cols_in = padded_in.dimension(dims_in.size() - 1);
		int rows_in = padded_in.dimension(dims_in.size() - 2);
		int fm_in = padded_in.dimension(dims_in.size() - 3);
		int upper_dims_in = std::accumulate(dims_in.begin(), dims_in.end(), 1, std::multiplies<int>()) / (rows_in * cols_in * fm_in);
		std::array<int, 4> dim4d_in{ { upper_dims_in, fm_in, rows_in, cols_in } };
		std::array<int, 4> shuffle_in{ { 0, 3, 2, 1 } };
		Eigen::Tensor<float, 4, Eigen::RowMajor> tmp_in = padded_in.reshape(dim4d_in).shuffle(shuffle_in);

		// Perform spatial convolution
		auto tmp_out = Eigen::SpatialConvolution(tmp_in, tmp_k, stride_y, stride_x, pad_type);

		// Reshape back output tensor (rows, cols, fmo, batch)
		std::array<int, 4> shuffle_out{ { 0, 3, 2, 1 } };
		Eigen::Tensor<float, 4, Eigen::RowMajor> tmp_out_final = tmp_out.shuffle(shuffle_out);

		// Reshape bias and add
		if(b != nullptr)
		{
			int bias_size = b->tensor.dimension(b->tensor.dimensions().size() - 1);
			int bias_broadcast_col = tmp_out_final.dimensions()[3];
			int bias_broadcast_row = tmp_out_final.dimensions()[2];
			std::array<int, 4> dim4d_b{ {1, bias_size, 1, 1 } };
			Eigen::Tensor<float, 4, Eigen::RowMajor> tmp_b = b->tensor.reshape(dim4d_b);
			Eigen::array<int, 4> bcast({ {upper_dims_in, 1, bias_broadcast_row, bias_broadcast_col} });
			Eigen::Tensor<float, 4, Eigen::RowMajor> tmp_bb = tmp_b.broadcast(bcast);
			tmp_out_final += tmp_bb;
		}

		// Reshape to final output
		auto dims_out = dims_in;
		dims_out[dims_out.size() - 1] = tmp_out_final.dimensions()[3];
		dims_out[dims_out.size() - 2] = tmp_out_final.dimensions()[2];
		dims_out[dims_out.size() - 3] = tmp_out_final.dimensions()[1];
		out.tensor = tmp_out_final.reshape(dims_out);
		_timer.stop();
		TestResults::report().add(_name, _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}
#else
	ERROR("Not implemented");
#endif
}

void EigenFunctions::locally_connected_layer(EigenTensor &in, EigenTensor &k, EigenTensor *bias, EigenTensor &out, PadStrideInfo conv_info)
{

#ifdef USE_EIGEN
	const auto &dims_in = in.tensor.dimensions();
	const auto &dims_k = k.tensor.dimensions();
	const auto &dims_out = out.tensor.dimensions();

	const int in_w = dims_in[dims_in.size() - 1];
	const int in_h = dims_in[dims_in.size() - 2];
	const int in_d = dims_in[dims_in.size() - 3];

	const int k_w = dims_k[dims_k.size() - 1];
	const int k_h = dims_k[dims_k.size() - 2];
	const int k_d = dims_k[dims_k.size() - 3];
	const int k_a = dims_k[dims_k.size() - 4];
	const int k_n = dims_k[dims_k.size() - 5];
	UNUSED(k_h);

	const int out_w = dims_out[dims_out.size() - 1];
	const int out_h = dims_out[dims_out.size() - 2];
	const int out_d = dims_out[dims_out.size() - 3];

	// reshape output as a 4D tensor (batches,d,h,w)
	constexpr int OUTPUT_RANK = 4;
	const int out_num_batches = std::accumulate(dims_out.begin(), dims_out.end(), 1, std::multiplies<int>()) / (out_w*out_h*out_d);
	const std::array<int,OUTPUT_RANK> dim4d_out {{out_num_batches,out_d,out_h,out_w}};
	Eigen::Tensor<float, OUTPUT_RANK, Eigen::RowMajor> out_4d = out.tensor.reshape(dim4d_out);

	ERROR_ON_MSG(dims_in.size() > 6, "Not implemented");
	ERROR_ON_MSG(dims_out.size() > 6, "Not implemented");

	// reshape input as a 4D tensor (batches,d,h,w)
	constexpr int INPUT_RANK = 4;
	const int in_num_batches = std::accumulate(dims_in.begin(), dims_in.end(), 1, std::multiplies<int>()) / (in_w*in_h*in_d);
	const std::array<int,INPUT_RANK> dim4d_in {{in_num_batches,in_d,in_h,in_w}};
	const Eigen::Tensor<float, INPUT_RANK, Eigen::RowMajor> in_4d = in.tensor.reshape(dim4d_in);

	// reshape kernel as a 5D tensor (num_patches, num_kernels,d,h,w)
	constexpr int KERNEL_RANK = 5;
	const Eigen::array<int, KERNEL_RANK> kernels_five_dim{ {k_n, k_a,k_d,k_h,k_w} };
	const Eigen::Tensor<float, KERNEL_RANK, Eigen::RowMajor> k_5d = k.tensor.reshape(kernels_five_dim);

	ERROR_ON_MSG(k_w != k_w, "Only square convolutions are supported");
	ERROR_ON_MSG(k_d != in_d, "Each kernel must have the same deph as the input volume");
	ERROR_ON_MSG(out_d != k_a, "Output deph must have the same as the number of kernels");
	ERROR_ON_MSG(in_num_batches != out_num_batches, "Number of batches must be the same in input and output tensors");


	// Adds padding to the input tensor: this is only needed if the kernel is > 1
	int pad_x = 0;
	int pad_y = 0;
	int stride_x = 0;
        int stride_y = 0;

	std::tie(pad_x, pad_y) = conv_info.pad();

	std::tie(stride_x, stride_y) = conv_info.stride();

	Eigen::array<std::pair<int, int>, INPUT_RANK> paddings;
	paddings[2] = std::make_pair(pad_y, pad_y);
	paddings[3] = std::make_pair(pad_x, pad_x);
	paddings[1] = paddings[0] =  std::make_pair(0,0);
	Eigen::Tensor<float, INPUT_RANK, Eigen::RowMajor> padded_in = in_4d.pad(paddings);
	Eigen::Tensor<float, 2, Eigen::RowMajor> tmp_b;

		// Reshape and add bias
		if(bias != nullptr)
		{
			int bias_size = bias->tensor.dimension(bias->tensor.dimensions().size() - 1);
			int bias_num = bias->tensor.dimension(bias->tensor.dimensions().size() - 2);
			std::array<int, 2> dim2d_b{ {bias_num, bias_size } };
			tmp_b = bias->tensor.reshape(dim2d_b);
		}

	for(uint i = 0; i < _ctx.num_iterations(); ++i)
	{
		_timer.start();
		for(int b = 0; b < out_num_batches; ++b) // iterate over each batch
		{
			for(int d = 0; d < k_a; ++d) // iterate over each OFM
			{
				// output volume's 'plane d' is be computed using kernel number 'd' in the kernels tensor.
				// output's 3rd dimension is the same as kernel's 3rd dimension.
				int oh, ih = 0;
				for(oh = 0, ih = -pad_y; oh < out_h; ++oh, ih+=stride_y) // iterate each patch in y dimension
				{
					int ow, iw =0;
					for(ow = 0, iw = -pad_x; ow < out_w; ++ow, iw+=stride_x) // iterate each patch in x dimension
					{
						float result = 0.f;
						for(int p = 0; p < in_d; ++p) // iterate kernel depth
						{
							for( int kh = 0; kh < k_h; ++kh) // iterate kernel height
							for( int kw = 0; kw < k_h; ++kw) // iterate kernel width
							{
								float padd_val = padded_in(b,p, ih + kh, iw + kw);
								float k_val = k_5d(ow + oh * out_w, d, p, kh, kw);
								float val_d = padd_val * k_val ;
								result += val_d;
							}
						}
						out_4d(b,d,oh,ow) = result;
						if(bias != nullptr)
						{
							out_4d(b,d,oh,ow) =  out_4d(b,d,oh,ow) + tmp_b(ow + oh * out_w, d);
						}
					}
				}
			}
		}

		// Reshape to final output
		out.tensor = out_4d.reshape(dims_out);
		_timer.stop();
		TestResults::report().add(_name, _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
#else
	ERROR("Not implemented");
#endif
}

void EigenFunctions::fully_connected_layer(EigenTensor &input, EigenTensor &weight, EigenTensor* bias, EigenTensor &output)
{
#ifdef USE_EIGEN
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();

		// Transpose weights
		auto dims_weight = weight.tensor.dimensions();
		int cols_weights = weight.tensor.dimension(dims_weight.size() - 1);
		int num_weights = weight.tensor.dimension(dims_weight.size() - 2);
		std::array<int, 2> dim2d_weight{ { num_weights, cols_weights } };
		Eigen::Tensor<float, 2, Eigen::RowMajor> transposed_weight;
		std::array<int, 2> shuffle_weights{ { 1, 0 } };
		transposed_weight = weight.tensor.reshape(dim2d_weight).shuffle(shuffle_weights);

		// Reshape input to 2D
		auto dims_in = input.tensor.dimensions();
		int input_w = input.tensor.dimension(dims_in.size() - 1);
		int input_h = input.tensor.dimension(dims_in.size() - 2);
		int input_d = input.tensor.dimension(dims_in.size() - 3);
		int cols_in, rows_in = 0;

		if(cols_weights == (input_w * input_h * input_d)) // Comes from CNN
		{
			cols_in = input_w * input_h * input_d;
		}
		else if(cols_weights == input_w) // Legacy FC
		{
			cols_in = input_w;
		}
		else
		{
			ERROR("Invalid input dimensions");
		}
		rows_in = std::accumulate(dims_in.begin(), dims_in.end(), 1, std::multiplies<int>()) / cols_in;
		std::array<int, 2> dim2d_in{ { rows_in, cols_in } };
		Eigen::Tensor<float, 2, Eigen::RowMajor> tmp_in = input.tensor.reshape(dim2d_in);

		// Perform contraction
		Eigen::array<Eigen::IndexPair<int>, 1> product_dims = {{Eigen::IndexPair<int>(1, 0)}};
		Eigen::Tensor<float, 2, Eigen::RowMajor> tmp_out = tmp_in.contract(transposed_weight, product_dims);

		// Reshape and add bias
		if(bias != nullptr)
		{
			int bias_size = bias->tensor.dimension(bias->tensor.dimensions().size() - 1);
			int bias_broadcast_row = rows_in;
			std::array<int, 2> dim2d_b{{1, bias_size}};
			Eigen::Tensor<float, 2, Eigen::RowMajor> tmp_b = bias->tensor.reshape(dim2d_b);
			Eigen::array<int, 2> bcast({{bias_broadcast_row, 1}});
			Eigen::Tensor<float, 2, Eigen::RowMajor> tmp_bb = tmp_b.broadcast(bcast);
			tmp_out += tmp_bb;
		}

		// Reshape to final output
		auto dims_out = output.tensor.dimensions();
		output.tensor = tmp_out.reshape(dims_out);

		_timer.stop();
		TestResults::report().add(_name, _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
#else
	ERROR("Not implemented");
#endif
}

void EigenFunctions::gemm(EigenTensor &a, EigenTensor &b, EigenTensor &c, EigenTensor &out, float alpha, float beta)
{
#ifdef USE_EIGEN
	// Get tensors dimensions
	auto dims_a = a.tensor.dimensions();
	auto dims_b = b.tensor.dimensions();

	int M = a.tensor.dimension(dims_a.size() - 2);
	int K = a.tensor.dimension(dims_b.size() - 1);
	int N = b.tensor.dimension(dims_b.size() - 1);

	Eigen::MatrixXf mtx_a(M, K);
	Eigen::MatrixXf mtx_b(K, N);
	Eigen::MatrixXf mtx_c(M, N);
	Eigen::MatrixXf mtx_out(M, N);

	// Copy values of Matrix A
	for(int row = 0; row < M; ++row)
	{
		for(int col = 0; col < K; ++col)
		{
			mtx_a(row, col) = a.tensor(0, 0, 0, 0, row, col);
		}
	}

	// Copy values of Matrix B
	for(int row = 0; row < K; ++row)
	{
		for(int col = 0; col < N; ++col)
		{
			mtx_b(row, col) = b.tensor(0, 0, 0, 0, row, col);
		}
	}

	// Copy values of Matrix C
	for(int row = 0; row < M; ++row)
	{
		for(int col = 0; col < N; ++col)
		{
			mtx_c(row, col) = c.tensor(0, 0, 0, 0, row, col);
		}
	}

	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		mtx_out = alpha * (mtx_a * mtx_b) + beta * mtx_c;
		_timer.stop();

		TestResults::report().add(_name, _timer.get_elapsed_ms() );
		_timer.print_elapsed_ms();
	}

	// Copy values to output tensor
	for(int row = 0; row < M; ++row)
	{
		for(int col = 0; col < N; ++col)
		{
			out.tensor(0, 0, 0, 0, row, col) = mtx_out(row, col);
		}
	}
#else
	ERROR("Not implemented");
#endif
}

void EigenFunctions::concatenate_depth(std::vector<EigenTensor*>& a,  EigenTensor &b)
{
#ifdef USE_EIGEN
	const unsigned int vector_size = a.size();

	std::vector<Eigen::Tensor<float, MAX_DIMS, Eigen::RowMajor>> in_vector(vector_size);

	const unsigned int x_diff = b.tensor.dimension( MAX_DIMS - 1) - a.at(0)->tensor.dimension(MAX_DIMS - 1);
	const unsigned int y_diff = b.tensor.dimension( MAX_DIMS - 2) - a.at(0)->tensor.dimension(MAX_DIMS - 2);

	Eigen::array<std::pair<int, int>, 6> paddings;
	paddings[0] = paddings[1] = paddings[2] = paddings[3] = std::make_pair(0, 0);
	paddings[4] = std::make_pair(x_diff/2, x_diff/2);
	paddings[5] = std::make_pair(y_diff/2, y_diff/2);

	Eigen::Tensor<float, 6, Eigen::RowMajor> padded_in = a.at(0)->tensor.pad(paddings);

	 in_vector.at(0) = padded_in;

	const int depth_dimension = b.tensor.dimensions().size() - 3;

	for( unsigned int i = 0; i < vector_size - 1; i ++)
	{
		const unsigned int x_diff = b.tensor.dimension( MAX_DIMS - 1) - a.at(i+1)->tensor.dimension(MAX_DIMS - 1);
		const unsigned int y_diff = b.tensor.dimension( MAX_DIMS - 2) - a.at(i+1)->tensor.dimension(MAX_DIMS - 2);

		Eigen::array<std::pair<int, int>, 6> paddings;
		paddings[0] = paddings[1] = paddings[2] = paddings[3] = std::make_pair(0, 0);
		paddings[4] = std::make_pair(x_diff/2, x_diff/2);
		paddings[5] = std::make_pair(y_diff/2, y_diff/2);

		Eigen::Tensor<float, 6, Eigen::RowMajor> padded_in = a.at(i + 1)->tensor.pad(paddings);
		in_vector.at(i + 1) = in_vector.at(i).concatenate(padded_in, depth_dimension);
	}

	b.tensor = in_vector.back();

#else
	ERROR("Not implemented");
#endif
}
