#include "mali_cv/ITest.h"
#include "TestResults.h"

#include <algorithm>
#include <iostream>

using namespace mali_cv;

#define SAFE_CALL( func )\
	try{\
		func;\
	}catch(std::exception &e)\
	{\
		std::cout<<"In "<< #func <<" : "<< e.what() << std::endl;\
	}

void mali_cv::record_test_result( const std::string test_name, int test_id, Result result, const char* target, const char *message)
{
	std::string result_str;

	switch(result)
	{
		case Result::PASSED:
			result_str = "PASSED";
			break;
		case Result::SKIPPED:
			result_str = "SKIPPED";
			break;
		case Result::FAILED_AS_EXPECTED:
			result_str = "FAILED_AS_EXPECTED";
			break;
		case Result::CRASHED:
			result_str = "CRASHED";
			break;
		case Result::VALID_OUTPUT:
			result_str = "VALID_OUTPUT";
			break;
		case Result::INVALID_OUTPUT:
			result_str = "INVALID_OUTPUT";
			break;
		default:
			ERROR("Unknown result");
	}
	printf("|TEST_RESULT|%03d|%s|%s|%s|%s|\n", test_id, test_name.c_str(), target ? target : "ALL", result_str.c_str(), message ? message : "");
}

Environment::Environment( int check_out, BorderMode border_mode, uint constant_border_value, int num_iterations):
	ctx(check_out ? 1 : num_iterations, border_mode, constant_border_value),
	print_all(0),env_threshold(0),
	stop_on_error(0),check_output(check_out),
	reference(TargetID::VX),to_check(TargetID::CL),acceptable_ratio_of_mismatches(0.0f),min_threshold(0.0f),targets()
{
#ifdef USE_OPENVX
	targets[(uint)TargetID::VX] = std::unique_ptr<VXTarget>(new VXTarget(ctx));
#else
	targets[(uint)TargetID::VX] = nullptr;
#endif

#ifdef USE_OPENCL
	targets[(uint)TargetID::CL] = std::unique_ptr<CLTarget>(new CLTarget(ctx));
#else
	targets[(uint)TargetID::CL] = nullptr;
#endif

#ifdef USE_OPENCV
	targets[(uint)TargetID::CV] = std::unique_ptr<CVTarget>(new CVTarget(ctx));
#else
	targets[(uint)TargetID::CV] = nullptr;
#endif

#ifdef USE_NEON
	targets[(uint)TargetID::NEON] = std::unique_ptr<NETarget>(new NETarget(ctx));
#else
	targets[(uint)TargetID::NEON] = nullptr;
#endif
}

uint Environment::border_to_exclude_if_undefined( uint border_width, uint other_border_width)
{
	if( ctx.border_mode() == BorderMode::UNDEFINED )
	{
		return  border_width;
	}
	else
		return  other_border_width;
}

ITarget* Environment::target( TargetID id)
{
	if(id >= TargetID::NUM_TARGETS)
	{
		return nullptr;
	}
	else
	{
		return targets[static_cast<unsigned int>(id)].get();
	}
}

CLTarget* Environment::cl_target()
{
	return static_cast<CLTarget*>(target( TargetID::CL));
}

VXTarget* Environment::vx_target()
{
	return static_cast<VXTarget*>(target( TargetID::VX));
}

CVTarget* Environment::cv_target()
{
	return static_cast<CVTarget*>(target( TargetID::CV));
}

NETarget* Environment::ne_target()
{
	return static_cast<NETarget*>(target( TargetID::NEON));
}

ITest::ITest(Environment &e): env(e)
{
}

void ITest::run_cl()
{
	run( env.target(TargetID::CL));
}

void ITest::run_cv()
{
	run( env.target(TargetID::CV));
}
void ITest::run_vx()
{
	run( env.target(TargetID::VX));
}

void ITest::run_ne()
{
	run( env.target(TargetID::NEON));
}

void ITest::validate_outputs()
{
	ERROR("Not implemented");
}

void ITest::run(ITarget *target)
{
	ERROR("Not implemented");
}

void ITest::run( TargetID id)
{
	/* Important: Using a switch/case here is required in case the children classes have overriden any of the run_* methods */
	switch( id )
	{
		case TargetID::VX:
			run_vx();
			break;
		case TargetID::CL:
			run_cl();
			break;
		case TargetID::CV:
			run_cv();
			break;
		case TargetID::NEON:
			run_ne();
			break;
		default:
			ERROR("Invalid target");
	}
}

void ITest::run_reference()
{
	ERROR_ON_MSG( env.target( env.reference ) == nullptr, "Reference target selected is not compiled in, check your build options or choose a different reference target");
	run( env.reference );
}

void ITest::run_test_on_target( int id, ExpectedFailures &failures, TargetID target, bool validate)
{
	const char* skip_reason = failures.expected_to_fail(id, target, validate, env.acceptable_ratio_of_mismatches, env.min_threshold);
	if(env.env_threshold != 0)
	{
		env.min_threshold = (float) env.env_threshold;
	}

	// If it's guaranteed to fail then don't run it:
	if( skip_reason && env.acceptable_ratio_of_mismatches == 0.0f && env.min_threshold == 0.0f )
	{
		record_test_result( name(), id, Result::SKIPPED, env.target(target)->name().c_str(), skip_reason);
		return;
	}

	bool crashed_during_run=true;
	try
	{
		Result result = Result::PASSED;
		ERROR_ON_MSG( env.target( target ) == nullptr, "The target to check selected is not compiled in, check your build options or choose a different target to check");

		run(target);

		if(validate)
		{
			crashed_during_run=false;
			run_reference();
			printf("To check= %s Reference = %s\n", env.target( env.to_check )->name().c_str(),env.target( env.reference )->name().c_str());
			validate_outputs();
			result = Result::VALID_OUTPUT;
		}

		record_test_result( name(), id, result, env.target(target)->name().c_str(), skip_reason);
	}
#ifdef USE_OPENCL
    catch(const cl::Error &err)
    {
		std::cout << "In "<< __FUNCTION__ << "[" << env.target( target )->name() << "]: " << err.what() << " with error " << err.err() << std::endl;

		Result result = Result::CRASHED;

		// If there is a reason and no threshold or ratio then we expected this test to fail:
		if( skip_reason && env.min_threshold == 0.0f && env.acceptable_ratio_of_mismatches == 0.0f)
		{
			result = Result::FAILED_AS_EXPECTED;
		}
		else if(!crashed_during_run) // Did it fail to run or did it fail validation ?
		{
			result = Result::INVALID_OUTPUT;
		}

		record_test_result( name(), id, result, env.target(target)->name().c_str(), skip_reason);

		if(env.stop_on_error && !skip_reason)
		{
			throw;
		}
    }
#endif
	catch(const std::exception &e)
	{
		std::cout << "In " << __FUNCTION__ << "[" << env.target( target )->name() << "]: " << e.what() << std::endl;

		Result result = Result::CRASHED;

		// If there is a reason and no threshold or ratio then we expected this test to fail:
		if( skip_reason && env.min_threshold == 0.0f && env.acceptable_ratio_of_mismatches == 0.0f)
		{
			result = Result::FAILED_AS_EXPECTED;
		}
		else if(!crashed_during_run) // Did it fail to run or did it fail validation ?
		{
			result = Result::INVALID_OUTPUT;
		}

		record_test_result( name(), id, result, env.target(target)->name().c_str(), skip_reason);

		if(env.stop_on_error && !skip_reason)
		{
			throw;
		}
	}
}

void ITest::run(int id, ExpectedFailures &failures)
{
	TestResults::report().set_current_test(id, name());
	if(!env.check_output)
	{
		// Run all targets except VX:
		for( uint i=1; i < (uint)TargetID::NUM_TARGETS; i++)
		{
			if(env.target((TargetID)i) != nullptr)
			{
				run_test_on_target( id, failures, (TargetID) i, false);
			}
		}
	}
	else
	{
		run_test_on_target( id, failures, env.to_check,true );
	}
}

ExpectedFailures::ExpectedFailures(Environment &env):_expected_failures(),_env(env)
{
	_expected_failures.resize( (uint) TargetID::NUM_TARGETS );
}

void ExpectedFailures::add_expected_failure(int test_id, FailureReason reason, TargetID target, const std::string& message, float acceptable_ratio_of_mismatches, float min_threshold)
{
	if( target == TargetID::NUM_TARGETS)
	{
		return;
	}
	Conditions cond;
	cond.reason = reason;
	cond.acceptable_ratio_of_mismatches = acceptable_ratio_of_mismatches;
	cond.min_threshold = min_threshold;
	cond.message = message;
	cond.borders.push_back(0);
	cond.borders.push_back(1);
	cond.borders.push_back(2);
	_expected_failures[(uint)target][test_id] = cond;
}

void ExpectedFailures::add_expected_failure(int test_id, FailureReason reason, TargetID target, BorderMode border_mode, const std::string& message, float acceptable_ratio_of_mismatches, float min_threshold)
{
	if( target == TargetID::NUM_TARGETS)
	{
		return;
	}
	auto& failures = _expected_failures[(uint)target];
	auto test = failures.find(test_id);
	if( failures.end() == test )
	{
		Conditions cond;
		cond.reason = reason;
		cond.min_threshold = min_threshold;
		cond.acceptable_ratio_of_mismatches = acceptable_ratio_of_mismatches;
		cond.borders.push_back( (int) border_mode );
		cond.message = message;
		failures[test_id] = cond;
	}
	else
	{
		test->second.borders.push_back( (int) border_mode );
	}
}

const char *ExpectedFailures::expected_to_fail(int id, TargetID target, bool validate, float &acceptable_ratio_of_mismatches, float &min_threshold)
{
	acceptable_ratio_of_mismatches = 0.0f;
	min_threshold = 0.0f;

	auto test = _expected_failures[(uint)target].find(id);
	if( test == _expected_failures[(uint)target].end()) // No expected failure for this test
		return NULL;
	// Is it going to fail for the current border mode ?
	if(test->second.is_going_to_fail(_env))
	{
		if( test->second.reason == FailureReason::NOT_IMPLEMENTED || test->second.reason == FailureReason::NOT_SUPPORTED)
			return test->second.message.c_str(); // If it's not implemented or supported: then it will fail.

		if( !validate && test->second.reason == FailureReason::CANNOT_BE_VALIDATED)
			return NULL; // If the test cannot be validated but its output won't be checked then run it

		acceptable_ratio_of_mismatches = test->second.acceptable_ratio_of_mismatches;
		min_threshold = test->second.min_threshold;
		return test->second.message.c_str(); // Yes
	}
	return NULL; // No
}

ExpectedFailures::Conditions::Conditions():borders(),message(),acceptable_ratio_of_mismatches(0.0f),min_threshold(0.0f),reason()
{
}

bool ExpectedFailures::Conditions::is_going_to_fail(Environment &env)
{
	if( 3 == borders.size() ) return true;

	for ( const auto& b : borders) if( b == (int) env.ctx.border_mode() ) return true;

	return false;
}

TestManager::TestManager(Environment &env):_failures(env),_tests()
{
}


void TestManager::add_core_test(int id, std::function<ITest*()> function)
{
	ERROR_ON_MSG( _tests.count(id) > 0, "ID already in use");
	_tests[id] = function;
}

void TestManager::add_test_failure(int id, const  TestFailure &tf)
{
	if( tf.has_border_mode() )
		_failures.add_expected_failure(id, tf.reason(), tf.target(), tf.border_mode(), tf.msg(), tf.ratio_of_mismatches(), tf.threshold() );
	else
		_failures.add_expected_failure(id, tf.reason(), tf.target(), tf.msg(), tf.ratio_of_mismatches(), tf.threshold() );
}


void TestManager::run(int id, const char*filter)
{
	if( _tests.count(id) < 1)
	{
		printf("Warning: test ID %d not found\n", id);
		return;
	}

	auto &test_builder = _tests[id];
	auto test = test_builder();

	if( filter && test->name().find(filter) == std::string::npos)
	{
		record_test_result( test->name(), id, Result::SKIPPED, nullptr, "Filtered out by TEST_FILTER" );
	}
	else
	{
		printf("\n\n----------\n");
		printf("[TEST] %02d: %s\n", id, test->name().c_str());
		printf("----------\n");

		test->run(id, _failures);
	}
	delete test;
}

void TestManager::run_range( const int first, const int last, const char* filter )
{
	for(auto& it : _tests)
	{
		if(it.first >= first && it.first <= last)
		{
			run(it.first, filter);
		}
	}
}

