#include "mali_cv/Image.h"

#include "arm_compute/core/Utils.h"
#include "mali_cv/Error.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/Context.h"
#include "mali_cv/Helper.h"
#include "NEComputeHelper.h"

#include <utility>

using namespace mali_cv;

#ifdef TEST_TENSORS
NEImageTensor::NEImageTensor():_buffer(nullptr),_info()
{
}

void NEImageTensor::init( arm_compute::ITensor& parent )
{
	const arm_compute::ITensorInfo* parent_info = parent.info();

	ERROR_ON( parent_info->num_dimensions() < 4 );
	ERROR_ON( dim2_to_use >= parent_info->dimension(2) );
	ERROR_ON( dim3_to_use >= parent_info->dimension(3) );

	_buffer = parent.buffer() + dim2_to_use * parent_info->strides_in_bytes()[2] + dim3_to_use * parent_info->strides_in_bytes()[3];

	arm_compute::TensorInfo info(
		arm_compute::TensorShape(parent_info->dimension(0), parent_info->dimension(1)),
		parent_info->format());

	info.auto_padding();

	info.init(
		info.tensor_shape(),
		info.format(),
		arm_compute::Strides(parent_info->strides_in_bytes()[0], parent_info->strides_in_bytes()[1]),
		parent_info->offset_first_element_in_bytes(),
		info.total_size());

	info.set_is_resizable(parent_info->is_resizable());

	_info = std::move(info);
}

arm_compute::ITensorInfo* NEImageTensor::info() const
{
	return &_info;
}

arm_compute::ITensorInfo* NEImageTensor::info()
{
	return &_info;
}

uint8_t* NEImageTensor::buffer() const
{
	return _buffer;
}
#endif /* TEST_TENSORS */

NEImage::NEImage(): IImage(),
#ifdef TEST_TENSORS
	tensor(),image()
#else /* TEST_TENSORS */
	multi_img(),
	tensor(*multi_img.plane(0)),
	image(*multi_img.plane(0))
#endif /* TEST_TENSORS */
{
}

void NEImage::internal_create(Context&, size_t w, size_t h, Format format)
{
	arm_compute::Format f = ARM_COMPUTEHelper::convert_to_arm_compute_format(format);

#ifdef TEST_TENSORS
	arm_compute::TensorInfo info(arm_compute::TensorShape(w, h, 3U, 2U), f);
	info.auto_padding();
	tensor.allocator()->init(info);
	tensor.allocator()->allocate();
	image.init(tensor);
#else /* TEST_TENSORS */
	/* The allocator takes care of the padding */
	multi_img.init_auto_padding(w, h, f);
	multi_img.allocate();
#endif /* TEST_TENSORS */
}

void NEImage::store(RawImage &out)
{
	out.create(width, height, format);

#ifdef TEST_TENSORS
	copy_plane(image, out.data());
#else /* TEST_TENSORS */
	unsigned char* tmp_ptr = out.data();
	for(int i = 0; i < 3 && (nullptr != multi_img.plane(i)->buffer()); ++i)
	{
		tmp_ptr = copy_plane(*multi_img.plane(i), tmp_ptr);
	}
#endif /* TEST_TENSORS */
}

void NEImage::load(const RawImage &in)
{
	ERROR_ON(in.width() != width);
	ERROR_ON(in.height() != height);
	ERROR_ON(in.format() != format);

#ifdef TEST_TENSORS
	fill_plane(in.data(),image);
#else /* TEST_TENSORS */
	unsigned char* tmp_ptr = in.data();
	for(int i = 0; i < 3 && (nullptr != multi_img.plane(i)->buffer()); ++i)
	{
		tmp_ptr = fill_plane(tmp_ptr , *multi_img.plane(i));
	}
#endif /* TEST_TENSORS */
}

arm_compute::ValidRegion NEImage::valid_region() const
{
    return image.info()->valid_region();
}
