#include "mali_cv/Target.h"

#include "arm_compute/runtime/NEON/NEFunctions.h"
#include "arm_compute/core/NEON/NEKernels.h" // Just to check this header is well formed
#include "arm_compute/runtime/Array.h"
#include "arm_compute/runtime/MultiHOG.h"
#include "arm_compute/runtime/Pyramid.h"
#include "arm_compute/core/Types.h"
#include "mali_cv/Error.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/Context.h"
#include "mali_cv/Convolution.h"
#include "mali_cv/Image.h"
#include "mali_cv/Pyramid.h"
#include "mali_cv/RemapTable.h"
#include "mali_cv/Timer.h"
#include "NEComputeHelper.h"
#include "TestResults.h"

#include <cstring>
#include <iostream>
#include <sstream>
#include <thread>

using namespace mali_cv;

NETarget::NETarget(Context& ctx): ITarget(ctx)
{
	_timer = Timer(name());

	unsigned int number_of_cpu_threads = 1;

#ifndef NO_MULTI_THREADING
	number_of_cpu_threads = std::thread::hardware_concurrency();
#endif

	std::cout << "Default number of CPU threads = " << number_of_cpu_threads << "\n";
}

void NETarget::absolute_difference( Image &a, Image &b, Image &c)
{
	arm_compute::NEAbsoluteDifference abs_diff;
	abs_diff.configure(&a.ne.tensor, &b.ne.tensor, &c.ne.tensor);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		abs_diff.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::bitwise_or( Image &a, Image &b, Image &c)
{
	arm_compute::NEBitwiseOr bitwise_or;
	bitwise_or.configure(&a.ne.tensor, &b.ne.tensor, &c.ne.tensor);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		bitwise_or.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::bitwise_and( Image &a, Image &b, Image &c)
{
	arm_compute::NEBitwiseAnd bitwise_and;
	bitwise_and.configure(&a.ne.tensor, &b.ne.tensor, &c.ne.tensor);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		bitwise_and.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::bitwise_not( Image &a,  Image &c)
{
	arm_compute::NEBitwiseNot bitwise_not;
	bitwise_not.configure(&a.ne.tensor, &c.ne.tensor);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		bitwise_not.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::accumulate( Image &a,  Image &c)
{
	arm_compute::NEAccumulate acc;
	acc.configure(&a.ne.tensor, &c.ne.tensor);

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		acc.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::accumulate_squared( Image &a, Image &c, uint32_t shift)
{
	arm_compute::NEAccumulateSquared acc_squared;
	acc_squared.configure(&a.ne.tensor, shift, &c.ne.tensor);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		acc_squared.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}

}

void NETarget::accumulate_weighted( Image &a, Image &c, float alpha)
{
	arm_compute::NEAccumulateWeighted acc_weighted;
	acc_weighted.configure(&a.ne.tensor, alpha, &c.ne.tensor);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		acc_weighted.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}


void NETarget::accumulate_weighted_fp16( Image &a, Image &c, float alpha)
{
	arm_compute::NEAccumulateWeighted acc_weighted_fp16;
	acc_weighted_fp16.configure(&a.ne.tensor, alpha, &c.ne.tensor,true);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		acc_weighted_fp16.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::arithmetic_addition( Image &a, Image &b, Image &c, uint8_t policy)
{
	arm_compute::NEArithmeticAddition add;
	add.configure( &a.ne.tensor, &b.ne.tensor, &c.ne.tensor, (arm_compute::ConvertPolicy) policy);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		add.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::harris_corners(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners,size_t& num_corners)
{
	arm_compute::NEHarrisCorners harris_corners;
	harris_corners.configure( &a.ne.image, threshold, min_dist, sensitivity, gradient, block_size, &corners.ne.array, _ctx.border_mode(),_ctx.constant_border_value() );
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		corners.ne.array.clear();

		_timer.start();
		harris_corners.run();
		_timer.stop();

		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}

	num_corners = corners.ne.array.num_values();
}

void NETarget::harris_corners_fp16(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners,size_t& num_corners)
{
	arm_compute::NEHarrisCorners harris_corners;
	harris_corners.configure( &a.ne.image, threshold, min_dist, sensitivity, gradient, block_size, &corners.ne.array, _ctx.border_mode(),_ctx.constant_border_value(),true);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		corners.ne.array.clear();

		_timer.start();
		harris_corners.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::fast_corners(Image &a,float threshold, bool nonmax_suppression, IKeyPointArray &corners, size_t &num_corners)
{
	arm_compute::NEFastCorners fast_corners;
	fast_corners.configure(&a.ne.image,threshold,nonmax_suppression, &corners.ne.array, _ctx.border_mode(),_ctx.constant_border_value() );
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		corners.ne.array.clear();

		_timer.start();
		fast_corners.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::arithmetic_subtraction( Image &a, Image &b, Image &c, uint8_t policy)
{
	arm_compute::NEArithmeticSubtraction sub;
	sub.configure( &a.ne.tensor, &b.ne.tensor, &c.ne.tensor, (arm_compute::ConvertPolicy) policy);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		sub.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::bitwise_xor( Image &a, Image &b, Image &c)
{
	arm_compute::NEBitwiseXor bitwise_xor;
	bitwise_xor.configure(&a.ne.tensor, &b.ne.tensor, &c.ne.tensor);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		bitwise_xor.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::pixelwise_mul( Image &a, Image &b, Image &c, float scale, uint8_t overflow_policy, uint8_t round_policy )
{
	arm_compute::NEPixelWiseMultiplication pixelwise_mul;
	arm_compute::ConvertPolicy op = (overflow_policy) ? arm_compute::ConvertPolicy::SATURATE : arm_compute::ConvertPolicy::WRAP;
	arm_compute::RoundingPolicy rp = (round_policy) ? arm_compute::RoundingPolicy::TO_NEAREST_EVEN : arm_compute::RoundingPolicy::TO_ZERO;
	pixelwise_mul.configure(&a.ne.tensor, &b.ne.tensor, &c.ne.tensor, scale, op, rp);
	for( unsigned int j= 0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		pixelwise_mul.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::erode( Image &a, Image &c)
{
	arm_compute::NEErode erode;
	erode.configure(&a.ne.tensor, &c.ne.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		erode.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::dilate( Image &a, Image &c)
{
	arm_compute::NEDilate dilate;
	dilate.configure(&a.ne.tensor, &c.ne.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		dilate.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::gaussian3x3( Image &a, Image &c)
{
	arm_compute::NEGaussian3x3 conv;
	conv.configure( &a.ne.tensor, &c.ne.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		conv.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}

}

void NETarget::gaussian5x5( Image &a, Image &c)
{
	arm_compute::NEGaussian5x5 conv;
	conv.configure( &a.ne.tensor, &c.ne.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		conv.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::warp_perspective(Image &a, Image &c, const float matrix[3][3], InterpolationMode interpolation)
{
	arm_compute::NEWarpPerspective warp;
	warp.configure( &a.ne.tensor, &c.ne.tensor, (const float*)matrix, interpolation, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		warp.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::warp_affine(Image &a, Image &c, const float matrix[3][2], InterpolationMode interpolation)
{
	arm_compute::NEWarpAffine warp;
	warp.configure( &a.ne.tensor, &c.ne.tensor, (const float*)matrix, interpolation, _ctx.border_mode(), _ctx.constant_border_value());
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		warp.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::box3x3( Image &a, Image &c)
{
	arm_compute::NEBox3x3 conv;
	conv.configure( &a.ne.tensor, &c.ne.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		conv.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::box3x3_fp16( Image &a, Image &c)
{
	arm_compute::NEBox3x3 conv;
	conv.configure( &a.ne.tensor, &c.ne.tensor, _ctx.border_mode(), _ctx.constant_border_value(), true);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		conv.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::median3x3( Image &a, Image &c)
{
	arm_compute::NEMedian3x3 filter;
	filter.configure( &a.ne.tensor, &c.ne.tensor, _ctx.border_mode(), _ctx.constant_border_value());
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		filter.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::convolution_rectangle(Image &a, Image &c, Convolution &matrix)
{
	ERROR_ON(matrix.ne.width != 3 && matrix.ne.width != 5 && matrix.ne.width != 7 && matrix.ne.width != 9);
	ERROR_ON(matrix.ne.height != 3 && matrix.ne.height != 5 && matrix.ne.height != 7 && matrix.ne.height != 9);
	ERROR_ON(matrix.ne.format != Format::LUMINANCE16);
	arm_compute::NEConvolutionRectangle conv;
	conv.configure(&a.ne.tensor, &c.ne.tensor, matrix.ne.convolution.get(), matrix.ne.width, matrix.ne.height, matrix.ne._scale , _ctx.border_mode(), _ctx.constant_border_value());
	conv.run();
}

void NETarget::convolution9x9(Image &a, Image &c, Convolution &matrix)
{
	ERROR_ON(matrix.ne.width != 9);
	ERROR_ON(matrix.ne.height != 9);
	ERROR_ON(matrix.ne.format != Format::LUMINANCE16);
	arm_compute::NEConvolution9x9 conv;
	conv.configure(&a.ne.tensor, &c.ne.tensor, matrix.ne.convolution.get(), matrix.ne._scale , _ctx.border_mode(), _ctx.constant_border_value());
	conv.run();
}


void NETarget::convolution7x7(Image &a, Image &c, Convolution &matrix)
{
	ERROR_ON(matrix.ne.width != 7);
	ERROR_ON(matrix.ne.height != 7);
	ERROR_ON(matrix.ne.format != Format::LUMINANCE16);
	arm_compute::NEConvolution7x7 conv;
	conv.configure(&a.ne.tensor, &c.ne.tensor, matrix.ne.convolution.get(), matrix.ne._scale , _ctx.border_mode(), _ctx.constant_border_value());
	conv.run();
}


void NETarget::convolution5x5(Image &a, Image &c, Convolution &matrix)
{
	ERROR_ON(matrix.ne.width != 5);
	ERROR_ON(matrix.ne.height != 5);
	ERROR_ON(matrix.ne.format != Format::LUMINANCE16);
	arm_compute::NEConvolution5x5 conv;
	conv.configure( &a.ne.tensor, &c.ne.tensor, matrix.ne.convolution.get(), matrix.ne._scale , _ctx.border_mode(), _ctx.constant_border_value());
	conv.run();
}

void NETarget::convolution3x3(Image &a, Image &c, Convolution &matrix)
{
	ERROR_ON(matrix.ne.width != 3);
	ERROR_ON(matrix.ne.height != 3);
	ERROR_ON(matrix.ne.format != Format::LUMINANCE16);
	arm_compute::NEConvolution3x3 conv;
	conv.configure( &a.ne.tensor, &c.ne.tensor, matrix.ne.convolution.get(), matrix.ne._scale , _ctx.border_mode(), _ctx.constant_border_value());
	conv.run();
}

void NETarget::convolution( Image &a, Image &c, Convolution &matrix)
{
	ERROR_ON(matrix.ne.width != 3 && matrix.ne.width != 5 && matrix.ne.width != 7 && matrix.ne.width != 9);
	ERROR_ON(matrix.ne.height != 3 && matrix.ne.height != 5 && matrix.ne.height != 7 && matrix.ne.height != 9);

	// Run test
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		if(matrix.ne.width == matrix.ne.height)
		{
		    switch(matrix.ne.width)
		    {
			case 3:
				convolution3x3(a, c, matrix);
				break;
			case 5:
				convolution5x5(a, c, matrix);
				break;
			case 7:
				convolution7x7(a, c, matrix);
				break;
			case 9:
				convolution9x9(a, c, matrix);
				break;
			default:
				ERROR("Not supported");
				break;
		    }
		}
		else
		{
		    convolution_rectangle(a, c, matrix);
		}
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::scale( Image &a, Image &c, InterpolationMode interpolation)
{
	arm_compute::NEScale scale;
	scale.configure(&a.ne.tensor, &c.ne.tensor, interpolation, _ctx.border_mode(), _ctx.constant_border_value());
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		scale.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::color_convert( Image &a, Image &c )
{
	arm_compute::NEColorConvert convert;

	switch(a.ne.format) // source format
	{
		case Format::RGBA8888:
		{
			switch(c.ne.format) // dst format
			{
				case Format::RGB888:
					convert.configure(&a.ne.tensor, &c.ne.tensor);
					break;
#ifndef TEST_TENSORS
				case Format::UYVY:
				case Format::NV12:
					convert.configure(&a.ne.image, &c.ne.multi_img);
					break;
				case Format::IYUV:
					convert.configure(&a.ne.image, &c.ne.multi_img);
					break;
				case Format::YUV444:
					convert.configure(&a.ne.image, &c.ne.multi_img);
					break;
#endif /* TEST_TENSORS */
				case Format::UNKNOWN:
				default:
					ERROR("Not supported");
					break;
			}
			break;
		}
		case Format::UYVY:
		{
			switch(c.ne.format)
			{
				case Format::RGB888:
					convert.configure(&a.ne.tensor, &c.ne.tensor);
					break;
				case Format::RGBA8888:
					convert.configure(&a.ne.tensor, &c.ne.tensor);
					break;
#ifndef TEST_TENSORS
				case Format::NV12:
					convert.configure(&a.ne.image, &c.ne.multi_img);
					break;
				case Format::IYUV:
					convert.configure(&a.ne.image, &c.ne.multi_img);
					break;
#endif /* TEST_TENSORS */
				case Format::YUV444:
				case Format::UNKNOWN:
				default:
					ERROR("Not supported");
					break;
			}
			break;
		}
		case Format::YUYV422:
		{
			switch(c.ne.format)
			{
				case Format::RGB888:
					convert.configure(&a.ne.tensor, &c.ne.tensor);
					break;
				case Format::RGBA8888:
					convert.configure(&a.ne.tensor, &c.ne.tensor);
					break;
#ifndef TEST_TENSORS
				case Format::NV12:
					convert.configure(&a.ne.image, &c.ne.multi_img);
					break;
				case Format::IYUV:
					convert.configure(&a.ne.image, &c.ne.multi_img);
					break;
#endif /* TEST_TENSORS */
				case Format::UNKNOWN:
				default:
					ERROR("Not supported");
					break;
			}
			break;
		}

		case Format::RGB888:
		{
			switch(c.ne.format)
			{
				case Format::RGBA8888:
					convert.configure(&a.ne.tensor, &c.ne.tensor);
					break;
#ifndef TEST_TENSORS
				case Format::NV12:
					convert.configure(&a.ne.image, &c.ne.multi_img);
					break;
				case Format::IYUV:
					convert.configure(&a.ne.image, &c.ne.multi_img);
					break;
				case Format::YUV444:
					convert.configure(&a.ne.image, &c.ne.multi_img);
					break;
#endif /* TEST_TENSORS */
				default:
					ERROR("Not supported");
					break;
			}
			break;
		}

		case Format::NV12:
		{
			switch(c.ne.format)
			{
#ifndef TEST_TENSORS
				case Format::RGBA8888:
					convert.configure(&a.ne.multi_img, &c.ne.image);
					break;
				case Format::RGB888:
					convert.configure(&a.ne.multi_img, &c.ne.image);
					break;
				case Format::IYUV:
					convert.configure(&a.ne.multi_img, &c.ne.multi_img);
					break;
				case Format::YUV444:
					convert.configure(&a.ne.multi_img, &c.ne.multi_img);
					break;
#endif /* TEST_TENSORS */
				default:
					ERROR("Not supported");
					break;

			}
			break;
		}

		case Format::NV21:
		{
			switch(c.ne.format)
			{
#ifndef TEST_TENSORS
				case Format::RGBA8888:
					convert.configure(&a.ne.multi_img, &c.ne.image);
					break;
				case Format::RGB888:
					convert.configure(&a.ne.multi_img, &c.ne.image);
					break;
				case Format::IYUV:
					convert.configure(&a.ne.multi_img, &c.ne.multi_img);
					break;
				case Format::YUV444:
					convert.configure(&a.ne.multi_img, &c.ne.multi_img);
					break;
#endif /* TEST_TENSORS */
				default:
					ERROR("Not supported");
					break;

			}
			break;
		}
		case Format::IYUV:
		{
			switch(c.ne.format)
			{
#ifndef TEST_TENSORS
				case Format::RGBA8888:
					convert.configure(&a.ne.multi_img, &c.ne.image);
					break;
				case Format::RGB888:
					convert.configure(&a.ne.multi_img, &c.ne.image);
					break;
				case Format::NV12:
					convert.configure(&a.ne.multi_img, &c.ne.multi_img);
					break;
				case Format::YUV444:
					convert.configure(&a.ne.multi_img, &c.ne.multi_img);
					break;
#endif /* TEST_TENSORS */
				default:
					ERROR("Not supported");
					break;

			}
			break;
		}
		case Format::YUV444:
		case Format::UNKNOWN:
		default:
			ERROR("Not supported");
			break;
	}

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		convert.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::sobel( Image &a, Image *c, Image *d)
{
	arm_compute::NESobel3x3 conv;
	conv.configure( &a.ne.tensor, c ? &c->ne.tensor : nullptr, d ? &d->ne.tensor : nullptr, _ctx.border_mode(), _ctx.constant_border_value());
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		conv.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::scharr( Image &a, Image *c, Image *d)
{
	arm_compute::NEScharr3x3 conv;

	conv.configure( &a.ne.tensor, c ? &c->ne.tensor : nullptr, d ? &d->ne.tensor : nullptr, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		conv.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::derivative( Image &a, Image *c, Image *d)
{
	arm_compute::NEDerivative der;

	der.configure( &a.ne.tensor, c ? &c->ne.tensor : nullptr, d ? &d->ne.tensor : nullptr, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		der.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::sobel5x5( Image &a, Image *c, Image *d)
{
	arm_compute::NESobel5x5 conv;

	conv.configure( &a.ne.tensor, c ? &c->ne.tensor : nullptr, d ? &d->ne.tensor : nullptr, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		conv.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::sobel7x7( Image &a, Image *c, Image *d)
{
	arm_compute::NESobel7x7 conv;

	conv.configure( &a.ne.tensor, c ? &c->ne.tensor : nullptr, d ? &d->ne.tensor : nullptr, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		conv.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::canny(Image &a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image &c)
{
	arm_compute::NECannyEdge canny;
	canny.configure(&a.ne.tensor, &c.ne.tensor, upper, lower, gradient_size, norm_type, _ctx.border_mode() );
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		canny.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::canny_fp16(Image &a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image &c)
{
	arm_compute::NECannyEdge canny;
	canny.configure(&a.ne.tensor, &c.ne.tensor, upper, lower, gradient_size, norm_type, _ctx.border_mode(),0,true );
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		canny.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::laplacian_reconstruct( Pyramid &pyramid, Image &a, Image &c)
{
	arm_compute::NELaplacianReconstruct lr;
	lr.configure(&pyramid.ne.pyramid, &a.ne.image, &c.ne.image,_ctx.border_mode(),_ctx.constant_border_value());

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		lr.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::laplacian_pyramid(Image &a, Pyramid &pyramid, Image &b, bool timers_enabled)
{
	arm_compute::NELaplacianPyramid lp;
	lp.configure(&a.ne.image, &pyramid.ne.pyramid, &b.ne.image, _ctx.border_mode(),_ctx.constant_border_value());

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		lp.run();
		if(timers_enabled)
		{
			_timer.stop();
			TestResults::report().add(name(), _timer.get_elapsed_ms());
			TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
			_timer.print_elapsed_ms();
		}
	}
}

void NETarget::gaussian_pyramid(Image &a, Pyramid &pyramid)
{
	ERROR_ON(Format::LUMINANCE != a.ne.format );

	std::unique_ptr<arm_compute::NEGaussianPyramid> gaussian_pyramid;

	if( arm_compute::SCALE_PYRAMID_HALF == pyramid.scale )
	{
		gaussian_pyramid.reset(new arm_compute::NEGaussianPyramidHalf());
	}
	else if( arm_compute::SCALE_PYRAMID_ORB == pyramid.scale )
	{
		gaussian_pyramid.reset(new arm_compute::NEGaussianPyramidOrb());
	}
	else
	{
		ERROR("Scale factor not supported for gaussian pyramid");
	}

	gaussian_pyramid->configure(&a.ne.image, &pyramid.ne.pyramid, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		gaussian_pyramid->run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::magnitude(Image &a, Image &b, Image &c)
{
	arm_compute::NEMagnitude mag;
	mag.configure(&a.ne.tensor, &b.ne.tensor, &c.ne.tensor);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		mag.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::magnitude_fp16(Image &a, Image &b, Image &c)
{
	arm_compute::NEMagnitude mag;
	mag.configure(&a.ne.tensor, &b.ne.tensor, &c.ne.tensor, arm_compute::MagnitudeType::L2NORM, true);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		mag.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::phase( Image &a, Image &b, Image &c)
{
	arm_compute::NEPhase phase;
	phase.configure(&a.ne.tensor, &b.ne.tensor, &c.ne.tensor);
	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		phase.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::threshold_binary( Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value)
{
	arm_compute::NEThreshold thresh;

	thresh.configure(&a.ne.tensor, &b.ne.tensor, threshold, false_value, true_value, arm_compute::ThresholdType::BINARY);
	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		thresh.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::threshold_range( Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value, uint8_t upper)
{
	arm_compute::NEThreshold thresh;

	thresh.configure(&a.ne.tensor, &b.ne.tensor, threshold, false_value, true_value, arm_compute::ThresholdType::RANGE, upper);
	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		thresh.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::remap(Image &a, RemapTable &table, Image &d, InterpolationMode interpolation)
{
	arm_compute::NERemap remap;
	remap.configure(&a.ne.tensor, &table.ne.table[0]->tensor, &table.ne.table[1]->tensor, &d.ne.tensor, interpolation, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		remap.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::channel_extract( Image &a, Channel channel, Image &c )
{
	arm_compute::NEChannelExtract ch_extract;

	// Initialize function appropriately.
	switch( a.ne.format )
	{
		case Format::NV12:
		case Format::NV21:
		case Format::IYUV:
		case Format::YUV444:
#ifdef TEST_TENSORS
			ERROR("Not supported");
#else /* TEST_TENSORS */
			ch_extract.configure(&a.ne.multi_img, channel, &c.ne.image);
#endif /* TEST_TENSORS */
			break;
		default:
			ch_extract.configure(&a.ne.tensor, channel, &c.ne.tensor);
			break;
	}

	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		ch_extract.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::mean_stddev(Image& a, Scalar<float>& mean, Scalar<float>& stddev)
{
	arm_compute::NEMeanStdDev mean_stddev;
	mean_stddev.configure(&a.ne.image, &mean.ne.scalar, &stddev.ne.scalar);
	for( unsigned int j = 0; j < _ctx.num_iterations(); j++ )
	{
		_timer.start();
		mean_stddev.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::tablelookup( Image &a , Lut &b, Image &c )
{
	arm_compute::NETableLookup lut;
	lut.configure( &a.ne.tensor, &b.ne.lut, &c.ne.tensor );
	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		lut.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::convert_depth( Image &a, Image &c, ConvertPolicy policy , uint32_t shift)
{
	arm_compute::NEDepthConvert dc;
	dc.configure( &a.ne.tensor , &c.ne.tensor, policy, shift );
	for(uint i = 0; i < _ctx.num_iterations(); i++)
	{
		_timer.start();
		dc.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::minmaxlocation( Image &a, Scalar<int32_t> &min, Scalar<int32_t> &max, ICoord2DArray *min_loc, ICoord2DArray *max_loc,
                               Scalar<uint32_t> *min_count, Scalar<uint32_t> *max_count)
{
	arm_compute::NEMinMaxLocation minMaxLoc;
	uint32_t* min_count_ptr = (min_count) ? &min_count->ne.scalar : nullptr;
	uint32_t* max_count_ptr = (max_count) ? &max_count->ne.scalar : nullptr;
	arm_compute::Coordinates2DArray* min_loc_ptr = (min_loc) ? &min_loc->ne.array : nullptr;
	arm_compute::Coordinates2DArray* max_loc_ptr = (max_loc) ? &max_loc->ne.array : nullptr;
	minMaxLoc.configure(&a.ne.image, &min.ne.scalar, &max.ne.scalar, min_loc_ptr, max_loc_ptr, min_count_ptr, max_count_ptr);
	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		minMaxLoc.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::histogram(Image& a, Distribution& b)
{
	arm_compute::NEHistogram hist;
	hist.configure(&a.ne.image, b.ne.dist.get());
	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		hist.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::channel_combine(Image &plane0, Image &plane1, Image &plane2, Image &plane3, Image &out)
{
	arm_compute::NEChannelCombine ch_combine;

	// Initialize function appropriately.
	switch( out.ne.format )
	{
		case Format::NV12:
		case Format::NV21:
		case Format::IYUV:
		case Format::YUV444:
#ifdef TEST_TENSORS
			ERROR("Not supported");
#else /* TEST_TENSORS */
			ch_combine.configure( &plane0.ne.image, &plane1.ne.image, &plane2.ne.image, &out.ne.multi_img );
#endif /* TEST_TENSORS */
			break;
		case Format::RGBA8888:
			ch_combine.configure( &plane0.ne.tensor, &plane1.ne.tensor, &plane2.ne.tensor, &plane3.ne.tensor, &out.ne.tensor );
			break;
		default:
			ch_combine.configure( &plane0.ne.tensor, &plane1.ne.tensor, &plane2.ne.tensor, nullptr, &out.ne.tensor );
			break;
	}

	for(unsigned int j=0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		ch_combine.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::optical_flow(Image &a, Image &b, IKeyPointArray *old_points, IKeyPointArray *new_points_estimates,
                            IKeyPointArray *new_points, Termination termination, float epsilon, uint32_t num_iterations,
                            bool use_initial_estimate, size_t window_dimension, size_t num_levels)
{
	arm_compute::NEGaussianPyramidHalf gp1;
	arm_compute::NEGaussianPyramidHalf gp2;
	arm_compute::NEOpticalFlow of;
	arm_compute::Pyramid old_pyramid;
	arm_compute::Pyramid new_pyramid;
	arm_compute::PyramidInfo pyramid_info;

	pyramid_info.init(num_levels, 0.5f, a.ne.width, a.ne.height, arm_compute::Format::U8);

	old_pyramid.init_auto_padding(pyramid_info);
	old_pyramid.allocate();
	new_pyramid.init_auto_padding(pyramid_info);
	new_pyramid.allocate();

	gp1.configure(&a.ne.image, &old_pyramid, _ctx.border_mode(), _ctx.constant_border_value());
	gp2.configure(&b.ne.image, &new_pyramid, _ctx.border_mode(), _ctx.constant_border_value());

	gp1.run();
	gp2.run();

	of.configure(&old_pyramid, &new_pyramid, &old_points->ne.array, &new_points_estimates->ne.array, &new_points->ne.array, termination, epsilon,
	             num_iterations, window_dimension, use_initial_estimate, _ctx.border_mode(), _ctx.constant_border_value());

	for(unsigned int j = 0; j < _ctx.num_iterations(); j++)
	{
		_timer.start();
		of.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::integral(Image &a, Image &b)
{
	arm_compute::NEIntegralImage integralimage;
	integralimage.configure(&a.ne.tensor, &b.ne.tensor);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		integralimage.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::equalize_histogram(Image& a, Image& b)
{
	arm_compute::NEEqualizeHistogram eq_hist;

	eq_hist.configure(&a.ne.image, &b.ne.image);
	for( unsigned int i = 0; i < _ctx.num_iterations(); ++i )
	{
		_timer.start();
		eq_hist.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::non_linear_filter(Image &a, Image &b, NonLinearFilterFunction function, MatrixPattern pattern, uint mask_size, uint8_t* mask)
{
	arm_compute::NENonLinearFilter nl_filter;

	nl_filter.configure(&a.ne.image, &b.ne.image, function, mask_size, pattern, mask, _ctx.border_mode(), _ctx.constant_border_value());
	for( unsigned int i = 0; i < _ctx.num_iterations(); ++i )
	{
		_timer.start();
		nl_filter.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::hog_descriptor(Image &a, HOG &b, Tensor &c)
{
	arm_compute::NEHOGDescriptor hog_descriptor;

	hog_descriptor.configure( &a.ne.tensor, &c.ne.tensor, &b.ne.hog, _ctx.border_mode(), _ctx.constant_border_value() );

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		hog_descriptor.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::hog_detector(Tensor &a, HOG &b, IKeyPointArray &keypoint_detection_windows)
{
	arm_compute::NEHOGDetector hog_detector;
	arm_compute::DetectionWindowArray detection_windows(10000);

	hog_detector.configure( &a.ne.tensor, &b.ne.hog, &detection_windows, b.ne.hog.info()->block_stride());

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		// Reset detection windows
		detection_windows.clear();

		_timer.start();
		hog_detector.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}

	// Copy the detection  windows
	std::vector<IKeyPoint> keypoints;

	for(uint i = 0; i < detection_windows.num_values(); i++)
	{
		IKeyPoint val;
		val.x = detection_windows.at(i).x;
		val.y = detection_windows.at(i).y;
		val.tracking_status = 1;
		val.strength = detection_windows.at(i).score;

		keypoints.push_back(val);
	}

	keypoint_detection_windows.ne.fill(keypoints);
}

void NETarget::hog_multi_detection(Image &a, HOG &b, IKeyPointArray &keypoint_detection_windows)
{
	arm_compute::NEHOGMultiDetection hog_multi_detection;
	arm_compute::MultiHOG multi_hog(1);
	arm_compute::DetectionWindowArray detection_windows(10000);
	arm_compute::Size2DArray detection_window_strides(1);

	detection_window_strides.resize(1);
	for(uint i = 0; i < 1; i++)
	{
		auto hog_model = reinterpret_cast<arm_compute::HOG*>(multi_hog.model(i));

		// Initialize HOG info
		hog_model->init(*(b.ne.hog.info()));

		// Initialize HOG descriptor
		std::memcpy(hog_model->descriptor(), b.ne.hog.descriptor(), b.descriptor_size * sizeof(float));

		// Initialize detection window stride
		detection_window_strides.at(i) = multi_hog.model(i)->info()->block_stride();
	}

	hog_multi_detection.configure( &a.ne.tensor, &multi_hog, &detection_windows, &detection_window_strides, arm_compute::BorderMode::REPLICATE, 0, 0.0f, false, 30.0f);

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		// Reset detection windows
		detection_windows.clear();

		_timer.start();
		hog_multi_detection.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}

	// Copy the detection  windows
	std::vector<IKeyPoint> keypoints;

	for(uint i = 0; i < detection_windows.num_values(); i++)
	{
		IKeyPoint val;
		val.x = detection_windows.at(i).x;
		val.y = detection_windows.at(i).y;
		val.tracking_status = 1;
		val.strength = detection_windows.at(i).score;

		keypoints.push_back(val);
	}

	keypoint_detection_windows.ne.fill(keypoints);
}

void NETarget::tensor_ops(Tensor &in1, Tensor &in2, Tensor &out, TensorOp op)
{
	arm_compute::IFunction* func = nullptr;

	arm_compute::NEArithmeticAddition add;
	arm_compute::NEArithmeticSubtraction sub;
	arm_compute::NEPixelWiseMultiplication mul;

	switch(op)
	{
		case TensorOp::ADD:
			add.configure(&in1.ne.tensor, &in2.ne.tensor, &out.ne.tensor, ConvertPolicy::WRAP);
			func = &add;
			break;
		case TensorOp::SUB:
			sub.configure(&in1.ne.tensor, &in2.ne.tensor, &out.ne.tensor, ConvertPolicy::WRAP);
			func = &sub;
			break;
		case TensorOp::MUL:
			mul.configure(&in1.ne.tensor, &in2.ne.tensor, &out.ne.tensor, 1.f, ConvertPolicy::WRAP, RoundingPolicy::TO_ZERO);
			func = &mul;
			break;
		case TensorOp::LUT:
		case TensorOp::TRANSPOSE:
		default:
			ERROR("Unsupported tensor operation");
	}

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		func->run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::activation_layer(Tensor &in, Tensor &out, ActivationLayerInfo act_info)
{
	arm_compute::NEActivationLayer actl;
	actl.configure(&in.ne.tensor, &out.ne.tensor, act_info);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		actl.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::pooling_layer(Tensor &in, Tensor &out, PoolingLayerInfo pool_info)
{
	arm_compute::NEPoolingLayer pool;

	pool.configure(&in.ne.tensor, &out.ne.tensor, pool_info);
	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		pool.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::softmax_layer(Tensor &in, Tensor &out)
{
	arm_compute::NESoftmaxLayer sml;

	sml.configure(&in.ne.tensor, &out.ne.tensor);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		sml.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::normalization_layer(Tensor &a, Tensor &b, NormalizationLayerInfo norm_info)
{
	arm_compute::NENormalizationLayer norm;

	norm.configure(&a.ne.tensor, &b.ne.tensor, norm_info);
	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		norm.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::batchnormalization_layer(Tensor &a, Tensor &b, Tensor &mean, Tensor &var, Tensor &beta, Tensor &gamma, float epsilon)
{
	arm_compute::NEBatchNormalizationLayer norm;

	norm.configure(&a.ne.tensor, &b.ne.tensor, &mean.ne.tensor, &var.ne.tensor, &beta.ne.tensor, &gamma.ne.tensor, epsilon);
	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		norm.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::concatenate_depth(std::vector<std::unique_ptr<Tensor>>& a, Tensor &b)
{
	arm_compute::NEDepthConcatenate cd;

	std::vector<arm_compute::ITensor*> input;

	for(auto& temp : a)
	{
		input.push_back(&(temp->ne.tensor));
	}

	cd.configure(input, &b.ne.tensor);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		cd.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::convolution_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info,Tensor *rw)
{
    if(rw)
    {
        arm_compute::NEConvolutionLayerReshapeWeights convlrshape;
        convlrshape.configure(&w.ne.tensor, b ? &b->ne.tensor : nullptr, rw ? &rw->ne.tensor : nullptr, true);
        convlrshape.run();
    }

	arm_compute::NEConvolutionLayer convl;

	convl.configure(&in.ne.tensor, rw ? &rw->ne.tensor : &w.ne.tensor, b ? &b->ne.tensor : nullptr, &out.ne.tensor, conv_info,
            rw ? arm_compute::WeightsInfo(true, w.ne.tensor.info()->dimension(0), w.ne.tensor.info()->dimension(1), w.ne.tensor.info()->dimension(3)) : arm_compute::WeightsInfo());

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		convl.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::locally_connected_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info)
{
	arm_compute::NELocallyConnectedLayer lc;

	lc.configure(&in.ne.tensor, &w.ne.tensor, b ? &b->ne.tensor : nullptr, &out.ne.tensor, conv_info);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		lc.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		_timer.print_elapsed_ms();
	}
}

void NETarget::gemm_fp32(Tensor &a, Tensor &b, Tensor &c, Tensor &d, float alpha, float beta)
{
	arm_compute::NEGEMM gemm;
	gemm.configure(&a.ne.tensor,  &b.ne.tensor,  &c.ne.tensor, &d.ne.tensor, alpha, beta);

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		gemm.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::gemm_fp16( Image &a, Image &b, Image &c, Image &d, float alpha, float beta)
{
	arm_compute::NEGEMM hgemm;
	hgemm.configure(&a.ne.tensor,  &b.ne.tensor,  &c.ne.tensor, &d.ne.tensor, alpha, beta);

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		hgemm.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::interleave_4x4(Image &in, Image &out)
{
	arm_compute::NEGEMMInterleave4x4 inter;
	inter.configure(&in.ne.tensor,  &out.ne.tensor);

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		inter.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::transpose_1xW(Image &in, Image &out)
{
	arm_compute::NEGEMMTranspose1xW tr;
	tr.configure(&in.ne.tensor,  &out.ne.tensor);

	for(unsigned int j=0; j < _ctx.num_iterations(); ++j )
	{
		_timer.start();
		tr.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::transpose( Image &a, Image &b)
{
	arm_compute::NETranspose transpose;

	transpose.configure(&a.ne.tensor, &b.ne.tensor);

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		transpose.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}

void NETarget::fully_connected_layer(Tensor &input, Tensor &weight, Tensor *bias, Tensor &output, bool are_weights_reshaped)
{
	arm_compute::Tensor                              weight_reshaped;
	arm_compute::NEFullyConnectedLayer               fc;
	arm_compute::NEFullyConnectedLayerReshapeWeights reshape_weights;

	const arm_compute::Tensor *weight_to_use = &weight.ne.tensor;
	if(are_weights_reshaped)
	{
		weight_to_use = &weight_reshaped;

		// Note: the tensor weight_reshaped will be initialized within the configuration of NEFullyConnectedLayerReshapeWeights
		reshape_weights.configure(&weight.ne.tensor, &weight_reshaped, true, output.ne._shape.y() > 1);
	}

	fc.configure(&input.ne.tensor, weight_to_use, bias ? &bias->ne.tensor : nullptr, &output.ne.tensor, true, are_weights_reshaped);

	if(are_weights_reshaped)
	{
		// Reshape the weight before running convolution layer
		weight_reshaped.allocator()->allocate();
		reshape_weights.run();
	}

	for(unsigned int j = 0; j < _ctx.num_iterations(); ++j)
	{
		_timer.start();
		fc.run();
		_timer.stop();
		TestResults::report().add(name(), _timer.get_elapsed_ms());
		TestResults::report().add_pmu_counter( TestResults::PMU_Cycles, _timer.get_cycles() );
		_timer.print_elapsed_ms();
	}
}
