#include "mali_cv/RawImage.h"

#include "mali_cv/Error.h"
#include "mali_cv/File.h"
#include "mali_cv/Helper.h"

#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstring>
#include <map>
#include <random>
#include <ctime>

#ifdef ANDROID
#include <cstdlib>
#endif

using namespace mali_cv;

using create_ptr = void (RawImage::*)(const std::string &filename);

void RawImage::create_from(const std::string &filename)
{
	static std::map<std::string, create_ptr> create_functions;

	if(create_functions.empty())
	{
		create_functions["ppm"] = &RawImage::create_from_ppm;
		create_functions["pgm"] = &RawImage::create_from_pgm;
		create_functions["bin"] = &RawImage::create_from_bin;
		create_functions["txt"] = &RawImage::create_from_txt;
	}

	size_t found = filename.find_last_of('.');
	ERROR_ON_MSG( found == std::string::npos, "Can't find extension of filename '%s'", filename.c_str());
	std::string extension = filename.substr(found + 1);
	auto it = create_functions.find(extension);
	ERROR_ON_MSG( it == create_functions.end(), "Can't find create function for extension '%s'", extension.c_str());

	((*this).*(it->second))(filename);
}

using save_ptr = void (RawImage::*)(const std::string &filename);

void RawImage::save_to(const std::string &filename)
{
	static std::map<std::string, save_ptr> save_functions;

	if(save_functions.empty())
	{
		save_functions["pgm"] = &RawImage::save_to_pgm;
		save_functions["ppm"] = &RawImage::save_to_ppm;
		save_functions["bin"] = &RawImage::save_to_bin;
		save_functions["txt"] = &RawImage::save_to_txt;
	}

	size_t found = filename.find_last_of('.');
	ERROR_ON_MSG( found == std::string::npos, "Can't find extension of filename '%s'", filename.c_str());
	std::string extension = filename.substr(found + 1);
	auto it = save_functions.find(extension);
	ERROR_ON_MSG( it == save_functions.end(), "Can't find save function for extension '%s'", extension.c_str());

	((*this).*(it->second))(filename);
}

namespace
{
using interleave_ptr = void (*)(RawImage &, std::vector<RawImage *> &channels);
using deinterleave_ptr = void (*)(RawImage &, std::vector<RawImage *> &channels);

void interleave_RGB888(RawImage &, std::vector<RawImage *> &channels);
void deinterleave_RGB888(RawImage &, std::vector<RawImage *> &channels);
void interleave_RGBA8888(RawImage &, std::vector<RawImage *> &channels);
void deinterleave_RGBA8888(RawImage &, std::vector<RawImage *> &channels);

interleave_ptr get_interleave_function( Format format )
{
	static std::map<Format, interleave_ptr> convert_functions;

	if( convert_functions.empty() )
	{
		/* Initialize all the possible conversions to NULL */
		for(unsigned int a = 0; a < (uint)Format::NUM_FORMATS; a++)
			convert_functions[(Format)a] = nullptr;

		/* Register the available conversion functions */
		convert_functions[Format::RGB888] = interleave_RGB888;
		convert_functions[Format::RGBA8888] = interleave_RGBA8888;
	}
	return convert_functions[format];
}

deinterleave_ptr get_deinterleave_function( Format format )
{
	static std::map<Format, deinterleave_ptr> convert_functions;

	if( convert_functions.empty() )
	{
		/* Initialize all the possible conversions to NULL */
		for(unsigned int a = 0; a < (uint)Format::NUM_FORMATS; a++)
			convert_functions[(Format)a] = nullptr;

		/* Register the available conversion functions */
		convert_functions[Format::RGB888] = deinterleave_RGB888;
		convert_functions[Format::RGBA8888] = deinterleave_RGBA8888;
	}
	return convert_functions[format];
}

void interleave_RGB888(RawImage &out, std::vector<RawImage *> &channels)
{
	ERROR_ON( out.width() != channels[0]->width() );
	ERROR_ON( out.width() != channels[1]->width() );
	ERROR_ON( out.width() != channels[2]->width() );
	ERROR_ON( out.height() != channels[0]->height() );
	ERROR_ON( out.height() != channels[1]->height() );
	ERROR_ON( out.height() != channels[2]->height() );
	ERROR_ON( Format::LUMINANCE != channels[0]->format() );
	ERROR_ON( Format::LUMINANCE != channels[1]->format() );
	ERROR_ON( Format::LUMINANCE != channels[2]->format() );

	unsigned char *out_ptr = out.data();
	unsigned char *red_ptr = channels[0]->data();
	unsigned char *green_ptr = channels[1]->data();
	unsigned char *blue_ptr = channels[2]->data();

	for(uint p = 0; p < out.width() * out.height(); p++)
	{
		out_ptr[0] = *red_ptr;
		out_ptr[1] = *green_ptr;
		out_ptr[2] = *blue_ptr;
		out_ptr += 3;
		red_ptr++;
		green_ptr++;
		blue_ptr++;
	}
}

void interleave_RGBA8888(RawImage &out, std::vector<RawImage *> &channels)
{
	ERROR_ON( out.width() != channels[0]->width() );
	ERROR_ON( out.width() != channels[1]->width() );
	ERROR_ON( out.width() != channels[2]->width() );
	ERROR_ON( out.width() != channels[3]->width() );
	ERROR_ON( out.height() != channels[0]->height() );
	ERROR_ON( out.height() != channels[1]->height() );
	ERROR_ON( out.height() != channels[2]->height() );
	ERROR_ON( out.height() != channels[3]->height() );
	ERROR_ON( Format::LUMINANCE != channels[0]->format() );
	ERROR_ON( Format::LUMINANCE != channels[1]->format() );
	ERROR_ON( Format::LUMINANCE != channels[2]->format() );
	ERROR_ON( Format::LUMINANCE != channels[3]->format() );

	unsigned char *out_ptr = out.data();
	unsigned char *red_ptr = channels[0]->data();
	unsigned char *green_ptr = channels[1]->data();
	unsigned char *blue_ptr = channels[2]->data();
	unsigned char *alpha_ptr = channels[3]->data();

	for(uint p = 0; p < out.width() * out.height(); p++)
	{
		out_ptr[0] = *red_ptr;
		out_ptr[1] = *green_ptr;
		out_ptr[2] = *blue_ptr;
		out_ptr[3] = *alpha_ptr;
		out_ptr += 4;
		red_ptr++;
		green_ptr++;
		blue_ptr++;
		alpha_ptr++;
	}
}

void deinterleave_RGB888(RawImage &src, std::vector<RawImage *> &channels)
{
	for(auto &c : channels )
	{
		c->create( src.width(), src.height(), Format::LUMINANCE);
	}

	unsigned char *src_ptr = src.data();
	unsigned char *red_ptr = channels[0]->data();
	unsigned char *green_ptr = channels[1]->data();
	unsigned char *blue_ptr = channels[2]->data();


	for(uint p = 0; p < src.width() * src.height(); p++)
	{
		*red_ptr = src_ptr[0];
		*green_ptr = src_ptr[1];
		*blue_ptr = src_ptr[2];
		src_ptr += 3;
		red_ptr++;
		green_ptr++;
		blue_ptr++;
	}
}

void deinterleave_RGBA8888(RawImage &src, std::vector<RawImage *> &channels)
{
	for(auto &c : channels )
	{
		c->create( src.width(), src.height(), Format::LUMINANCE);
	}

	unsigned char *src_ptr = src.data();
	unsigned char *red_ptr = channels[0]->data();
	unsigned char *green_ptr = channels[1]->data();
	unsigned char *blue_ptr = channels[2]->data();
	unsigned char *alpha_ptr = channels[3]->data();

	for(uint p = 0; p < src.width() * src.height(); p++)
	{
		*red_ptr = src_ptr[0];
		*green_ptr = src_ptr[1];
		*blue_ptr = src_ptr[2];
		*alpha_ptr = src_ptr[3];
		src_ptr += 4;
		red_ptr++;
		green_ptr++;
		blue_ptr++;
		alpha_ptr++;
	}
}
}

void RawImage::deinterleave( std::vector<RawImage *> &channels_out)
{
	ERROR_ON( num_channels() < 2 );
	ERROR_ON( channels_out.size() != num_channels());
	deinterleave_ptr func = get_deinterleave_function( format() );
	ERROR_ON_MSG( !func, "There is no deinterleave function for this image format");
	(*func)(*this, channels_out);
}

void RawImage::interleave( std::vector<RawImage *> &channels_in)
{
	ERROR_ON( num_channels() < 2 );
	ERROR_ON( channels_in.size() != num_channels());
	interleave_ptr func = get_interleave_function( format() );
	ERROR_ON_MSG( !func, "There is no interleave function for this image format");
	(*func)(*this, channels_in);
}

namespace
{
using convert_ptr = void (*)(const unsigned char *&, unsigned char *&);
using convert_all_ptr = void (*)(const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding);

void convert_RGB888_to_RGBA8888( const unsigned char *&src, unsigned char *&dst);
void convert_RGBA8888_to_RGB888( const unsigned char *&src, unsigned char *&dst);
void convert_YUV444_to_RGB888( const unsigned char *&src, unsigned char *&dst);
void convert_RGB888_to_YUV444( const unsigned char *&src, unsigned char *&dst);
void convert_RGBA8888_to_YUV444( const unsigned char *&src, unsigned char *&dst);
void convert_YUV444_to_RGBA8888( const unsigned char *&src, unsigned char *&dst);
void convert_RGBA_FLOAT_to_RGB888( const unsigned char *&src, unsigned char *&dst);
void convert_RGB888_to_RGBA_FLOAT( const unsigned char *&src, unsigned char *&dst);
void convert_RGBA16_to_RGB888( const unsigned char *&src, unsigned char *&dst);
void convert_RGB888_to_RGBA16( const unsigned char *&src, unsigned char *&dst);
void convert_RGB888_to_LUMINANCE( const unsigned char *&src, unsigned char *&dst);
void convert_RGB888_to_LUMINANCE16( const unsigned char *&src, unsigned char *&dst);
void convert_RGB888_to_U16( const unsigned char *&src, unsigned char *&dst);
void convert_RGB888_to_U8( const unsigned char *&src, unsigned char *&dst);

void convert_RGBA8888_to_LUMINANCE( const unsigned char *&src, unsigned char *&dst);
void convert_LUMINANCE_to_RGB888( const unsigned char *&src, unsigned char *&dst);
void convert_LUMINANCE_to_LUMINANCE16( const unsigned char *&src, unsigned char *&dst);
void convert_LUMINANCE_to_FP32C1( const unsigned char *&src, unsigned char *&dst);
void convert_FP32C1_to_FP16C1( const unsigned char *&src, unsigned char *&dst);
void convert_FP16C1_to_FP32C1( const unsigned char *&src, unsigned char *&dst);
void convert_LUMINANCE16_to_RGB888( const unsigned char *&src, unsigned char *&dst);

void convert_all_RGB888_to_NV12( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding);
void convert_all_RGB888_to_NV21( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding);
void convert_all_RGB888_to_IYUV( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding);
void convert_all_RGB888_to_YUYV422( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding);
void convert_all_RGB888_to_UYVY( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding);

template<convert_ptr func>
void for_each_pixel( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding)
{
	for(size_t h = 0; h < height; h++)
	{
		for( size_t w = 0; w < width; w++)
		{
			func(src, dst);
		}
		src += src_row_padding;
	}
}

convert_all_ptr getconvert_function(Format src_format, Format dst_format)
{
	static std::map<Format, std::map<Format, convert_all_ptr> > convert_functions;

	if( convert_functions.empty() )
	{
		/* Initialize all the possible conversions to NULL */
		for(unsigned int a = 0; a < (uint)Format::NUM_FORMATS; a++)
			for(unsigned int b = 0; b < (uint)Format::NUM_FORMATS; b++)
				convert_functions[(Format)a][(Format)b] = nullptr;

		/* Register the available conversion functions */
		convert_functions[Format::RGB888][Format::RGBA8888] = for_each_pixel<convert_RGB888_to_RGBA8888>;
		convert_functions[Format::RGBA8888][Format::RGB888] = for_each_pixel<convert_RGBA8888_to_RGB888>;
		convert_functions[Format::YUV444][Format::RGB888] = for_each_pixel<convert_YUV444_to_RGB888>;
		convert_functions[Format::RGB888][Format::YUV444] = for_each_pixel<convert_RGB888_to_YUV444>;
		convert_functions[Format::RGBA8888][Format::YUV444] = for_each_pixel<convert_RGBA8888_to_YUV444>;
		convert_functions[Format::YUV444][Format::RGBA8888] = for_each_pixel<convert_YUV444_to_RGBA8888>;
		convert_functions[Format::RGBA_FLOAT][Format::RGB888] = for_each_pixel<convert_RGBA_FLOAT_to_RGB888>;
		convert_functions[Format::RGB888][Format::RGBA_FLOAT] = for_each_pixel<convert_RGB888_to_RGBA_FLOAT>;
		convert_functions[Format::RGBA16][Format::RGB888] = for_each_pixel<convert_RGBA16_to_RGB888>;
		convert_functions[Format::RGB888][Format::RGBA16] = for_each_pixel<convert_RGB888_to_RGBA16>;
		convert_functions[Format::RGB888][Format::LUMINANCE] = for_each_pixel<convert_RGB888_to_LUMINANCE>;
		convert_functions[Format::RGB888][Format::LUMINANCE16] = for_each_pixel<convert_RGB888_to_LUMINANCE16>;
		convert_functions[Format::RGB888][Format::U16] = for_each_pixel<convert_RGB888_to_U16>;
		convert_functions[Format::RGB888][Format::U8] = for_each_pixel<convert_RGB888_to_U8>;
		convert_functions[Format::RGBA8888][Format::LUMINANCE] = for_each_pixel<convert_RGBA8888_to_LUMINANCE>;
		convert_functions[Format::LUMINANCE][Format::RGB888] = for_each_pixel<convert_LUMINANCE_to_RGB888>;
		convert_functions[Format::LUMINANCE][Format::LUMINANCE16] = for_each_pixel<convert_LUMINANCE_to_LUMINANCE16>;
		convert_functions[Format::LUMINANCE][Format::FP32C1] = for_each_pixel<convert_LUMINANCE_to_FP32C1>;
		convert_functions[Format::FP32C1][Format::FP16C1] = for_each_pixel<convert_FP32C1_to_FP16C1>;
		convert_functions[Format::FP16C1][Format::FP32C1] = for_each_pixel<convert_FP16C1_to_FP32C1>;
		convert_functions[Format::LUMINANCE16][Format::RGB888] = for_each_pixel<convert_LUMINANCE16_to_RGB888>;
		convert_functions[Format::RGB888][Format::NV12] = convert_all_RGB888_to_NV12;
		convert_functions[Format::RGB888][Format::NV21] = convert_all_RGB888_to_NV21;
		convert_functions[Format::RGB888][Format::IYUV] = convert_all_RGB888_to_IYUV;
		convert_functions[Format::RGB888][Format::YUYV422] = convert_all_RGB888_to_YUYV422;
		convert_functions[Format::RGB888][Format::UYVY] = convert_all_RGB888_to_UYVY;
	}

	return convert_functions[src_format][dst_format];
}

void convertAll( Format src_format, Format dst_format, const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding = 0)
{
	if( dst_format == Format::SAME_AS_SRC )
	{
		size_t row_data_size = width * RawImage::get_pixel_size(src_format);
		for(size_t h = 0; h < height; h++)
		{
			memcpy(dst, src, row_data_size);
			dst += row_data_size;
			src += src_row_padding + row_data_size;
		}
	}
	else
	{
		convert_all_ptr func = getconvert_function( src_format, dst_format);
		if( func == nullptr )
		{
			ERROR("Conversion from %s to %s is not supported by this library\n", RawImageBase::format_str( src_format ), RawImageBase::format_str( dst_format ));
		}
		func(src, dst, width, height, src_row_padding);
	}
}

void convert_RGBA16_to_RGB888( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = reinterpret_cast<const short*>(src)[0];
	dst[1] = reinterpret_cast<const short*>(src)[1];
	dst[2] = reinterpret_cast<const short*>(src)[2];
	dst += 3;
	src += sizeof(short) * 4;
}

void convert_RGB888_to_RGBA16( const unsigned char *&src, unsigned char *&dst)
{
	reinterpret_cast<short*>(dst)[0] = src[0];
	reinterpret_cast<short*>(dst)[1] = src[1];
	reinterpret_cast<short*>(dst)[2] = src[2];
	reinterpret_cast<short*>(dst)[3] = 255;
	dst += sizeof(short) * 4;
	src += 3;
}

void convert_RGBA_FLOAT_to_RGB888( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = reinterpret_cast<const float*>(src)[0];
	dst[1] = reinterpret_cast<const float*>(src)[1];
	dst[2] = reinterpret_cast<const float*>(src)[2];
	dst += 3;
	src += sizeof(float) * 4;
}

void convert_RGB888_to_RGBA_FLOAT( const unsigned char *&src, unsigned char *&dst)
{
	reinterpret_cast<float*>(dst)[0] = src[0];
	reinterpret_cast<float*>(dst)[1] = src[1];
	reinterpret_cast<float*>(dst)[2] = src[2];
	reinterpret_cast<float*>(dst)[3] = 255;
	dst += sizeof(float) * 4;
	src += 3;

}

inline uint8_t clamp( int v )
{
	return std::max(0, std::min(v, 255));
}

#if 0
void yuv2rgb_bt709(uint8_t y, uint8_t cb, uint8_t cr, uint8_t *r, uint8_t *g, uint8_t *b)
{
	// R'= Y' + 0.0000*U + 1.5748*V
	// G'= Y' - 0.1873*U - 0.4681*V
	// B'= Y' + 1.8556*U + 0.0000*V
	float f_y = (float)y;
	float f_u = (float)cb - (float)128;
	float f_v = (float)cr - (float)128;

	float f_r = f_y + 0.0000f * f_u + 1.5748f * f_v;
	float f_g = f_y - 0.1873f * f_u - 0.4681f * f_v;
	float f_b = f_y + 1.8556f * f_u + 0.0000f * f_v;

	int i_r = (int)f_r;
	int i_g = (int)f_g;
	int i_b = (int)f_b;

	*r = clamp(i_r);
	*g = clamp(i_g);
	*b = clamp(i_b);
}
#endif

void rgb2yuv_bt709(uint8_t r, uint8_t g, uint8_t b, uint8_t *y, uint8_t *cb, uint8_t *cr)
{
	// Y'= 0.2126*R' + 0.7152*G' + 0.0722*B'
	// U'=-0.1146*R' - 0.3854*G' + 0.5000*B'
	// V'= 0.5000*R' - 0.4542*G' - 0.0458*B'

	double f_y = 0 + 0.2126f * (double)r + 0.7152f * (double)g + 0.0722f * (double)b;
	double f_u = 0 - 0.1146f * (double)r - 0.3854f * (double)g + 0.5000f * (double)b;
	double f_v = 0 + 0.5000f * (double)r - 0.4542f * (double)g - 0.0458f * (double)b;

	int i_y = f_y;
	int i_u = f_u + 128;
	int i_v = f_v + 128;

	*y	= clamp(i_y);
	*cb = clamp(i_u);
	*cr = clamp(i_v);
}

void convert_all_RGB888_to_NV12( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding)
{
	ERROR_ON_MSG( width % 2 != 0, "Width has to be a multiple of 2");
	ERROR_ON_MSG( height % 2 != 0, "Height has to be a multiple of 2");

	for(size_t h = 0; h < height; h += 2)
	{
		for( size_t w = 0; w < width; w += 2)
		{
			uint8_t y[4], cb[4], cr[4];
			int idx    = (h    ) * width + w;
			int idx_1  = (h + 1) * width + w;
			int uv_idx = width * height + h/2 * width  + w;
			int src_idx =	(h	 ) * (width * 3 + src_row_padding) + w * 3;
			int src_idx_1 = (h + 1) * (width * 3 + src_row_padding) + w * 3;

			rgb2yuv_bt709(src[src_idx], src[src_idx + 1], src[src_idx + 2], &y[0], &cb[0], &cr[0]);
			src_idx += 3;
			rgb2yuv_bt709(src[src_idx], src[src_idx + 1], src[src_idx + 2], &y[1], &cb[1], &cr[1]);
			rgb2yuv_bt709(src[src_idx_1], src[src_idx_1 + 1], src[src_idx_1 + 2], &y[2], &cb[2], &cr[2]);
			src_idx_1 += 3;
			rgb2yuv_bt709(src[src_idx_1], src[src_idx_1 + 1], src[src_idx_1 + 2], &y[3], &cb[3], &cr[3]);

			dst[idx		 ] = y[0];
			dst[idx   + 1] = y[1];
			dst[idx_1	 ] = y[2];
			dst[idx_1 + 1] = y[3];

			dst[uv_idx	  ] =(cb[0] + cb[1] + cb[2] + cb[3]) / 4;
			dst[uv_idx + 1] = (cr[0] + cr[1] + cr[2] + cr[3]) / 4;
		}
	}
}

void convert_all_RGB888_to_NV21( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding)
{
	ERROR_ON_MSG( width % 2 != 0, "Width has to be a multiple of 2");
	ERROR_ON_MSG( height % 2 != 0, "Height has to be a multiple of 2");

	for(size_t h = 0; h < height; h += 2)
	{
		for( size_t w = 0; w < width; w += 2)
		{
			uint8_t y[4], cb[4], cr[4];
			int idx    = (h    ) * width + w;
			int idx_1  = (h + 1) * width + w;
			int uv_idx = width * height + h * width / 2 + w;
			int src_idx =	(h	 ) * (width * 3 + src_row_padding) + w * 3;
			int src_idx_1 = (h + 1) * (width * 3 + src_row_padding) + w * 3;

			rgb2yuv_bt709(src[src_idx], src[src_idx + 1], src[src_idx + 2], &y[0], &cb[0], &cr[0]);
			src_idx += 3;
			rgb2yuv_bt709(src[src_idx], src[src_idx + 1], src[src_idx + 2], &y[1], &cb[1], &cr[1]);
			rgb2yuv_bt709(src[src_idx_1], src[src_idx_1 + 1], src[src_idx_1 + 2], &y[2], &cb[2], &cr[2]);
			src_idx_1 += 3;
			rgb2yuv_bt709(src[src_idx_1], src[src_idx_1 + 1], src[src_idx_1 + 2], &y[3], &cb[3], &cr[3]);

			dst[idx		 ] = y[0];
			dst[idx   + 1] = y[1];
			dst[idx_1	 ] = y[2];
			dst[idx_1 + 1] = y[3];

			dst[uv_idx	  ] = (cr[0] + cr[1] + cr[2] + cr[3]) / 4;
			dst[uv_idx + 1] = (cb[0] + cb[1] + cb[2] + cb[3]) / 4;
		}
	}
}

void convert_all_RGB888_to_IYUV( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding)
{
	ERROR_ON_MSG( width % 2 != 0, "Width has to be a multiple of 2");
	ERROR_ON_MSG( height % 2 != 0, "Height has to be a multiple of 2");

	for(size_t h = 0; h < height; h += 2)
	{
		for( size_t w = 0; w < width; w += 2)
		{
			uint8_t y[4], cb[4], cr[4];
			int idx    = (h    ) * width + w;
			int idx_1  = (h + 1) * width + w;
			int u_idx  = width * height + h * width / 4 + w / 2 ;
			int v_idx  = width * height / 4 + u_idx;
			int src_idx =	(h	 ) * (width * 3 + src_row_padding) + w * 3;
			int src_idx_1 = (h + 1) * (width * 3 + src_row_padding) + w * 3;

			rgb2yuv_bt709(src[src_idx], src[src_idx + 1], src[src_idx + 2], &y[0], &cb[0], &cr[0]);
			src_idx += 3;
			rgb2yuv_bt709(src[src_idx], src[src_idx + 1], src[src_idx + 2], &y[1], &cb[1], &cr[1]);
			rgb2yuv_bt709(src[src_idx_1], src[src_idx_1 + 1], src[src_idx_1 + 2], &y[2], &cb[2], &cr[2]);
			src_idx_1 += 3;
			rgb2yuv_bt709(src[src_idx_1], src[src_idx_1 + 1], src[src_idx_1 + 2], &y[3], &cb[3], &cr[3]);

			dst[idx		 ] = y[0];
			dst[idx   + 1] = y[1];
			dst[idx_1	 ] = y[2];
			dst[idx_1 + 1] = y[3];

			dst[u_idx] = (cb[0] + cb[1] + cb[2] + cb[3]) / 4;
			dst[v_idx] = (cr[0] + cr[1] + cr[2] + cr[3]) / 4;
		}
	}
}

void convert_all_RGB888_to_YUYV422( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding)
{
	ERROR_ON_MSG( width % 2 != 0, "Width has to be a multiple of 2");

	for(size_t h = 0; h < height; h++)
	{
		for( size_t w = 0; w < width; w += 2)
		{
			uint8_t y[2], cb[2], cr[2];
			int idx    = 2 * (h * width + w);
			int src_idx = h * (width * 3 + src_row_padding) + w * 3;

			rgb2yuv_bt709(src[src_idx], src[src_idx + 1], src[src_idx + 2], &y[0], &cb[0], &cr[0]);
			src_idx += 3;
			rgb2yuv_bt709(src[src_idx], src[src_idx + 1], src[src_idx + 2], &y[1], &cb[1], &cr[1]);

			dst[idx    ] = y[0];
			dst[idx + 1] = (cb[0] + cb[1]) / 2;
			dst[idx + 2] = y[1];
			dst[idx + 3] = (cr[0] + cr[1]) / 2;
		}
	}
}

void convert_all_RGB888_to_UYVY( const unsigned char *src, unsigned char *dst, size_t width, size_t height, size_t src_row_padding)
{
	ERROR_ON_MSG( width % 2 != 0, "Width has to be a multiple of 2");

	for(size_t h = 0; h < height; h++)
	{
		for( size_t w = 0; w < width; w += 2)
		{
			uint8_t y[2], cb[2], cr[2];
			int idx    = 2 * (h * width + w);
			int src_idx = h * (width * 3 + src_row_padding) + w * 3;

			rgb2yuv_bt709(src[src_idx], src[src_idx + 1], src[src_idx + 2], &y[0], &cb[0], &cr[0]);
			src_idx += 3;
			rgb2yuv_bt709(src[src_idx], src[src_idx + 1], src[src_idx + 2], &y[1], &cb[1], &cr[1]);

			dst[idx    ] = (cb[0] + cb[1]) / 2;
			dst[idx + 1] = y[0];
			dst[idx + 2] = (cr[0] + cr[1]) / 2;
			dst[idx + 3] = y[1];
		}
	}
}

void convert_RGB888_to_RGBA8888( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = src[0];
	dst[1] = src[1];
	dst[2] = src[2];
	dst[3] = 255;
	dst += 4;
	src += 3;
}

void convert_RGBA8888_to_RGB888( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = src[0];
	dst[1] = src[1];
	dst[2] = src[2];
	dst += 3;
	src += 4;
}

void convert_YUV444_to_RGB888( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = 1.00000f * src[0] +  0.00000f * src[1] + 1.00000f  * src[2];
	dst[1] = 1.00000f * src[0] + -0.19421f * src[1] + -0.50937f * src[2];
	dst[2] = 1.00000f * src[0] +  1.00000f * src[1] + 0.00000f * src[2];
	dst += 3;
	src += 3;
}

void convert_RGB888_to_YUV444( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = 0.29900f * src[0] +  0.58700f * src[1] + 0.11400f  * src[2];
	dst[1] = -0.29900f * src[0] + -0.58700f * src[1] + 0.88600f * src[2];
	dst[2] = 0.70100f * src[0] +  -0.5870f * src[1] + -0.11400 * src[2];
	dst += 3;
	src += 3;
}

void convert_RGBA8888_to_YUV444( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = 0.29900f * src[0] +  0.58700f * src[1] + 0.11400f  * src[2];
	dst[1] = -0.29900f * src[0] + -0.58700f * src[1] + 0.88600f * src[2];
	dst[2] = 0.70100f * src[0] +  -0.5870f * src[1] + -0.11400 * src[2];
	dst += 3;
	src += 4;
}

void convert_YUV444_to_RGBA8888( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = 1.00000f * src[0] +  0.00000f * src[1] + 1.00000f	* src[2];
	dst[1] = 1.00000f * src[0] + -0.19421f * src[1] + -0.50937f * src[2];
	dst[2] = 1.00000f * src[0] +  1.00000f * src[1] + 0.00000f * src[2];
	dst[3] = 255;
	dst += 4;
	src += 3;
}

void convert_RGB888_to_LUMINANCE( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = 0.2126f * src[0] +  0.7152f * src[1] +  0.0722f  * src[2];
	dst += 1;
	src += 3;
}

void convert_RGB888_to_U16( const unsigned char *&src, unsigned char *&dst)
{
	reinterpret_cast<uint16_t*>(dst)[0] = 0.2126f * src[0] +  0.7152f * src[1] +  0.0722f  * src[2];
	dst += 2;
	src += 3;
}

void convert_RGB888_to_LUMINANCE16( const unsigned char *&src, unsigned char *&dst)
{
	((int16_t*)dst)[0] = 0.2126f * src[0] +  0.7152f * src[1] +  0.0722f  * src[2];
	dst += 2;
	src += 3;
}
void convert_RGBA8888_to_LUMINANCE( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = (0.2126f * src[0] +  0.7152f * src[1] +  0.0722f  * src[2]) * src[3] / 255.0f;
	dst += 1;
	src += 4;
}

void convert_RGB888_to_U8( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = 0.2126f * src[0] +  0.7152f * src[1] +  0.0722f  * src[2];
	dst += 1;
	src += 3;
}

void convert_LUMINANCE_to_RGB888( const unsigned char *&src, unsigned char *&dst)
{
	dst[0] = src[0];
	dst[1] = src[0];
	dst[2] = src[0];
	dst += 3;
	src += 1;
}

void convert_LUMINANCE_to_LUMINANCE16( const unsigned char *&src, unsigned char *&dst)
{
	reinterpret_cast<int16_t*>(dst)[0] = src[0];
	src += 1;
	dst += sizeof(int16_t);
}

void convert_LUMINANCE_to_FP32C1( const unsigned char *&src, unsigned char *&dst)
{
	reinterpret_cast<float*>(dst)[0] = src[0];
	src += 1;
	dst += sizeof(float);
}

void convert_FP32C1_to_FP16C1( const unsigned char *&src, unsigned char *&dst)
{
	reinterpret_cast<mali_cv::float16_t*>(dst)[0] = reinterpret_cast<const float*>(src)[0];
	src += sizeof(float);
	dst += sizeof(mali_cv::float16_t);
}

void convert_FP16C1_to_FP32C1( const unsigned char *&src, unsigned char *&dst)
{
	reinterpret_cast<float*>(dst)[0] = reinterpret_cast<const mali_cv::float16_t*>(src)[0];
	src += sizeof(mali_cv::float16_t);
	dst += sizeof(float);
}

void convert_LUMINANCE16_to_RGB888( const unsigned char *&src, unsigned char *&dst)
{
	uint8_t val = clamp(*reinterpret_cast<const int16_t*>(src));

	dst[0] = val;
	dst[1] = val;
	dst[2] = val;
	dst += 3;
	src += 2;
}
}

void RawImage::copy( size_t width, size_t height, const unsigned char *src, Format src_format, Format dst_format, size_t src_row_padding)
{
	if( dst_format == src_format)
		dst_format = Format::SAME_AS_SRC;

	create(width, height, dst_format == Format::SAME_AS_SRC ? src_format : dst_format);
	if( src_row_padding == 0 && dst_format == Format::SAME_AS_SRC)
	{
		memcpy(_data.get(), src, width * height * get_pixel_size(src_format));
	}
	else
	{
		convertAll(src_format, dst_format, src, _data.get(), width, height, src_row_padding);
	}
}

RawImage::RawImage(): RawImageBase(), _data()
{
}

RawImage::RawImage( size_t width, size_t height, Format format): RawImage()
{
	create( width, height, format );
}

RawImage::RawImage( const RawImage &src, CopyData copy_data ): RawImage()
{
	if( copy_data == CopyData::YES )
		copy_from(src);
	else
		allocate_from_image(src);
}

unsigned char *RawImage::data()const
{
	return _data.get();
}

void RawImage::random_init(float min, float max)
{
	ERROR_ON( max > 255.0f );

	uint seed = random_seed();
	std::mt19937 rng( seed );

	if( Format::FP32C1 == format() || Format::FP32C2 == format() )
	{
		std::uniform_real_distribution<float> gen( min, max ); // uniform, unbiased between min and max

		const auto data = reinterpret_cast<float*>(_data.get());

		for( uint i = 0; i < num_pixels() ; i++ )
		{
			data[i] = gen(rng);
		}
	}
	else if( Format::FP16C1 == format() || Format::FP16C2 == format() )
	{
		std::uniform_real_distribution<float> gen( min, max ); // uniform, unbiased between min and max

		const auto data = reinterpret_cast<float16_t*>(_data.get());

		for( uint i = 0; i < num_pixels() ; i++ )
		{
			data[i] = static_cast<float16_t>(gen(rng));
		}
	}
	else
	{
		ERROR_ON( min < 0.0f );

		std::uniform_int_distribution<int> gen( (int)min, (int)max ); // uniform, unbiased between min and max

		switch(format())
		{
			case Format::S16:
			{
				int16_t *data = reinterpret_cast<int16_t*>(_data.get());
				unsigned int data_elements = num_pixels();

				for( size_t i = 0; i < data_elements; ++i )
				{
					data[i] = gen(rng);
				}
				break;
			}
			default:
			{
				unsigned char *data = _data.get();
				unsigned int data_elements = num_pixels() * pixel_size();

				for( size_t i = 0; i < data_elements; ++i )
				{
					data[i] = gen(rng);
				}
				break;
			}
		}
	}
}

void RawImage::create_diff( const RawImage &r, float amplify )
{
	ERROR_ON( r.width() != width() );
	ERROR_ON( r.height() != height() );
	ERROR_ON( r.format() != format() );

	unsigned char *a = r._data.get();
	unsigned char *b = _data.get();
	for( uint y = 0; y < height() ; y++ )
	{
		for( uint x = 0; x < width() ; x++ )
		{
			for(uint c = 0; c < pixel_size(); c++)
			{
				int delta = ((int) * a) - ((int) * b);
				delta += 128;
				delta *= amplify;

				if( delta < 0 ) delta = 0;
				if( delta > 255 ) delta = 255;

				*b = delta;

				a++;
				b++;
			}
		}
	}
}

void RawImage::create( size_t w, size_t h, Format fmt )
{
	if( _data )
	{
		ERROR_ON_MSG( width() != w, "RawImage already allocated but width doesn't match" );
		ERROR_ON_MSG( height() != h, "RawImage already allocated but height doesn't match" );
		ERROR_ON_MSG( format() != fmt, "RawImage already allocated but format doesn't match" );
	}
	else
	{
		RawImageBase::create(w, h, fmt);
		_data = std::unique_ptr<unsigned char[]>(new unsigned char[num_pixels() * pixel_size()]);
		ERROR_ON_MSG(!_data, "Failed to allocate memory for the image");
	}
}
void RawImage::allocate_from_image( const RawImage &src )
{
	create(src.width(), src.height(), src.format());
}

void RawImage::copy_from( const RawImage &src, Format fmt )
{
	copy( src.width(), src.height(), src._data.get(), src.format(), fmt );
}

void RawImage::save_to_txt( const std::string &filename )
{
	File fpout (filename, "w");
	fpout.printf("%zu %zu %d\n", width(), height(), (int)format());
	fpout.write(_data.get(), num_pixels() * pixel_size());
}

void RawImage::create_from_txt( const std::string &filename )
{
	int format;
	size_t width, height;

	File fp(filename, "r");
	fp.scanf( "%zu %zu %d\n", &width, &height, &format);
	create( width, height, (Format) format );
	fp.read(_data.get(), num_pixels() * pixel_size());
}

void RawImage::create_from_bin( const std::string &filename )
{
	int format;
	size_t width, height;

	File fp (filename, "rb");
	fp.scanf( "%zu %zu %d\n", &width, &height, &format);
	create( width, height, (Format) format );
	fp.read(_data.get(), num_pixels() * pixel_size());
}

void RawImage::save_to_bin( const std::string &filename )
{
	File fpout(filename, "wb");
	fpout.printf("%zu %zu %d\n", width(), height(), (int)format());
	fpout.write(_data.get(), num_pixels() * pixel_size());
}

//http://stackoverflow.com/questions/2693631/read-ppm-file-and-store-it-in-an-array-coded-with-c
void RawImage::create_from_ppm( const std::string &filename )
{
	int w, h;
	char buff[2];
	int c, rgb_comp_color;

	File fp(filename, "rb");
	fp.read(buff, sizeof(buff));

	ERROR_ON_MSG(!(buff[0] == 'P' && buff[1] == '6'), "Invalid image format");

	while (fp.getc() != '\n') ;
	c = fp.getc();
	while (c == '#')
	{
		while (fp.getc() != '\n') ;
		c = fp.getc();
	}
	fp.ungetc(c);

	ERROR_ON_MSG(fscanf(fp.fd(), "%d %d", &w, &h) != 2, "Error parsing image size");
	ERROR_ON_MSG(fscanf(fp.fd(), "%d", &rgb_comp_color) != 1, "Invalid rgb component");
	ERROR_ON_MSG(rgb_comp_color != 255, "'%s' does not have 8-bits compoents", filename.c_str());

	while (fp.getc() != '\n') ;

	create( w, h, Format::RGB888 );
	fp.read(_data.get(), num_pixels() * pixel_size());
}

void RawImage::create_from_pgm( const std::string &filename)
{
	int w, h;
	char buff[2];
	int c, grey_comp_color;

	File fp(filename, "rb");
	fp.read(buff, sizeof(buff));

	ERROR_ON_MSG(!(buff[0] == 'P' && buff[1] == '5'), "Invalid image format");
	while (fp.getc() != '\n') ;
	c = fp.getc();
	while (c == '#')
	{
		while (fp.getc() != '\n') ;
		c = fp.getc();
	}
	fp.ungetc(c);

	ERROR_ON_MSG(fscanf(fp.fd(), "%d %d", &w, &h) != 2, "Error parsing image size");
	ERROR_ON_MSG(fscanf(fp.fd(), "%d", &grey_comp_color) != 1, "Invalid number of components");
	ERROR_ON_MSG(grey_comp_color != 255, "'%s' does not have 8-bits components", filename.c_str());

	while (fp.getc() != '\n') ;

	create( w, h, Format::LUMINANCE );
	fp.read(_data.get(), num_pixels() * pixel_size());
}

void RawImage::save_to_ppm( const std::string &filename )
{
	ERROR_ON_MSG( format() != Format::RGB888, "Only RGB images can be saved as PPM");

	File fpout(filename, "wb");
	fpout.printf("P6\n%d %d 255\n", (int)width(), (int)height());
	fpout.write(_data.get(), num_pixels() * pixel_size());
}

void RawImage::save_to_pgm( const std::string &filename )
{
	ERROR_ON_MSG( format() != Format::LUMINANCE, "Only LUMINANCE images can be saved as PGM");

	File fpout(filename, "wb");
	fpout.printf("P5\n%d %d 255\n", (int)width(), (int)height());
	fpout.write(_data.get(), num_pixels() * pixel_size());
}

#if 0
void RawImage::createFromBMP( const char *filename )
{
	FILE *fp;
	int filelen;
	int res;
	int i;
	unsigned char *px;
	assert( header == NULL );
	assert( data == NULL );

	// Open BMP file
	fp = fopen(filename, "rb");

	// Skip the signature BM
	fseek(fp, 2, SEEK_SET);

	// Get the file size lenght
	res = fread(&filelen, 1, 4, fp);


	fseek(fp, 10, SEEK_SET);

	// Get the header lenght (also it is the starting address of the byte where the bitmap image data can be found.
	res = fread(&headlen, 1, 4, fp);

	// Allocate mermory for the header
	header = new unsigned char[headlen];

	// Come back to the start of file
	fseek(fp, 0, SEEK_SET);

	// Get the header
	res = fread(header, 1, headlen, fp);

	// size of image RGB 888
	data_size = filelen - headlen;

	// Allocate memory for the image
	data = new unsigned char[data_size];
	res = fread(data, 1, data_size, fp);


	// Skip to the width image
	fseek(fp, 0x12, SEEK_SET);

	// Get the width image
	res = fread(&width, 1, 4, fp);

	// Skip to height image
	fseek(fp, 0x16, SEEK_SET);

	// Get height image
	res = fread(&height, 1, 4, fp);

	// Skip to image data
	fseek(fp, headlen, SEEK_SET);


	// Work only without padding
	for(i = 0; i < height; i++)
	{

		px = (unsigned char *)(data + (height - 1 - i) * width * 3);

		res = fread(px, 1, width * 3, fp);

	}



	fclose(fp);
	printf("Loaded '%s' Resolution: %dx%d\n", filename, width, height);
}

void RawImage::saveToBMP( const char *filename )
{
	FILE *fpout;

	assert( header );

	fpout = fopen(filename, "wb");
	fwrite(header, 1, headlen, fpout);
	fwrite(data, 1, data_size, fpout);
	fclose(fpout);
}
#endif
