#include "mali_cv/RawImageBase.h"

#include "mali_cv/Error.h"

#include <map>

using namespace mali_cv;

RawImageBase::RawImageBase( size_t width, size_t height, Format format):_format(Format::UNKNOWN),_width(0),_height(0),_pixel_size(0)
{
	create( width, height, format );
}

void RawImageBase::create( size_t width, size_t height, Format format )
{
	_width = width;
	_height = height;
	_format = format;
	_pixel_size = get_pixel_size( format );
}

const char *RawImageBase::format_str( Format format )
{
	static std::map<Format, const char *> formats;

	if( formats.empty())
	{
		formats[Format::UNKNOWN]     = "UNKNOWN";
		formats[Format::SAME_AS_SRC] = "SAME_AS_SRC";
		formats[Format::RGB888]      = "RGB888";
		formats[Format::YUV444]      = "YUV444";
		formats[Format::YUYV422]     = "YUYV422";
		formats[Format::LUMINANCE]   = "LUMINANCE";
		formats[Format::RGBA8888]    = "RGBA8888";
		formats[Format::RGBA_FLOAT]  = "RGBA_FLOAT";
		formats[Format::LUMINANCE16] = "LUMINANCE16";
		formats[Format::U16]         = "U16";
		formats[Format::S16]         = "S16";
		formats[Format::S32]         = "S32";
		formats[Format::U32]         = "U32";
		formats[Format::S8]          = "S8";
		formats[Format::U8]          = "U8";
		formats[Format::LUMINANCE32] = "LUMINANCE32";
		formats[Format::RGBA16]      = "RGBA16";
		formats[Format::NV12]        = "NV12";
		formats[Format::NV21]        = "NV21";
		formats[Format::UYVY]        = "UYVY";
		formats[Format::IYUV]        = "IYUV";
		formats[Format::FP16C1]      = "FP16C1";
		formats[Format::FP32C1]      = "FP32C1";
		formats[Format::FP16C2]      = "FP16C2";
		formats[Format::FP32C2]      = "FP32C2";

		ERROR_ON( formats.size() != (size_t) Format::NUM_FORMATS );
	}
	return formats[format];
}
size_t RawImageBase::get_pixel_size(Format format)
{
	size_t pixel_size = 0;
	switch(format)
	{
		case Format::LUMINANCE /* Format::U8 */:
		case Format::S8:
			pixel_size = 1;
			break;
		case Format::LUMINANCE16:
		case Format::FP16C1:
		case Format::U16:
			pixel_size = 2;
			break;
		case Format::LUMINANCE32:
		case Format::FP32C1:
		case Format::FP16C2:
		case Format::U32:
			pixel_size = 4;
			break;
		case Format::RGB888:
			pixel_size = 3;
			break;
		case Format::UYVY:
		case Format::YUYV422:
			pixel_size = 2;
			break;
		case Format::YUV444:
			pixel_size = 3;
			break;
		case Format::RGBA8888:
			pixel_size = 4;
			break;
		case Format::RGBA_FLOAT:
			pixel_size = 16;
			break;
		case Format::RGBA16:
		case Format::FP32C2:
			pixel_size = 8;
			break;
		case Format::NV12:
		case Format::NV21:
		case Format::IYUV:
			/* TODO */
			pixel_size = 2;
			break;
		default:
			ERROR("Format 0x%x not supported", format);
			break;
	}
	return pixel_size;
}

uint RawImageBase::get_num_channels(Format format)
{
	size_t num_channels = 0;
	switch(format)
	{
		case Format::LUMINANCE:
		case Format::LUMINANCE16:
		case Format::LUMINANCE32:
		case Format::FP16C1:
		case Format::FP32C1:
		case Format::U32:
		case Format::U16:
		case Format::S8:
			num_channels = 1;
			break;
		case Format::FP16C2:
		case Format::FP32C2:
			num_channels = 2;
			break;
		case Format::RGB888:
			num_channels = 3;
			break;
		//Because the U and V channels are subsampled
		//These formats appear like having only 2 channels:
		case Format::YUYV422:
		case Format::UYVY:
			num_channels = 2;
			break;
		case Format::RGBA8888:
		case Format::RGBA_FLOAT:
		case Format::RGBA16:
			num_channels = 4;
			break;
		case Format::YUV444:
		case Format::NV12:
		case Format::NV21:
		case Format::IYUV:
			num_channels = 3;
			break;
		default:
			ERROR("Format 0x%x not supported", format);
			break;
	}
	return num_channels;
}

uint RawImageBase::get_element_size(Format format)
{
	size_t element_size = 0;
	switch(format)
	{
		case Format::LUMINANCE:
		case Format::RGB888:
		case Format::RGBA8888:
		case Format::UYVY:
		case Format::YUYV422:
		case Format::S8:
			element_size = 1;
			break;
		case Format::LUMINANCE16:
		case Format::RGBA16:
		case Format::FP16C1:
		case Format::FP16C2:
		case Format::U16:
			element_size = 2;
			break;
		case Format::RGBA_FLOAT:
		case Format::FP32C1:
		case Format::FP32C2:
		case Format::LUMINANCE32:
		case Format::U32:
			element_size = 4;
			break;
		/* Planar formats */
		case Format::IYUV:
		case Format::NV12:
		case Format::NV21:
		case Format::YUV444:
			element_size = 1;
			break;
		default:
			ERROR("Format 0x%x not supported", format);
			break;
	}
	return element_size;
}
RawImageBase::RawImageBase():_format(Format::UNKNOWN),_width(0), _height(0), _pixel_size(0)
{
}

size_t RawImageBase::width()const
{
	return _width;
}
size_t RawImageBase::height()const
{
	return _height;
}

uint RawImageBase::element_size()const
{
	return get_element_size(_format);
}

uint RawImageBase::num_channels()const
{
	return get_num_channels(_format);
}

Format RawImageBase::format()const
{
	return _format;
}
size_t RawImageBase::pixel_size()const
{
	return _pixel_size;
}

size_t RawImageBase::num_pixels()const
{
	return _width * _height;
}
