#include "TestResults.h"

#include "mali_cv/Context.h"
#include "mali_cv/Error.h"
#include "mali_cv/ITest.h"

#include <algorithm>
#include <ctime>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <numeric>
#include "arm_compute/runtime/Scheduler.h"
#include "arm_compute/runtime/Utils.h"
#include <sstream>
#include <vector>

#ifdef ANDROID
#include <sys/system_properties.h>
#endif


using namespace mali_cv;

TestResults &TestResults::report()
{
	static TestResults _report = TestResults();
	return _report;
}

TestResults::TestResults():_tests(),_targets(),_num_iterations(0),_current_test_id(-1),_current_test_name()
{
}

void TestResults::init(uint num_iterations)
{
	_num_iterations = num_iterations;
}

uint TestResults::TimeData::add_time(float time)
{
	_times.push_back(std::move(time));

	return _times.size();
}

void TestResults::TimeData::compute_stats(float reference_time)
{
	if (_times.size() == 0)
	{
		return;
	}

	const float sum = std::accumulate(_times.begin(), _times.end(), 0.f);

	// Don't exclude min / max if you don't have enough values
	if (_times.size() < 3)
	{
		_average = sum / _times.size();
	}
	else
	{
		const float min = *std::min_element(_times.begin(), _times.end());
		const float max = *std::max_element(_times.begin(), _times.end());

		_average = (sum - min - max) / (_times.size() - 2);
	}

	const auto center_element = std::distance(_times.begin(), _times.end()) / 2;
	auto sorted_times(_times);
	std::sort(sorted_times.begin(), sorted_times.end());

	_median = sorted_times.at(center_element);
	_speedup = reference_time / _average;
}

void TestResults::TestData::compute_stats(TargetID reference)
{
	auto &ref = _data[(int)reference];
	ref.compute_stats(0.0f);

	for (auto &target : _data)
	{
		if (&target != &ref)
		{
			target.compute_stats(ref.average());
		}
	}
}

float TestResults::TimeData::speedup() const
{
	return _speedup;
}

float TestResults::TimeData::average() const
{
	return _average;
}

const std::vector<float> &TestResults::TimeData::times() const
{
	return _times;
}

TestResults::TestData::TestData(int test_id, const std::string &name)
	:  _data(),_pmu_counters(), _name(name), _full_name()
{
	std::ostringstream tmp;
	tmp << test_id << "_" << name;
	_full_name = tmp.str();
}

TestResults::TimeData &TestResults::TestData::data(TargetID target)
{
	return _data[(int)target];
}

TestResults::PMUData & TestResults::TestData::pmu_counters()
{
        return _pmu_counters;
}

const TestResults::PMUData & TestResults::TestData::pmu_counters() const
{
       return _pmu_counters;
}

size_t TestResults::PMUData::add_counter( const std::string& name, long long value )
{
       auto it = _counters.find(name);
       if( _counters.end() == it)
       {
               _counters[name].push_back(value);
               return 1;
       }
       else
       {
               it->second.push_back(value);
               return it->second.size();
       }
}


const std::string TestResults::PMU_Cycles("cycles");

long long TestResults::PMUData::get_median( const std::string & name ) const
{
       auto it = _median.find(name);
       if( _median.end() == it) return 0;
       return it->second;
}

void TestResults::PMUData::compute_stats()
{
       //for now we just compute the median for the Cycles counters
       //TODO: compute median for other counters
       auto it = _counters.find(TestResults::PMU_Cycles);
       if( _counters.end() == it) return;
       else
       {
               ERROR_ON(it->second.size() == 0);
               // just use the first value
               _median[TestResults::PMU_Cycles] = it->second[0];
       }
}


void TestResults::add_pmu_counter(const std::string &name, long long value)
{
       ERROR_ON_MSG(_current_test_id < 0, "Current test id not set");
       if (_tests.find(_current_test_id) == _tests.end())
       {
               _tests.emplace(_current_test_id, TestData(_current_test_id, _current_test_name));
       }
       auto &test = _tests.find(_current_test_id)->second;
       PMUData &data = test.pmu_counters();
       const size_t n_values = data.add_counter(name,value);
       if (n_values == _num_iterations)
       {
               data.compute_stats();
       }
       ERROR_ON(n_values > _num_iterations);
}

const TestResults::TimeData &
TestResults::TestData::data(TargetID target) const
{
	return _data[(int)target];
}

float TestResults::TimeData::median() const
{
	return _median;
}

void TestResults::set_current_test(int test_id, const std::string &name)
{
	_current_test_id = test_id;
	_current_test_name = name;
}

void TestResults::add(const std::string &target, float time)
{
#ifdef ENABLE_TIMERS
	TargetID id = TargetID::NUM_TARGETS;
	if (target == "OpenVX")
	{
		id = TargetID::VX;
	}
	else if (target == "OpenCV")
	{
		id = TargetID::CV;
	}
	else if (target == "NEON")
	{
		id = TargetID::NEON;
	}
	else if (target == "OpenCL")
	{
		id = TargetID::CL;
	}
	else
	{
		ERROR("Unknown target %s", target.c_str());
	}
	if (_targets[(int)id].size() == 0)
	{
		_targets[(int)id] = target;
	}

	ERROR_ON_MSG(_current_test_id < 0, "Current test id not set");
	if (_tests.find(_current_test_id) == _tests.end())
	{
		_tests.emplace(_current_test_id, TestData(_current_test_id, _current_test_name));
	}

	auto &test = _tests.find(_current_test_id)->second;
	auto &data = test.data(id);

	uint n_values = data.add_time(time);

	if (n_values == _num_iterations)
	{
		data.compute_stats(0.f);

		std::cout << "|AVERAGE_TIMER|" << target << "|" << data.average()
				  << " ms|\n";
		std::cout << "|MEDIAN_TIMER|" << target << "|" << data.median() << " ms|\n";
	}

	ERROR_ON(n_values > _num_iterations);
#endif
}

inline std::string _get_cur_time()
{
	time_t rawtime = time(NULL);
	struct tm *local = localtime(&rawtime);
	char tmp[128];
	strftime(tmp, sizeof(tmp), "%c", local);
	return tmp;
}

// some android compilers don't support std::to_string()
#include <sstream>
#include <string>
namespace std
{
template <typename T> std::string to_string(T value)
{
	std::ostringstream os;
	os << value;
	return os.str();
}
}

void TestResults::save_to_file(const char *image, Environment &env)
{
#ifdef ENABLE_TIMERS
	for (auto &test : _tests)
	{
		test.second.compute_stats(env.reference);
	}
	save_to_html("html_report.html", image, env);
	save_to_json("report.json", image, env);
#endif
}

void TestResults::save_to_json( const std::string &filename, const char *image,
                                Environment &env)
{
	const ITarget *ref_target = nullptr;
	if (env.reference !=
		TargetID::VX) // Reference can't be VX as VX doesn't have timers
	{
		ref_target = env.target(env.reference);
	}
	std::ofstream s;
	s.open(filename);
	s << "{";
	s << "\"date\": \"" << _get_cur_time() << "\",";
	s << "\"OS\": \"";
#ifdef ANDROID
		s<< "Android";
#else /* ANDROID */
		s<< "Linux";
#endif /* ANDROID */
	if( sizeof(size_t) == 4)
	{
		s<<"32";
	}
	else
	{
		ERROR_ON( sizeof(size_t) != 8 );
		s<<"64";
	}
	s<<"\",";
	s << "\"border_mode\": \"";
	if (BorderMode::UNDEFINED == env.ctx.border_mode())
		s << "UNDEFINED";
	if (BorderMode::CONSTANT == env.ctx.border_mode())
		s << "CONSTANT";
	if (BorderMode::REPLICATE == env.ctx.border_mode())
		s << "REPLICATE";
	s << "\",";
	s << "\"iterations\": " << _num_iterations << ",";
	s << "\"image\": \"" << image << "\",";
	if (ref_target != nullptr)
	{
		s << "\"reference\": \"" << ref_target->name() << "\",";
	}
	s << "\"tests\": [";

	bool first_test = true;
	for (const auto &test : _tests)
	{
		if (!first_test)
		{
			s << ",";
		}
		first_test = false;
		s << "{";
		s << "\"id\": " << test.first << ",";
		s << "\"name\": \"" << test.second.name() << "\",";

		s << "\"targets\": [";
		bool first_target = true;
		for (uint i = 1; i < (uint)TargetID::NUM_TARGETS;
			 i++) // i=1 : Skip VX as timings are irrelevant for this target
		{
			const auto &data = test.second.data((TargetID)i);
			if (_targets[i].size() > 0 && data.times().size() > 0)
			{
				if (!first_target)
				{
					s << ",";
				}
				first_target = false;
				s << "{";
				s << "\"name\": \"" << _targets[i] << "\",";
				s << "\"average\": " << data.average() << ",";
				s << "\"median\": " << data.median() << ",";
				s << "\"num_times\": " << data.times().size() << ",";
				s << "\"times\": [";
				bool first_time = true;
				for (float time : data.times())
				{
					if (!first_time)
					{
						s << ",";
					}
					first_time = false;
					s << time;
				}
				s << "]}";
			}
		}
		s << "]";
		s << "}";
	}
	s << "]";
	s << "}";
	s.close();
}

void TestResults::save_to_html(const std::string &filename, const char *image,
							   Environment &env)
{
	std::ofstream s;
	s.open(filename);
	s << "<!DOCTYPE html>\n";
	s << "<html>\n";
	s << "<head>\n";
	s << "<title> Computer Vision Results - " << _get_cur_time() << "</title>\n";
	s << "</head>\n";
	s << "<body>\n";
	s << "<h2>Computer Vision Results - " << _get_cur_time() << "</h2>\n";

#ifdef ANDROID
	char osVersion[PROP_VALUE_MAX+1];
	char osBase[PROP_VALUE_MAX+1];
	char codeName[PROP_VALUE_MAX+1];
	s << "<h3>";
	if( __system_property_get("ro.build.version.base_os", osBase) )
		s << osBase << " ";
	if( __system_property_get("ro.build.version.release", osVersion) )
		s << osVersion << " ";
	if( __system_property_get("ro.build.version.codename", codeName) )
		s << codeName << "</h3>\n";
#endif
	s << "<h4>";
	s << "Image " << image;
	s << "<br> Border mode = ";
	if (BorderMode::UNDEFINED == env.ctx.border_mode())
		s << "UNDEFINED";
	if (BorderMode::CONSTANT == env.ctx.border_mode())
		s << "CONSTANT";
	if (BorderMode::REPLICATE == env.ctx.border_mode())
		s << "REPLICATE";
	s << "<br> Iterations = " << _num_iterations;
	s << "<br> Scheduler = " << arm_compute::string_from_scheduler_type(arm_compute::Scheduler::get_type());

	const ITarget *ref_target = nullptr;
	if (env.reference !=
		TargetID::VX) // Reference can't be VX as VX doesn't have timers
	{
		ref_target = env.target(env.reference);
	}
	if (ref_target != nullptr)
	{
		s << "<br> Reference = " << ref_target->name();
	}

	s << "</h4>\n";
	s << "<table border=\"1\">\n";
	s << "<tr>\n";
	s << "<th> Function </th>\n";

	for (uint i = 1; i < (uint)TargetID::NUM_TARGETS;
		 i++) // i=1 : Skip VX as timings are irrelevant for this target
	{
		if (_targets[i].size() > 0)
		{
			s << "<th>" << _targets[i] << " ms"
			  << "</th>\n";
		}
	}

	if (ref_target)
	{
		for (uint i = 1; i < (uint)TargetID::NUM_TARGETS;
			 i++) // i=1 : Skip VX as timings are irrelevant for this target
		{
			if (env.reference != (TargetID)i)
			{
				if (_targets[i].size() > 0)
				{
					s << "<th> " << _targets[i] << " speedup</th>\n";
				}
			}
		}
	}

	s << "<th> PMU Cycles {NEON} </th>\n";

	s << "</tr>\n";

	const char *fncolor[] =
	{
		"#ffff00", "#00ff00", "#ff00ff", "#00ffff", "#ADFF2F", "#FF8C00",
		"#FFF8DC", "#A52A2A", "#FF00FF", "#A9A9A9", "#008B8B", "#6495ED",
		"#E9967A", "#DAA520", "#FFD700", "#F0FFF0", "#D8BFD8", "#F0E68C",
		"#E6E6FA", "#BA55D3", "#9370D8", "#FFA500", "#FF4500", "#98FB98",
		"#DDA0DD", "#CD853F", "#4169E1", "#00FF7F", "#ffff00", "#00ff00",
		"#ff00ff", "#00ffff", "#ADFF2F", "#FF8C00", "#FFF8DC", "#A52A2A",
		"#FF00FF", "#A9A9A9", "#008B8B", "#6495ED",

	};
	int32_t n_color = sizeof(fncolor) / sizeof(fncolor[0]);

	int color = 0;
	for (const auto &test : _tests)
	{
		s << "<tr>\n";
		s << "<td BGCOLOR=\"" << fncolor[color] << "\">" << test.second.full_name()
		  << "</td>\n";
		color = (color + 1) % n_color;

		for (uint i = 1; i < (uint)TargetID::NUM_TARGETS;
			 i++) // i=1 : Skip VX as timings are irrelevant for this target
		{
			if (_targets[i].size() > 0)
			{
				s << "<td align=\"right\">" << std::setw(8) << std::fixed
				  << std::setprecision(6)
				  << test.second.data((TargetID)i).average() << "</td>\n";
			}
		}

		if (ref_target)
		{
			for (uint i = 1; i < (uint)TargetID::NUM_TARGETS;
				 i++) // i=1 : Skip VX as timings are irrelevant for this target
			{
				if (env.reference != (TargetID)i && _targets[i].size() > 0)
				{
					s << "<td align=\"right\">" << std::setw(8) << std::fixed
					  << std::setprecision(6)
					  << test.second.data((TargetID)i).speedup() << "</td>\n";
				}
			}
		}


		{ // outputs PMU cycles
			s << "<td align=\"right\">" <<  test.second.pmu_counters().get_median("cycles") << "</td>\n";
		}

		s << "</tr>\n";
	}

	s << "</table>\n";
	s << "<br><br>\n";
	const int bar_width(30);
	const int canvas_height(400);
	const int canvas_width(bar_width * _tests.size() + 32 );
	s << "<canvas id=\"myCanvas\" width=\"" << std::to_string(canvas_width) <<"\" height=\"" << std::to_string(canvas_height)<<"\"style=\"border:4px solid ##2F4F4F;\">Your browser does not support the HTML5 canvas tag.</canvas>\n";
	s << "<script>\n";
	s << "var c = document.getElementById(\"myCanvas\");";
	s << "var ctx = c.getContext(\"2d\");";
	s << "ctx.fillStyle = '#DCDCDC'\n";
	s << "ctx.fillRect(0,0," << std::to_string(canvas_width) << "," << std::to_string(canvas_height) <<");\n" ;
	s << "ctx.font = \"bold 15px Arial\"\n";
	s << "ctx.fillStyle = 'black'\n";
	s << "ctx.fillText(\"Speedup\",0,15);\n";
	int x(0);
	color = 0;
	for( const auto &test : _tests )
	{
		s << "ctx.fillStyle = \"" <<  fncolor[color++% n_color] << "\";";
		const float bar_height = test.second.data(TargetID::NEON).speedup() / 200.F * 5000.F;
		s << "ctx.fillRect(" << std::to_string(x) << "," <<  std::to_string(canvas_height-bar_height)  <<"," << std::to_string(bar_width) <<"," << std::to_string(bar_height) <<");" ;
		s << "ctx.fillStyle = 'black'\n";
		s << "ctx.fillText(\"" << std::setprecision(1) << test.second.data(TargetID::NEON).speedup()  <<"\"," << x+bar_width/4 << "," << "200);\n";
		x+=bar_width;
	}
	s << "</script>";
	s << "</body>\n";
	s << "</html>\n";
	s.close();
}


