#include "mali_cv/Timer.h"

#include <iostream>

#if ARM_CV_ENABLE_PMU
#define _GNU_SOURCE 1
#include <unistd.h>
#include <asm/unistd.h>
#include <fcntl.h>
#include <linux/perf_event.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#endif



using namespace mali_cv;

const std::chrono::high_resolution_clock::time_point Timer::ZERO_TIME = std::chrono::high_resolution_clock::time_point() ;

Timer::Timer(const std::string &msg): _start(std::chrono::high_resolution_clock::now()), _stop(), _msg(msg)
#if ARM_CV_ENABLE_PMU
	,_pe(),_fd_instr(-1),_fd_cycles(-1),_inst_count(0),_cycles_count(0)
#endif
{
}

void Timer::start(void)
{
#ifdef ENABLE_TIMERS
	_stop = ZERO_TIME;
	_start = std::chrono::high_resolution_clock::now();
#endif
#if ARM_CV_ENABLE_PMU
	if( _fd_instr  == -1)
	{
		pid_t mypid = getpid();
		memset(&_pe, 0, sizeof(struct perf_event_attr) );
		_pe.size = sizeof(struct perf_event_attr);
		_pe.type = PERF_TYPE_HARDWARE;
		_pe.config = PERF_COUNT_HW_INSTRUCTIONS;
		_pe.inherit = 1; // The inherit bit specifies that this counter should count events of child tasks as well as the task specified
		_pe.inherit_stat = 1; // enables saving of event counts on context switch for inherited tasks
		_fd_instr = syscall(__NR_perf_event_open, &_pe , mypid , -1 ,  -1, 0);
		if( _fd_instr < 0 )
		{
			perror("per_event_open for inst_count failed\n\n\n\n");
		}
		_pe.config = PERF_COUNT_HW_CPU_CYCLES;
		_fd_cycles = syscall(__NR_perf_event_open, &_pe , mypid , -1 ,  -1,0);
		if( _fd_cycles < 0 )
		{
			perror("per_event_open for cycles failed\n\n\n\n");
		}
	}
	ioctl(_fd_instr, PERF_EVENT_IOC_RESET,0);
	ioctl(_fd_instr, PERF_EVENT_IOC_ENABLE,0);
	ioctl(_fd_cycles, PERF_EVENT_IOC_RESET,0);
	ioctl(_fd_cycles, PERF_EVENT_IOC_ENABLE,0);
#endif
}

std::chrono::high_resolution_clock::duration Timer::stop(void)
{
#ifdef ENABLE_TIMERS
#if ARM_CV_ENABLE_PMU
	ioctl(_fd_instr, PERF_EVENT_IOC_DISABLE,0);
	ioctl(_fd_cycles, PERF_EVENT_IOC_DISABLE,0);
	read(_fd_instr, &_inst_count, sizeof(long long) );
	read(_fd_cycles, &_cycles_count, sizeof(long long) );
#endif
	_stop = std::chrono::high_resolution_clock::now();
#endif

	return _stop - _start;
}

void Timer::reset(void)
{
#ifdef ENABLE_TIMERS
	_start = std::chrono::high_resolution_clock::now();
#if ARM_CV_ENABLE_PMU
	close(_fd_instr);
	close(_fd_cycles);
	_fd_cycles = -1;
	_fd_cycles = -1;
#endif
#endif

}

long long Timer::get_cycles(bool reset)
{
	if(reset)
	{
		this->reset();
	}
#if ARM_CV_ENABLE_PMU
	return _cycles_count;
#else
	return 0;
#endif
}

float Timer::get_elapsed_ms(const bool reset)
{
#ifdef ENABLE_TIMERS
	std::chrono::duration<float, std::milli> delta;

	if(_stop == ZERO_TIME)
	{
		delta = std::chrono::high_resolution_clock::now() - _start;
	}
	else
	{
		delta = _stop - _start;
	}

	if(reset)
	{
		this->reset();
	}

	return delta.count();
#else
	return 0.0f;
#endif
}

void Timer::print_elapsed_ms(const bool reset)
{
#ifdef ENABLE_TIMERS
	if(!_msg.empty())
	{
		std::cout << "|CPU_TIMER|" << _msg << "|" << get_elapsed_ms(false) << " ms|\n";
#if ARM_CV_ENABLE_PMU
		std::cout << "|PMU_CYCLES|" << _msg << "|" << _cycles_count << "|" << std::endl;
		std::cout << "|PMU_INSTR|" << _msg << "|" << _inst_count << "|" << std::endl;
#endif
	}

	if(reset)
	{
		this->reset();
	}
#endif
}
