#include "mali_cv/Image.h"

#include "mali_cv/Error.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/Context.h"
#include "mali_cv/Helper.h"

using namespace mali_cv;

VXImage::VXImage():IImage(),img(nullptr),type()
{
}

void VXImage::internal_create( Context &ctx, size_t w, size_t h, Format format)
{
	type = VXHelper::convert_to_vx_format( format );
	width = w;
	height = h;
	img = vxCreateImage(ctx.vx.context(), width, height, type );
	ERROR_ON_MSG( 0 == img, "Cannot create VX image" );
}

VXImage::~VXImage()
{
	if( img )
	{
		CHECK_VX( vxReleaseImage( &img ) );
	}
}

void VXImage::store( RawImage &out )
{
	Format format = Format::UNKNOWN;
	switch( type )
	{
		case VX_DF_IMAGE_U8:
			format = Format::LUMINANCE;
			break;
		case VX_DF_IMAGE_U16:
			format = Format::U16;
			break;
		case VX_DF_IMAGE_S16:
			format = Format::LUMINANCE16;
			break;
		case VX_DF_IMAGE_RGB:
			format = Format::RGB888;
			break;
		case VX_DF_IMAGE_RGBX:
			format = Format::RGBA8888;
			break;
		case VX_DF_IMAGE_NV12:
			format = Format::NV12;
			break;
		case VX_DF_IMAGE_NV21:
			format = Format::NV21;
			break;
		case VX_DF_IMAGE_UYVY:
			format = Format::UYVY;
			break;
		case VX_DF_IMAGE_YUYV:
			format = Format::YUYV422;
			break;
		case VX_DF_IMAGE_IYUV:
			format = Format::IYUV;
			break;
		case VX_DF_IMAGE_YUV4:
			format = Format::YUV444;
			break;
		case VX_DF_IMAGE_S32:
			format = Format::LUMINANCE32;
			break;
		case VX_DF_IMAGE_U32:
			format = Format::U32;
			break;
		default:
			ERROR("Not supported");
			break;
	}

	out.create( width, height, format);

	vx_rectangle_t rect = {0, 0, (vx_uint32)width, (vx_uint32)height};
	vx_imagepatch_addressing_t addr[4];
	void *ptr[4] = {NULL};
	vx_map_id map_id[4];
	uint8_t *data = out.data();
	uint8_t *uv_data = NULL, *u_data = NULL, *v_data = NULL;
	vx_uint32 x, y, p;
	vx_size planes;

	CHECK_VX( vxQueryImage(img, VX_IMAGE_PLANES, &planes, sizeof(planes)));
	for(p = 0; p < planes; p++)
	{
		CHECK_VX( vxMapImagePatch(img, &rect, p, &map_id[p], &addr[p],&ptr[p], VX_READ_ONLY, VX_MEMORY_TYPE_NONE, VX_NOGAP_X));
	}

	switch( type )
	{
		case VX_DF_IMAGE_U32:
			for( uint i = 0; i < width * height; i++)
			{
				void *d = vxFormatImagePatchAddress1d(ptr[0], i, &addr[0]);
				((uint32_t *)data)[i] = ((uint32_t *)d)[0];
			}
			break;
		case VX_DF_IMAGE_S32:
			for( uint i = 0; i < width * height; i++)
			{
				void *d = vxFormatImagePatchAddress1d(ptr[0], i, &addr[0]);
				((int32_t *)data)[i] = ((int32_t *)d)[0];
			}
			break;
		case VX_DF_IMAGE_S16:
			for( uint i = 0; i < width * height; i++)
			{
				void *d = vxFormatImagePatchAddress1d(ptr[0], i, &addr[0]);
				((int16_t *)data)[i] = ((int16_t *)d)[0];
			}
			break;
		case VX_DF_IMAGE_U16:
			for( uint i = 0; i < width * height; i++)
			{
				void *d = vxFormatImagePatchAddress1d(ptr[0], i, &addr[0]);
				((uint16_t *)data)[i] = ((uint16_t *)d)[0];
			}
			break;
		case VX_DF_IMAGE_U8:
			for( uint i = 0; i < width * height; i++)
			{
				void *d = vxFormatImagePatchAddress1d(ptr[0], i, &addr[0]);
				((uint8_t *)data)[i] = ((uint8_t *)d)[0];
			}
			break;
		case VX_DF_IMAGE_RGB:
			for (y = 0; y < addr[0].dim_y; y++)
			{
				for (x = 0; x < addr[0].dim_x; x++)
				{
					vx_uint8 *rgb = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					data[0] = rgb[0];
					data[1] = rgb[1];
					data[2] = rgb[2];
					data += 3;
				}
			}
			break;
		case VX_DF_IMAGE_RGBX:
			for (y = 0; y < addr[0].dim_y; y++)
			{
				for (x = 0; x < addr[0].dim_x; x++)
				{
					vx_uint8 *rgbx = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					data[0] = rgbx[0];
					data[1] = rgbx[1];
					data[2] = rgbx[2];
					data[3] = rgbx[3];
					data += 4;
				}
			}
			break;
		case VX_DF_IMAGE_NV12:
		case VX_DF_IMAGE_NV21:
			uv_data = data + addr[0].dim_y * addr[0].dim_x;
			for (y = 0; y < addr[0].dim_y; y+=addr[0].step_y)
			{
				for (x = 0; x < addr[0].dim_x; x+=addr[0].step_x)
				{
					vx_uint8 *luma = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					data[0] = luma[0];
					data++;
				}
			}

			for (y = 0; y < addr[1].dim_y; y+=addr[1].step_y)
			{
				for (x = 0; x < addr[1].dim_x; x+=addr[1].step_x)
				{
					vx_uint8 *crcb = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[1], x, y, &addr[1]);
					uv_data[0] = crcb[0];
					uv_data[1] = crcb[1];
					uv_data += 2;
				}
			}
			break;
		case VX_DF_IMAGE_UYVY:
		case VX_DF_IMAGE_YUYV:
			for (y = 0; y < addr[0].dim_y; y++)
			{
				for (x = 0; x < addr[0].dim_x; x += 2)
				{
					vx_uint8 *yuyv = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					data[0] = yuyv[0];
					data[1] = yuyv[1];
					data[2] = yuyv[2];
					data[3] = yuyv[3];
					data += 4;
				}
			}
			break;
		case VX_DF_IMAGE_IYUV:
			u_data = data + addr[0].dim_y * addr[0].dim_x;
			v_data = data + addr[0].dim_y * addr[0].dim_x + addr[0].dim_y * addr[0].dim_x / 4;
			for (y = 0; y < addr[0].dim_y; y++)
			{
				for (x = 0; x < addr[0].dim_x; x++)
				{
					vx_uint8 *luma = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					vx_uint8 *cb   = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[1], x, y, &addr[1]);
					vx_uint8 *cr   = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[2], x, y, &addr[2]);
					data[0] = luma[0];
					data++;
					if((x % 2 == 0) && (y % 2 == 0))
					{
						u_data[0] = cb[0];
						v_data[0] = cr[0];
						u_data++;
						v_data++;
					}
				}
			}
			break;
		case VX_DF_IMAGE_YUV4:
			u_data = data + addr[0].dim_y * addr[0].dim_x;
			v_data = data + addr[0].dim_y * addr[0].dim_x * 2;
			for (y = 0; y < addr[0].dim_y; y++)
			{
				for (x = 0; x < addr[0].dim_x; x++)
				{
					vx_uint8 *luma = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					vx_uint8 *cb   = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[1], x, y, &addr[1]);
					vx_uint8 *cr   = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[2], x, y, &addr[2]);
					data[0]   = luma[0];
					u_data[0] = cb[0];
					v_data[0] = cr[0];
					data++;
					u_data++;
					v_data++;
				}
			}
			break;
		default:
			ERROR("Not supported");
			break;
	}

	for(p = 0; p < planes; p++)
	{
		CHECK_VX( vxUnmapImagePatch(img, map_id[p]));
	}
}

void VXImage::load(const RawImage &in )
{
	ERROR_ON( in.width() != width );
	ERROR_ON( in.height() != height );
	vx_rectangle_t rect = {0, 0, (vx_uint32)width, (vx_uint32)height};
	vx_imagepatch_addressing_t addr[4];
	void *ptr[4] = {NULL};
	vx_map_id map_id[4];
	uint8_t *data = NULL;
	uint8_t *uv_data = NULL, *u_data = NULL, *v_data = NULL;
	vx_uint32 x, y, p;
	vx_size planes;

	CHECK_VX( vxQueryImage(img, VX_IMAGE_PLANES, &planes, sizeof(planes)));
	for(p = 0; p < planes; p++)
	{
		CHECK_VX( vxMapImagePatch(img, &rect, p, &map_id[p], &addr[p],&ptr[p], VX_WRITE_ONLY, VX_MEMORY_TYPE_NONE, VX_NOGAP_X));
	}

	switch( type )
	{
		case VX_DF_IMAGE_U16:
			for( uint i = 0; i < width * height; i++)
			{
				vx_uint16 *d = (vx_uint16 *) vxFormatImagePatchAddress1d(ptr[0], i, &addr[0]);
				*d = (vx_uint16) ((uint16_t *)in.data())[i];
			}
			break;
		case VX_DF_IMAGE_S16:
			for( uint i = 0; i < width * height; i++)
			{
				vx_int16 *d = (vx_int16 *) vxFormatImagePatchAddress1d(ptr[0], i, &addr[0]);
				if( in.pixel_size() == 2)
				{
					*d = (vx_int16) ((int16_t *)in.data())[i];
				}
				else
				{
					*d = (vx_int16) ((uint8_t *)in.data())[i];
				}
			}
			break;
		case VX_DF_IMAGE_U8:
			for( uint i = 0; i < width * height; i++)
			{
				data = in.data();
				vx_uint8 *d = (vx_uint8 *) vxFormatImagePatchAddress1d(ptr[0], i, &addr[0]);
				*d = data[i];
			}
			break;
		case VX_DF_IMAGE_RGB:
			data = in.data();
			for (y = 0; y < addr[0].dim_y; y++)
			{
				for (x = 0; x < addr[0].dim_x; x++)
				{
					vx_uint8 *rgb = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					rgb[0] = data[0];
					rgb[1] = data[1];
					rgb[2] = data[2];
					data += 3;
				}
			}
			break;
		case VX_DF_IMAGE_RGBX:
			data = in.data();
			for (y = 0; y < addr[0].dim_y; y++)
			{
				for (x = 0; x < addr[0].dim_x; x++)
				{
					vx_uint8 *rgbx = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					rgbx[0] = data[0];
					rgbx[1] = data[1];
					rgbx[2] = data[2];
					rgbx[3] = data[3];
					data += 4;
				}
			}
			break;
		case VX_DF_IMAGE_NV12:
		case VX_DF_IMAGE_NV21:
			data = in.data();
			uv_data = data + addr[0].dim_y * addr[0].dim_x;
			for (y = 0; y < addr[0].dim_y; y+=addr[0].step_y)
			{
				for (x = 0; x < addr[0].dim_x; x+=addr[0].step_x)
				{
					vx_uint8 *luma = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					luma[0] = data[0];
					data++;
				}
			}

			for (y = 0; y < addr[1].dim_y; y+=addr[1].step_y)
			{
				for (x = 0; x < addr[1].dim_x; x+=addr[1].step_x)
				{
					vx_uint8 *cbcr = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[1], x, y, &addr[1]);
					cbcr[0] = uv_data[0];
					cbcr[1] = uv_data[1];
					uv_data += 2;
				}
			}
			break;
		case VX_DF_IMAGE_UYVY:
		case VX_DF_IMAGE_YUYV:
			data = in.data();
			for (y = 0; y < addr[0].dim_y; y++)
			{
				for (x = 0; x < addr[0].dim_x; x += 2)
				{
					vx_uint8 *yuyv = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					yuyv[0] = data[0];
					yuyv[1] = data[1];
					yuyv[2] = data[2];
					yuyv[3] = data[3];
					data += 4;
				}
			}
			break;
		case VX_DF_IMAGE_IYUV:
			data = in.data();
			u_data = data + addr[0].dim_y * addr[0].dim_x;
			v_data = data + addr[0].dim_y * addr[0].dim_x + addr[0].dim_y * addr[0].dim_x / 4;
			for (y = 0; y < addr[0].dim_y; y++)
			{
				for (x = 0; x < addr[0].dim_x; x++)
				{
					vx_uint8 *luma = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					vx_uint8 *cb   = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[1], x, y, &addr[1]);
					vx_uint8 *cr   = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[2], x, y, &addr[2]);
					luma[0] = data[0];
					data++;
					if((x % 2 == 0) && (y % 2 == 0))
					{
						cb[0] = u_data[0];
						cr[0] = v_data[0];
						u_data++;
						v_data++;
					}
				}
			}
			break;
		case VX_DF_IMAGE_YUV4:
			data = in.data();
			u_data = data + addr[0].dim_y * addr[0].dim_x;
			v_data = data + addr[0].dim_y * addr[0].dim_x * 2;
			for (y = 0; y < addr[0].dim_y; y++)
			{
				for (x = 0; x < addr[0].dim_x; x++)
				{
					vx_uint8 *luma = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[0], x, y, &addr[0]);
					vx_uint8 *cb   = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[1], x, y, &addr[1]);
					vx_uint8 *cr   = (vx_uint8 *) vxFormatImagePatchAddress2d(ptr[2], x, y, &addr[2]);
					luma[0] = data[0];
					cb[0]   = u_data[0];
					cr[0]   = v_data[0];
					data++;
					u_data++;
					v_data++;
				}
			}
			break;
		default:
			for( uint i = 0; i < width * height; i++)
			{
				data = in.data();
				vx_uint8 *d = (vx_uint8 *) vxFormatImagePatchAddress1d(ptr[0], i, &addr[0]);
				*d = data[i];
			}
			break;
	}

	for(p = 0; p < planes; p++)
	{
		CHECK_VX( vxUnmapImagePatch(img, map_id[p]));
	}
}

arm_compute::ValidRegion VXImage::valid_region() const
{
    vx_rectangle_t valid_region;

    CHECK_VX(vxGetValidRegionImage(img, &valid_region));

    return arm_compute::ValidRegion(arm_compute::Coordinates(static_cast<int>(valid_region.start_x), static_cast<int>(valid_region.start_y)),
                               arm_compute::TensorShape(valid_region.end_x, valid_region.end_y));
}
