#include "mali_cv/Target.h"

#include "mali_cv/Convolution.h"
#include "mali_cv/Error.h"
#include "mali_cv/Image.h"
#include "mali_cv/Pyramid.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/RemapTable.h"
#include "TestResults.h"

#include <numeric>
#include <sstream>

using namespace mali_cv;

VXTarget::VXTarget(Context &ctx): ITarget(ctx),_eigen(ctx,name())
{
	_timer = Timer(name());
}

void VXTarget::absolute_difference( Image &a, Image &b, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxAbsDiffNode(graph, a.vx.img, b.vx.img, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),"abs_diff");
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::gemm_lowp(Image &a, Image &b, Image &d,int32_t a_offset, int32_t b_offset, int32_t c_offset, int32_t c_mult_int, int32_t shift)
{
	/*Note: We need this empty function otherwise GEMM test crashes. The current reference implementation is within the GEMM test*/
	return;
}

void VXTarget::interleave_4x4(Image &in, Image &out)
{
	/*Note: We need this empty function otherwise GEMM test crashes. The current reference implementation is within the GEMMInterleave4x4 test*/
	return;
}

void VXTarget::transpose_1xW(Image &in, Image &out)
{
	/*Note: We need this empty function otherwise GEMM test crashes. The current reference implementation is within the GEMMTranspose1xW test*/
	return;
}

void VXTarget::bitwise_or( Image &a, Image &b, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxOrNode(graph, a.vx.img, b.vx.img, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),"bit_or");
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::bitwise_not( Image &a, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxNotNode(graph, a.vx.img, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(), "bit_not");
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::accumulate( Image &a, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxAccumulateImageNode(graph, a.vx.img, c.vx.img) ,  _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),"accumulate");
	CHECK_VX( vxReleaseGraph( &graph ) );

}

void VXTarget::accumulate_squared( Image &a, Image &c, uint32_t shift)
{
	vx_graph graph = _ctx.vx.create_graph();
	vx_scalar s = vxCreateScalar( _ctx.vx.context(), VX_TYPE_UINT32, &shift);
	ERROR_ON( 0 == s );
	vxAccumulateSquareImageNode(graph, a.vx.img, s ,c.vx.img);
	_ctx.vx.run_graph(graph, _ctx.num_iterations(),"accumulate_sq");
	CHECK_VX( vxReleaseGraph( &graph ) );
	CHECK_VX( vxReleaseScalar( &s ) );
}

void VXTarget::accumulate_weighted( Image &a, Image &c, float alpha)
{
	vx_graph graph = _ctx.vx.create_graph();
	vx_scalar s = vxCreateScalar( _ctx.vx.context(), VX_TYPE_FLOAT32, &alpha);
	ERROR_ON( 0 == s );
	vxAccumulateWeightedImageNode(graph, a.vx.img, s ,c.vx.img);
	_ctx.vx.run_graph(graph, _ctx.num_iterations(), "accumulate_we");
	CHECK_VX( vxReleaseGraph( &graph ) );
	CHECK_VX( vxReleaseScalar( &s ) );
}

void VXTarget::accumulate_weighted_fp16( Image &a, Image &c, float alpha)
{
	VXTarget::accumulate_weighted(a,c,alpha);
}

void VXTarget::warp_perspective(Image &a, Image &c, const float matrix[3][3], InterpolationMode mode)
{
	vx_graph graph = _ctx.vx.create_graph();
	vx_matrix vxmatrix = vxCreateMatrix(_ctx.vx.context(), VX_TYPE_FLOAT32, 3, 3);
	vxCopyMatrix(vxmatrix, (void*) matrix, VX_WRITE_ONLY, VX_MEMORY_TYPE_HOST);

	const vx_enum interpolation = VXHelper::convert_to_vx_interpolation(mode);

	_ctx.vx.set_border_mode( vxWarpPerspectiveNode( graph, a.vx.img, vxmatrix, interpolation, c.vx.img), _ctx.border_mode() );

	CHECK_VX( vxVerifyGraph(graph) );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(), "warp_perspective");

	CHECK_VX( vxReleaseGraph( &graph ) );
	CHECK_VX( vxReleaseMatrix( &vxmatrix ) );
}

void VXTarget::warp_affine(Image &a, Image &c, const float matrix[3][2], InterpolationMode mode)
{
	vx_graph graph = _ctx.vx.create_graph();
	vx_matrix vxmatrix = vxCreateMatrix(_ctx.vx.context(), VX_TYPE_FLOAT32, 2, 3);
	vxCopyMatrix(vxmatrix, (void*) matrix, VX_WRITE_ONLY, VX_MEMORY_TYPE_HOST);

	const vx_enum interpolation = VXHelper::convert_to_vx_interpolation(mode);

	_ctx.vx.set_border_mode( vxWarpAffineNode( graph, a.vx.img, vxmatrix, interpolation, c.vx.img), _ctx.border_mode() );

	CHECK_VX( vxVerifyGraph(graph) );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(), "warp_affine");

	CHECK_VX( vxReleaseGraph( &graph ) );
	CHECK_VX( vxReleaseMatrix( &vxmatrix ) );
}

void VXTarget::pixelwise_mul( Image &a, Image &b, Image &c, float scale, uint8_t overflow_policy, uint8_t round_policy )
{
	vx_graph  graph = _ctx.vx.create_graph();
	vx_scalar vx_scale =  vxCreateScalar( _ctx.vx.context(), VX_TYPE_FLOAT32, &scale);
	vx_enum op = (overflow_policy) ? VX_CONVERT_POLICY_SATURATE : VX_CONVERT_POLICY_WRAP;
	vx_enum rp = (round_policy) ? VX_ROUND_POLICY_TO_NEAREST_EVEN : VX_ROUND_POLICY_TO_ZERO;

	vxMultiplyNode( graph, a.vx.img, b.vx.img, vx_scale, op, rp, c.vx.img );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(), "pixelwise_mul");
	CHECK_VX( vxReleaseGraph( &graph ) );
	CHECK_VX( vxReleaseScalar( &vx_scale ) );
}

void VXTarget::harris_corners(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners, size_t& num_corners)
{
	vx_int32  gradient_size = gradient;
	vx_int32  bsize         = block_size;

	vx_scalar vx_threshold   = vxCreateScalar( _ctx.vx.context(), VX_TYPE_FLOAT32, &threshold);
	vx_scalar vx_min_dist    = vxCreateScalar( _ctx.vx.context(), VX_TYPE_FLOAT32, &min_dist);
	vx_scalar vx_num_corners = vxCreateScalar( _ctx.vx.context(), VX_TYPE_SIZE, &num_corners);
	vx_scalar vx_sensitivity = vxCreateScalar( _ctx.vx.context(), VX_TYPE_FLOAT32, &sensitivity);

	vx_graph graph = _ctx.vx.create_graph();
	vx_node harris_node = vxHarrisCornersNode(graph, a.vx.img, vx_threshold, vx_min_dist, vx_sensitivity, gradient_size, bsize, corners.vx.a, vx_num_corners);

	_ctx.vx.set_border_mode( harris_node, _ctx.border_mode() );

	CHECK_VX( vxVerifyGraph(graph) );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(), "harris");

	CHECK_VX( vxCopyScalar(vx_num_corners, &num_corners, VX_READ_ONLY, VX_MEMORY_TYPE_HOST) );

	CHECK_VX( vxReleaseGraph( &graph ) );
	CHECK_VX( vxReleaseScalar( &vx_threshold ) );
	CHECK_VX( vxReleaseScalar( &vx_min_dist ) );
	CHECK_VX( vxReleaseScalar( &vx_num_corners ) );
	CHECK_VX( vxReleaseScalar( &vx_sensitivity ) );
}

void VXTarget::harris_corners_fp16(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners,size_t& num_corners)
{
	VXTarget::harris_corners(a,threshold,min_dist,sensitivity,gradient,block_size,corners,num_corners);
}

void VXTarget::fast_corners(Image &a,float threshold, bool nonmax_suppression, IKeyPointArray &corners, size_t &num_corners)
{
	ERROR_ON_MSG( BorderMode::UNDEFINED != _ctx.border_mode(), "VX only supports UNDEFINED border mode" );
	vx_graph graph = _ctx.vx.create_graph();
	vx_scalar vx_threshold = vxCreateScalar( _ctx.vx.context(), VX_TYPE_FLOAT32, &threshold);
	vx_bool vx_nonmax = nonmax_suppression?vx_true_e:vx_false_e;
	_ctx.vx.set_border_mode( vxFastCornersNode(graph, a.vx.img, vx_threshold, vx_nonmax, corners.vx.a, NULL),   _ctx.border_mode() );
	CHECK_VX( vxVerifyGraph(graph) );
	_ctx.vx.run_graph(graph, _ctx.num_iterations(), "fast_corners");
	CHECK_VX( vxReleaseGraph( &graph ) );
	CHECK_VX( vxReleaseScalar( &vx_threshold ) );
}

void VXTarget::arithmetic_addition( Image &a, Image &b, Image &c, uint8_t policy)
{
	vx_graph graph = _ctx.vx.create_graph();
	if( policy )
		vxAddNode (graph, a.vx.img, b.vx.img, VX_CONVERT_POLICY_SATURATE, c.vx.img);
	else
		vxAddNode (graph, a.vx.img, b.vx.img, VX_CONVERT_POLICY_WRAP, c.vx.img);
	_ctx.vx.run_graph(graph, _ctx.num_iterations(), "addition");
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::arithmetic_subtraction( Image &a, Image &b, Image &c, uint8_t policy)
{
	vx_graph graph = _ctx.vx.create_graph();
	if( policy )
		vxSubtractNode (graph, a.vx.img, b.vx.img, VX_CONVERT_POLICY_SATURATE, c.vx.img);
	else
		vxSubtractNode (graph, a.vx.img, b.vx.img, VX_CONVERT_POLICY_WRAP, c.vx.img);
	_ctx.vx.run_graph(graph, _ctx.num_iterations(), "subtraction");
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::bitwise_and( Image &a, Image &b, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxAndNode(graph, a.vx.img, b.vx.img, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),"bit_and");
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::bitwise_xor( Image &a, Image &b, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxXorNode(graph, a.vx.img, b.vx.img, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),"bit_xor");
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::erode( Image &a, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxErode3x3Node(graph, a.vx.img, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::dilate( Image &a, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxDilate3x3Node(graph, a.vx.img, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::gaussian3x3( Image &a, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxGaussian3x3Node(graph, a.vx.img, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::box3x3( Image &a, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxBox3x3Node(graph, a.vx.img, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::box3x3_fp16( Image &a, Image &c)
{
	VXTarget::box3x3(a,c);
}

void VXTarget::median3x3( Image &a, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxMedian3x3Node(graph, a.vx.img, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::gaussian5x5( Image &a, Image &c)
{
	VXConvolution conv;
	const int16_t gaussian5x5 []=
	{
		1,  4,  6,  4, 1,
		4, 16, 24, 16, 4,
		6, 24, 36, 24, 6,
		4, 16, 24, 16, 4,
		1,  4,  6,  4, 1
	};
	RawImage coeffs(5,5, Format::LUMINANCE16);
	memcpy( coeffs.data(), gaussian5x5, sizeof(gaussian5x5));
	conv.create( _ctx, coeffs, CopyData::YES);

	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxConvolveNode(graph, a.vx.img, conv.convolution, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::convolution( Image &a, Image &c, Convolution &convolution_matrix)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxConvolveNode(graph, a.vx.img, convolution_matrix.vx.convolution, c.vx.img), _ctx.border_mode() );

	std::stringstream tag;
	tag << "conv" << convolution_matrix.vx.width << "x" << convolution_matrix.vx.height;
	if( Format::LUMINANCE16 == c.vx.format )
	{
		tag << "u16";
	}
	else
	{
		tag << "u8";

	}
	_ctx.vx.run_graph(graph, _ctx.num_iterations(), tag.str().c_str());

	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::scale( Image &a, Image &c, InterpolationMode mode)
{
	const vx_enum interpolation = VXHelper::convert_to_vx_interpolation(mode);

	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxScaleImageNode(graph, a.vx.img, c.vx.img, interpolation), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::color_convert( Image &a, Image &c )
{
	vx_graph graph = _ctx.vx.create_graph();

	vxColorConvertNode(graph, a.vx.img, c.vx.img);

	const char *tag=NULL;
	switch(a.vx.format)
	{
		case Format::RGBA8888:
		{
			switch(c.vx.format)
			{
				case Format::RGB888:
					tag ="rgba2rgb";
					break;
				case Format::NV12:
					tag ="rgba2nv12";
					break;
				case Format::IYUV:
					tag ="rgba2iyuv";
					break;
				case Format::YUV444:
					tag = "rgba2yuv4";
					break;
				case Format::UNKNOWN:
				default:
					ERROR("Not supported");
					break;

			}
			break;
		}
		case Format::UYVY:
		{
			switch(c.vx.format)
			{
				case Format::RGBA8888:
					tag = "uyvy2rgba";
					break;
				case Format::RGB888:
					tag = "uyvy2rgb";
					break;
				case Format::NV12:
					tag = "uyvy2nv12";
					break;
				case Format::IYUV:
					tag = "uyvy2iyuv";
					break;
				case Format::YUV444:
				case Format::UNKNOWN:
				default:
					ERROR("Not supported");
					break;
			}
			break;
		}
		case Format::YUYV422:
		{
			switch(c.vx.format)
			{
				case Format::RGBA8888:
					tag = "yuyv2rgba";
					break;
				case Format::RGB888:
					tag = "yuyv2rgb";
					break;
				case Format::NV12:
					tag = "yuyv2nv12";
					break;
				case Format::IYUV:
					tag = "yuyv2iyuv";
					break;
				case Format::YUV444:
				case Format::UNKNOWN:
				default:
					ERROR("Not supported");
					break;
			}
			break;
		}

		case Format::RGB888:
		{
			switch(c.vx.format)
			{
				case Format::RGBA8888:
					tag = "rgb2rgba";
					break;
				case Format::NV12:
					tag = "rgb2nv12";
					break;
				case Format::IYUV:
					tag = "rgb2iyuv";
					break;
				case Format::YUV444:
					tag = "rgb2yuv4";
					break;
				case Format::UNKNOWN:
				default:
					ERROR("Not supported");
					break;

			}
			break;
		}
		case Format::NV12:
		{
			switch(c.vx.format)
			{
				case Format::RGBA8888:
					tag = "nv122rgba";
					break;
				case Format::RGB888:
					tag = "nv122rgb";
					break;
				case Format::IYUV:
					tag = "nv122iyuv";
					break;
				case Format::YUV444:
					tag = "nv212yuv4";
					break;
				case Format::UNKNOWN:
				default:
					ERROR("Not supported");
					break;

			}
			break;
		}
		case Format::NV21:
		{
			switch(c.vx.format)
			{
				case Format::RGBA8888:
					tag = "nv212rgba";
					break;
				case Format::RGB888:
					tag = "nv212rgb";
					break;
				case Format::IYUV:
					tag = "nv212iyuv";
					break;
				case Format::YUV444:
					tag = "nv212yuv4";
					break;
				case Format::UNKNOWN:
				default:
					ERROR("Not supported");
					break;

			}
			break;
		}
		case Format::IYUV:
		{
			switch(c.vx.format)
			{
				case Format::RGBA8888:
					tag = "iyuv2rgba";
					break;
				case Format::RGB888:
					tag = "iyuv2rgb";
					break;
				case Format::NV12:
					tag = "iyuv2nv12";
					break;
				case Format::YUV444:
					tag = "iyuv2yuv4";
					break;
				case Format::UNKNOWN:
				default:
					ERROR("Not supported");
					break;

			}
			break;
		}
		case Format::UNKNOWN:
		default:
		{
			ERROR("Not supported");
			break;
		}
	}
	_ctx.vx.run_graph(graph, _ctx.num_iterations(),tag);
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::sobel( Image &a, Image *c, Image *d)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxSobel3x3Node(graph, a.vx.img,
	                                        c ? c->vx.img : NULL,
	                                        d ? d->vx.img : NULL), _ctx.border_mode() );

	const char* tag=NULL;
	if(c)
	{
		if(d)
			tag = "sobelxy_3x3";
		else
			tag = "sobelx_3x3";
	}
	if(d)
	{
		tag = "sobely_3x3";
	}

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),tag);
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::sobel5x5( Image &a, Image *c, Image *d)
{
	if(c)
	{
		VXConvolution convX;
		const int16_t sobelX5x5[]=
		{
			-1, -2, 0, 2,1,
			-4, -8, 0, 8,4,
			-6,-12, 0,12,6,
			-4, -8, 0, 8,4,
			-1, -2, 0, 2,1
		};
		RawImage coeffsX(5,5, Format::LUMINANCE16);
		memcpy( coeffsX.data(), sobelX5x5, sizeof(sobelX5x5));

		vx_graph graphx = _ctx.vx.create_graph();

		convX.create( _ctx, coeffsX, CopyData::YES);
		_ctx.vx.set_border_mode( vxConvolveNode(graphx, a.vx.img, convX.convolution, c->vx.img), _ctx.border_mode() );
		_ctx.vx.run_graph(graphx, _ctx.num_iterations(),__FUNCTION__);
		CHECK_VX( vxReleaseGraph( &graphx ) );
	}
	if(d)
	{
		VXConvolution convY;

		const int16_t sobelY5x5[]=
		{
			-1, -4,  -6, -4, -1,
			-2, -8, -12, -8, -2,
			 0,  0,   0,  0,  0,
			 2,  8,  12,  8,  2,
			 1,  4,   6,  4,  1
		};
		RawImage coeffsY(5,5, Format::LUMINANCE16);
		memcpy( coeffsY.data(), sobelY5x5, sizeof(sobelY5x5));
		convY.create( _ctx, coeffsY, CopyData::YES);

		vx_graph graphy = _ctx.vx.create_graph();
		_ctx.vx.set_border_mode( vxConvolveNode(graphy, a.vx.img, convY.convolution, d->vx.img), _ctx.border_mode() );
		_ctx.vx.run_graph(graphy, _ctx.num_iterations(),__FUNCTION__);
		CHECK_VX( vxReleaseGraph( &graphy ) );
	}

}

void VXTarget::sobel7x7( Image &a, Image *c, Image *d)
{
	if(c)
	{
		VXConvolution convX;
		const int16_t sobelX7x7[]=
		{
			-1,  -4,  -5, 0,  5,  4, 1,
			-6, -24, -30, 0, 30, 24, 6,
			-15,-60, -75, 0, 75, 60,15,
			-20,-80,-100, 0,100, 80,20,
			-15,-60, -75, 0, 75, 60,15,
			-6, -24, -30, 0, 30, 24, 6,
			-1,  -4,  -5, 0,  5,  4, 1,
		};
		RawImage coeffsX(7,7, Format::LUMINANCE16);
		memcpy( coeffsX.data(), sobelX7x7, sizeof(sobelX7x7));

		vx_graph graphx = _ctx.vx.create_graph();

		convX.create( _ctx, coeffsX, CopyData::YES);
		_ctx.vx.set_border_mode( vxConvolveNode(graphx, a.vx.img, convX.convolution, c->vx.img), _ctx.border_mode() );
		_ctx.vx.run_graph(graphx, _ctx.num_iterations(),__FUNCTION__);
		CHECK_VX( vxReleaseGraph( &graphx ) );
	}
	if(d)
	{
		VXConvolution convY;
		const int16_t sobelY7x7[]=
		{
			-1,  -6, -15,  -20, -15,  -6, -1,
			-4, -24, -60,  -80, -60, -24, -4,
			-5, -30, -75, -100, -75, -30, -5,
			 0,   0,   0,    0,   0,   0,  0,
			 5,  30,  75,  100,  75,  30,  5,
			 4,  24,  60,   80,  60,  24,  4,
			 1,   6,  15,   20,  15,   6,  1,
		};
		RawImage coeffsY(7,7, Format::LUMINANCE16);
		memcpy( coeffsY.data(), sobelY7x7, sizeof(sobelY7x7));
		convY.create( _ctx, coeffsY, CopyData::YES);

		vx_graph graphy = _ctx.vx.create_graph();
		_ctx.vx.set_border_mode( vxConvolveNode(graphy, a.vx.img, convY.convolution, d->vx.img), _ctx.border_mode() );
		_ctx.vx.run_graph(graphy, _ctx.num_iterations(),__FUNCTION__);
		CHECK_VX( vxReleaseGraph( &graphy ) );
	}
}


void VXTarget::scharr( Image &a, Image *c, Image *d)
{
	if(c)
	{
		VXConvolution convX;
		const int16_t scharrX3x3[]=
		{
			-3,  0,  3,
			-10, 0, 10,
			-3,  0,  3
		};
		RawImage coeffsX(3,3, Format::LUMINANCE16);
		memcpy( coeffsX.data(), scharrX3x3, sizeof(scharrX3x3));
		vx_graph graphx = _ctx.vx.create_graph();

		convX.create( _ctx, coeffsX, CopyData::YES);
		_ctx.vx.set_border_mode( vxConvolveNode(graphx, a.vx.img, convX.convolution, c->vx.img), _ctx.border_mode() );
		_ctx.vx.run_graph(graphx, _ctx.num_iterations(),__FUNCTION__);
		CHECK_VX( vxReleaseGraph( &graphx ) );
	}
	if(d)
	{
		VXConvolution convY;

		const int16_t scharrY3x3[]=
		{
			-3, -10, -3,
			 0,   0,  0,
			 3,  10,  3
		};
		RawImage coeffsY(3,3, Format::LUMINANCE16);
		memcpy( coeffsY.data(), scharrY3x3, sizeof(scharrY3x3));
		convY.create( _ctx, coeffsY, CopyData::YES);

		vx_graph graphy = _ctx.vx.create_graph();
		_ctx.vx.set_border_mode( vxConvolveNode(graphy, a.vx.img, convY.convolution, d->vx.img), _ctx.border_mode() );
		_ctx.vx.run_graph(graphy, _ctx.num_iterations(),__FUNCTION__);
		CHECK_VX( vxReleaseGraph( &graphy ) );
	}
}

void VXTarget::derivative( Image &a, Image *c, Image *d)
{
	if(c)
	{
		VXConvolution convX;
		const int16_t derivativeX3x3[]=
		{
			0,  0, 0,
			-1, 0, 1,
			0,  0, 0
		};
		RawImage coeffsX(3,3, Format::LUMINANCE16);
		memcpy( coeffsX.data(), derivativeX3x3, sizeof(derivativeX3x3));
		vx_graph graphx = _ctx.vx.create_graph();

		convX.create( _ctx, coeffsX, CopyData::YES);
		_ctx.vx.set_border_mode( vxConvolveNode(graphx, a.vx.img, convX.convolution, c->vx.img), _ctx.border_mode() );
		_ctx.vx.run_graph(graphx, _ctx.num_iterations(),__FUNCTION__);
		CHECK_VX( vxReleaseGraph( &graphx ) );
	}
	if(d)
	{
		VXConvolution convY;

		const int16_t derivativeY3x3[]=
		{
			0, -1, 0,
			0,  0, 0,
			0,  1, 0
		};
		RawImage coeffsY(3,3, Format::LUMINANCE16);
		memcpy( coeffsY.data(), derivativeY3x3, sizeof(derivativeY3x3));
		convY.create( _ctx, coeffsY, CopyData::YES);

		vx_graph graphy = _ctx.vx.create_graph();
		_ctx.vx.set_border_mode( vxConvolveNode(graphy, a.vx.img, convY.convolution, d->vx.img), _ctx.border_mode() );
		_ctx.vx.run_graph(graphy, _ctx.num_iterations(),__FUNCTION__);
		CHECK_VX( vxReleaseGraph( &graphy ) );
	}
}

void VXTarget::laplacian_reconstruct( Pyramid &pyramid, Image &a, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();
	_ctx.vx.set_border_mode( vxLaplacianReconstructNode(graph,pyramid.vx.pyramid, a.vx.img, c.vx.img), _ctx.border_mode() );
	_ctx.vx.run_graph(graph, _ctx.num_iterations(),"laplace_reconstruct");
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::laplacian_pyramid( Image &a, Pyramid &pyramid, Image &b, bool timers_enabled)
{
	vx_graph graph = _ctx.vx.create_graph();
	_ctx.vx.set_border_mode( vxLaplacianPyramidNode(graph, a.vx.img, pyramid.vx.pyramid, b.vx.img), _ctx.border_mode() );
	_ctx.vx.run_graph(graph, _ctx.num_iterations(),"laplace_pyr");
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::gaussian_pyramid( Image &a, Pyramid &pyramid)
{
	vx_graph graph = _ctx.vx.create_graph();

	_ctx.vx.set_border_mode( vxGaussianPyramidNode(graph, a.vx.img, pyramid.vx.pyramid), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),"gaussian_pyramid");
	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::canny(Image &a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	vx_threshold hyst = vxCreateThreshold(_ctx.vx.context(), VX_THRESHOLD_TYPE_RANGE, VX_TYPE_UINT8);
	ERROR_ON(!hyst);
	CHECK_VX(vxSetThresholdAttribute(hyst, VX_THRESHOLD_THRESHOLD_UPPER, &upper, sizeof(upper)));
	CHECK_VX(vxSetThresholdAttribute(hyst, VX_THRESHOLD_THRESHOLD_LOWER, &lower, sizeof(lower)));

	_ctx.vx.set_border_mode(vxCannyEdgeDetectorNode(graph, a.vx.img, hyst, gradient_size, norm_type == 1 ? VX_NORM_L1 : VX_NORM_L2, c.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseThreshold(&hyst));
	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::canny_fp16(Image &a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image &c)
{
	VXTarget::canny(a,upper,lower,gradient_size,norm_type,c);
}

void VXTarget::mean_stddev(Image &a, Scalar<float> &mean, Scalar<float> &stddev)
{
	vx_graph graph = _ctx.vx.create_graph();

	vx_scalar sc_mean   = vxCreateScalar(_ctx.vx.context(), VX_TYPE_FLOAT32, &mean.vx.scalar);
	vx_scalar sc_stddev = vxCreateScalar(_ctx.vx.context(), VX_TYPE_FLOAT32, &stddev.vx.scalar);

	_ctx.vx.set_border_mode(vxMeanStdDevNode(graph, a.vx.img, sc_mean, sc_stddev), _ctx.border_mode());

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseGraph(&graph));

	CHECK_VX(vxCopyScalar(sc_mean, &mean.vx.scalar, VX_READ_ONLY, VX_MEMORY_TYPE_HOST));
	CHECK_VX(vxCopyScalar(sc_stddev, &stddev.vx.scalar, VX_READ_ONLY, VX_MEMORY_TYPE_HOST));
}

void VXTarget::magnitude( Image &a, Image &b, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	vxMagnitudeNode(graph, a.vx.img, b.vx.img, c.vx.img);

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::magnitude_fp16(Image &a, Image &b, Image &c)
{
	VXTarget::magnitude(a,b,c);
}

void VXTarget::phase( Image &a, Image &b, Image &c)
{
	vx_graph graph = _ctx.vx.create_graph();

	vxPhaseNode(graph, a.vx.img, b.vx.img, c.vx.img);

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::threshold_binary( Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value)
{
	vx_graph graph = _ctx.vx.create_graph();
	vx_int32 vxthreshold = threshold;
	vx_int32 vxfalse = false_value;
	vx_int32 vxtrue = true_value;

	vx_threshold thr = vxCreateThreshold(_ctx.vx.context(), VX_THRESHOLD_TYPE_BINARY, VX_TYPE_UINT8);
	ERROR_ON(!thr);

	CHECK_VX(vxSetThresholdAttribute(thr, VX_THRESHOLD_THRESHOLD_VALUE, &vxthreshold, sizeof(vxthreshold)));
	CHECK_VX(vxSetThresholdAttribute(thr, VX_THRESHOLD_TRUE_VALUE, &vxtrue, sizeof(vxtrue)));
	CHECK_VX(vxSetThresholdAttribute(thr, VX_THRESHOLD_FALSE_VALUE, &vxfalse, sizeof(vxfalse)));

	vxThresholdNode(graph, a.vx.img, thr, b.vx.img);
	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseThreshold(&thr));
	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::threshold_range( Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value, uint8_t upper)
{
	vx_graph graph = _ctx.vx.create_graph();
	vx_int32 vxthreshold = threshold;
	vx_int32 vxfalse = false_value;
	vx_int32 vxtrue = true_value;
	vx_int32 vxupper = upper;

	vx_threshold thr = vxCreateThreshold(_ctx.vx.context(), VX_THRESHOLD_TYPE_RANGE, VX_TYPE_UINT8);
	ERROR_ON(!thr);

	CHECK_VX(vxSetThresholdAttribute(thr, VX_THRESHOLD_THRESHOLD_LOWER, &vxthreshold, sizeof(vxthreshold)));
	CHECK_VX(vxSetThresholdAttribute(thr, VX_THRESHOLD_THRESHOLD_UPPER, &vxupper, sizeof(vxupper)));
	CHECK_VX(vxSetThresholdAttribute(thr, VX_THRESHOLD_TRUE_VALUE, &vxtrue, sizeof(vxtrue)));
	CHECK_VX(vxSetThresholdAttribute(thr, VX_THRESHOLD_FALSE_VALUE, &vxfalse, sizeof(vxfalse)));

	vxThresholdNode(graph, a.vx.img, thr, b.vx.img);
	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseThreshold(&thr));
	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::remap(Image &a, RemapTable &table, Image &d, InterpolationMode policy)
{
	vx_graph graph = _ctx.vx.create_graph();

	const vx_enum interpolation = VXHelper::convert_to_vx_interpolation(policy);

	_ctx.vx.set_border_mode( vxRemapNode( graph, a.vx.img, table.vx.table, interpolation, d.vx.img), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, _ctx.num_iterations(), __FUNCTION__);

	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::channel_extract( Image &a, Channel channel, Image &c )
{
	vx_graph graph = _ctx.vx.create_graph();

	vx_channel_e vx_channel = VX_CHANNEL_0;
	switch( channel )
	{
		case Channel::R:
			vx_channel = VX_CHANNEL_R ;
			break;
		case Channel::G:
			vx_channel = VX_CHANNEL_G ;
			break;
		case Channel::B:
			vx_channel = VX_CHANNEL_B ;
			break;
		case Channel::A:
			vx_channel = VX_CHANNEL_A ;
			break;
		case Channel::Y:
			vx_channel = VX_CHANNEL_Y ;
			break;
		case Channel::U:
			vx_channel = VX_CHANNEL_U ;
			break;
		case Channel::V:
			vx_channel = VX_CHANNEL_V ;
			break;
		default:
			ERROR("Invalid channel.");
			break;
	}

	vxChannelExtractNode(graph, a.vx.img, vx_channel, c.vx.img);

	_ctx.vx.run_graph(graph, _ctx.num_iterations(), __FUNCTION__);

	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::tablelookup( Image &a , Lut &b, Image &c )
{
	vx_graph graph = _ctx.vx.create_graph();

	vxTableLookupNode(graph, a.vx.img, b.vx.lut, c.vx.img);

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::convert_depth( Image &a, Image &c, ConvertPolicy policy , uint32_t shift)
{
	vx_graph graph = _ctx.vx.create_graph();
	vx_scalar s = vxCreateScalar( _ctx.vx.context(), VX_TYPE_INT32, &shift);
	switch(policy)
	{
		case ConvertPolicy::SATURATE:
			vxConvertDepthNode(graph, a.vx.img, c.vx.img, VX_CONVERT_POLICY_SATURATE, s);
			break;
		case ConvertPolicy::WRAP:
			vxConvertDepthNode(graph, a.vx.img, c.vx.img, VX_CONVERT_POLICY_WRAP, s);
			break;
		default:
			ERROR("Invalid policy");
			break;
	}
	_ctx.vx.run_graph(graph, _ctx.num_iterations(), __FUNCTION__);
	CHECK_VX(vxReleaseGraph(&graph));
	CHECK_VX( vxReleaseScalar( &s ) );
}

template <class T, enum vx_type_e type>
static inline void minmaxlocation_vx( Context &_ctx, Image &a, Scalar<int32_t> &min, Scalar<int32_t> &max,
                                      ICoord2DArray *min_loc, ICoord2DArray *max_loc,
                                      Scalar<uint32_t> *min_count, Scalar<uint32_t> *max_count)
{
	T tmp_min = (T)min.vx.scalar;
	T tmp_max = (T)max.vx.scalar;

	vx_graph graph = _ctx.vx.create_graph();

	vx_scalar vx_min = vxCreateScalar(_ctx.vx.context(), type, &tmp_min);
	vx_scalar vx_max = vxCreateScalar(_ctx.vx.context(), type, &tmp_max);
	vx_scalar vx_min_loc = ( min_count ) ? vxCreateScalar(_ctx.vx.context(), VX_TYPE_UINT32 , &min_count->vx.scalar) : nullptr;
	vx_scalar vx_max_loc = ( max_count ) ? vxCreateScalar(_ctx.vx.context(), VX_TYPE_UINT32 , &max_count->vx.scalar) : nullptr;
	vx_array vx_min_loc_array = ( min_loc ) ? min_loc->vx.a : nullptr;
	vx_array vx_max_loc_array = ( max_loc ) ? max_loc->vx.a : nullptr;

	vxMinMaxLocNode(graph, a.vx.img, vx_min, vx_max, vx_min_loc_array, vx_max_loc_array, vx_min_loc, vx_max_loc);

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseGraph(&graph));

	// Copy scalars and release
	CHECK_VX(vxCopyScalar(vx_min, &tmp_min, VX_READ_ONLY, VX_MEMORY_TYPE_HOST));
	CHECK_VX(vxCopyScalar(vx_max, &tmp_max, VX_READ_ONLY, VX_MEMORY_TYPE_HOST));
	CHECK_VX( vxReleaseScalar( &vx_min ) );
	CHECK_VX( vxReleaseScalar( &vx_max ) );

	// Release min and max counter if created
	if ( min_count )
	{
		CHECK_VX(vxCopyScalar(vx_min_loc, &min_count->vx.scalar, VX_READ_ONLY, VX_MEMORY_TYPE_HOST));
		CHECK_VX( vxReleaseScalar( &vx_min_loc ) );
	}
	if ( max_count )
	{
		CHECK_VX(vxCopyScalar(vx_max_loc, &max_count->vx.scalar, VX_READ_ONLY, VX_MEMORY_TYPE_HOST));
		CHECK_VX( vxReleaseScalar( &vx_max_loc ) );
	}

	min.vx.scalar = (int32_t)tmp_min;
	max.vx.scalar = (int32_t)tmp_max;
}

void VXTarget::minmaxlocation( Image &a, Scalar<int32_t> &min, Scalar<int32_t> &max, ICoord2DArray *min_loc, ICoord2DArray *max_loc,
                               Scalar<uint32_t> *min_count, Scalar<uint32_t> *max_count)
{
	if( Format::LUMINANCE16 == a.vx.format )
		minmaxlocation_vx<uint16_t, VX_TYPE_INT16>(_ctx, a, min, max, min_loc, max_loc, min_count, max_count);
	else
		minmaxlocation_vx<uint8_t, VX_TYPE_UINT8>(_ctx, a, min, max, min_loc, max_loc, min_count, max_count);
}

void VXTarget::optical_flow(Image &a, Image &b, IKeyPointArray *old_points, IKeyPointArray *new_points_estimates,
                            IKeyPointArray *new_points, Termination termination, float epsilon, uint32_t num_iterations,
                            bool use_initial_estimate, size_t window_dimension, size_t num_levels)
{
	vx_graph graph = _ctx.vx.create_graph();
	vx_enum vxtermination = VX_TERM_CRITERIA_ITERATIONS;

	vx_pyramid old_pyramid;
	vx_pyramid new_pyramid;

	old_pyramid = vxCreatePyramid(_ctx.vx.context(), num_levels, VX_SCALE_PYRAMID_HALF, a.vx.width, a.vx.height, VX_DF_IMAGE_U8);
	new_pyramid = vxCreatePyramid(_ctx.vx.context(), num_levels, VX_SCALE_PYRAMID_HALF, a.vx.width, a.vx.height, VX_DF_IMAGE_U8);

	_ctx.vx.set_border_mode( vxGaussianPyramidNode(graph, a.vx.img, old_pyramid), _ctx.border_mode() );
	_ctx.vx.set_border_mode( vxGaussianPyramidNode(graph, b.vx.img, new_pyramid), _ctx.border_mode() );

	_ctx.vx.run_graph(graph, 1, "");

	switch(termination)
	{
		case Termination::TERM_CRITERIA_ITERATIONS:
			vxtermination = VX_TERM_CRITERIA_ITERATIONS;
			break;
		case Termination::TERM_CRITERIA_EPSILON:
			vxtermination = VX_TERM_CRITERIA_EPSILON;
			break;
		case Termination::TERM_CRITERIA_BOTH:
			vxtermination = VX_TERM_CRITERIA_BOTH;
			break;
		default:
			ERROR("Not supported");
	}

	vx_scalar vxepsilon = vxCreateScalar( _ctx.vx.context(), VX_TYPE_FLOAT32, &epsilon);
	vx_scalar vxnum_iterations = vxCreateScalar( _ctx.vx.context(), VX_TYPE_UINT32, &num_iterations);
	vx_scalar vxuse_initial_estimate = vxCreateScalar( _ctx.vx.context(), VX_TYPE_BOOL, &use_initial_estimate);

	_ctx.vx.set_border_mode(vxOpticalFlowPyrLKNode(graph, old_pyramid, new_pyramid, old_points->vx.a, new_points_estimates->vx.a, new_points->vx.a,
	                                               vxtermination, vxepsilon, vxnum_iterations, vxuse_initial_estimate, window_dimension), _ctx.border_mode() ) ;
	_ctx.vx.run_graph(graph, _ctx.num_iterations(), __FUNCTION__);
	CHECK_VX( vxReleaseGraph( &graph ) );
	CHECK_VX( vxReleaseScalar( &vxepsilon ) );
	CHECK_VX( vxReleaseScalar( &vxnum_iterations ) );
	CHECK_VX( vxReleaseScalar( &vxuse_initial_estimate ) );
}

void VXTarget::histogram(Image &a, Distribution &b)
{
	vx_graph graph = _ctx.vx.create_graph();

	vxHistogramNode(graph, a.vx.img, b.vx.dist);

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::channel_combine(Image &plane0, Image &plane1, Image &plane2, Image &plane3, Image &out)
{
	vx_graph graph = _ctx.vx.create_graph();

	vxChannelCombineNode(graph, plane0.vx.img, plane1.vx.img, plane2.vx.img, plane3.vx.img, out.vx.img);

	_ctx.vx.run_graph(graph, _ctx.num_iterations(), __FUNCTION__);

	CHECK_VX( vxReleaseGraph( &graph ) );
}

void VXTarget::integral(Image &a, Image &b)
{
	vx_graph graph = _ctx.vx.create_graph();

	vxIntegralImageNode(graph, a.vx.img, b.vx.img);

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::equalize_histogram(Image &a, Image &b)
{
	vx_graph graph = _ctx.vx.create_graph();

	vxEqualizeHistNode(graph, a.vx.img, b.vx.img);

	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);

	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::non_linear_filter(Image &a, Image &b, NonLinearFilterFunction function, MatrixPattern pattern, uint mask_size, uint8_t* mask)
{
	// Create  vx_matrix
	vx_enum matrix_pattern = VX_PATTERN_BOX;
	switch(pattern)
	{
		case MatrixPattern::BOX:
			matrix_pattern = VX_PATTERN_BOX;
			break;
		case MatrixPattern::CROSS:
			matrix_pattern = VX_PATTERN_CROSS;
			break;
		case MatrixPattern::DISK:
			matrix_pattern = VX_PATTERN_DISK;
			break;
		case MatrixPattern::OTHER:
			matrix_pattern = VX_PATTERN_OTHER;
			break;
		default:
			ERROR("Unknown matrix pattern.");
			break;
	}
	vx_matrix vx_mask = vxCreateMatrixFromPattern(_ctx.vx.context(), matrix_pattern, mask_size, mask_size);
	if(MatrixPattern::OTHER == pattern)
	{
		ERROR_ON(nullptr == mask);
		CHECK_VX(vxCopyMatrix(vx_mask, (void*) mask, VX_WRITE_ONLY, VX_MEMORY_TYPE_HOST));
	}
	else
	{
		std::unique_ptr<uint8_t[]> mask_data(new uint8_t[mask_size * mask_size]());
		CHECK_VX(vxCopyMatrix(vx_mask, (void*) mask_data.get(), VX_READ_ONLY, VX_MEMORY_TYPE_HOST));
		ERROR_ON(!VXHelper::pattern_check(mask_data.get(), mask_size, mask_size, matrix_pattern));
	}

	// Get filter function
	vx_enum filter_function = VX_NONLINEAR_FILTER_MEDIAN;
	switch(function)
	{
		case NonLinearFilterFunction::MEDIAN:
			filter_function = VX_NONLINEAR_FILTER_MEDIAN;
			break;
		case NonLinearFilterFunction::MIN:
			filter_function = VX_NONLINEAR_FILTER_MIN;
			break;
		case NonLinearFilterFunction::MAX:
			filter_function = VX_NONLINEAR_FILTER_MAX;
			break;
		default:
			ERROR("Unknown non linear function. ");
			break;
	}

	// Run operation
	vx_graph graph = _ctx.vx.create_graph();
	_ctx.vx.set_border_mode( vxNonLinearFilterNode(graph, filter_function, a.vx.img, vx_mask, b.vx.img), _ctx.border_mode() );
	_ctx.vx.run_graph(graph, _ctx.num_iterations(),__FUNCTION__);
	CHECK_VX(vxReleaseGraph(&graph));
}

void VXTarget::activation_layer(Tensor &in, Tensor &out, ActivationLayerInfo act_info)
{
	_eigen.activation_layer(in.vx, out.vx, act_info);
}

void VXTarget::normalization_layer(Tensor &a, Tensor &b, NormalizationLayerInfo norm_info)
{
	ERROR_ON("NOT IMPLEMENTED");
}

void VXTarget::batchnormalization_layer(Tensor &a, Tensor &b, Tensor &mean, Tensor &var, Tensor &beta, Tensor &gamma, float epsilon)
{
	_eigen.batchnormalization_layer( a.vx, b.vx, mean.vx, var.vx, beta.vx, gamma.vx, epsilon);
}

void VXTarget::pooling_layer(Tensor &in, Tensor &out, PoolingLayerInfo pool_info)
{
	_eigen.pooling_layer( in.vx, out.vx, pool_info);
}

void VXTarget::softmax_layer(Tensor &in, Tensor &out)
{
	_eigen.softmax_layer( in.vx, out.vx );
}

void VXTarget::tensor_ops(Tensor &in1, Tensor &in2, Tensor &out, TensorOp op)
{
	_eigen.tensor_ops( in1.vx, in2.vx, out.vx, op);
}

void VXTarget::convolution_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info,Tensor *rw)
{
	_eigen.convolution_layer( in.vx, w.vx, b ? &b->vx : nullptr, out.vx, conv_info);
}

void VXTarget::hog_descriptor(Image &a, HOG &b, Tensor &c)
{
	/* Note: We need this empty function otherwise HOGDescriptor test crashes.*/
	return;
}

void VXTarget::fully_connected_layer(Tensor &input, Tensor &weight, Tensor* bias, Tensor &output, bool are_weights_reshaped)
{
	_eigen.fully_connected_layer( input.vx, weight.vx, bias ? &bias->vx : nullptr, output.vx);
}

void VXTarget::locally_connected_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info)
{
	_eigen.locally_connected_layer( in.vx, w.vx, b ? &b->vx : nullptr, out.vx, conv_info);
}

void VXTarget::transpose( Image &a, Image &b)
{
	/*Note: We need this empty function otherwise the transpose test crashes. The current reference implementation is within the Transpose test*/
	return;
}

void VXTarget::gemm_fp32(Tensor &a, Tensor &b, Tensor &c, Tensor &d, float alpha, float beta)
{
	_eigen.gemm(a.vx, b.vx, c.vx, d.vx, alpha, beta);
}

void VXTarget::gemm_fp16(Image &a, Image &b, Image &c, Image &d, float alpha, float beta)
{
	/*Note: We need this empty function otherwise the gemm_fp16 test crashes. The current reference implementation is within the gemm test*/
	return;
}

void VXTarget::concatenate_depth(std::vector<std::unique_ptr<Tensor>> &a, Tensor &b)
{
	std::vector<EigenTensor*> input;

	for(auto& temp : a)
	{
		input.push_back(&(temp->vx));
	}

	_eigen.concatenate_depth(input, b.vx);
}
