#include "mali_cv/Validate.h"

#include "mali_cv/Error.h"
#include "mali_cv/ITest.h"
#include "mali_cv/Scalar.h"

#include <algorithm>
#include <cmath>
#include <iostream>
#include <limits>

using namespace mali_cv;

namespace
{
arm_compute::ValidRegion get_valid_region(const Image &in, TargetID target)
{
	switch(target)
	{
		case TargetID::CL:
			return in.cl.valid_region();
		case TargetID::CV:
			return in.cv.valid_region();
		case TargetID::VX:
			return in.vx.valid_region();
		case TargetID::NEON:
			return in.ne.valid_region();
		default:
			ERROR("Target not supported");
	}
}

RawImage get_image(Image &in, TargetID target)
{
	RawImage out;

	switch(target)
	{
		case TargetID::CL:
			in.cl.store(out);
			break;
		case TargetID::CV:
			in.cv.store(out);
			break;
		case TargetID::VX:
			in.vx.store(out);
			break;
		case TargetID::NEON:
			in.ne.store(out);
			break;
		default:
			ERROR("Target not supported");
	}

	return out;
}

std::vector<float> get_tensor(Tensor &in, TargetID target)
{
	std::vector<float> out;

	switch(target)
	{
		case TargetID::VX:
			out.resize(in.vx.size());
			in.vx.store(reinterpret_cast<uint8_t*>(out.data()));
			break;
		case TargetID::CL:
			out.resize(in.cl.size());
			in.cl.store(reinterpret_cast<uint8_t*>(out.data()));
			break;
		case TargetID::NEON:
			out.resize(in.ne.size());
			in.ne.store(reinterpret_cast<uint8_t*>(out.data()));
			break;
		case TargetID::CV:
			out.resize(in.cv.size());
			in.cv.store(reinterpret_cast<uint8_t*>(out.data()));
			break;
		default:
			ERROR("Target not supported");
	}

	return out;
}

void read_corners( std::vector<IKeyPoint> &out, const IKeyPointArray &in, TargetID target)
{
	switch( target )
	{
		case TargetID::VX:
			in.vx.read(out);
			break;
		case TargetID::NEON:
			in.ne.read(out);
			break;
		case TargetID::CV:
			in.cv.read(out);
			break;
		case TargetID::CL:
			in.cl.read(out);
			break;
		default:
			ERROR("Target not supported");
	}

	std::sort(out.begin(), out.end(), [] (const IKeyPoint &a, const IKeyPoint &b) -> bool
	{
		return a.strength > b.strength;
	});
}

void read_distribution( std::vector<uint32_t> &out, const Distribution &in, TargetID target)
{
	switch( target )
	{
		case TargetID::VX:
			in.vx.read(out);
			break;
		case TargetID::NEON:
			in.ne.read(out);
			break;
		case TargetID::CV:
			in.cv.read(out);
			break;
		case TargetID::CL:
			in.cl.read(out);
			break;
		default:
			ERROR("Target not supported");
	}
}

void read_coords2D( std::vector<ICoordinates2D> &out, const ICoord2DArray &in, TargetID target )
{
	switch( target )
	{
		case TargetID::VX:
			in.vx.read(out);
			break;
		case TargetID::NEON:
			in.ne.read(out);
			break;
		case TargetID::CV:
			in.cv.read(out);
			break;
		case TargetID::CL:
			in.cl.read(out);
			break;
		default:
			ERROR("Target not supported");
	}
	std::sort(out.begin(),out.end(),
		[](const ICoordinates2D & a, const ICoordinates2D & b) -> bool
		{
			if( a.x == b.x )
				return a.y < b.y;
			else
				return a.x < b.x;
		});
}

float convert_to_float(const uint8_t *ptr, Format format)
{
	switch(format)
	{
		case Format::LUMINANCE:
		case Format::RGB888:
		case Format::RGBA8888:
		case Format::UYVY:
		case Format::YUYV422:
			return ptr[0];
		case Format::U16:
			return *reinterpret_cast<const uint16_t*>(ptr);
		case Format::LUMINANCE16:
		case Format::RGBA16:
			return *reinterpret_cast<const int16_t*>(ptr);
		case Format::LUMINANCE32:
			return *reinterpret_cast<const int32_t*>(ptr);
		case Format::U32:
			return *reinterpret_cast<const uint32_t*>(ptr);
		case Format::FP16C1:
		case Format::FP16C2:
			return *reinterpret_cast<const mali_cv::float16_t*>(ptr);
		case Format::RGBA_FLOAT:
		case Format::FP32C1:
		case Format::FP32C2:
			return *reinterpret_cast<const float*>(ptr);
		/* Planar formats */
		case Format::NV12:
		case Format::IYUV:
		case Format::NV21:
		case Format::YUV444:
			return ptr[0];
		default:
			ERROR("Format 0x%x not supported", format);
			return 0.f;
	}
}
}

unsigned int Validate::compare(const RawImage &img1, const RawImage &img2,
                               float threshold, float acceptable_ratio_of_mismatches, bool print_all)
{
	ERROR_ON(img1.width() != img2.width());
	ERROR_ON(img1.height() != img2.height());

	return compare(img1, img2, arm_compute::ValidRegion(arm_compute::Coordinates(0, 0),
	                                               arm_compute::TensorShape(img1.width(), img1.height())),
	               threshold, acceptable_ratio_of_mismatches, print_all);
}

unsigned int Validate::compare(const RawImage &img1, const RawImage &img2, const arm_compute::ValidRegion &valid_region,
                               float threshold, float acceptable_ratio_of_mismatches, bool print_all)
{
	ERROR_ON(img1.width() < valid_region.anchor[0] + valid_region.shape[0]);
	ERROR_ON(img1.height() < valid_region.anchor[1] + valid_region.shape[1]);
	ERROR_ON(img2.width() < valid_region.anchor[0] + valid_region.shape[0]);
	ERROR_ON(img2.height() < valid_region.anchor[1] + valid_region.shape[1]);
	ERROR_ON(img1.format() != img2.format());
	ERROR_ON(img1.num_channels() != img2.num_channels());
	ERROR_ON(img1.element_size() != img2.element_size());
	ERROR_ON(img1.num_channels() > 4);

	std::array<unsigned int, 4> errors = {0, 0, 0, 0};
	unsigned int num_errors = 0;
	unsigned int num_values = 0;

	for(int y = valid_region.anchor[1], end_y = y + valid_region.shape[1]; y < end_y; ++y)
	{
	    for(int x = valid_region.anchor[0], end_x = x + valid_region.shape[0]; x < end_x; ++x)
		{
			bool valid = true;

	        for(unsigned int c = 0; c < img1.num_channels(); ++c)
			{
	            const uint8_t * const val1_ptr = img1.data() + (y * img1.width() + x * img1.num_channels() + c) * img1.element_size();
	            const uint8_t * const val2_ptr = img2.data() + (y * img2.width() + x * img2.num_channels() + c) * img2.element_size();

				const float val1 = convert_to_float(val1_ptr, img1.format());
				const float val2 = convert_to_float(val2_ptr, img2.format());

				if(std::fabs(val1 - val2) > threshold)
				{
					++errors[c];

					valid = false;

					if(print_all)
					{
						printf("[%d,%d][%d] %.2f != %.2f\n", x, y, c, val1, val2);
					}
				}

				++num_values;
			}

			if(!valid)
			{
				++num_errors;
			}
		}

		// Since the actual element_size() of NV12/IYUV/NV21 is 1.5, and it's treated as 2, so need to skip the redundant pixels
		if(img1.format() == Format::NV12 || img1.format() == Format::IYUV || img1.format() == Format::NV21)
		{
			if(num_values == (img1.width() * img1.height() * 3) / 2)
	        {
				break;
	        }
		}
	}

	if(num_errors > num_values * acceptable_ratio_of_mismatches)
	{
		ERROR("ERROR: %u values (%2.2f%%) mismatched (maximum tolerated %2.2f%%)\n", num_errors, 100.0f * num_errors / num_values, 100.0f * acceptable_ratio_of_mismatches);
	}
	else if(num_errors > 0)
	{
		printf("WARNING: %u values (%2.2f%%) mismatched (maximum tolerated %2.2f%%):", num_errors, 100.0f * num_errors / num_values, 100.0f * acceptable_ratio_of_mismatches);

	    for(unsigned int c = 0; c < errors.size(); ++c)
	    {
			printf(" %d", errors[c]);
	    }

		printf("\n");
	}
	else
	{
		printf("RawImages match\n");
	}

	return num_errors;
}

bool Validate::compare(const arm_compute::ValidRegion &region1, const arm_compute::ValidRegion &region2)
{
	if(region1.anchor.num_dimensions() != region2.anchor.num_dimensions())
	{
		ERROR("ERROR: Valid region's anchors have different number of dimensions (%u != %u)!\n", region1.anchor.num_dimensions(), region2.anchor.num_dimensions());
		return false;
	}

	if(region1.shape.num_dimensions() != region2.shape.num_dimensions())
	{
		ERROR("ERROR: Valid region's shapes have different number of dimensions (%u != %u)!\n", region1.shape.num_dimensions(), region2.shape.num_dimensions());
		return false;
	}

	bool valid = true;

	for(unsigned int d = 0; d < region1.anchor.num_dimensions(); ++d)
	{
		if(region1.anchor[d] != region2.anchor[d])
		{
			ERROR("ERROR: Valid region's anchors mismatch in dimension %d (%d != %d)!\n", d, region1.anchor[d], region2.anchor[d]);
			valid = false;
		}
	}

	for(unsigned int d = 0; d < region1.shape.num_dimensions(); ++d)
	{
		if(region1.shape[d] != region2.shape[d])
		{
			ERROR("ERROR: Valid region's shapes mismatch in dimension %d (%d != %d)!\n", d, region1.shape[d], region2.shape[d]);
			valid = false;
		}
	}

	return valid;
}


void Validate::single_image_output(const Environment &env, Image &img, float ref_lo, float ref_hi, const arm_compute::ValidRegion &valid_region, float threshold_f)
{
	RawImage rimg = get_image(img, env.to_check);
	ERROR_ON(rimg.width() < valid_region.anchor[0] + valid_region.shape[0]);
	ERROR_ON(rimg.height() < valid_region.anchor[1] + valid_region.shape[1]);
	unsigned int num_errors = 0;
	unsigned int num_values = 0;

	for(int y = valid_region.anchor[1], end_y = y + valid_region.shape[1]; y < end_y; ++y)
	{
		for(int x = valid_region.anchor[0], end_x = x + valid_region.shape[0]; x < end_x; ++x)
		{
			const uint8_t * const val1_ptr = rimg.data() + (y * rimg.width() + x * rimg.num_channels()) * rimg.element_size();
			const float fval = convert_to_float(val1_ptr, rimg.format());
			if(fval > ref_hi || fval < ref_lo )
			{
				++num_errors;
				if(env.print_all)
				{
					printf("[%d,%d] %.2f not in valid range [%.2f,%.2f]\n", x, y, fval, ref_lo, ref_hi);
				}

			}
			++num_values;
		}
	}

	if(num_errors)
	{
		ERROR("ERROR: %u values (%2.2f%%) mismatched (maximum tolerated %2.2f%%)\n", num_errors, 100.0f * num_errors / num_values, 0.f);
	}
	else if(num_errors > 0)
	{
		printf("WARNING: %u values (%2.2f%%) mismatched (maximum tolerated %2.2f%%):", num_errors, 100.0f * num_errors / num_values, 0.f);
		printf("\n");
	}
	else
	{
		printf("RawImages match\n");
	}
}

void Validate::single_image_output(const Environment &env, Image &img, const RawImage &ref, const arm_compute::ValidRegion &valid_region, float threshold_f)
{
	RawImage other = get_image(img, env.to_check);

	compare(other, ref, valid_region, threshold_f, env.acceptable_ratio_of_mismatches, env.print_all);
}

void Validate::single_image_output(Image &img, const RawImage &ref, float threshold_f, const Environment &env, uint border_width)
{
	single_image_output(env, img, ref, arm_compute::ValidRegion(arm_compute::Coordinates(static_cast<int>(border_width), static_cast<int>(border_width)),
	                                                       arm_compute::TensorShape(ref.width() - 2 * border_width, ref.height() - 2 * border_width)),
	                    threshold_f);
}

void Validate::single_image_output(const Environment &env, Image &img)
{
	arm_compute::ValidRegion other_valid_region = get_valid_region(img, env.to_check);
	arm_compute::ValidRegion ref_valid_region = get_valid_region(img, env.reference);

	if(compare(other_valid_region, ref_valid_region))
	{
		single_image_output(env, img, ref_valid_region);
	}
}

void Validate::single_image_output( Image &img, const Environment &env, uint border_width )
{
	RawImage ref = get_image(img, env.reference);

	single_image_output(env, img, arm_compute::ValidRegion(arm_compute::Coordinates(static_cast<int>(border_width), static_cast<int>(border_width)),
	                                                  arm_compute::TensorShape(ref.width() - 2 * border_width, ref.height() - 2 * border_width)));
}

void Validate::single_image_output(const Environment &env, Image &img, const arm_compute::ValidRegion &valid_region)
{
	RawImage ref = get_image(img, env.reference);
	RawImage other = get_image(img, env.to_check);

	compare(other, ref, valid_region, env.min_threshold, env.acceptable_ratio_of_mismatches, env.print_all);
}

void Validate::double_image_output( Image &img1, Image &img2, const Environment &env, uint border_width )
{
	single_image_output(img1, env, border_width);
	single_image_output(img2, env, border_width);
}

void Validate::double_image_output(const Environment &env, Image &img1, Image &img2)
{
	single_image_output(env, img1);
	single_image_output(env, img2);
}

void Validate::double_image_output(const Environment &env, Image &img1, Image &img2, const arm_compute::ValidRegion &valid_region)
{
	single_image_output(env, img1, valid_region);
	single_image_output(env, img2, valid_region);
}

void Validate::single_tensor_output(Tensor &tensor, const Environment &env, uint border_width, float custom_threshold)
{
	single_tensor_output(env, tensor, custom_threshold);
}

void Validate::single_tensor_output(const Environment &env, Tensor &tensor, float threshold)
{
	std::vector<float> ref = get_tensor(tensor, env.reference);
	std::vector<float> other = get_tensor(tensor, env.to_check);

	unsigned int mismatches = 0;

	if(threshold == 0.f)
	{
		threshold = std::numeric_limits<float>::epsilon() + 1e-4f + env.min_threshold;
	}

	if(ref.size() != other.size())
	{
		ERROR("Reference and output have different sizes");
	}

	for(unsigned int i = 0; i < ref.size(); ++i)
	{
		if(std::fabs(ref[i] - other[i]) > threshold)
		{
			if(env.print_all)
			{
				printf("Value[%u] does not match! Ref: [%03f] -- Other: [%03f]\n", i, ref[i], other[i]);
			}

			++mismatches;
		}
	}

	std::cout << "Mismatches " << mismatches << "\n";

	ERROR_ON(mismatches > (ref.size() * env.acceptable_ratio_of_mismatches));
}

void Validate::corners_output(const IKeyPointArray &corners, const Environment &env, int32_t threshold_distance, float threshold_strength, std::pair<unsigned int, unsigned int> image_size, unsigned int border)
{
	std::vector<IKeyPoint> ref, other;

	read_corners(other, corners, env.to_check);
	read_corners(ref, corners, env.reference);

	uint mismatches = 0;

	for(const IKeyPoint& r : ref)
	{
		if( r.x < static_cast<int32_t>(border) || r.x >= static_cast<int32_t>(image_size.first) - static_cast<int32_t>(border) ||
		    r.y < static_cast<int32_t>(border) || r.y >= static_cast<int32_t>(image_size.second) - static_cast<int32_t>(border))
		{
			continue;
		}

		bool found = false;

		for(const IKeyPoint& o : other)
		{
			if(std::fabs(o.x - r.x) <= threshold_distance && std::fabs(o.y - r.y) <= threshold_distance)
			{
				found = true;
				break;
			}
		}

		if(!found)
		{
			if(std::fabs(r.strength) > threshold_strength)
			{
				if(env.print_all)
				{
					std::cout << "Missing from To_Check: " << r << "\n";
				}
				++mismatches;
			}
		}
	}

	std::cout << "\n";

	for(const IKeyPoint& o : other)
	{
		if( o.x < static_cast<int32_t>(border) || o.x >= (static_cast<int32_t>(image_size.first) - static_cast<int32_t>(border)) ||
		    o.y < static_cast<int32_t>(border) || o.y >= (static_cast<int32_t>(image_size.second) - static_cast<int32_t>(border)))
		{
			continue;
		}

		bool found = false;

		for(const IKeyPoint& r : ref)
		{
			if(std::fabs(o.x - r.x) <= threshold_distance && std::fabs(o.y - r.y) <= threshold_distance)
			{
				found = true;
				break;
			}
		}

		if(!found)
		{
			if(std::fabs(o.strength) > threshold_strength)
			{
				if(env.print_all)
				{
					std::cout << "Missing in Reference: " << o << "\n";
				}
				++mismatches;
			}
		}
	}

	printf(" %d mismatches out of %zu (%2.2f %%, maximum tolerated %2.2f %%)\n", mismatches, ref.size(), 100.0f * mismatches / ref.size(), 100.0f * env.acceptable_ratio_of_mismatches);
	ERROR_ON(mismatches > (ref.size() * env.acceptable_ratio_of_mismatches));
}

void Validate::coords2D_output( const ICoord2DArray &coords, int32_t num_coords, const Environment &env)
{
	std::vector<ICoordinates2D> ref,other;
	read_coords2D( ref, coords, env.reference );
	read_coords2D( other, coords, env.to_check);
	uint mismatches = 0;
	for( const ICoordinates2D& k : ref ) {
		bool found(false);
		for( const auto w : other ) if( w.x == k.x && w.y == k.y ) { found = true; break; }
		if( ! found )
		{
			if(env.print_all)
			{
				std::cout << "Missing from To_Check:" << k << std::endl;
			}
			++mismatches;
		}
	}
	std::cout << std::endl;

	for( const ICoordinates2D& k : other ) {
		bool found(false);
		for( const auto w : ref ) if( w.x == k.x && w.y == k.y ) { found = true; break; }
		if( ! found )
		{
			if(env.print_all)
			{
				std::cout << "Missing in Reference" << k << std::endl;
			}
			++mismatches;
		}
	}
	printf(" %d mismatches out of %zu (%2.2f %%, maximum tolerated %2.2f %%)\n", mismatches, ref.size(), 100.0f * mismatches / ref.size(), 100.0f * env.acceptable_ratio_of_mismatches);
	ERROR_ON(mismatches > (ref.size() * env.acceptable_ratio_of_mismatches));
}

void Validate::distribution_output(const Distribution& dist, const Environment& env)
{
	std::vector<uint32_t> ref,other;
	read_distribution( ref, dist, env.reference );
	read_distribution( other, dist, env.to_check);
	uint mismatches = 0;

	if( ref.size() != other.size() )
	{
		ERROR("Reference and output have different sizes");
	}

	for( unsigned int i = 0; i < ref.size(); ++i )
	{
		if( ref[i] != other[i] )
		{
			if( env.print_all )
			{
				printf("Values do not match! Pos[%d] -> Ref: [%d] -- Other: [%d]\n", i, ref[i], other[i]);
			}
			mismatches++;
		}
	}
	std::cout << "Mismatches  " << mismatches << std::endl;
	ERROR_ON(mismatches > (ref.size() * env.acceptable_ratio_of_mismatches));
}

template<class T>
static float read_scalar( Scalar<T> &in, TargetID target)
{
	switch( target )
	{
		case TargetID::CL:
			return (float) in.cl.read();
			break;
		case TargetID::CV:
			return (float) in.cv.read();
			break;
		case TargetID::VX:
			return (float) in.vx.read();
		case TargetID::NEON:
			return (float) in.ne.read();
			break;
		default:
			ERROR("Target not supported");
	}
	return 0.0f;
}

template <class T>
void Validate::single_scalar_output( Scalar<T> &scalar, const Environment &env)
{
	float ref, other;

	ref = read_scalar( scalar, env.reference );
	other = read_scalar( scalar, env.to_check );

	if(std::fabs(ref - other) <= (float)(std::numeric_limits<float>::epsilon() + 1e-4 + env.min_threshold) )
	{
		printf("Scalar match!\n");
	}
	else
	{
		ERROR("ERROR: Scalars do not match! Ref: [%g] -- Other: [%g]\n", ref, other);
	}
}

template <class T>
void Validate::double_scalar_output( Scalar<T> &scalar1, Scalar<T> &scalar2, const Environment &env )
{
	single_scalar_output(scalar1, env);
	single_scalar_output(scalar2, env);
}

/* Allowed scalar validation */
template void Validate::double_scalar_output<uint8_t> ( Scalar<uint8_t>  &, Scalar<uint8_t>  &, const Environment &env );
template void Validate::double_scalar_output<uint16_t>( Scalar<uint16_t> &, Scalar<uint16_t> &, const Environment &env );
template void Validate::double_scalar_output<int16_t> ( Scalar<int16_t>  &, Scalar<int16_t>  &, const Environment &env );
template void Validate::double_scalar_output<uint32_t>( Scalar<uint32_t> &, Scalar<uint32_t> &, const Environment &env );
template void Validate::double_scalar_output<int32_t> ( Scalar<int32_t>  &, Scalar<int32_t>  &, const Environment &env );
template void Validate::double_scalar_output<uint64_t>( Scalar<uint64_t> &, Scalar<uint64_t> &, const Environment &env );
template void Validate::double_scalar_output<int64_t> ( Scalar<int64_t>  &, Scalar<int64_t>  &, const Environment &env );
template void Validate::double_scalar_output<float>   ( Scalar<float>    &, Scalar<float>    &, const Environment &env );
