#include "mali_cv/Error.h"
#include "mali_cv/Context.h"
#include "mali_cv/Image.h"
#include "mali_cv/Convolution.h"
#include "mali_cv/RemapTable.h"
#include "mali_cv/Pyramid.h"
#include "mali_cv/ITest.h"
#include "mali_cv/Types.h"
#include "mali_cv/Validate.h"
#include "mali_cv/Helper.h"

#include "tests/gemm.h"
#include "tests/sobel.h"
#include "tests/scharr.h"
#include "tests/pyramid.h"
#include "tests/convolution.h"
#include "tests/scale.h"
#include "tests/color_convert.h"
#include "tests/canny.h"
#include "tests/accumulate.h"
#include "tests/meanstddev.h"
#include "tests/arithmetic.h"
#include "tests/harris.h"
#include "tests/warp.h"
#include "tests/magnitude.h"
#include "tests/phase.h"
#include "tests/threshold.h"
#include "tests/derivative.h"
#include "tests/fast_corners.h"
#include "tests/remap.h"
#include "tests/pixelwise_mul.h"
#include "tests/channel_extract.h"
#include "tests/tablelookup.h"
#include "tests/depth.h"
#include "tests/minmaxlocation.h"
#include "tests/histogram.h"
#include "tests/channel_combine.h"
#include "tests/optical_flow.h"
#include "tests/integral_image.h"
#include "tests/non_linear_filter.h"
#include "tests/batchnormalization_layer.h"
#include "tests/softmax_layer.h"
#include "tests/tensor_ops.h"
#include "tests/hog.h"
#include "tests/transpose.h"
#include "tests/concatenate_depth.h"
#include "tests/locally_connected_layer.h"

#include "TestResults.h"

#include <iostream>

#include "arm_compute/runtime/Scheduler.h"
#include "arm_compute/runtime/Utils.h"

#ifdef USE_OPENCL
#include "arm_compute/core/CL/CLKernelLibrary.h"
#include "arm_compute/runtime/CL/CLScheduler.h"
#endif /* USE_OPENCL */

using namespace mali_cv;

/* Tests with 2 inputs images and 1 output image : */
TEST_2_IN_1_OUT( AbsoluteDifference, "AbsoluteDifference", absolute_difference)
TEST_2_IN_1_OUT( BitwiseAnd, "BitwiseAnd", bitwise_and)
TEST_2_IN_1_OUT( BitwiseOr, "BitwiseOr", bitwise_or)
TEST_2_IN_1_OUT( BitwiseXor, "BitwiseXor", bitwise_xor)

/* Tests with 1 input image and 1 output image : */
TEST_1_IN_1_OUT( BitwiseNot, "BitwiseNot", bitwise_not,0)
TEST_1_IN_1_OUT( Erode, "Erode", erode,1)
TEST_1_IN_1_OUT( Dilate, "Dilate", dilate,1)
TEST_1_IN_1_OUT( Gaussian3x3, "GaussianFilter3x3", gaussian3x3,1)
TEST_1_IN_1_OUT( Gaussian5x5, "GaussianFilter5x5", gaussian5x5,2)
TEST_1_IN_1_OUT( Box3x3, "BoxFilter3x3", box3x3,1)
TEST_1_IN_1_OUT( Box3x3_FP16, "BoxFilter3x3_FP16", box3x3_fp16,1)

TEST_1_IN_1_OUT( Median3x3, "MedianFilter3x3", median3x3,1)

namespace
{
RawImage &get_image( RawImage &input, Format dst_format)
{
	static std::map<Format, RawImage> images;
	auto it = images.find( dst_format );
	if( it == images.end())
	{
		images[dst_format].copy_from(input, dst_format);
	}
	return images[dst_format];
}

#if defined(DEBUG_ENABLED) && defined(USE_OPENCL)    // Remove DEBUG_ENABLED define to use printf with OpenCL
void printf_callback(const char *buffer, unsigned int len, size_t complete, void *user_data)
{
    printf("%.*s", len, buffer);
}
#endif /* defined(DEBUG_ENABLED) && defined(USE_OPENCL) */

}

int32_t main(int32_t argc,  const char **argv)
{
	const char *default_image = "data/images/800x600.ppm";
	const char *default_old_image = "data/images/opticalflow_old.pgm";
	const char *default_new_image = "data/images/opticalflow_new.pgm";
	RawImage ppmImage;         // Input PPM image
	RawImage red, green, blue;
	RawImage u_w2, v_w2, u_wh2, v_wh2;
	RawImage newImage, oldImage;

	RawImage gx, gy; //to test phase and magnitude
	RawImage mapx, mapy; //used for remap

	int test_id;
	int first_test_id;
	int last_test_id;
	int check_output;
	const char* test_filter = getenv("TEST_FILTER");

	int border_mode;
	int interpolation_mode; // Scale only
	/* Canny edge parameters: */
	int gradient;
	int threshold_upper;
	int threshold_lower;
	int norm;
	size_t corner_threshold;
	size_t pyramid_level;
	const char* tensor_dims_str = nullptr;
	const char* kernel_dims_str = nullptr;
	int stride_x;
	int stride_y;
	int num_iterations=1;
	int use_image;

	READ_CONFIG( use_image,"USE_IMAGE", 1);
	READ_CONFIG( test_id, "TEST", -1);
	READ_CONFIG(first_test_id, "FIRST_TEST", -1);
	READ_CONFIG(last_test_id, "LAST_TEST", INT_MAX);
	READ_CONFIG( check_output, "CHECK_OUTPUT", 0);
	if(check_output == 0)
		READ_CONFIG( num_iterations, "NUM_ITERATIONS", 10);

	READ_CONFIG( border_mode, "BORDER_MODE", 1);
	READ_CONFIG( interpolation_mode, "INTERPOLATION_MODE", 0);
	READ_CONFIG( gradient, "GRADIENT", 3);
	READ_CONFIG( threshold_upper, "THRESHOLD_UPPER", 70);
	READ_CONFIG( threshold_lower, "THRESHOLD_LOWER", 50);
	READ_CONFIG( norm, "NORMALIZATION", 1);
	READ_CONFIG( pyramid_level, "PYRAMID_LEVEL", 3 );
	if( (179 == test_id) || (test_id < 0))
		READ_CONFIG( corner_threshold, "CORNER_THRESHOLD", 3);
	else
		READ_CONFIG( corner_threshold, "CORNER_THRESHOLD", 0);

	// Read tensor dims;
	READ_CONFIG(stride_x, "STRIDE_X", 1);
	READ_CONFIG(stride_y, "STRIDE_Y", 1);
	READ_CONFIG_STR(tensor_dims_str, "TENSOR_DIMS", "128,64,16,2");
	READ_CONFIG_STR(kernel_dims_str, "KERNEL_DIMS", "5,5,16,64");
	std::vector<int> tensor_dims = split_string_to_int(std::string(tensor_dims_str), ',');
	std::vector<int> kernel_dims = split_string_to_int(std::string(kernel_dims_str), ',');
	ERROR_ON_MSG( 0 == tensor_dims.size() || tensor_dims.size() > 6, "Invalid number of tensor dimensions.");
	ERROR_ON_MSG( 4 != kernel_dims.size(), "Invalid number of kernel dimensions. Should be four.");
	TensorShape tensor_shape = vector_to_tensorshape(tensor_dims);
	TensorShape kernel_shape = vector_to_tensorshape(kernel_dims);
	UNUSED(tensor_shape);
	UNUSED(kernel_shape);
	UNUSED(stride_x);
	UNUSED(stride_y);
	int sch;
	READ_CONFIG( sch, "SCHEDULER", -1);
	if( sch > -1)
	{
		arm_compute::Scheduler::set((arm_compute::Scheduler::Type)sch);
	}
	std::cout << "SCHEDULER: " << arm_compute::string_from_scheduler_type(arm_compute::Scheduler::get_type()) << std::endl;
#ifdef USE_NEON
	// Read number of threads
	unsigned int num_threads;
	READ_CONFIG( num_threads, "NUM_THREADS", 0);
	if( num_threads )
	{
		arm_compute::Scheduler::get().set_num_threads(num_threads);
	}
#endif /* USE_NEON */

#ifdef USE_EIGEN
	Eigen::initParallel();
	Eigen::setNbThreads(arm_compute::Scheduler::get().num_threads());
	std::cout<< "Number of threads used by EIGEN: "<< Eigen::nbThreads() << std::endl;
#endif /* USE_EIGEN */

#ifdef USE_OPENCL
#ifdef DEBUG_ENABLED    // Remove this define to use printf with OpenCL
	// Create a cl_context with a printf_callback and user specified buffer size.
	cl_context_properties properties[] =
	{
		// Enable a printf callback function for this context.
		CL_PRINTF_CALLBACK_ARM, reinterpret_cast<cl_context_properties>(printf_callback),
		// Request a minimum printf buffer size of 4MiB for devices in the
		// context that support this extension.
		CL_PRINTF_BUFFERSIZE_ARM, static_cast<cl_context_properties>(0x100000),
		CL_CONTEXT_PLATFORM, reinterpret_cast<cl_context_properties>(cl::Platform::get()()),
		0
	};

	cl::Context::setDefault(cl::Context(CL_DEVICE_TYPE_DEFAULT, properties));
#endif
	arm_compute::CLKernelLibrary::get().init("./cl_kernels/", cl::Context::getDefault(), cl::Device::getDefault());
	arm_compute::CLScheduler::get().init(cl::Context::getDefault(), cl::CommandQueue::getDefault(), cl::Device::getDefault());

#endif

	Environment env(check_output, (BorderMode) border_mode, CONSTANT_BORDER_VALUE, num_iterations);
	int to_check, reference;
	READ_CONFIG( reference, "REFERENCE", (int)TargetID::VX);
	env.to_check= TargetID::NUM_TARGETS;
	printf("List of targets available:\n");
	for(uint i=0; i < (uint) TargetID::NUM_TARGETS; i++)
	{
		const ITarget *target = env.target((TargetID)i);

		if(target != nullptr)
		{
			printf("\t[%d] %s\n", i, target->name().c_str());
			if(env.to_check == TargetID::NUM_TARGETS && i != (uint)reference )
			{
				env.to_check = (TargetID)i;
			}
		}
	}
	printf("\n");

	READ_CONFIG( env.print_all, "PRINT_ALL", 0);
	READ_CONFIG( env.env_threshold, "THRESHOLD", 0);
	READ_CONFIG( env.stop_on_error, "STOP_ON_ERROR", 0);
	READ_CONFIG( to_check, "TO_CHECK", (int)env.to_check);
	env.to_check = (TargetID) to_check;
	env.reference = (TargetID) reference;

	ERROR_ON_MSG( check_output && env.target(env.to_check) == nullptr, "Invalid TO_CHECK target");
	ERROR_ON_MSG( check_output && env.target(env.reference) == nullptr, "Invalid REFERENCE target");

	/*******************************************************************************
	 Help command
	*******************************************************************************/
	IKeyPointArray old_points(env.ctx, 1000); // To test optical flow
	IKeyPointArray new_points_estimates(env.ctx, 1000);
	std::vector<IKeyPoint> old_array;
	std::vector<IKeyPoint> new_array_estimates;
	std::vector<RawImage *> channels;

	if( use_image )
	{
		if(argc < 2)
		{
			// Print help
			printf("%s [image_name]\n\n", argv[0]);

			printf("Image processed: %s\n", default_image);
			ppmImage.create_from(default_image);
		}
		else if ( 2 == argc )
		{
			printf("Image processed: %s\n", argv[1]);
			ppmImage.create_from(argv[1]);
		}

		if( (179 == test_id) || (test_id < 0))
		{
			if(argc == 1)
			{
				// Print help
				printf("%s [image_name]\n\n", argv[0]);
				printf("Image processed: old image %s, new image %s\n", default_old_image, default_new_image );
				oldImage.create_from_pgm(default_old_image);
				newImage.create_from_pgm(default_new_image);
			}
			else if ( argc == 3)
			{
				printf("Image processed: old image %s, new image %s\n", argv[1], argv[2]);
				oldImage.create_from_pgm(argv[1]);
				newImage.create_from_pgm(argv[2]);
				ERROR_ON_MSG(oldImage.width() != newImage.width(), "old image and new image have different width");
				ERROR_ON_MSG(oldImage.height() != oldImage.height(), "old image and new image have different height");
			}
			for( int i = 0; i < 1000; i++)
			{
				IKeyPoint keypoint;
				/* VX doesn't handle the left border correctly. Therefore here it adjust the test data not touch
				 * * the border in each level of the pyramid. */
				size_t max = oldImage.width() - 1;
				size_t min = pow(2, pyramid_level);
				ERROR_ON(min > max);
				keypoint.x = rand() % (max - min + 1) + min;
				max = oldImage.height() - 1 ;
				keypoint.y = rand() % (max - min + 1) + min;
				keypoint.tracking_status = 1;
				old_array.push_back(keypoint);
				new_array_estimates.push_back(keypoint);
			}

			old_points.ne.fill(old_array);
			new_points_estimates.ne.fill(new_array_estimates);
			old_points.cv.fill(old_array);
			new_points_estimates.cv.fill(new_array_estimates);
			old_points.vx.fill(old_array);
			new_points_estimates.vx.fill(new_array_estimates);
			old_points.cl.fill(old_array);
			new_points_estimates.cl.fill(new_array_estimates);
		}
		if ( argc < 3 )
		{
			channels.push_back( &red );
			channels.push_back( &green );
			channels.push_back( &blue );
			ppmImage.deinterleave( channels );

			gx.copy_from(red, Format::LUMINANCE16);
			gy.copy_from(green, Format::LUMINANCE16);

			// Create yuv images
			u_w2.create(red.width()/2, red.height(), Format::LUMINANCE);
			v_w2.create(red.width()/2, red.height(), Format::LUMINANCE);
			u_wh2.create(red.width()/2, red.height()/2, Format::LUMINANCE);
			v_wh2.create(red.width()/2, red.height()/2, Format::LUMINANCE);
			u_w2.random_init(  0.0f, 255.0f );
			v_w2.random_init(  0.0f, 255.0f );
			u_wh2.random_init( 0.0f, 255.0f );
			v_wh2.random_init( 0.0f, 255.0f );
		}
	}

	// TODO: move the matrices into their respective CustomConvolution tests as they are only needed there and should not be any need of them here
	// e.g : look at the tests: CustomConvolutionShort3x3Sobelx && CustomConvolution3x3Laplacian
	int16_t highpass5x5[5][5] =
	{
		{ 0, -1, -1, -1,  0},
		{ -1,  2, -4,  2, -1},
		{ -1, -4, 13, -4, -1},
		{ -1,  2, -4,  2, -1},
		{ 0, -1, -1, -1,  0}
	};

	int16_t gaussian3x3[3][3] =
	{
		{1, 2, 1},
		{2, 4, 2},
		{1, 2, 1}
	};

	int16_t gaussian5x5[5][5] =
	{
		{ 1, 4, 6, 4, 1 },
		{ 4, 16, 24, 16, 4 },
		{ 6, 24, 36, 24, 6 },
		{ 4, 16, 24, 16, 4 },
		{ 1, 4, 6, 4, 1 }
	};

	int16_t _gaus1x3[3] =
	{
		1, 2, 1
	};

	int16_t _gaus3x1[3] =
	{
		1, 2, 1
	};

	int16_t sobel_x_5x5[5][5]
	{
		{1, 2, 0, -2, -1 },
		{4, 8, 0, -8, -4 },
		{6, 12, 0, -12, -6 },
		{4, 8, 0, -8, -4 },
		{1, 2, 0, -2, -1 }
	};

	const int16_t sobel_x_7x7[7][7]=
	{
		{ 1,  4,  5, 0,  -5,  -4, -1},
		{ 6, 24, 30, 0, -30, -24, -6},
		{15, 60, 75, 0, -75, -60,-15},
		{20, 80,100, 0,-100, -80,-20},
		{15, 60, 75, 0, -75, -60,-15},
		{ 6, 24, 30, 0, -30, -24, -6},
		{ 1,  4,  5, 0,  -5,  -4, -1},
	};

	const int16_t sep_9x9[9][9]=
	{
		{ -1,  -2,  -4,  -6,  -8,  -6,  -4,  -2, -1 },
		{ -2,  -4,  -8, -12, -16, -12,  -8,  -4, -2 },
		{ -4,  -8, -16, -24, -32, -24, -16,  -8, -4 },
		{ -6, -12, -24, -36, -48, -36, -24, -12, -6 },
		{ -8, -16, -32, -48, -64, -48, -32, -16, -8 },
		{  6,  12,  24,  36,  48,  36,  24,  12,  6 },
		{  4,   8,  16,  24,  32,  24,  16,   8,  4 },
		{  2,   4,   8,  12,  16,  12,   8,   4,  2 },
		{  1,   2,   4,   6,   8,   6,   4,   2,  1 },
	};

	std::vector<int16_t> gaus1x3(_gaus1x3, _gaus1x3 + 3);
	std::vector<int16_t> gaus3x1(_gaus3x1, _gaus3x1 + 3);

	const float warp_matrix3x2[3][2] = { {1.f,0.f}, {0.f, 1.f}, {0.f,0.f} };
	const float warp_matrix3x3[3][3] = { {1.f,0.f,0.0f}, {0.f, 1.0f,0.0f}, {0.f,0.f,1.f} };

	// Was the ppmImage initialized ?
	bool use_single_image = ppmImage.format() != Format::UNKNOWN;
	TestManager tests(env);
	if(use_single_image)
	{
		mapx.create(red.width(), red.height(), Format::FP32C1);
		mapy.create(red.width(), red.height(), Format::FP32C1);

		const int32_t width  = mapx.width();
		const int32_t height = mapx.height();

		const auto mapx_ptr = reinterpret_cast<float*>(mapx.data());
		const auto mapy_ptr = reinterpret_cast<float*>(mapy.data());

		/* Initialize mapx and mapy using the warp affine matrix transformation  */
		for(int32_t y = 0; y < height; ++y)
		{
			for(int32_t x = 0; x < width; ++x)
			{
				float x0 = x * warp_matrix3x2[0][0] + y * warp_matrix3x2[1][0] + warp_matrix3x2[2][0];
				float y0 = x * warp_matrix3x2[0][1] + y * warp_matrix3x2[1][1] + warp_matrix3x2[2][1];

				x0 = (x0 < 0.0f?    0.0f       : x0);
				y0 = (y0 < 0.0f?    0.0f       : y0);
				x0 = (x0 >= width?  width - 1  : x0);
				y0 = (y0 >= height? height - 1 : y0);

				*(mapx_ptr + x + y * width) = x0;
				*(mapy_ptr + x + y * width) = y0;
			}
		}

		tests.add_core_test(  0,  [&] { return new Sobel( env, red); } );
		tests.add_core_test(  1,  [&] { return new SobelX( env, red); } );
		tests.add_core_test(  2,  [&] { return new SobelY( env, red); } );
		tests.add_core_test(  3,  [&] { return new CustomConvolutionShort3x3Sobelx( env, red); } );
		tests.add_core_test(  4,  [&] { return new CustomConvolution3x3Laplacian( env, red); } );

		// Arithmetic Addition tests
		tests.add_core_test(  5,  [&] { return new ArithmeticAddition( env, red, blue , 1); } );
		tests.add_core_test(  6,  [&] { return new ArithmeticAdditionU16U8U16( env, get_image(ppmImage,Format::LUMINANCE16), blue , 1); } );
		tests.add_core_test(  7,  [&] { return new ArithmeticAdditionU8U16U16( env, red, get_image(ppmImage,Format::LUMINANCE16) , 0); } );
		tests.add_core_test(  8,  [&] { return new ArithmeticAdditionU8U8U16( env, red, blue , 1); } );

		// Accumulate tests
		tests.add_core_test(  9,  [&] { return new Accumulate( env, red, get_image(ppmImage,Format::LUMINANCE16)); });
		tests.add_core_test( 10,  [&] { return new Gaussian3x3( env, red); } );
		tests.add_core_test( 11,  [&] { return new Box3x3( env, red); } );
		tests.add_core_test( 12,  [&] { return new CustomConvolution3x3Laplacian( env, red); } );
		tests.add_core_test( 14,  [&] { return new CustomConvolutionShort3x3Laplacian( env, blue); } );
		tests.add_core_test( 16,  [&] { return new CustomConvolution3x3Sobelx( env, red); } );
		tests.add_core_test( 18,  [&] { return new CustomConvolution5x5( env, red, highpass5x5); } );
		tests.add_core_test( 20,  [&] { return new CustomConvolution5x5Short( env, red, highpass5x5); } );
		//Separable 3x3 matrices are not useful.
		tests.add_core_test( 24,  [&] { return new CustomConvolution( env, red, gaussian3x3); } );
		tests.add_core_test( 25,  [&] { return new CustomConvolutionSeparable5x5( env, red, sobel_x_5x5); } );
		tests.add_core_test( 28,  [&] { return new CustomConvolution5x5( env, red, sobel_x_5x5); } );
		tests.add_core_test( 29,  [&] { return new CustomConvolution( env, red, highpass5x5); } );

		tests.add_core_test( 30,  [&] { return new HarrisCorners(env,red, 0.00115f , 2.f, 0.04f, 3, 3, corner_threshold); },
				TestFailure(TargetID::CV, FailureReason::CANNOT_BE_VALIDATED, "OpenCV detects the same but a different number of corners as the threshold is handled differently"));

		if( ppmImage.width() % 2 == 0) // The width of the image must be a multiple of 2 for these image formats to be used
		{
			tests.add_core_test( 31,  [&] { return new ColorConvert( env, ppmImage, Format::RGBA8888); } );
			tests.add_core_test( 32,  [&] { return new ColorConvert( env, get_image(ppmImage, Format::RGBA8888), Format::RGB888); } );
			tests.add_core_test( 33,  [&] { return new ColorConvert( env, get_image(ppmImage,Format::YUYV422), Format::RGB888); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"));
			tests.add_core_test( 34,  [&] { return new ColorConvert( env, get_image(ppmImage,Format::YUYV422), Format::RGBA8888); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"));
			tests.add_core_test( 35,  [&] { return new ColorConvert( env, get_image(ppmImage,Format::UYVY), Format::RGB888); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"));
			tests.add_core_test( 36,  [&] { return new ColorConvert( env, get_image(ppmImage,Format::UYVY), Format::RGBA8888); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"));
		}

		tests.add_core_test( 37,  [&] { return new Gaussian5x5( env, red); } );
		tests.add_core_test( 38,  [&] { return new WarpAffine(env,red,warp_matrix3x2, (InterpolationMode) interpolation_mode); } );

		tests.add_core_test( 39,  [&] { return new WarpPerspective(env,red,warp_matrix3x3, (InterpolationMode) interpolation_mode); },
				TestFailure(TargetID::CL, "To investigate: differences of up to 4 on mate9? (Threshold of 1 tolerated + 0.02% of mismatches)",0.0002f, 2.0f));

		tests.add_core_test( 40,  [&] { return new CannyEdge( env, red, threshold_upper, threshold_lower, gradient, norm); },
				TestFailure(TargetID::CV, FailureReason::CANNOT_BE_VALIDATED, "OpenCV only supports float phase"),
				TestFailure(TargetID::CL, "COMPMID-328: Mismatches due to the static stack depths", 0.05f),
				TestFailure(TargetID::NEON,"COMPMID-328: Mismatches because OpenVX internally uses float to compute magnitude and phase", 0.02f));

		tests.add_core_test( 41,  [&] { return new Sobel5x5( env, red); } );
		tests.add_core_test( 42,  [&] { return new Sobel5x5X( env, red); } );
		tests.add_core_test( 43,  [&] { return new Sobel5x5Y( env, red); } );
		tests.add_core_test( 44,  [&] { return new Scharr( env, red); } );
		tests.add_core_test( 45,  [&] { return new ScharrX( env, red); } );
		tests.add_core_test( 46,  [&] { return new ScharrY( env, red); } );
		tests.add_core_test( 47,  [&] { return new Median3x3( env, red); } );

		tests.add_core_test( 48,
				[&] { return new Sobel7x7( env, red); },
				TestFailure( TargetID::NEON, FailureReason::NOT_IMPLEMENTED, "COMPMID-272 arm_compute only supports S32 output and not S16"),
				TestFailure( TargetID::CL, FailureReason::NOT_IMPLEMENTED, "COMPMID-272 arm_compute only supports S32 output and not S16")
		);


		tests.add_core_test( 49,  [&] { return new Sobel7x7X( env, red); },
				TestFailure( TargetID::NEON, FailureReason::NOT_IMPLEMENTED, "COMPMID-272 arm_compute only supports S32 output and not S16"),
				TestFailure( TargetID::CL, FailureReason::NOT_IMPLEMENTED, "COMPMID-272 arm_compute only supports S32 output and not S16")
		);

		tests.add_core_test( 50,  [&] { return new Sobel7x7Y( env, red); },
				TestFailure( TargetID::NEON, FailureReason::NOT_IMPLEMENTED, "COMPMID-272 arm_compute only supports S32 output and not S16"),
				TestFailure( TargetID::CL, FailureReason::NOT_IMPLEMENTED, "COMPMID-272 arm_compute only supports S32 output and not S16")
		);

		// Bitwise Operation Tests
		tests.add_core_test( 51,  [&] { return new BitwiseOr( env, red, green); } );
		tests.add_core_test( 52,  [&] { return new BitwiseAnd( env, red, green); } );
		tests.add_core_test( 53,  [&] { return new BitwiseXor( env, red, green); } );
		tests.add_core_test( 54,  [&] { return new BitwiseNot( env, red); } );

		// Absolute Difference Tests
		tests.add_core_test( 55,  [&] { return new AbsoluteDifference( env, red, green); } );
		tests.add_core_test( 56,  [&] { return new AbsoluteDifference( env, get_image(red,Format::LUMINANCE16), get_image(green,Format::LUMINANCE16)); } );

		// Erode tests
		tests.add_core_test( 57,  [&] { return new Erode(env, red); } );

		// Dilate tests
		tests.add_core_test( 58,  [&] { return new Dilate(env, red); } );

		// Accumulate weighted tests
		tests.add_core_test( 59,  [&] { return new AccumulateWeighted(env, red, blue, 0.1F); },
					TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "OpenCV only supports floating point input/output images for accumulate squared"),
					TestFailure(TargetID::NEON, "Rounding error? (Threshold of 1 tolerated)",0.0f, 1.0f),
					TestFailure(TargetID::CL, "Rounding error? (Threshold of 1 tolerated)",0.0f, 1.0f));

		// Accumulate squared tests
		tests.add_core_test( 60,  [&] { return new AccumulateSquared(env, red, get_image(ppmImage,Format::LUMINANCE16), 3); },
						TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "OpenCV only supports floating point input/output images for accumulate weighted"));
		// Magnitude tests
		tests.add_core_test( 61,  [&] { return new Magnitude( env, gx, gy); } );

		// Arithmetic Subtraction tests
		tests.add_core_test( 62,  [&] { return new ArithmeticSubtraction( env, red, blue , 1); } );
		tests.add_core_test( 63,  [&] { return new ArithmeticSubtractionU16U8U16( env, get_image(ppmImage,Format::LUMINANCE16), blue , 1); } );
		tests.add_core_test( 64,  [&] { return new ArithmeticSubtractionU8U16U16( env, red, get_image(ppmImage,Format::LUMINANCE16) , 0); } );
		tests.add_core_test( 65,  [&] { return new ArithmeticSubtractionU8U8U16( env, red, blue , 1); } );

		// Phase tests
		tests.add_core_test( 66,  [&] { return new Phase( env, gx, gy); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "OpenCV doesn't support float phase."),
			TestFailure(TargetID::NEON, "Mismatches < 1 between arm_compute and openvx due to different approximation of atan2.", 0.0f, 1.0f));

		// Thresholding tests
		tests.add_core_test( 67,  [&] { return new ThresholdBinary(env, red, 200, 3, 210); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "No proper equivalent in OpenCV"));
		tests.add_core_test( 68,  [&] { return new ThresholdRange(env, red, 150,220, 3, 210); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "No proper equivalent in OpenCV"));

#ifndef TEST_TENSORS
		if( ppmImage.width() % 2 == 0) // The width of the image must be a multiple of 2 for these image formats to be used
		{
			tests.add_core_test( 69,  [&] { return new ColorConvert( env, get_image(ppmImage, Format::NV12), Format::RGB888); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"));
		}
#endif
		tests.add_core_test( 70,  [&] { return new GaussianPyramid( env, red, 3, 0.5f); },
				TestFailure(TargetID::CV, env.ctx.border_mode() == BorderMode::UNDEFINED ? FailureReason::CANNOT_BE_VALIDATED : FailureReason::NOT_SUPPORTED, "OpenCV version has mismatch and crashed when border mode is defined")
					);

		tests.add_core_test( 71, [&] { return new FastCorners( env, green, 61.1f, false, corner_threshold); },
					TestFailure( TargetID::NEON, FailureReason::NOT_IMPLEMENTED, BorderMode::REPLICATE, "COMPMID-273 Fast9: border mode not supported"),
					TestFailure( TargetID::NEON, FailureReason::NOT_IMPLEMENTED, BorderMode::CONSTANT,"COMPMID-273 Fast9: border mode not supported"),
					TestFailure( TargetID::CL, FailureReason::NOT_IMPLEMENTED, BorderMode::REPLICATE, "COMPMID-273 Fast9: border mode not supported in VX sample implementation"),
					TestFailure( TargetID::CL, FailureReason::NOT_IMPLEMENTED, BorderMode::CONSTANT,"COMPMID-273 Fast9: border mode not supported in VX sample implementation"),
					TestFailure( TargetID::VX, FailureReason::NOT_SUPPORTED, BorderMode::REPLICATE, "Fast9: border mode not supported" ),
					TestFailure( TargetID::VX, FailureReason::NOT_SUPPORTED, BorderMode::CONSTANT, "Fast9: border mode not supported" )
		);

		tests.add_core_test( 72,  [&] { return new FastCornersNonMaxSuppressed( env, green, 61.1f, corner_threshold); },
						TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "Fast9 MaxSup: Not implemented"),
						TestFailure(TargetID::VX, FailureReason::NOT_SUPPORTED, BorderMode::REPLICATE, "Fast9: border mode not supported"),
						TestFailure(TargetID::VX, FailureReason::NOT_SUPPORTED, BorderMode::CONSTANT, "Fast9: border mode not supported" ),
						TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED, BorderMode::REPLICATE, "COMPMID-273 Fast9: border mode not supported"),
						TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED, BorderMode::CONSTANT, "COMPMID-273 Fast9: border mode not supported" ),
						TestFailure(TargetID::NEON, FailureReason::NOT_IMPLEMENTED, BorderMode::REPLICATE, "COMPMID-273 Fast9: border mode not supported"),
						TestFailure(TargetID::NEON, FailureReason::NOT_IMPLEMENTED, BorderMode::CONSTANT, "COMPMID-273 Fast9: border mode not supported"));

		tests.add_core_test( 73,  [&] { return new Derivative( env, red); } );
		tests.add_core_test( 74,  [&] { return new DerivativeX( env, red); } );
		tests.add_core_test( 75,  [&] { return new DerivativeY( env, red); } );
		tests.add_core_test( 76,  [&] { return new Remap( env, red, mapx, mapy, (InterpolationMode) interpolation_mode); } );
	}

	if(use_single_image)
	{
		// Pixelwise Multiplication tests
		tests.add_core_test( 83,  [&] { return new PixelWiseMultiplication( env, red, blue , 1.f, 1, 0); } );
		tests.add_core_test( 84,  [&] { return new PixelWiseMultiplication( env, red, blue , 1.f/255.f, 1, 0); },
			TestFailure(TargetID::CV, FailureReason::CANNOT_BE_VALIDATED, "CV supports round to nearest mode while VX round to zero mode"),
			TestFailure(TargetID::NEON, FailureReason::NOT_SUPPORTED, "COMPMID-274: OpenVX only supports round towards zero and we can only support round to nearest even."));

		tests.add_core_test( 85,  [&] { return new PixelWiseMultiplicationS16S16S16( env, get_image(ppmImage, Format::LUMINANCE16), get_image(ppmImage, Format::LUMINANCE16) , 1.f/32.f, 0, 0); },
						TestFailure(TargetID::CV, FailureReason::CANNOT_BE_VALIDATED, "Test supports only saturated mode"));

		tests.add_core_test( 86,  [&] { return new PixelWiseMultiplicationU8U8S16( env, red, blue , 1.f/32.f, 1, 0); },
						TestFailure(TargetID::CV, FailureReason::CANNOT_BE_VALIDATED, "CV supports round to nearest mode while VX round to zero mode"));

		tests.add_core_test( 87,  [&] { return new PixelWiseMultiplicationU8S16S16( env, blue , get_image(ppmImage, Format::LUMINANCE16), 1.f/32.f, 0, 0); },
						TestFailure(TargetID::CV, FailureReason::CANNOT_BE_VALIDATED, "Test supports only saturated mode"));

		tests.add_core_test( 88,  [&] { return new Scale( env, red, red.width()*0.86, red.height()*0.74, (InterpolationMode)interpolation_mode); },
						TestFailure(TargetID::CL, "COMPMID-300 failing on mate9: to investigate", 2.0f));
		// Channel Extract tests
		tests.add_core_test( 89,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::RGB888), Channel::R ); } );
		tests.add_core_test( 90,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::RGB888), Channel::G ); } );
		tests.add_core_test( 91,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::RGB888), Channel::B ); } );
		tests.add_core_test( 92,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::RGBA8888), Channel::R ); } );
		tests.add_core_test( 93,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::RGBA8888), Channel::G ); } );
		tests.add_core_test( 94,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::RGBA8888), Channel::B ); } );
		tests.add_core_test( 95,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::RGBA8888), Channel::A ); } );

		tests.add_core_test( 96,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::YUYV422), Channel::Y ); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ));
		tests.add_core_test( 97,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::YUYV422), Channel::U ); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ));
		tests.add_core_test( 98,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::YUYV422), Channel::V ); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ));
		tests.add_core_test( 99,   [&] { return new ChannelExtract( env, get_image(ppmImage,Format::UYVY), Channel::Y ); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ));
		tests.add_core_test( 100,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::UYVY), Channel::U ); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ));
#ifndef TEST_TENSORS
		tests.add_core_test( 101,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::UYVY), Channel::V ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 102,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::NV12), Channel::Y ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 103,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::NV12), Channel::U ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 104,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::NV12), Channel::V ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 105,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::NV21), Channel::Y ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 106,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::NV21), Channel::U ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 107,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::NV21), Channel::V ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 108,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::YUV444), Channel::Y ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 109,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::YUV444), Channel::U ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 110,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::YUV444), Channel::V ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 111,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::IYUV), Channel::Y ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 112,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::IYUV), Channel::U ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 113,  [&] { return new ChannelExtract( env, get_image(ppmImage,Format::IYUV), Channel::V ); },TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
#endif
		// 114 -> 121 used to be the ChannelExtract tests where planes were moved, but this feature has been removed

		tests.add_core_test( 122,  [&] { return new DepthConvert_S16_U8_SATURATE( env, get_image(ppmImage, Format::LUMINANCE16), 7 ); },
						TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "DepthConvert_S16_U8_SATURATE not supported"));

		tests.add_core_test( 123,  [&] { return new DepthConvert_S16_U8_SATURATE( env, get_image(ppmImage, Format::LUMINANCE16), 5 ); },
						TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "DepthConvert_S16_U8_SATURATE not supported"));

		tests.add_core_test( 124,  [&] { return new DepthConvert_S16_U8_SATURATE( env, get_image(ppmImage, Format::LUMINANCE16), 1 ); },
						TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "DepthConvert_S16_U8_SATURATE not supported"));

		tests.add_core_test( 125,  [&] { return new DepthConvert_S16_U8_WRAP( env, get_image(ppmImage, Format::LUMINANCE16) , 3 ); },
						TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "DepthConvert_S16_U8_SATURATE not supported"));

		tests.add_core_test( 126,  [&] { return new DepthConvert_U8_S16( env, red , 2 ); });
		tests.add_core_test( 127,  [&] { return new DepthConvert_U8_S16( env, red , 0 ); });

		// Must compile VX with EXPERIMENTAL_USE_S16=1 to use other formats in the depth conversion tests
		tests.add_core_test( 128,  [&] { return new DepthConvert_S16_S32( env, get_image(ppmImage, Format::LUMINANCE16) , 1 ); },
						TestFailure(TargetID::CV, FailureReason::CANNOT_BE_VALIDATED, "DepthConvert_S16_S32 not supported"));

		tests.add_core_test( 129,  [&] { return new DepthConvert_U16_U32( env, get_image(ppmImage, Format::U16) , 2 ); },
						TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "DepthConvert_U16_U32: not supported"));

		tests.add_core_test( 130,  [&] { return new DepthConvert_U16_U8_SATURATE( env, get_image(ppmImage, Format::U16) , 4 ); },
						TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "DepthConvert_U16_U8_SATURATE: not supported"));

		tests.add_core_test( 131,  [&] { return new DepthConvert_U16_U8_WRAP( env, get_image(ppmImage, Format::U16) , 5 ); },
						TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "DepthConvert_U16_U8_WRAP: not supported"));

		tests.add_core_test( 132,  [&] { return new DepthConvert_U8_U16( env, red , 2 ); },
						TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "DepthConvert_U8_U16: not supported"));

		// Mean and Standard deviation test.
		tests.add_core_test( 133,  [&] { return new MeanStdDev( env, red ); } );

		// TableLookup tests
		tests.add_core_test( 134,  [&] { return new TableLookup( env, red ); } );
		tests.add_core_test( 135,  [&] { return new TableLookupS16( env, get_image(ppmImage, Format::LUMINANCE16) ); },
							TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ));
		// Minmax Location tests
		tests.add_core_test( 136,  [&] { return new MinMaxLocation( env, red ); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ),
						TestFailure(TargetID::CL, FailureReason::CANNOT_BE_VALIDATED, "Coords list might not match when number of coords are greater than the max number of elements the array can hold."));

		tests.add_core_test( 137,  [&] { return new MinMaxLocationS16( env, get_image(ppmImage, Format::LUMINANCE16) ); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ),
						TestFailure(TargetID::CL, FailureReason::CANNOT_BE_VALIDATED, "Coords list might not match when number of coords are greater than the max number of elements the array can hold."));

		tests.add_core_test( 138,  [&] { return new MinMaxLocationNoLoc( env, red ); } );
		tests.add_core_test( 139,  [&] { return new MinMaxLocationNoLocS16( env, get_image(ppmImage, Format::LUMINANCE16) ); } );

#ifndef TEST_TENSORS
		tests.add_core_test( 140, [&] { return new ColorConvert( env, ppmImage, Format::NV12); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "RGB to NV12 not supported"),
			TestFailure(TargetID::NEON, "Rounding error? (Threshold of 1 tolerated)",0.0f, 1.0f),
			TestFailure(TargetID::CL, "Precision cause tiny mismatch, VX uses double precision while CL uses float(Threshold of 1 tolerated)",0.0f, 1.0f));

		tests.add_core_test( 141, [&] { return new ColorConvert( env, ppmImage, Format::IYUV); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"),
			TestFailure(TargetID::NEON, "Rounding error? (Threshold of 1 tolerated)",0.0f, 1.0f),
			TestFailure(TargetID::CL, "Precision cause tiny mismatch, VX uses double precision while CL uses float(Threshold of 1 tolerated)",0.0f, 1.0f));

		tests.add_core_test( 142, [&] { return new ColorConvert( env, ppmImage, Format::YUV444); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "RGB to YUV4 not supported"),
			TestFailure(TargetID::NEON, "Rounding error?(Threshold of 2 tolerated)",0.0f,2.0f),
			TestFailure(TargetID::CL, "Precision cause tiny mismatch, VX uses double precision while CL uses float(Threshold of 1 tolerated)",0.0f, 1.0f));

		tests.add_core_test( 143, [&] { return new ColorConvert( env, get_image(ppmImage, Format::RGBA8888), Format::NV12); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "RGBX to NV12 not supported"),
			TestFailure(TargetID::NEON, "Rounding error? (Threshold of 1 tolerated)",0.0f, 1.0f),
			TestFailure(TargetID::CL, "Precision cause tiny mismatch, VX uses double precision while CL uses float(Threshold of 1 tolerated)",0.0f, 1.0f));

		tests.add_core_test( 144, [&] { return new ColorConvert( env, get_image(ppmImage, Format::RGBA8888), Format::IYUV); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"),
			TestFailure(TargetID::NEON, "Rounding error? (Threshold of 1 tolerated)",0.0f, 1.0f),
			TestFailure(TargetID::CL, "Precision cause tiny mismatch, VX uses double precision while CL uses float(Threshold of 1 tolerated)",0.0f, 1.0f));

		tests.add_core_test( 145, [&] { return new ColorConvert( env, get_image(ppmImage, Format::RGBA8888), Format::YUV444); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "RGBX to YUV4 not supported"),
			TestFailure(TargetID::NEON, "Rounding error? (Threshold of 2 tolerated)",0.0f, 2.0f),
			TestFailure(TargetID::CL, "Precision cause tiny mismatch, VX uses double precision while CL uses float(Threshold of 1 tolerated)",0.0f, 1.0f));

		tests.add_core_test( 146, [&] { return new ColorConvert( env, get_image(ppmImage, Format::NV12), Format::RGBA8888); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"));
		tests.add_core_test( 147, [&] { return new ColorConvert( env, get_image(ppmImage, Format::NV12), Format::IYUV); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "YUV to YUV not supported"));
		tests.add_core_test( 148, [&] { return new ColorConvert( env, get_image(ppmImage, Format::NV12), Format::YUV444); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "YUV to YUV not supported"));
		tests.add_core_test( 149, [&] { return new ColorConvert( env, get_image(ppmImage, Format::NV21), Format::RGB888); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"));
		tests.add_core_test( 150, [&] { return new ColorConvert( env, get_image(ppmImage, Format::NV21), Format::RGBA8888); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"));
		tests.add_core_test( 151, [&] { return new ColorConvert( env, get_image(ppmImage, Format::NV21), Format::IYUV); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "YUV to YUV not supported"));
		tests.add_core_test( 152, [&] { return new ColorConvert( env, get_image(ppmImage, Format::NV21), Format::YUV444); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "YUV to YUV not supported"));
		tests.add_core_test( 153, [&] { return new ColorConvert( env, get_image(ppmImage, Format::UYVY), Format::IYUV); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "YUV to YUV not supported"));
		tests.add_core_test( 154, [&] { return new ColorConvert( env, get_image(ppmImage, Format::YUYV422), Format::IYUV); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "YUV to YUV not supported"));
		tests.add_core_test( 155, [&] { return new ColorConvert( env, get_image(ppmImage, Format::IYUV), Format::RGB888); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"));
		tests.add_core_test( 156, [&] { return new ColorConvert( env, get_image(ppmImage, Format::IYUV), Format::RGBA8888); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "BT.709 not supported"));
		tests.add_core_test( 157, [&] { return new ColorConvert( env, get_image(ppmImage, Format::IYUV), Format::YUV444); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "YUV to YUV not supported"));
		tests.add_core_test( 158, [&] { return new ColorConvert( env, get_image(ppmImage, Format::IYUV), Format::NV12); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "YUV to YUV not supported") );
		tests.add_core_test( 159, [&] { return new ColorConvert( env, get_image(ppmImage, Format::YUYV422), Format::NV12); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "YUV to YUV not supported") );
		tests.add_core_test( 160, [&] { return new ColorConvert( env, get_image(ppmImage, Format::UYVY), Format::NV12); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "YUV to YUV not supported") );
#endif

		// Distribution tests
		tests.add_core_test( 161, [&] { return new Histogram( env, red, 256, 0, 256 ); });
		tests.add_core_test( 162, [&] { return new Histogram( env, red, 256, 50, 256 ); });
		tests.add_core_test( 163, [&] { return new Histogram( env, red, 64, 50, 128 ); });

		// Channel Combine tests
#ifndef TEST_TENSORS
		tests.add_core_test( 164,  [&] { return new ChannelCombine( env, red, green, blue, nullptr, Format::RGB888); } );
		tests.add_core_test( 165,  [&] { return new ChannelCombine( env, red, green, blue, &red, Format::RGBA8888); } );

		tests.add_core_test( 166,  [&] { return new ChannelCombine( env, red, u_w2, v_w2, nullptr, Format::YUYV422); },  TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 167,  [&] { return new ChannelCombine( env, red, u_w2, v_w2, nullptr, Format::UYVY); } , TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ));
		tests.add_core_test( 168,  [&] { return new ChannelCombine( env, red, u_wh2, v_wh2, nullptr, Format::NV12); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED )  );
		tests.add_core_test( 169,  [&] { return new ChannelCombine( env, red, u_wh2, v_wh2, nullptr, Format::NV21); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 170,  [&] { return new ChannelCombine( env, red, u_wh2, v_wh2, nullptr, Format::IYUV); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 171,  [&] { return new ChannelCombine( env, red, green, blue, nullptr, Format::YUV444); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
#endif
		// 172 -> 175 used to be the ChannelCombine tests where planes were moved, but this feature has been removed

		// Arithmetic Addition tests
		tests.add_core_test( 176,  [&] { return new ArithmeticAdditionU16U16U16Saturate( env, get_image(ppmImage,Format::LUMINANCE16), get_image(ppmImage,Format::LUMINANCE16)); } );
		tests.add_core_test( 177,  [&] { return new ArithmeticAdditionU16U16U16Wrap( env, get_image(ppmImage,Format::LUMINANCE16), get_image(ppmImage,Format::LUMINANCE16)); } );

		tests.add_core_test( 178,  [&] { return new CustomConvolutionScale( env, red, gaussian3x3,3); }, TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED, "Only in cl_framework"));
	}

	if( oldImage.format() != Format::UNKNOWN)
	{
		tests.add_core_test( 179,  [&] { return new OpticalFlow( env, oldImage, newImage, &old_points, &new_points_estimates,
							 Termination::TERM_CRITERIA_BOTH, 0.01f, 3, true, 5, pyramid_level, corner_threshold);},
							 TestFailure(TargetID::CL, FailureReason::CANNOT_BE_VALIDATED, BorderMode::UNDEFINED,"COMPMID-276 VX Optical flow scharr stage doesn't handle UNDEFINED mode correctly"),
							 TestFailure(TargetID::NEON, FailureReason::CANNOT_BE_VALIDATED, BorderMode::UNDEFINED,"COMPMID-276 VX Optical flow scharr stage doesn't handle UNDEFINED mode correctly")
		);// the last four are num_iterations, use initial estimates, dimension, level, corner_threshold
	}


	if(use_single_image)
	{
		// Integral Image tests
		tests.add_core_test( 180, [&] { return new IntegralImage( env, red ); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "Does not support U32 data type.") );

		tests.add_core_test( 182,  [&] { return new ScaleUpNearest( env, get_image(ppmImage,Format::LUMINANCE16), 2.f); });
		tests.add_core_test( 183,  [&] { return new LaplacianPyramid( env, blue, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "Laplacian Pyramid: Not Implemented"));
		tests.add_core_test( 184,  [&] { return new LaplacianPyramid( env, green, 2); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "Laplacian Pyramid: Not Implemented"));
		tests.add_core_test( 185,  [&] { return new LaplacianReconstruct( env, red, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED, "Laplacian reconstruct: Not Implemented"));

		// Equalize Histogram tests
		tests.add_core_test( 186, [&] { return new EqualizeHistogram( env, red ); });

		// Gaussian Pyramid test
		tests.add_core_test( 187, [&] { return new GaussianPyramid( env, red, 3, arm_compute::SCALE_PYRAMID_ORB); },
			TestFailure(TargetID::NEON, FailureReason::CANNOT_BE_VALIDATED, "COMPMID-277 Propagation of undefined pixels at lower levels of the pyramid and mismatches between arm_compute and openvx due to different calculation of image resolution.(https://cvs.khronos.org/bugzilla/show_bug.cgi?id=16075). "),
			TestFailure(TargetID::CL, FailureReason::CANNOT_BE_VALIDATED, "COMPMID-277 Propagation of undefined pixels at lower levels of the pyramid and mismatches between arm_compute and openvx due to different calculation of image resolution.(https://cvs.khronos.org/bugzilla/show_bug.cgi?id=16075). "));


		tests.add_core_test( 188,  [&] { return new CustomConvolutionSeparable5x5( env, red, gaussian5x5); } );

		// Additional Convolution tests
		tests.add_core_test( 189,  [&] { return new CustomConvolutionRandom( env, red, 3, 3, 10); } );
		tests.add_core_test( 190,  [&] { return new CustomConvolutionRandom( env, red, 5, 5, 10); } );
		tests.add_core_test( 191,  [&] { return new CustomConvolutionRandom( env, red, 7, 7, 10); } );
		tests.add_core_test( 192,  [&] { return new CustomConvolutionRandom( env, red, 9, 9, 10); } );
		tests.add_core_test( 193,  [&] { return new CustomConvolutionRandom( env, red, 3, 5, 10); } );
		tests.add_core_test( 194,  [&] { return new CustomConvolutionRandom( env, red, 3, 7, 10); } );
		tests.add_core_test( 195,  [&] { return new CustomConvolutionRandom( env, red, 3, 9, 10); } );
		tests.add_core_test( 196,  [&] { return new CustomConvolutionRandom( env, red, 5, 3, 10); } );
		tests.add_core_test( 197,  [&] { return new CustomConvolutionRandom( env, red, 5, 7, 10); } );
		tests.add_core_test( 198,  [&] { return new CustomConvolutionRandom( env, red, 5, 9, 10); } );
		tests.add_core_test( 199,  [&] { return new CustomConvolutionRandom( env, red, 7, 3, 10); } );
		tests.add_core_test( 200,  [&] { return new CustomConvolutionRandom( env, red, 7, 5, 10); } );
		tests.add_core_test( 201,  [&] { return new CustomConvolutionRandom( env, red, 7, 9, 10); } );
		tests.add_core_test( 202,  [&] { return new CustomConvolutionRandom( env, red, 9, 3, 10); } );
		tests.add_core_test( 203,  [&] { return new CustomConvolutionRandom( env, red, 9, 5, 10); } );
		tests.add_core_test( 204,  [&] { return new CustomConvolutionRandom( env, red, 9, 7, 10); } );
		tests.add_core_test( 205,  [&] { return new CustomConvolutionSeparable5x5( env, red, sobel_x_5x5); } );
		tests.add_core_test( 206,  [&] { return new CustomConvolutionSeparable7x7( env, red, sobel_x_7x7); } );
		tests.add_core_test( 207,  [&] { return new CustomConvolutionSeparable9x9( env, red, sep_9x9); } );

		// Non linear filter tests
		tests.add_core_test( 208,  [&] { return new NonLinearFilterBox( env, red, NonLinearFilterFunction::MEDIAN, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 209,  [&] { return new NonLinearFilterBox( env, red, NonLinearFilterFunction::MIN, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 210,  [&] { return new NonLinearFilterBox( env, red, NonLinearFilterFunction::MAX, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 211,  [&] { return new NonLinearFilterBox( env, red, NonLinearFilterFunction::MEDIAN, 5); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 212,  [&] { return new NonLinearFilterBox( env, red, NonLinearFilterFunction::MIN, 5); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 213,  [&] { return new NonLinearFilterBox( env, red, NonLinearFilterFunction::MAX, 5); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 214,  [&] { return new NonLinearFilterCross( env, red, NonLinearFilterFunction::MEDIAN, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 215,  [&] { return new NonLinearFilterCross( env, red, NonLinearFilterFunction::MIN, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 216,  [&] { return new NonLinearFilterCross( env, red, NonLinearFilterFunction::MAX, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 217,  [&] { return new NonLinearFilterCross( env, red, NonLinearFilterFunction::MEDIAN, 5); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 218,  [&] { return new NonLinearFilterCross( env, red, NonLinearFilterFunction::MIN, 5); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 219,  [&] { return new NonLinearFilterCross( env, red, NonLinearFilterFunction::MAX, 5); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 220,  [&] { return new NonLinearFilterDisk( env, red, NonLinearFilterFunction::MEDIAN, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 221,  [&] { return new NonLinearFilterDisk( env, red, NonLinearFilterFunction::MIN, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 222,  [&] { return new NonLinearFilterDisk( env, red, NonLinearFilterFunction::MAX, 3); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 223,  [&] { return new NonLinearFilterDisk( env, red, NonLinearFilterFunction::MEDIAN, 5); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 224,  [&] { return new NonLinearFilterDisk( env, red, NonLinearFilterFunction::MIN, 5); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );
		tests.add_core_test( 225,  [&] { return new NonLinearFilterDisk( env, red, NonLinearFilterFunction::MAX, 5); }, TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ) );

		tests.add_core_test( 226,  [&] { return new NonLinearFilterOther( env, red, NonLinearFilterFunction::MEDIAN, 3); },
				TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ), TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED, "Pattern other not implemented.") );

		tests.add_core_test( 227,  [&] { return new NonLinearFilterOther( env, red, NonLinearFilterFunction::MIN, 3); },
				TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ), TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED, "Pattern other not implemented.") );
		tests.add_core_test( 228,  [&] { return new NonLinearFilterOther( env, red, NonLinearFilterFunction::MAX, 3); },
				TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ), TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED, "Pattern other not implemented.") );
		tests.add_core_test( 229,  [&] { return new NonLinearFilterOther( env, red, NonLinearFilterFunction::MEDIAN, 5); },
				TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ), TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED, "Pattern other not implemented.") );
		tests.add_core_test( 230,  [&] { return new NonLinearFilterOther( env, red, NonLinearFilterFunction::MIN, 5); },
				TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ), TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED, "Pattern other not implemented.") );
		tests.add_core_test( 231,  [&] { return new NonLinearFilterOther( env, red, NonLinearFilterFunction::MAX, 5); },
				TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED ), TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED, "Pattern other not implemented.") );
	}
	if(!use_image)
	{
#ifdef USE_EIGEN
		// Tensor Operations
		tests.add_core_test( 258,  [&] { return new TensorOpTest( env, tensor_shape, DataType::F32, TensorOp::ADD); });
		tests.add_core_test( 259,  [&] { return new TensorOpTest( env, tensor_shape, DataType::F32, TensorOp::SUB); });
		tests.add_core_test( 260,  [&] { return new TensorOpTest( env, tensor_shape, DataType::F32, TensorOp::MUL); });

		// Softmax Layer
		tests.add_core_test( 261,  [&] { return new SoftmaxLayer( env, tensor_shape, DataType::F32); });
#endif
	}

	if(use_single_image)
	{
		// HOG descriptor
		tests.add_core_test( 262,  [&] { return new HOGDescriptor( env, red, Size2D(8, 8), Size2D(16, 16), Size2D(8, 8), 9, mali_cv::HOGNormType::L2HYS_NORM, 0.2f); },
			TestFailure(TargetID::VX, FailureReason::NOT_IMPLEMENTED),
			TestFailure(TargetID::CV, FailureReason::CANNOT_BE_VALIDATED, "Reference implementation not available."),
			TestFailure(TargetID::CL, FailureReason::CANNOT_BE_VALIDATED, "Reference implementation not available."),
			TestFailure(TargetID::NEON, FailureReason::CANNOT_BE_VALIDATED, "Reference implementation not available."));
	}
	if(!use_image)
	{
		// Transpose
		tests.add_core_test( 274,  [&] { return new Transpose(env, 102, 326, 0.0f, 255.0f, Format::FP32C1); });
		tests.add_core_test( 275,  [&] { return new Transpose(env, 95,  53,  0.0f, 255.0f, Format::FP16C1); });
		tests.add_core_test( 276,  [&] { return new Transpose(env, 150, 69,  0.0f, 255.0f, Format::U8); },
				TestFailure(TargetID::CL, FailureReason::NOT_SUPPORTED, "Not enough auto padding for 16x16 transpose"));
	}
	if(use_single_image)
	{
#ifdef __ARM_FEATURE_FP16_VECTOR_ARITHMETIC
		tests.add_core_test( 289,  [&] { return new AccumulateWeighted_FP16(env, red, blue, 0.1F); },
			TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED),
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));
		tests.add_core_test( 290,  [&] { return new Box3x3_FP16(env, red); },
			TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED),
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));
		tests.add_core_test( 291,  [&] { return new Magnitude_FP16( env, gx, gy); },
			TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED),
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));
		tests.add_core_test( 292,  [&] { return new CannyEdge_FP16( env, red, threshold_upper, threshold_lower, gradient, norm); },
			TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED),
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));
		tests.add_core_test( 293,  [&] { return new HarrisCorners_FP16(env,red, 0.00115f , 2.f, 0.04f, 3, 3, corner_threshold) ; },
			TestFailure(TargetID::CL, FailureReason::NOT_IMPLEMENTED),
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));
#endif /* __ARM_FEATURE_FP16_VECTOR_ARITHMETIC */
	}
	if(!use_image)
	{
		tests.add_core_test( 294, [&] { return new GEMMInterleave4x4( env, /*M*/ 73, /*N*/ 51, 2.0f, 12.0f, Format::FP32C1); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));

		tests.add_core_test( 295, [&] { return new GEMMInterleave4x4( env, /*M*/ 31, /*N*/ 83, 2.0f, 12.0f, Format::U8); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));

		tests.add_core_test( 296, [&] { return new GEMMInterleave4x4( env, /*M*/ 91, /*N*/ 49, 2.0f, 12.0f, Format::U16); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));

		tests.add_core_test( 297, [&] { return new GEMMTranspose1xW( env, /*M*/ 73, /*N*/ 51, 0.0f, 255.0f, Format::U8); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));

		tests.add_core_test( 298, [&] { return new GEMMTranspose1xW( env, /*M*/ 31, /*N*/ 83, 2.0f, 4.0f, Format::U16); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));

		tests.add_core_test( 299, [&] { return new GEMMTranspose1xW( env, /*M*/ 91, /*N*/ 49, 12.0f, 14.0f, Format::U32); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));

		tests.add_core_test( 302,  [&] { return new DepthConcatenate(env, {TensorShape( 14u, 14u, 100u), TensorShape(14u, 14u, 125u), TensorShape(14u, 14u, 31u)}); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));
		tests.add_core_test( 303,  [&] { return new DepthConcatenate(env, {TensorShape( 12u, 12u, 100u, 8u), TensorShape(14u, 14u, 125u, 8u), TensorShape(16u, 16u, 31u, 8u)}); },
			TestFailure(TargetID::CV, FailureReason::NOT_SUPPORTED));

#ifdef USE_EIGEN
		// Batch Normalization layer
		tests.add_core_test( 304,  [&] { return new BatchNormalizationLayer( env, tensor_shape, DataType::F32, 1E-50 ); });

		// Locally connected Layer
		tests.add_core_test( 321,  [&] { return new LocallyConnectedLayer(env, TensorShape(16U, 16U, 3U, 2U), TensorShape(4U, 4U, 3U, 4U, 16u), DataType::F32, 4, 4, 0, 0, DimensionRoundingType::FLOOR); },
			TestFailure(TargetID::VX, FailureReason::NOT_SUPPORTED));
		tests.add_core_test( 322,  [&] { return new LocallyConnectedLayer(env, TensorShape(25U, 25U, 3U, 16U), TensorShape(7U, 7U, 3U, 16U, 49U), DataType::F32, 3, 3, 0, 0, DimensionRoundingType::FLOOR); },
			TestFailure(TargetID::VX, FailureReason::NOT_SUPPORTED));

		// Additional CNN tests
		// Softmax Layer
		tests.add_core_test( 600,  [&] { return new SoftmaxLayer( env, TensorShape(10U), DataType::F32); });
#endif
	}

	if(!use_image)
	{
		// HOG detector
		tests.add_core_test( 308,  [&] { return new HOGDetector( env, 320U, 240U, Size2D(8, 8), Size2D(16, 16), Size2D(64, 128), Size2D(8, 8), 9); },
			TestFailure(TargetID::VX, FailureReason::NOT_IMPLEMENTED),
			TestFailure(TargetID::CV, FailureReason::NOT_IMPLEMENTED),
			TestFailure(TargetID::CL, FailureReason::CANNOT_BE_VALIDATED, "Reference implementation not available."),
			TestFailure(TargetID::NEON, FailureReason::CANNOT_BE_VALIDATED, "Reference implementation not available."));
	}

	if(use_image)
	{
		// HOG multi detection
		tests.add_core_test( 309,  [&] { return new HOGMultiDetection( env, red, Size2D(8, 8), Size2D(16, 16), Size2D(32, 32), Size2D(8, 8), 9, mali_cv::HOGNormType::L2HYS_NORM, 0.2f); },
			TestFailure(TargetID::VX, FailureReason::NOT_IMPLEMENTED),
			TestFailure(TargetID::CV, FailureReason::NOT_IMPLEMENTED),
			TestFailure(TargetID::CL, FailureReason::CANNOT_BE_VALIDATED, "Reference implementation not available."),
			TestFailure(TargetID::NEON, FailureReason::CANNOT_BE_VALIDATED, "Reference implementation not available."));
	}

	TestResults::report().init( env.ctx.num_iterations() );
	if( test_id >= 0 )
	{
		tests.run( test_id );
	}
	else
	{
		if ( 3 == argc)
			tests.run(179);
		else
		{
			tests.run_range(first_test_id, last_test_id, test_filter);
		}
	}

	printf("\n\n[ALL_TESTS_RUN]\n\n");
	TestResults::report().save_to_file( argv[1] ? argv[1]: default_image, env );
}
