#ifndef __KEY_POINT_H__
#define __KEY_POINT_H__

#ifdef USE_NEON
#include "arm_compute/runtime/Array.h"
#endif

#ifdef USE_OPENCL
#include "arm_compute/runtime/CL/CLArray.h"
#endif

#include "mali_cv/Error.h"
#include "mali_cv/Context.h"
#include "mali_cv/Types.h"

#include <vector>

namespace mali_cv
{
constexpr size_t MAX_NUM_VALUES = 20000;

template <class T>
class IArray
{
	public:
		virtual ~IArray() = default;
		void create( Context &ctx,size_t size )
		{
			internal_create( ctx, size );
		}
		virtual void read( std::vector<T> &out)const
		{
		};
		virtual void fill( std::vector<T> &in)
		{
		};
	protected:
		virtual void internal_create( Context &ctx, size_t s )
		{
		}
};

std::ostream& operator<<(std::ostream& os, const IKeyPoint& k);
std::ostream& operator<<(std::ostream& os, const ICoordinates2D& k);

#ifdef USE_OPENVX
template <class T>
class VXArray: public IArray<T>
{
	public:
		VXArray() : IArray<T>(), a(nullptr)
		{
		}
		VXArray(const VXArray&)=delete;
		const VXArray& operator=(const VXArray&)=delete;
		virtual void read( std::vector<T> &out ) const
		{
			vx_size num_items = 0;
			vxQueryArray(this->a, VX_ARRAY_NUMITEMS, &num_items, sizeof(num_items));

			if(num_items > 0)
			{
				vx_size stride = sizeof(vx_size);
				void *base = nullptr;
				vx_map_id map_id;
				CHECK_VX( vxMapArrayRange(this->a, 0, num_items, &map_id, &stride, &base, VX_READ_ONLY, VX_MEMORY_TYPE_NONE, VX_NOGAP_X));
				read_internal( out, base, num_items, stride );
				CHECK_VX( vxUnmapArrayRange( this->a, map_id));
			}
		}

		virtual void fill( std::vector<T> &in)
		{
			vx_size capacity = 0;
			vxQueryArray(this->a, VX_ARRAY_CAPACITY, &capacity, sizeof(capacity));
			ERROR_ON( in.size() > capacity);
			fill_internal(in);
		}
		vx_array a;
		~VXArray()
		{
			CHECK_VX( vxReleaseArray( &a ) );
		}
	protected:
		virtual void internal_create( Context &ctx, size_t s )
		{
			enum vx_type_e element_type_enum = get_element_type_enum();
			a = vxCreateArray( ctx.vx.context(), element_type_enum, s);
		}

		void read_internal( std::vector<T> &out, void* base, const vx_size num_items, const vx_size stride )const;
		void fill_internal( std::vector<T> &in);
	private:
		enum vx_type_e get_element_type_enum(void)
		{
			if (std::is_same<T, IKeyPoint>::value)
				return VX_TYPE_KEYPOINT;
			if (std::is_same<T, ICoordinates2D>::value)
				return VX_TYPE_COORDINATES2D;
			else
				return VX_TYPE_INVALID;
		}
};

#else /* USE_OPENVX */
template<typename T>
using VXArray = IArray<T>;
#endif /* USE_OPENVX */

#ifdef USE_NEON
template <class T>
class NEArray: public IArray<T>
{
	public:
		NEArray() : array(MAX_NUM_VALUES)
		{
		}
		virtual void read( std::vector<T> &out )const
		{
			size_t total_values = array.overflow() ? array.max_num_values() : array.num_values();
			out.resize( total_values );
			std::copy( array.buffer(), array.buffer() + total_values, out.begin());
		}
		virtual void fill( std::vector<T> &in)
		{
			ERROR_ON((in.size() > array.max_num_values()));
			std::copy(in.begin(), in.end(), array.buffer());
			array.resize(in.size());
		}
		arm_compute::Array<T> array;
};
#else /* USE_NEON */
template<typename T>
using NEArray = IArray<T>;
#endif /* USE_NEON */

#ifdef USE_OPENCV
template <class T>
class CVArray: public IArray<T>
{
	public:
		CVArray() : array()
		{
		}
		virtual void read( std::vector<T> &out )const
		{
			for ( const auto w : array )
			{
				out.push_back(std::move(w));
			}
		}
		std::vector<T> array;
};

template <>
class CVArray<IKeyPoint>: public IArray<IKeyPoint>
{
	public:
		CVArray() : array()
		{
		}
		virtual void read( std::vector<IKeyPoint> &out )const
		{
			for ( const auto& w : array )
			{
				IKeyPoint kp;
				kp.x = w.pt.x;
				kp.y = w.pt.y;
				out.push_back(kp);
			}
		}

		virtual void fill( std::vector<IKeyPoint> &in)
		{
			array.resize(in.size());

			for( unsigned int i = 0; i < in.size(); i++)
			{
				array[i].pt.x = in[i].x;
				array[i].pt.y = in[i].y;
			}
		}
		std::vector<cv::KeyPoint> array;
};
#else /* USE_OPENCV */
template<typename T>
using CVArray = IArray<T>;
#endif

#ifdef USE_OPENCL
template <class T>
class CLArray: public IArray<T>
{
	public:
		CLArray() : array(MAX_NUM_VALUES)
		{
		}
		void read( std::vector<T> &out )const
		{
			const size_t total_values = array.overflow() ? array.max_num_values() : array.num_values();
			out.resize( total_values );
			cl::CommandQueue q = cl::CommandQueue::getDefault();
			ERROR_ON_MSG( nullptr != array.buffer(),"The CL buffer is already mapped");
			array.map(q, true);
			std::copy( array.buffer(), array.buffer() + total_values, out.begin());
			array.unmap(q);
			q.finish();
		}
		void fill(std::vector<T> &in)
		{
			cl::CommandQueue q = cl::CommandQueue::getDefault();
			array.resize(in.size());
			ERROR_ON_MSG(array.buffer() != nullptr, "The CL buffer is already mapped");
			array.map(q, true);
			std::copy(in.begin(), in.end(), array.buffer());
			array.unmap(q);
			q.finish();
		}
		mutable arm_compute::CLArray<T> array;
};
#else /* USE_OPENCL */
template<typename T>
using CLArray = IArray<T>;
#endif /* USE_OPENCL */

template <class T>
struct Array final
{
	Array() : cl(), cv(), vx(), ne()
	{
	}
	Array( Context &ctx, size_t size ) : cl(), cv(), vx(), ne()
	{
		cl.create(ctx, size);
		cv.create(ctx, size);
		vx.create(ctx, size);
		ne.create(ctx, size);
	}

	CLArray<T> cl;
	CVArray<T> cv;
	VXArray<T> vx;
	NEArray<T> ne;
};

using IKeyPointArray = Array<IKeyPoint>;
using ICoord2DArray = Array<ICoordinates2D>;
}
#endif
