#ifndef __DISTRIBUTION_H__
#define __DISTRIBUTION_H__

#ifdef USE_NEON
#include "arm_compute/runtime/Distribution1D.h"
#endif

#ifdef USE_OPENCL
#include "arm_compute/runtime/CL/CLDistribution1D.h"
#endif

#include "mali_cv/Context.h"
#include "mali_cv/Types.h"

#include <memory>
#include <vector>

namespace mali_cv
{
class IDistribution
{
	public:
		virtual ~IDistribution() = default;
		void create(Context &ctx, size_t bins, int32_t offset, uint32_t range);
		virtual void read(std::vector<uint32_t> &out)const;
	protected:
		virtual void internal_create(Context &ctx, size_t bins, int32_t offset, uint32_t range)
		{
		}
};

#ifdef USE_OPENVX
class VXDistribution: public IDistribution
{
	public:
		VXDistribution();
		~VXDistribution();
		VXDistribution(const VXDistribution&)=delete;
		const VXDistribution& operator=(const VXDistribution&)=delete;
		virtual void read(std::vector<uint32_t> &out)const override;
	public:
		vx_distribution dist;
	protected:
		void internal_create(Context &ctx, size_t bins, int32_t offset, uint32_t range)override;
};

#else /* USE_OPENVX */
using VXDistribution = IDistribution;
#endif /* USE_OPENVX */

#ifdef USE_NEON
class NEDistribution: public IDistribution
{
	public:
		NEDistribution();
		virtual void read(std::vector<uint32_t> &out)const override;
	public:
        std::unique_ptr<arm_compute::Distribution1D> dist;
	protected:
		void internal_create(Context &ctx, size_t bins, int32_t offset, uint32_t range) override;
};
#else /* USE_NEON */
using NEDistribution = IDistribution;
#endif /* USE_NEON */

#ifdef USE_OPENCV
class CVDistribution: public IDistribution
{
	public:
		CVDistribution();
		~CVDistribution() = default;
		CVDistribution(const CVDistribution&) = delete;
		CVDistribution& operator=(const CVDistribution&) = delete;
		virtual void read( std::vector<uint32_t> &out)const override;
	public:
		int h_size;
		int h_offset;
		float h_range;
		cv::MatND hist;
	protected:
		void internal_create(Context &ctx, size_t bins, int32_t offset, uint32_t range) override;

};
#else /* USE_OPENCV */
using CVDistribution = IDistribution;
#endif

#ifdef USE_OPENCL
class CLDistribution: public IDistribution
{
	public:
		CLDistribution();
		virtual void read(std::vector<uint32_t> &out)const override;
	public:
        std::unique_ptr<arm_compute::CLDistribution1D> dist;
	protected:
		void internal_create(Context &ctx, size_t bins, int32_t offset, uint32_t range)override;
};
#else /* USE_OPENCL */
using CLDistribution = IDistribution;
#endif /* USE_OPENCL */

struct Distribution final
{
	Distribution(Context &ctx, size_t bins, int32_t offset, uint32_t range);
	CLDistribution cl;
	CVDistribution cv;
	VXDistribution vx;
	NEDistribution ne;
};
}
#endif
