#ifndef HELPER_H
#define HELPER_H

#include "arm_compute/core/TensorInfo.h"
#include "mali_cv/Error.h"
#include "mali_cv/Types.h"

#include <sstream>
#include <vector>

namespace mali_cv
{
struct ARM_COMPUTEHelper
{
	static arm_compute::Format convert_to_arm_compute_format(Format format)
	{
		switch(format)
		{
			case Format::LUMINANCE:
				return arm_compute::Format::U8;
			case Format::LUMINANCE16:
				return arm_compute::Format::S16;
			case Format::U16:
				return arm_compute::Format::U16;
			case Format::U32:
				return arm_compute::Format::U32;
			case Format::LUMINANCE32:
				return arm_compute::Format::S32;
			case Format::RGB888:
				return arm_compute::Format::RGB888;
			case Format::RGBA8888:
				return arm_compute::Format::RGBA8888;
			case Format::YUV444:
				return arm_compute::Format::YUV444;
			case Format::IYUV:
				return arm_compute::Format::IYUV;
			case Format::YUYV422:
				return arm_compute::Format::YUYV422;
			case Format::UYVY:
				return arm_compute::Format::UYVY422;
			case Format::NV12:
				return arm_compute::Format::NV12;
			case Format::NV21:
				return arm_compute::Format::NV21;
			case Format::FP32C1:
				return arm_compute::Format::F32;
			case Format::FP16C1:
				return arm_compute::Format::F16;
			default:
				return arm_compute::Format::UNKNOWN;
		}
	}
};

#ifdef USE_OPENVX
struct VXHelper
{
	static enum vx_type_e convert_to_vx_data_type(DataType type)
	{
		switch(type)
		{
			case DataType::U8:
				return VX_TYPE_UINT8;
			case DataType::S8:
				return VX_TYPE_INT8;
			case DataType::U16:
				return VX_TYPE_UINT16;
			case DataType::S16:
				return VX_TYPE_INT16;
			case DataType::U32:
				return VX_TYPE_UINT32;
			case DataType::S32:
				return VX_TYPE_INT32;
			case DataType::U64:
				return VX_TYPE_UINT64;
			case DataType::S64:
				return VX_TYPE_INT64;
			case DataType::F32:
				return VX_TYPE_FLOAT32;
			case DataType::F64:
				return VX_TYPE_FLOAT64;
			case DataType::SIZET:
				return VX_TYPE_SIZE;
			default:
				return VX_TYPE_INVALID;
		}
	}

	static vx_df_image convert_to_vx_format(Format format)
	{
		switch(format)
		{
			case Format::LUMINANCE32:
				return VX_DF_IMAGE_S32;
			case Format::U32:
				return VX_DF_IMAGE_U32;
			case Format::RGB888:
				return VX_DF_IMAGE_RGB;
			case Format::RGBA8888:
				return VX_DF_IMAGE_RGBX;
			case Format::LUMINANCE:
				return VX_DF_IMAGE_U8;
			case Format::LUMINANCE16:
				return VX_DF_IMAGE_S16;
			case Format::U16:
				return VX_DF_IMAGE_U16;
			case Format::NV12:
				return VX_DF_IMAGE_NV12;
			case Format::NV21:
				return VX_DF_IMAGE_NV21;
			case Format::UYVY:
				return VX_DF_IMAGE_UYVY;
			case Format::YUYV422:
				return VX_DF_IMAGE_YUYV;
			case Format::IYUV:
				return VX_DF_IMAGE_IYUV;
			case Format::YUV444:
				return VX_DF_IMAGE_YUV4;
			case Format::FP32C1:
				return VX_DF_IMAGE_S32; // VX does not support F32
			case Format::FP16C1:
				return VX_DF_IMAGE_S16; // VX does not support F16
			default:
				return VX_DF_IMAGE_VIRT;
		}
	}

	static vx_enum convert_to_vx_interpolation(InterpolationMode mode)
	{
		switch(mode)
		{
			case InterpolationMode::NEAREST_NEIGHBOR:
				return VX_INTERPOLATION_NEAREST_NEIGHBOR;
			case InterpolationMode::BILINEAR:
				return VX_INTERPOLATION_BILINEAR;
			case InterpolationMode::AREA:
				return VX_INTERPOLATION_AREA;
			default:
				ERROR("Not supported");
}
	}

	static bool pattern_check(vx_uint8* mask, vx_size cols, vx_size rows, vx_enum pattern)
	{
		if(VX_PATTERN_OTHER == pattern)
		{
			return true;
		}

		vx_size x, y;
		for (y = 0; y < rows; ++y)
		{
			for (x = 0; x < cols; ++x)
			{
				vx_uint8 value = mask[x + y * cols];
				vx_uint8 ref = 0;
				switch (pattern)
				{
					case VX_PATTERN_BOX:
						ref = 255;
						break;
					case VX_PATTERN_CROSS:
						ref = ((y == (rows / 2)) || (x == (cols / 2))) ? 255 : 0;
						break;
					case VX_PATTERN_DISK:
						ref = (((y - rows / 2.0 + 0.5) * (y - rows / 2.0 + 0.5)) / ((rows / 2.0) * (rows / 2.0)) +
						    ((x - cols / 2.0 + 0.5) * (x - cols / 2.0 + 0.5)) / ((cols / 2.0) * (cols / 2.0)))
						    <= 1 ? 255 : 0;
						break;
					default:
						break;
				}
				if(ref != value)
				{
					return false;
				}
			}
		}
		return true;
	}
};
#endif

#ifdef USE_OPENCV
struct CVHelper
{
	static int convert_to_cv_data_type(DataType type)
	{
		switch(type)
		{
			case DataType::U8:
				return CV_8U;
			case DataType::S8:
				return CV_8S;
			case DataType::U16:
				return CV_16U;
			case DataType::S16:
				return CV_16S;
			default:
				return 0;
		}
	}

	static int convert_to_cv_format(Format format)
	{
		switch(format)
		{
			case Format::RGB888:
				return CV_8UC3;
			case Format::RGBA8888:
				return CV_8UC4;
			case Format::LUMINANCE:
			case Format::IYUV:
			case Format::NV12:
			case Format::NV21:
				return CV_8UC1;
			case Format::LUMINANCE16:
				return CV_16S;
			case Format::U16:
				return CV_16U;
			case Format::LUMINANCE32:
				return CV_32S;
			case Format::UYVY:
			case Format::YUYV422:
				return CV_8UC2;
			case Format::YUV444:
				return CV_8UC1;
			case Format::FP32C1:
				return CV_32FC1;
			default:
				return 0;
		}
	}

	static int convert_to_cv_interpolation(InterpolationMode mode)
	{
		switch(mode)
		{
			case InterpolationMode::NEAREST_NEIGHBOR:
				return cv::INTER_NEAREST;
			case InterpolationMode::BILINEAR:
				return cv::INTER_LINEAR;
			case InterpolationMode::AREA:
			default:
				ERROR("Not supported");
		}
	}
};
#endif

std::vector<int> split_string_to_int(std::string input, char delimiter);
TensorShape vector_to_tensorshape(std::vector<int> v);
unsigned int random_seed();
std::string format_to_string(const Format &fmt);

}
#endif /* HELPER_H */
