#ifndef __CV_ITEST_H__
#define __CV_ITEST_H__

#include "mali_cv/Context.h"
#include "mali_cv/Image.h"
#include "mali_cv/Target.h"
#include "mali_cv/Validate.h"

#include <array>
#include <memory>

namespace mali_cv
{
enum class TargetID
{
	VX = 0,
	CL,
	CV,
	NEON,
	NUM_TARGETS
};

enum class Result
{
	PASSED,               /* Test passed, output not validated */
	FAILED_AS_EXPECTED,   /* Test failed to run or pass validation as expected */
	CRASHED,              /* Test crashed during execution */
	VALID_OUTPUT,         /* Test ran and its output has been validated */
	INVALID_OUTPUT,       /* Test ran but its output is invalid */
	SKIPPED               /* Test not run */
};

enum class FailureReason
{
	PASS_WITH_MISMATCHES, /* Test passes if a threshold or a percentage of mismatches is tolerated */
	NOT_IMPLEMENTED,      /* Test not currently implemented: could be added in the future */
	NOT_SUPPORTED,        /* Test not supported by the target: no way around it. */
	CANNOT_BE_VALIDATED   /* Test can run but the output can't be validated */
};


void record_test_result( const std::string name, int id, Result result, const char *target=nullptr, const char *message=nullptr);

class Environment
{
	public:
		Environment( int check_output, BorderMode border_mode, uint constant_border_value, int num_iterations);
		ITarget* target( TargetID id);
		CLTarget* cl_target();
		VXTarget* vx_target();
		CVTarget* cv_target();
		NETarget*  ne_target();
		uint border_to_exclude_if_undefined( uint border_width, uint other_border_width=0);
	public:
		Context ctx;
		int print_all;
		int env_threshold;
		int stop_on_error;
		const int check_output;
		TargetID reference, to_check;
		float acceptable_ratio_of_mismatches,min_threshold;
	private:
		std::array<std::unique_ptr<ITarget>, (uint)TargetID::NUM_TARGETS> targets;
};

class ITest;
class ExpectedFailures
{
	public:
		class Conditions
		{
			public:
				Conditions();
				bool is_going_to_fail(Environment &env);
			public:
				std::vector<int> borders;
				std::string message;
				float acceptable_ratio_of_mismatches,min_threshold;
				FailureReason reason;
		};
		ExpectedFailures(Environment &env);
		void add_expected_failure(int test_id, FailureReason reason, TargetID target, const std::string& message, float acceptable_ratio_of_mismatches=0.0f, float min_threshold=0.0f);
		void add_expected_failure(int test_id, FailureReason reason, TargetID target, BorderMode border_mode, const std::string& message, float acceptable_ratio_of_mismatches =0.0f, float threshold=0.0f);
		const char *expected_to_fail(int id, TargetID target, bool validate, float &acceptable_ratio_of_mismatches, float &min_threshold);
	private:
		std::vector< std::map<int,Conditions> > _expected_failures;
		Environment &_env;
};

class TestFailure
{
public:
	TestFailure():_reason(),_target(TargetID::NUM_TARGETS),_msg(),_acceptable_ratio_of_mismatches(0.0f),_min_threshold(0.0f),_border(std::pair<bool,BorderMode>(false,BorderMode::UNDEFINED))
    {
    }

	TestFailure(TargetID target, std::string msg, float acceptable_ratio_of_mismatches, float min_threshold=0.0f) :
		_target(target), _msg(msg), _acceptable_ratio_of_mismatches(acceptable_ratio_of_mismatches), _min_threshold(min_threshold),
		_border(std::pair<bool,BorderMode>(false,BorderMode::UNDEFINED)), _reason(FailureReason::PASS_WITH_MISMATCHES) {}

	TestFailure(TargetID target, BorderMode border_mode, std::string msg, float acceptable_ratio_of_mismatches, float min_threshold=0.0f) :
		_target(target), _msg(msg), _acceptable_ratio_of_mismatches(acceptable_ratio_of_mismatches), _min_threshold(min_threshold),
		_border(std::pair<bool,BorderMode>(true,border_mode)), _reason(FailureReason::PASS_WITH_MISMATCHES) {}
	TestFailure(TargetID target, FailureReason reason, std::string msg=""):_target(target),_msg(msg),_acceptable_ratio_of_mismatches(0.0f),
		_min_threshold(0.0f), _border(std::pair<bool,BorderMode>(false,BorderMode::UNDEFINED)),_reason(reason)
		{
			ERROR_ON( _reason == FailureReason::PASS_WITH_MISMATCHES);
			if(_msg == "" )
			{
				switch(_reason)
				{
					case FailureReason::NOT_IMPLEMENTED:
						_msg = "Not implemented";
						break;
					case FailureReason::CANNOT_BE_VALIDATED:
						_msg = "Implementation cannot be validated";
						break;
					case FailureReason::NOT_SUPPORTED:
						_msg = "Not supported";
						break;
					default:
						ERROR("Reason not supported");
				}
			}
		}
	TestFailure(TargetID target, FailureReason reason, BorderMode border_mode, std::string msg=""):_target(target),_msg(msg),_acceptable_ratio_of_mismatches(0.0f),
		_min_threshold(0.0f), _border(std::pair<bool,BorderMode>(true,border_mode)),_reason(reason)
		{
			ERROR_ON( _reason == FailureReason::PASS_WITH_MISMATCHES);
			if(_msg == "" )
			{
				switch(_reason)
				{
					case FailureReason::NOT_IMPLEMENTED:
						_msg = "Not implemented";
						break;
					case FailureReason::CANNOT_BE_VALIDATED:
						_msg = "Implementation cannot be validated";
						break;
					case FailureReason::NOT_SUPPORTED:
						_msg = "Not supported";
						break;
					default:
						ERROR("Reason not supported");
				}
			}
		}

	TargetID target() const { return _target; }
	std::string msg() const { return _msg; }
	float ratio_of_mismatches() const { return _acceptable_ratio_of_mismatches; }
	float threshold() const { return _min_threshold; }
	bool has_border_mode() const  { return _border.first; }
	BorderMode border_mode() const { return _border.second; }
	FailureReason reason()const { return _reason;}
private:
	FailureReason _reason;
	TargetID _target;
	std::string _msg;
	float _acceptable_ratio_of_mismatches;
	float _min_threshold;
	std::pair<bool,BorderMode> _border;
};



class TestManager
{
	public:
		TestManager(Environment &env);
		void add_core_test(int id, std::function<ITest*()> function);

		template<typename T, typename... Ts>
		void add_core_test(int id, std::function<ITest*()> function, T && exp_failure, Ts&&... other_exp_failures)
		{
			add_core_test(id,function);
			const std::array<T, sizeof...(Ts)> exp_fails{ { std::forward<Ts>(other_exp_failures)... } };
			add_test_failure(id, exp_failure);
			for( auto tf : exp_fails ) add_test_failure(id, tf);
		}

		void run(int id, const char*filter=nullptr);
		void run_range( const int first, const int last, const char* filter);
	private:
		void add_test_failure(int id, const  TestFailure &tf);
		ExpectedFailures _failures;
		std::map<int,std::function<ITest*()> > _tests;
};

class ITest
{
	public:
		virtual const std::string name() const = 0;
		ITest(Environment &env);
		virtual ~ITest() = default;

		virtual void run_cl();
		virtual void run_cv();
		virtual void run_vx();
		virtual void run_ne();
		void run_reference();
		void run(int id, ExpectedFailures& failures);
		virtual void validate_outputs();
	protected:
		void run_test_on_target( int id, ExpectedFailures &failures, TargetID target, bool validate);
		virtual void run( ITarget* target );
		void run( TargetID id);
		Environment &env;
};

#define TEST_2_IN_1_OUT( class_name, test_name, context_func )\
	class class_name:public ITest\
	{\
		public:\
			const std::string name() const override { return test_name;}\
			class_name( Environment &env, RawImage &in, RawImage &in2):ITest(env),a(env.ctx,in,CopyData::YES),b(env.ctx,in2,CopyData::YES),c(env.ctx,in,CopyData::NO) {}\
			virtual void run(ITarget *target)override { target-> context_func (a,b,c); }\
			virtual void validate_outputs( ) override { Validate::single_image_output( c, env ); }\
		private:\
			Image a,b,c;\
	};

#define TEST_1_IN_1_OUT( class_name, test_name, context_func, border_width )\
	class class_name:public ITest\
	{\
		public:\
			const std::string name() const override { return test_name;}\
			class_name( Environment &env, RawImage &in):ITest(env),a(env.ctx,in,CopyData::YES),c(env.ctx,in,CopyData::NO) {}\
			virtual void run(ITarget *target)override { target-> context_func (a,c); }\
			virtual void validate_outputs( ) override { Validate::single_image_output( c, env, env.border_to_exclude_if_undefined( border_width ) ); }\
		private:\
			Image a,c;\
	};

}
#endif /* __CV_ITEST_H__ */
