#ifndef __CV_IMAGE_H__
#define __CV_IMAGE_H__

#include "mali_cv/Context.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/Types.h"

#ifdef USE_NEON
#include "arm_compute/runtime/MultiImage.h"
#include "arm_compute/runtime/Tensor.h"
#endif

#ifdef USE_OPENCL
#include "arm_compute/core/ITensor.h"
#include "arm_compute/runtime/CL/CLTensor.h"
#include "arm_compute/runtime/CL/CLMultiImage.h"
#endif

namespace mali_cv
{
class IImage
{
	public:
		IImage();
		virtual ~IImage() = default;
		void create( Context &ctx, const RawImage &in, CopyData copy);
		void create( Context &ctx, size_t width, size_t height, Format format);
		virtual void load(const RawImage &in)
		{
		}
		virtual void store(RawImage &out)
		{
		}
		virtual arm_compute::ValidRegion valid_region() const
		{
		    return arm_compute::ValidRegion(arm_compute::Coordinates(0, 0), arm_compute::TensorShape(width, height));
		}
	protected:
		virtual void internal_create( Context &ctx, size_t w, size_t h, Format format)
		{
		}
#if defined(USE_NEON) || defined(USE_OPENCL)
		static unsigned char* copy_plane(arm_compute::IImage& in, unsigned char* out_buffer);
		static unsigned char* fill_plane(unsigned char* in_buffer, arm_compute::IImage& out);
#endif /* defined(USE_NEON) || defined(USE_OPENCL) */
	public:
		size_t width, height;
		Format format;
};

#ifdef USE_OPENVX
class VXImage: public IImage
{
	public:
		VXImage();
		~VXImage();
		VXImage(const VXImage&)=delete;
		const VXImage& operator=(const VXImage&)=delete;

		void store(RawImage &out) override;
		void load(const RawImage &in) override;
		arm_compute::ValidRegion valid_region() const override;
	protected:
		void internal_create( Context &ctx, size_t w, size_t h, Format format)override;
	public:
		vx_image img;
		vx_df_image type;
};

#else /* USE_OPENVX */
using VXImage = IImage;
#endif /* USE_OPENVX */

#ifdef USE_OPENCV
class CVImage: public IImage
{
	public:
		CVImage();
		void store(RawImage &out) override;
		void load(const RawImage &in) override;
	protected:
		void internal_create( Context &ctx, size_t w, size_t h, Format format)override;
	public:
		cv::Mat img;
};
#else /* USE_OPENCV */
using CVImage = IImage;
#endif /* USE_OPENCV */

#ifdef USE_NEON
#ifdef TEST_TENSORS
class NEImageTensor: public arm_compute::IImage
{
	public:
		static constexpr int dim2_to_use = 2;
		static constexpr int dim3_to_use = 1;
		/** @brief Constructor
		 */
		NEImageTensor();
		void init( arm_compute::ITensor& parent );
		NEImageTensor& operator=(NEImageTensor&) = delete;
		NEImageTensor(const NEImageTensor&) = delete;
		/** @brief Return a pointer to the image's metadata
		 *
		 * @return A pointer to the image's metadata.
		 */
		arm_compute::ITensorInfo* info() const override;
		/** @brief Return a pointer to the image's metadata
		 *
		 * @return A pointer to the image's metadata.
		 */
		arm_compute::ITensorInfo* info() override;
		/** @brief Return a pointer to the beginning of the image's allocation
		 *
		 * @return A pointer to the beginning of the image's allocation.
		 */
		uint8_t* buffer() const override;
	private:
		mutable arm_compute::TensorInfo _info;
		uint8_t *_buffer;
};
#endif /* TEST_TENSORS */

class NEImage: public IImage
{
	public:
		NEImage();
		void store(RawImage &out) override;
		void load(const RawImage &in) override;
		arm_compute::ValidRegion valid_region() const override;
	protected:
		void internal_create( Context &ctx, size_t w, size_t h, Format format)override;
	public:
#ifdef TEST_TENSORS
		arm_compute::Tensor tensor;
		NEImageTensor image;
#else /* TEST_TENSORS */
		arm_compute::MultiImage multi_img;
		arm_compute::ITensor& tensor;
		arm_compute::IImage& image;
#endif /* TEST_TENSORS */
};
#else /* USE_NEON */
using NEImage = IImage;
#endif /* USE_NEON */

#ifdef USE_OPENCL
#ifdef TEST_TENSORS
class CLImageTensor: public arm_compute::ICLImage
{
	public:
		static constexpr int dim2_to_use = 2;
		static constexpr int dim3_to_use = 1;
		/** @brief Constructor
		 */
		CLImageTensor();
		void init( arm_compute::ICLTensor& parent );
		CLImageTensor& operator=(CLImageTensor&) = delete;
		CLImageTensor(const CLImageTensor&) = delete;

		//Methods to override:
		arm_compute::ITensorInfo* info() const override;
		arm_compute::ITensorInfo* info() override;
		const cl::Buffer &cl_buffer() const override;
	protected:
		uint8_t *do_map(cl::CommandQueue &q, bool blocking) override;
		void do_unmap(cl::CommandQueue &q) override;
	private:
		mutable arm_compute::TensorInfo _info;
		cl::Buffer _buffer;
};
#endif /* TEST_TENSORS */
class CLImage: public IImage
{
	public:
		CLImage();
		using IImage::create;
		void store(RawImage &out) override;
		void load(const RawImage &in) override;
		arm_compute::ValidRegion valid_region() const override;
	protected:
		void internal_create( Context &ctx, size_t w, size_t h, Format format)override;
	public:
#ifdef TEST_TENSORS
		arm_compute::CLTensor tensor;
		CLImageTensor image;
#else
		arm_compute::CLMultiImage multi_img;
		arm_compute::ICLTensor &tensor;
		arm_compute::ICLImage &image;
#endif
};
#else /* USE_OPENCL */
using CLImage = IImage;
#endif /* USE_OPENCL */

class Image final
{
public:
	Image( Context &ctx, const RawImage &in, CopyData copy);
	Image( Context &ctx, const RawImage *in, CopyData copy);
	Image( Context &ctx, size_t width, size_t height, Format format);
	void load(const RawImage &in);

	CVImage cv;
	VXImage vx;
	CLImage cl;
	NEImage ne;

    size_t width{0};
    size_t height{0};
};

}
#endif /* __CV_IMAGE_H__ */
