#ifndef __LUT_H__
#define __LUT_H__

#ifdef USE_NEON
#include "arm_compute/runtime/Lut.h"
#endif

#ifdef USE_OPENCL
#include "arm_compute/runtime/CL/CLLut.h"
#endif

#include "mali_cv/Error.h"
#include "mali_cv/Context.h"
#include "mali_cv/Helper.h"
#include "mali_cv/Types.h"

#include <vector>

namespace mali_cv
{
class ILut
{
	public:
		virtual ~ILut() = default;
		void create( Context &ctx, size_t size, DataType type )
		{
			internal_create( ctx, size, type );
		}
		void init( void *data )
		{
			internal_init( data );
		}
	protected:
		virtual void internal_create( Context &ctx, size_t s, DataType type )
		{
		}
		virtual void internal_init( void *data )
		{
		}
};

#ifdef USE_OPENVX
class VXLut: public ILut
{
	public:
		VXLut() : ILut(), lut(nullptr)
		{
		};
		~VXLut() {
			CHECK_VX( vxReleaseLUT( &lut ) );
		}
		VXLut(const VXLut&)=delete;
		const VXLut& operator=(const VXLut&)=delete;
	protected:
		void internal_create( Context &ctx, size_t s, DataType type ) override
		{
			enum vx_type_e element_type_enum = VXHelper::convert_to_vx_data_type(type);
			lut = vxCreateLUT( ctx.vx.context(), element_type_enum, s );
		}
		void internal_init( void *data ) override
		{
			vxCopyLUT(lut, data, VX_WRITE_ONLY, VX_MEMORY_TYPE_NONE);
		}
	public:
		vx_lut lut;
};

#else /* USE_OPENVX */
using VXLut = ILut;
#endif /* USE_OPENVX */

#ifdef USE_NEON
class NELut: public ILut
{
	public:
		NELut() : lut()
		{
		}
		void internal_create( Context &ctx, size_t s, DataType type ) override
		{
			lut.allocator()->init( s, type );
		}
	protected:
		void internal_init( void *data ) override
		{
			std::memcpy(lut.buffer(), data, lut.size_in_bytes());
		}
	public:
		arm_compute::Lut lut;
};
#else /* USE_NEON */
using NELut = ILut;
#endif /* USE_NEON */

#ifdef USE_OPENCV
class CVLut: public ILut
{
	public:
		CVLut() : lut()
		{
		}
		void internal_create( Context &ctx, size_t s, DataType type ) override
		{
			lut = cv::Mat(1, s, CVHelper::convert_to_cv_data_type(type) );
		}
	protected:
		void internal_init( void *data ) override
		{
			std::memcpy(lut.data, data, 256);
		}
	public:
		cv::Mat lut;

};
#else /* USE_OPENCV */
using CVLut = ILut;
#endif

#ifdef USE_OPENCL
class CLLut: public ILut
{
	public:
		CLLut() : lut()
		{
		}
		void internal_create( Context &ctx, size_t s, DataType type ) override
		{
			lut.allocator()->init( s, type );
		}
	protected:
		void internal_init( void *data ) override
		{
			cl::CommandQueue q = cl::CommandQueue::getDefault();
			lut.map(q, true);
			std::memcpy(lut.buffer(), data, lut.size_in_bytes());
			lut.unmap(q);
			q.finish();
		}
	public:
		arm_compute::CLLut lut;
};
#else /* USE_OPENCL */
using CLLut = ILut;
#endif /* USE_OPENCL */

struct Lut final
{
	Lut( Context &ctx, size_t size, DataType type ) : cl(), cv(), vx(), ne()
	{
		cl.create(ctx, size, type);
		cv.create(ctx, size, type);
		vx.create(ctx, size, type);
		ne.create(ctx, size, type);
	}
	void init( void* data )
	{
		cl.init(data);
		cv.init(data);
		vx.init(data);
		ne.init(data);
	}

	CLLut cl;
	CVLut cv;
	VXLut vx;
	NELut ne;
};
}
#endif
