#ifndef __TARGET_H__
#define __TARGET_H__

#include "mali_cv/Array.h"
#include "mali_cv/Distribution.h"
#include "mali_cv/EigenFunctions.h"
#include "mali_cv/HOG.h"
#include "mali_cv/Lut.h"
#include "mali_cv/Scalar.h"
#include "mali_cv/Tensor.h"
#include "mali_cv/Timer.h"
#include "mali_cv/Types.h"

namespace mali_cv
{
class ITarget
{
	public:
		virtual std::string name() const = 0;
		ITarget(Context& ctx);
		virtual ~ITarget() = default;
		virtual void absolute_difference(Image &a, Image &b, Image &c);
		virtual void accumulate(Image &a, Image &c);
		virtual void accumulate_squared(Image &a, Image &c, uint32_t shift);
		virtual void accumulate_weighted(Image &a, Image &c, float alpha);
		virtual void accumulate_weighted_fp16(Image &a, Image &c, float alpha);
		virtual void arithmetic_addition(Image &a, Image &b, Image &c, uint8_t policy);
		virtual void arithmetic_subtraction(Image &a, Image &b, Image &c, uint8_t policy);
		virtual void bitwise_or(Image &a, Image &b, Image &c);
		virtual void bitwise_and(Image &a, Image &b, Image &c);
		virtual void bitwise_xor(Image &a, Image &b, Image &c);
		virtual void bitwise_not(Image &a, Image &c);
		virtual void erode(Image &a, Image &c);
		virtual void dilate(Image &a, Image &c);
		virtual void gaussian3x3(Image &a, Image &c);
		virtual void gaussian5x5(Image &a, Image &c);
		virtual void box3x3(Image &a, Image &c);
		virtual void box3x3_fp16(Image &a, Image &c);
		virtual void median3x3(Image &a, Image &c);
		virtual void convolution( Image &a, Image &c, Convolution &matrix);
		virtual void scale( Image &a, Image &c, InterpolationMode mode);
		virtual void color_convert( Image &a, Image &c);
		virtual void sobel(Image &a, Image *c, Image *d);
		virtual void sobel5x5(Image &a, Image *c, Image *d);
		virtual void sobel7x7(Image &a, Image *c, Image *d);
		virtual void scharr(Image &a, Image *c, Image *d);
		virtual void derivative( Image &a, Image *c, Image *d);
		virtual void gaussian_pyramid( Image &a, Pyramid &pyramid);
		virtual void laplacian_pyramid( Image &a, Pyramid &pyramid, Image &b, bool timers_enabled=true);
		virtual void laplacian_reconstruct( Pyramid &pyramid, Image &a, Image &c);
		virtual void canny(Image& a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image& c);
		virtual void canny_fp16(Image& a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image& c);
		virtual void harris_corners(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners,size_t& num_corners);
		virtual void harris_corners_fp16(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners,size_t& num_corners);
		virtual void fast_corners(Image &a,float threshold, bool nonmax_suppression, IKeyPointArray& corners,size_t& num_corners);
		virtual void warp_affine(Image &a, Image &c, const float matrix[3][2], InterpolationMode mode);
		virtual void warp_perspective(Image &a, Image &c, const float matrix[3][3], InterpolationMode mode);
		virtual void mean_stddev(Image &a, Scalar<float> &mean, Scalar<float> &stddev);
		virtual void magnitude(Image &a, Image &b, Image &c);
		virtual void magnitude_fp16(Image &a, Image &b, Image &c);
		virtual void phase(Image &a, Image &b, Image &c);
		virtual void threshold_binary(Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value);
		virtual void threshold_range(Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value, uint8_t upper);
		virtual void remap(Image &a, RemapTable &table, Image &d, InterpolationMode policy);
		virtual void pixelwise_mul(Image &a , Image &b, Image &c, float scale, uint8_t overflow_policy, uint8_t round_policy );
		virtual void convert_depth(Image &a, Image &c, ConvertPolicy policy , uint32_t shift);
		virtual void channel_extract( Image &a, Channel channel, Image &c );
		virtual void tablelookup( Image &a , Lut &b, Image &c );
		virtual void minmaxlocation( Image &a, Scalar<int32_t> &min, Scalar<int32_t> &max,
		                             ICoord2DArray *min_loc = nullptr, ICoord2DArray *max_loc = nullptr,
		                             Scalar<uint32_t> *min_count = nullptr, Scalar<uint32_t> *max_count = nullptr);
		virtual void histogram(Image &a, Distribution &b);
		virtual void channel_combine(Image &plane0, Image &plane1, Image &plane2, Image &plane3, Image &out);
		virtual void optical_flow(Image &a, Image &b, IKeyPointArray *old_points, IKeyPointArray *new_points_estimates,
		                          IKeyPointArray *new_points, Termination termination, float epsilon, uint32_t num_iterations,
		                          bool use_initial_estimate, size_t window_dimension, size_t num_levels);
		virtual void integral(Image &a, Image &b);
		virtual void equalize_histogram(Image &a, Image &b);
		virtual void non_linear_filter(Image &a, Image &b, NonLinearFilterFunction function, MatrixPattern pattern, uint mask_size, uint8_t* mask = nullptr);
		virtual void activation_layer(Tensor &in, Tensor &out, ActivationLayerInfo act_info);
		virtual void normalization_layer(Tensor &a, Tensor &b, NormalizationLayerInfo norm_info);
		virtual void batchnormalization_layer(Tensor &a, Tensor &b, Tensor &mean, Tensor &var, Tensor &beta, Tensor &gamma, float epsilon);
		virtual void pooling_layer(Tensor &in, Tensor &out, PoolingLayerInfo pool_info);
		virtual void softmax_layer(Tensor &in, Tensor &out);
		virtual void tensor_ops(Tensor &in1, Tensor &in2, Tensor &out, TensorOp op);
		virtual void convolution_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info, Tensor *rw = nullptr);
		virtual void fully_connected_layer(Tensor &input, Tensor &weight, Tensor *bias, Tensor &output, bool are_weights_reshaped = false);
		virtual void locally_connected_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info);
		virtual void hog_descriptor(Image &a, HOG &b, Tensor &c);
		virtual void hog_detector(Tensor &a, HOG &b, IKeyPointArray &keypoint_detection_windows);
		virtual void hog_multi_detection(Image &a, HOG &b, IKeyPointArray &keypoint_detection_windows);
		virtual void transpose(Image &a, Image &b);
		virtual void gemm_fp32(Tensor &a, Tensor &b, Tensor &c, Tensor &d, float alpha, float beta);
		virtual void gemm_fp16(Image &a, Image &b, Image &c, Image &d, float alpha, float beta);
		virtual void gemm_lowp(Image &a, Image &b, Image &d, int32_t a_offset, int32_t b_offset, int32_t c_offset, int32_t c_mult_int, int32_t shift);
		virtual void interleave_4x4(Image &in, Image &out);
		virtual void concatenate_depth(std::vector<std::unique_ptr<Tensor>> &a, Tensor &b);
		virtual void transpose_1xW(Image &in, Image &out);

	protected:
		Context &_ctx;
		Timer _timer;
};

#ifdef USE_OPENVX
class VXTarget: public ITarget
{
	public:
		VXTarget(Context& ctx);
		std::string name() const override
		{
			return "OpenVX";
		}
		void absolute_difference(Image &a, Image &b, Image &c)override;
		void accumulate(Image &a, Image &c)override;
		void accumulate_squared(Image &a, Image &c, uint32_t shift)override;
		void accumulate_weighted(Image &a, Image &c, float alpha)override;
		void accumulate_weighted_fp16(Image &a, Image &c, float alpha)override;
		void arithmetic_addition(Image &a, Image &b, Image &c, uint8_t policy)override;
		void arithmetic_subtraction(Image &a, Image &b, Image &c, uint8_t policy)override;
		void bitwise_or(Image &a, Image &b, Image &c)override;
		void bitwise_and(Image &a, Image &b, Image &c)override;
		void bitwise_xor(Image &a, Image &b, Image &c)override;
		void bitwise_not(Image &a, Image &c)override;
		void erode(Image &a, Image &c)override;
		void dilate(Image &a, Image &c)override;
		void gaussian3x3(Image &a, Image &c)override;
		void gaussian5x5(Image &a, Image &c)override;
		void box3x3(Image &a, Image &c)override;
		void box3x3_fp16(Image &a, Image &c)override;
		void median3x3(Image &a, Image &c)override;
		void convolution( Image &a, Image &c, Convolution &matrix)override;
		void scale( Image &a, Image &c, InterpolationMode mode)override;
		void color_convert( Image &a, Image &c)override;
		void sobel(Image &a, Image *c, Image *d)override;
		void sobel7x7(Image &a, Image *c, Image *d)override;
		void sobel5x5(Image &a, Image *c, Image *d)override;
		void scharr(Image &a, Image *c, Image *d)override;
		void derivative( Image &a, Image *c, Image *d)override;
		void gaussian_pyramid( Image &a, Pyramid &pyramid)override;
		void laplacian_pyramid( Image &a, Pyramid &pyramid, Image &b, bool timers_enabled=true)override;
		void laplacian_reconstruct( Pyramid &pyramid, Image &a, Image &c)override;
		void canny(Image& a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image& c)override;
		void canny_fp16(Image& a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image& c)override;
		void harris_corners(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners, size_t& num_corners) override;
		void harris_corners_fp16(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners, size_t& num_corners) override;
		void fast_corners(Image &a,float threshold, bool nonmax_suppression, IKeyPointArray& corners, size_t& num_corners)override;
		void warp_affine(Image &a, Image &c, const float matrix[3][2], InterpolationMode mode)override;
		void warp_perspective(Image &a, Image &c, const float matrix[3][3], InterpolationMode mode)override;
		void mean_stddev(Image &a, Scalar<float> &mean, Scalar<float> &stddev) override;
		void magnitude(Image &a, Image &b, Image &c) override;
		void magnitude_fp16(Image &a, Image &b, Image &c) override;
		void phase(Image &a, Image &b, Image &c) override;
		void threshold_binary(Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value) override;
		void threshold_range(Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value, uint8_t upper) override;
		void remap(Image &a, RemapTable &table, Image &d, InterpolationMode policy) override;
		void pixelwise_mul(Image &a , Image &b, Image &c, float scale, uint8_t overflow_policy, uint8_t round_policy )override;
		void convert_depth(Image &a, Image &c, ConvertPolicy policy , uint32_t shift)override;
		void channel_extract( Image &a, Channel channel, Image &c )override;
		void tablelookup( Image &a , Lut &b, Image &c )override;
		void minmaxlocation( Image &a, Scalar<int32_t> &min, Scalar<int32_t> &max,
		                     ICoord2DArray *min_loc = nullptr, ICoord2DArray *max_loc = nullptr,
		                     Scalar<uint32_t> *min_count = nullptr, Scalar<uint32_t> *max_count = nullptr)override;
		void histogram(Image &a, Distribution &b)override;
		void channel_combine(Image &plane0, Image &plane1, Image &plane2, Image &plane3, Image &out)override;
		void optical_flow(Image &a, Image &b, IKeyPointArray *old_points, IKeyPointArray *new_points_estimates,
		                  IKeyPointArray *new_points, Termination termination, float epsilon, uint32_t num_iterations,
		                  bool use_initial_estimate, size_t window_dimension, size_t num_levels)override;
		void integral(Image &a, Image &b)override;
		void equalize_histogram(Image &a, Image &b)override;
		void non_linear_filter(Image &a, Image &b, NonLinearFilterFunction function, MatrixPattern pattern, uint mask_size, uint8_t* mask = nullptr) override;
		void hog_descriptor(Image &a, HOG &b, Tensor &c)override;
		void activation_layer(Tensor &in, Tensor &out, ActivationLayerInfo act_info) override;
		void normalization_layer(Tensor &a, Tensor &b, NormalizationLayerInfo norm_info) override;
		void batchnormalization_layer(Tensor &a, Tensor &b, Tensor &mean, Tensor &var, Tensor &beta, Tensor &gamma, float epsilon) override;
		void pooling_layer(Tensor &in, Tensor &out, PoolingLayerInfo pool_info) override;
		void softmax_layer(Tensor &in, Tensor &out) override;
		void tensor_ops(Tensor &in1, Tensor &in2, Tensor &out, TensorOp op) override;
		void convolution_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info, Tensor *rw = nullptr) override;
		void fully_connected_layer(Tensor &input, Tensor &weight, Tensor *bia, Tensor &output, bool are_weights_reshaped = false)override;
		void locally_connected_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info)override;
		void transpose(Image &a, Image &b) override;
		void gemm_fp32(Tensor &a, Tensor &b, Tensor &c, Tensor &d, float alpha, float beta) override;
		void gemm_fp16(Image &a, Image &b, Image &c, Image &d, float alpha, float beta) override;
		void gemm_lowp(Image &a, Image &b, Image &d, int32_t a_offset, int32_t b_offset, int32_t c_offset, int32_t c_mult_int, int32_t shift) override;
		void interleave_4x4(Image &in, Image &out)override;
		void concatenate_depth(std::vector<std::unique_ptr<Tensor>> &a, Tensor &b)override;
		void transpose_1xW(Image &in, Image &out)override;

	private:
		EigenFunctions _eigen;
};
#else/* USE_OPENVX */
using VXTarget = ITarget;
#endif /* USE_OPENVX */

#ifdef USE_OPENCV
class CVTarget: public ITarget
{
	public:
		CVTarget(Context& ctx);
		std::string name() const override
		{
			return "OpenCV";
		}
		void absolute_difference(Image &a, Image &b, Image &c)override;
		void accumulate(Image &a, Image &c)override;
		void accumulate_squared(Image &a, Image &c, uint32_t shift)override;
		void accumulate_weighted(Image &a, Image &c, float alpha)override;
		void arithmetic_addition(Image &a, Image &b, Image &c, uint8_t policy)override;
		void arithmetic_subtraction(Image &a, Image &b, Image &c, uint8_t policy)override;
		void bitwise_or(Image &a, Image &b, Image &c)override;
		void bitwise_and(Image &a, Image &b, Image &c)override;
		void bitwise_xor(Image &a, Image &b, Image &c)override;
		void bitwise_not(Image &a, Image &c)override;
		void erode(Image &a, Image &c)override;
		void dilate(Image &a, Image &c)override;
		void gaussian3x3(Image &a, Image &c)override;
		void gaussian5x5(Image &a, Image &c)override;
		void box3x3(Image &a, Image &c)override;
		void median3x3(Image &a, Image &c)override;
		void convolution( Image &a, Image &c, Convolution &matrix)override;
		void scale( Image &a, Image &c, InterpolationMode mode)override;
		void color_convert( Image &a, Image &c)override;
		void sobel(Image &a, Image *c, Image *d)override;
		void sobel5x5(Image &a, Image *c, Image *d)override;
		void sobel7x7(Image &a, Image *c, Image *d)override;
		void scharr(Image &a, Image *c, Image *d)override;
		void derivative( Image &a, Image *c, Image *d)override;
		void gaussian_pyramid( Image &a, Pyramid &pyramid)override;
		void laplacian_pyramid( Image &a, Pyramid &pyramid, Image &b, bool timers_enabled=true)override;
		void laplacian_reconstruct( Pyramid &pyramid, Image &a, Image &c)override;
		void canny(Image& a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image& c)override;
		void harris_corners(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners,size_t& num_corners)override;
		void fast_corners(Image &a,float threshold, bool nonmax_suppression, IKeyPointArray& corners, size_t& num_corners)override;
		void warp_affine(Image &a, Image &c, const float matrix[3][2], InterpolationMode mode)override;
		void warp_perspective(Image &a, Image &c, const float matrix[3][3], InterpolationMode mode)override;
		void mean_stddev(Image &a, Scalar<float> &mean, Scalar<float> &stddev)override;
		void magnitude(Image &a, Image &b, Image &c) override;
		void phase(Image &a, Image &b, Image &c) override;
		void remap(Image &a, RemapTable &table, Image &d, InterpolationMode policy)override;
		void pixelwise_mul(Image &a , Image &b, Image &c, float scale, uint8_t overflow_policy, uint8_t round_policy)override;
		void channel_extract( Image &a, Channel channel, Image &c )override;
		void tablelookup( Image &a , Lut &b, Image &c )override;
		void convert_depth(Image &a, Image &c, ConvertPolicy policy , uint32_t shift)override;
		void minmaxlocation( Image &a, Scalar<int32_t> &min, Scalar<int32_t> &max,
		                     ICoord2DArray *min_loc = nullptr, ICoord2DArray *max_loc = nullptr,
		                     Scalar<uint32_t> *min_count = nullptr, Scalar<uint32_t> *max_count = nullptr)override;
		void histogram(Image &a, Distribution &b)override;
		void channel_combine(Image &plane0, Image &plane1, Image &plane2, Image &plane3, Image &out)override;
		void optical_flow(Image &a, Image &b, IKeyPointArray *old_points, IKeyPointArray *new_points_estimates,
		                  IKeyPointArray *new_points, Termination termination, float epsilon, uint32_t num_iterations,
		                  bool use_initial_estimate, size_t window_dimension, size_t num_levels)override;
		void integral(Image &a, Image &b)override;
		void equalize_histogram(Image &a, Image &b)override;
		void non_linear_filter(Image &a, Image &b, NonLinearFilterFunction function, MatrixPattern pattern, uint mask_size, uint8_t* mask = nullptr) override;
		void hog_descriptor(Image &a, HOG &b, Tensor &c)override;
		void transpose(Image &a, Image &b) override;
		void activation_layer(Tensor &in, Tensor &out, ActivationLayerInfo act_info) override;
		void normalization_layer(Tensor &a, Tensor &b, NormalizationLayerInfo norm_info) override;
		void batchnormalization_layer(Tensor &a, Tensor &b, Tensor &mean, Tensor &var, Tensor &beta, Tensor &gamma, float epsilon) override;
		void pooling_layer(Tensor &in, Tensor &out, PoolingLayerInfo pool_info) override;
		void softmax_layer(Tensor &in, Tensor &out) override;
		void tensor_ops(Tensor &in1, Tensor &in2, Tensor &out, TensorOp op) override;
		void convolution_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info, Tensor *rw = nullptr) override;
		void fully_connected_layer(Tensor &input, Tensor &weight, Tensor *bia, Tensor &output, bool are_weights_reshaped = false)override;
		void locally_connected_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info)override;
		void gemm_fp32(Tensor &a, Tensor &b, Tensor &c, Tensor &d, float alpha, float beta) override;
	private:
		EigenFunctions _eigen;
};
#else /* USE_OPENCV */
using CVTarget = ITarget;
#endif /* USE_OPENCV */

#ifdef USE_NEON
class NETarget: public ITarget
{
	public:
		NETarget(Context& ctx);
		std::string name() const override
		{
			return "NEON";
		}
		void absolute_difference(Image &a, Image &b, Image &c)override;
		void accumulate(Image &a, Image &c)override;
		void accumulate_squared(Image &a, Image &c, uint32_t shift)override;
		void accumulate_weighted(Image &a, Image &c, float alpha)override;
		void accumulate_weighted_fp16(Image &a, Image &c, float alpha)override;
		void arithmetic_addition(Image &a, Image &b, Image &c, uint8_t policy)override;
		void arithmetic_subtraction(Image &a, Image &b, Image &c, uint8_t policy)override;
		void bitwise_or(Image &a, Image &b, Image &c)override;
		void bitwise_and(Image &a, Image &b, Image &c)override;
		void bitwise_xor(Image &a, Image &b, Image &c)override;
		void bitwise_not(Image &a, Image &c)override;
		void erode(Image &a, Image &c)override;
		void dilate(Image &a, Image &c)override;
		void gaussian3x3(Image &a, Image &c)override;
		void gaussian5x5(Image &a, Image &c)override;
		void box3x3(Image &a, Image &c)override;
		void box3x3_fp16(Image &a, Image &c)override;
		void median3x3(Image &a, Image &c)override;
		void convolution( Image &a, Image &c, Convolution &matrix)override;
		void scale( Image &a, Image &c, InterpolationMode mode)override;
		void color_convert( Image &a, Image &c)override;
		void sobel(Image &a, Image *c, Image *d)override;
		void sobel5x5(Image &a, Image *c, Image *d)override;
		void sobel7x7(Image &a, Image *c, Image *d)override;
		void scharr(Image &a, Image *c, Image *d)override;
		void derivative( Image &a, Image *c, Image *d)override;
		void gaussian_pyramid( Image &a, Pyramid &pyramid)override;
		void laplacian_pyramid( Image &a, Pyramid &pyramid, Image &b, bool timers_enabled=true)override;
		void laplacian_reconstruct( Pyramid &pyramid, Image &a, Image &c)override;
		void canny(Image& a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image& c)override;
		void canny_fp16(Image& a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image& c)override;
		void harris_corners(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners,size_t& num_corners)override;
		void harris_corners_fp16(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners,size_t& num_corners)override;
		void fast_corners(Image &a,float threshold, bool nonmax_suppression, IKeyPointArray& corners,size_t& num_corners)override;
		void warp_affine(Image &a, Image &c, const float matrix[3][2], InterpolationMode mode)override;
		void warp_perspective(Image &a, Image &c, const float matrix[3][3], InterpolationMode mode)override;
		void mean_stddev(Image &a, Scalar<float> &mean, Scalar<float> &stddev)override;
		void magnitude(Image &a, Image &b, Image &c) override;
		void magnitude_fp16(Image &a, Image &b, Image &c) override;
		void phase(Image &a, Image &b, Image &c) override;
		void threshold_binary(Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value)override;
		void threshold_range(Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value, uint8_t upper)override;
		void remap(Image &a, RemapTable &table, Image &d, InterpolationMode policy)override;
		void pixelwise_mul(Image &a , Image &b, Image &c, float scale, uint8_t overflow_policy, uint8_t round_policy)override;
		void channel_extract( Image &a, Channel channel, Image &c )override;
		void tablelookup( Image &a , Lut &b, Image &c )override;
		void convert_depth(Image &a, Image &c, ConvertPolicy policy , uint32_t shift)override;
		void minmaxlocation( Image &a, Scalar<int32_t> &min, Scalar<int32_t> &max,
		                     ICoord2DArray *min_loc = nullptr, ICoord2DArray *max_loc = nullptr,
		                     Scalar<uint32_t> *min_count = nullptr, Scalar<uint32_t> *max_count = nullptr)override;
		void histogram(Image &a, Distribution &b)override;
		void channel_combine(Image &plane0, Image &plane1, Image &plane2, Image &plane3, Image &out)override;
		void optical_flow(Image &a, Image &b, IKeyPointArray *old_points, IKeyPointArray *new_points_estimates,
		                  IKeyPointArray *new_points, Termination termination, float epsilon, uint32_t num_iterations,
		                  bool use_initial_estimate, size_t window_dimension, size_t num_levels)override;
		void integral(Image &a, Image &b)override;
		void equalize_histogram(Image &a, Image &b)override;
		void non_linear_filter(Image &a, Image &b, NonLinearFilterFunction function, MatrixPattern pattern, uint mask_size, uint8_t* mask = nullptr) override;
		void hog_descriptor(Image &a, HOG &b, Tensor &c)override;
		void hog_detector(Tensor &a, HOG &b, IKeyPointArray &keypoint_detection_windows)override;
		void hog_multi_detection(Image &a, HOG &b, IKeyPointArray &keypoint_detection_windows)override;
		void tensor_ops(Tensor &in1, Tensor &in2, Tensor &out, TensorOp op) override;
		void activation_layer(Tensor &in, Tensor &out, ActivationLayerInfo act_info) override;
		void pooling_layer(Tensor &in, Tensor &out, PoolingLayerInfo pool_info) override;
		void softmax_layer(Tensor &in, Tensor &out)override;
		void normalization_layer(Tensor &a, Tensor &b, NormalizationLayerInfo norm_info) override;
		void batchnormalization_layer(Tensor &a, Tensor &b, Tensor &mean, Tensor &var, Tensor &beta, Tensor &gamma, float epsilon) override;
		void locally_connected_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info) override;
		void convolution_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info, Tensor *rw = nullptr) override;
		void gemm_fp32(Tensor &a, Tensor &b, Tensor &c, Tensor &d, float alpha, float beta)override;
		void gemm_fp16( Image &a, Image &b, Image &c, Image &d, float alpha, float beta)override;
		void transpose(Image &a, Image &b) override;
		void fully_connected_layer(Tensor &input, Tensor &weight, Tensor *bia, Tensor &output, bool are_weights_reshaped = false)override;
		void interleave_4x4(Image &in, Image &out)override;
		void transpose_1xW(Image &in, Image &out)override;
		void concatenate_depth(std::vector<std::unique_ptr<Tensor>> &a, Tensor &b) override;

	private:
		void convolution3x3(Image &a, Image &c, Convolution &matrix);
		void convolution5x5(Image &a, Image &c, Convolution &matrix);
		void convolution7x7(Image &a, Image &c, Convolution &matrix);
		void convolution9x9(Image &a, Image &c, Convolution &matrix);
		void convolution_rectangle(Image &a, Image &c, Convolution &matrix);



};
#else /* USE_NEON */
using NETarget = ITarget;
#endif /* USE_NEON */

#ifdef USE_OPENCL
class CLTarget: public ITarget
{
	public:
		CLTarget(Context& ctx);
		std::string name() const override
		{
			return "OpenCL";
		}
		void absolute_difference(Image &a, Image &b, Image &c)override;
		void accumulate(Image &a, Image &c)override;
		void accumulate_squared(Image &a, Image &c, uint32_t shift)override;
		void accumulate_weighted(Image &a, Image &c, float alpha)override;
		void arithmetic_addition(Image &a, Image &b, Image &c, uint8_t policy)override;
		void arithmetic_subtraction(Image &a, Image &b, Image &c, uint8_t policy)override;
		void bitwise_or(Image &a, Image &b, Image &c)override;
		void bitwise_and(Image &a, Image &b, Image &c)override;
		void bitwise_xor(Image &a, Image &b, Image &c)override;
		void bitwise_not(Image &a, Image &c)override;
		void erode(Image &a, Image &c)override;
		void dilate(Image &a, Image &c)override;
		void gaussian3x3(Image &a, Image &c)override;
		void gaussian5x5(Image &a, Image &c)override;
		void box3x3(Image &a, Image &c)override;
		void median3x3(Image &a, Image &c)override;
		void convolution( Image &a, Image &c, Convolution &matrix)override;
		void scale( Image &a, Image &c, InterpolationMode mode)override;
		void color_convert( Image &a, Image &c)override;
		void sobel(Image &a, Image *c, Image *d)override;
		void sobel5x5(Image &a, Image *c, Image *d)override;
		void sobel7x7(Image &a, Image *c, Image *d)override;
		void scharr(Image &a, Image *c, Image *d)override;
		void gaussian_pyramid( Image &a, Pyramid &pyramid)override;
		void laplacian_pyramid( Image &a, Pyramid &pyramid, Image &b, bool timers_enabled=true)override;
		void laplacian_reconstruct( Pyramid &pyramid, Image &a, Image &c)override;
		void canny(Image& a, int32_t upper, int32_t lower, int32_t gradient_size, int32_t norm_type, Image& c)override;
		void mean_stddev(Image &a, Scalar<float> &mean, Scalar<float> &stddev)override;
		void harris_corners(Image &a,float threshold,float min_dist,float sensitivity,uint8_t gradient,uint8_t block_size, IKeyPointArray& corners,size_t& num_corners)override;
		void fast_corners(Image &a,float threshold, bool nonmax_suppression, IKeyPointArray& corners,size_t& num_corners)override;
		void warp_affine(Image &a, Image &c, const float matrix[3][2], InterpolationMode mode)override;
		void warp_perspective(Image &a, Image &c, const float matrix[3][3], InterpolationMode mode)override;
		void convert_depth(Image &a, Image &c, ConvertPolicy policy , uint32_t shift)override;
		void threshold_binary(Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value)override;
		void threshold_range(Image &a, Image &b, uint8_t threshold, uint8_t false_value, uint8_t true_value, uint8_t upper)override;
		void optical_flow(Image &a, Image &b, IKeyPointArray *old_points, IKeyPointArray *new_points_estimates,
		                  IKeyPointArray *new_points, Termination termination, float epsilon, uint32_t num_iterations,
		                  bool use_initial_estimate, size_t window_dimension, size_t num_levels)override;
		void integral(Image &a, Image &b)override;
		void pixelwise_mul(Image &a , Image &b, Image &c, float scale, uint8_t overflow_policy, uint8_t round_policy)override;
		void histogram(Image &a, Distribution &b)override;
		void magnitude(Image &a, Image &b, Image &c) override;
		void phase(Image &a, Image &b, Image &c) override;
		void equalize_histogram(Image &a, Image &b)override;
		void tablelookup( Image &a , Lut &b, Image &c ) override;
		void channel_extract(Image &a, Channel channel, Image &c) override;
		void channel_combine(Image &plane0, Image &plane1, Image &plane2, Image &plane3, Image &out) override;
		void derivative( Image &a, Image *c, Image *d) override;
		void remap(Image &a, RemapTable &table, Image &d, InterpolationMode policy)override;
		void minmaxlocation( Image &a, Scalar<int32_t> &min, Scalar<int32_t> &max,
		                     ICoord2DArray *min_loc = nullptr, ICoord2DArray *max_loc = nullptr,
		                     Scalar<uint32_t> *min_count = nullptr, Scalar<uint32_t> *max_count = nullptr) override;
		void activation_layer(Tensor &in, Tensor &out, ActivationLayerInfo act_info) override;
		void normalization_layer(Tensor &a, Tensor &b, NormalizationLayerInfo norm_info) override;
		void batchnormalization_layer(Tensor &a, Tensor &b, Tensor &mean, Tensor &var, Tensor &beta, Tensor &gamma, float epsilon) override;
		void pooling_layer(Tensor &in, Tensor &out, PoolingLayerInfo pool_info) override;
		void softmax_layer(Tensor &in, Tensor &out) override;
		void tensor_ops(Tensor &in1, Tensor &in2, Tensor &out, TensorOp op) override;
		void convolution_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info, Tensor *rw = nullptr) override;

		void fully_connected_layer(Tensor &input, Tensor &weight, Tensor *bia, Tensor &output, bool are_weights_reshaped = false)override;
		void convolution_separable3x3( Image &a, Image &c, Convolution &matrix);
		void convolution_separable5x5( Image &a, Image &c, Convolution &matrix);
		void non_linear_filter(Image &a, Image &b, NonLinearFilterFunction function, MatrixPattern pattern, uint mask_size, uint8_t* mask = nullptr) override;
		void transpose(Image &a, Image &b) override;
		void gemm_fp32(Tensor &a, Tensor &b, Tensor &c, Tensor &d, float alpha, float beta)override;
		void gemm_fp16(Image &a, Image &b, Image &c, Image &d, float alpha, float beta) override;
		void gemm_lowp(Image &a, Image &b, Image &d, int32_t a_offset, int32_t b_offset, int32_t c_offset, int32_t c_mult_int, int32_t shift) override;
		void concatenate_depth(std::vector<std::unique_ptr<Tensor>> &a, Tensor &b) override;
		void hog_descriptor(Image &a, HOG &b, Tensor &c)override;
		void locally_connected_layer(Tensor &in, Tensor &w, Tensor *b, Tensor &out, PadStrideInfo conv_info)override;
		void hog_detector(Tensor &a, HOG &b, IKeyPointArray &keypoint_detection_windows)override;
		void hog_multi_detection(Image &a, HOG &b, IKeyPointArray &keypoint_detection_windows)override;
		void transpose_1xW(Image &in, Image &out)override;
	private:
		void convolution_separable3x3( Image &a, Image &c, const int16_t convolution_matrix_x[3], const int16_t convolution_matrix_y[3]);
		void convolution_separable5x5( Image &a, Image &c, const int16_t convolution_matrix_x[5], const int16_t convolution_matrix_y[5]);
		void fill_image_borders( CLImage &a, uint num_pixels = 1);
		void interleave_4x4(Image &in, Image &out)override;
};
#else /* USE_OPENCL */
using CLTarget = ITarget;
#endif /* USE_OPENCL */
}
#endif /* __TARGET_H__ */
