#ifndef __CV_TENSOR_H__
#define __CV_TENSOR_H__

#include "mali_cv/Context.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/Types.h"

#ifdef USE_NEON
#include "arm_compute/runtime/Tensor.h"
#endif

#ifdef USE_OPENCL
#include "arm_compute/runtime/CL/CLTensor.h"
#endif

#ifdef USE_EIGEN
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-overflow"
#pragma GCC diagnostic ignored "-Weffc++"
#pragma GCC diagnostic ignored "-Wctor-dtor-privacy"
#include "unsupported/Eigen/CXX11/Tensor"
#include "unsupported/Eigen/CXX11/NeuralNetworks"
#pragma GCC diagnostic pop
#endif

namespace mali_cv
{
class ITensor
{
	public:
		ITensor();
		virtual ~ITensor() = default;
		void create(Context &ctx, TensorShape shape, DataType data_type, size_t num_channels = 1);
		virtual void load(uint8_t *in) {}
		virtual void store(uint8_t *out) {}
		virtual int size() { return 0; }
	protected:
		virtual void internal_create(Context &ctx, TensorShape shape, DataType data_type, size_t num_channels = 1)
		{
		}
	public:
		TensorShape _shape;
		DataType _dt;
		size_t _num_channels;
};

#ifdef USE_OPENCL
class CLTensor: public ITensor
{
	public:
		CLTensor();
		using ITensor::create;
		void load(uint8_t *in) override;
		void store(uint8_t *out) override;
		int size() override;
	protected:
		void internal_create(Context &ctx, TensorShape shape, DataType data_type, size_t num_channels = 1) override;
	public:
		arm_compute::CLTensor tensor;
};
#else /* USE_OPENCL */
using CLTensor = ITensor;
#endif /* USE_OPENCL */

#ifdef USE_EIGEN
class EigenTensor: public ITensor
{
	public:
		EigenTensor();
		using ITensor::create;
		void load(uint8_t *in) override;
		void store(uint8_t *out) override;
		int size() override;
	protected:
		void internal_create(Context &ctx, TensorShape shape, DataType data_type, size_t num_channels = 1) override;
	public:
		Eigen::Tensor<float, MAX_DIMS, Eigen::RowMajor> tensor;
};
#else /* USE_EIGEN */
using EigenTensor = ITensor;
#endif /* USE_EIGEN */

#ifdef USE_NEON
class NETensor: public ITensor
{
	public:
		NETensor();
		using ITensor::create;
		void load(uint8_t *in) override;
		void store(uint8_t *out) override;
		int size() override;
	protected:
		void internal_create(Context &ctx, TensorShape shape, DataType data_type, size_t num_channels = 1) override;
	public:
		arm_compute::Tensor tensor;
};
#else /* USE_NEON */
using NETensor = ITensor;
#endif /* USE_NEON */

using VXTensor = EigenTensor;
using CVTensor = EigenTensor;

class Tensor final
{
public:
	Tensor();
	Tensor(Context &ctx, TensorShape shape, DataType dt, size_t num_channels = 1);
	void init(Context &ctx, TensorShape shape, DataType dt, size_t num_channels = 1);
	void random_fill(float min, float max);
private:
	TensorShape _shape;
	DataType _dt;
	size_t   _num_channels;
public:
	CVTensor cv;
	VXTensor vx;
	CLTensor cl;
	NETensor ne;
};
}
#endif /* __CV_TENSOR_H__ */
