#include <cstring>
#include <cmath>
#include <sstream>

namespace
{
	const int16_t sobel3x3_x[3][3] =
	{
		{ 1,  -31,  -1 },
		{ 2,  31,  -2 },
		{ 1,  10,  -1 }
	};

	const int16_t laplacian3x3[3][3] =
	{
		{  0, -1,  0 },
		{ -1,  4, -1 },
		{  0, -1,  0 }
	};

}

namespace mali_cv
{

class CustomConvolutionScale: public ITest
{
	public:
		const std::string name() const override
		{
			std::ostringstream s;
			s << "CustomConvolutionScale" << static_cast<unsigned int>(matrix_size) << "x" << static_cast<unsigned int>(matrix_size);
			return s.str();
		}
		CustomConvolutionScale( Environment &env, RawImage &in, const short matrix[3][3], uint32_t in_scale) : ITest(env), a(env.ctx, in, CopyData::YES), c(env.ctx, in, CopyData::NO),
			convolution(env.ctx, 3, 3, Format::LUMINANCE16),matrix_size(3)
		{
			RawImage m( 3, 3, Format::LUMINANCE16);
			memcpy(m.data(), (void *)matrix, m.num_pixels() * m.pixel_size());
			convolution.load( m, in_scale );
		}
		CustomConvolutionScale( Environment &env, RawImage &in, const short matrix[5][5], uint32_t in_scale)
			: ITest(env), a(env.ctx, in, CopyData::YES), c(env.ctx, in, CopyData::NO), convolution(env.ctx, 5, 5, Format::LUMINANCE16), matrix_size(5)
		{
			RawImage m( 5, 5, Format::LUMINANCE16);
			memcpy(m.data(), (void *)matrix, m.num_pixels() * m.pixel_size());
			convolution.load( m, in_scale );
		}
		CustomConvolutionScale( Environment &env, RawImage &in, const short matrix[7][7], uint32_t in_scale):ITest(env), a(env.ctx, in, CopyData::YES), c(env.ctx, in, CopyData::NO),
		convolution(env.ctx, 7, 7, Format::LUMINANCE16),matrix_size(7)
		{
			RawImage m( 7, 7, Format::LUMINANCE16);
			memcpy(m.data(), (void *)matrix, m.num_pixels() * m.pixel_size());
			convolution.load( m, in_scale );
		}
		CustomConvolutionScale( Environment &env, RawImage &in, const short matrix[9][9], uint32_t in_scale):ITest(env), a(env.ctx, in, CopyData::YES), c(env.ctx, in, CopyData::NO),
		convolution(env.ctx, 9, 9, Format::LUMINANCE16),matrix_size(9)
		{
			RawImage m( 9, 9, Format::LUMINANCE16);
			memcpy(m.data(), (void *)matrix, m.num_pixels() * m.pixel_size());
			convolution.load( m, in_scale );
		}

		virtual void run(ITarget *target)override
		{
			target->convolution(a, c, convolution);
		}
		virtual void validate_outputs() override
		{
			Validate::single_image_output( c, env, env.border_to_exclude_if_undefined( matrix_size / 2) );
		}
	protected:
		Image a, c;
		Convolution convolution;
		const uint8_t matrix_size;
};


class CustomConvolution: public CustomConvolutionScale
{
	public:
		const std::string name() const override
		{
			std::ostringstream s;
			s << "CustomConvolution" << static_cast<unsigned int>(matrix_size) << "x" << static_cast<unsigned int>(matrix_size);
			return s.str();
		}
		CustomConvolution( Environment &env, RawImage &in, const short matrix[3][3]) : CustomConvolutionScale(env,in,matrix,0) {}
		CustomConvolution( Environment &env, RawImage &in, const short matrix[5][5]) : CustomConvolutionScale(env,in,matrix,0) {}
		CustomConvolution( Environment &env, RawImage &in, const short matrix[7][7]) : CustomConvolutionScale(env,in,matrix,0) {}
		CustomConvolution( Environment &env, RawImage &in, const short matrix[9][9]) : CustomConvolutionScale(env,in,matrix,0) {}
};

class CustomConvolutionRandom: public ITest
{
	public:
		const std::string name() const override
		{
			return _name;
		}
		CustomConvolutionRandom( Environment &env, RawImage &in, uint32_t width, uint32_t height, uint32_t in_scale, int16_t min_val = 0, int16_t max_val = 255)
		: ITest(env), _a(env.ctx, in, CopyData::YES), _c(env.ctx, in.width(), in.height(), Format::LUMINANCE16),
		_convolution(env.ctx, width, height, Format::LUMINANCE16), _matrix_size(0), _name()
		{
			RawImage m(width, height, Format::LUMINANCE16);
			m.random_init(min_val, max_val);
			_convolution.load(m, in_scale);

			// Create test name
			std::stringstream tmp_name;
			tmp_name << "CustomConvolutionRandom" << width << "x" << height;
			_name = tmp_name.str();

			// Calculate matrix size TODO: Fix to support different borders on top/bottom and right/left
			_matrix_size = std::max(width, height);
		}

		virtual void run(ITarget *target)override
		{
			target->convolution(_a, _c, _convolution);
		}
		virtual void validate_outputs() override
		{
			Validate::single_image_output(_c, env, env.border_to_exclude_if_undefined(_matrix_size / 2));
		}
	protected:
		Image _a, _c;
		Convolution _convolution;
		uint8_t _matrix_size;
		std::string _name;
};


class CustomConvolution3x3Sobelx: public CustomConvolution
{
	public:
		const std::string name() const override
		{
			return "CustomConvolution3x3_Sobelx";
		}
		CustomConvolution3x3Sobelx( Environment &env, RawImage &in) : CustomConvolution(env,in,sobel3x3_x) {}
};

class CustomConvolution3x3Laplacian: public CustomConvolution
{
	public:
		const std::string name() const override
		{
			return "CustomConvolution3x3_Laplacian";
		}
		CustomConvolution3x3Laplacian( Environment &env, RawImage &in) : CustomConvolution(env,in,laplacian3x3) {}
};

class CustomConvolutionShortScale: public ITest
{
	public:
		const std::string name() const override
		{
			return "CustomConvolutionScale3x3_S16";
		}
		CustomConvolutionShortScale( Environment &env, RawImage &in, const short matrix[3][3], uint32_t in_scale) :
			ITest(env), a(env.ctx, in, CopyData::YES), c(env.ctx, in.width(), in.height(), Format::LUMINANCE16),
			convolution(env.ctx, 3, 3, Format::LUMINANCE16), matrix_size(3)
		{
			RawImage m( 3, 3, Format::LUMINANCE16);
			memcpy(m.data(), (void *)matrix, m.num_pixels() * m.pixel_size());
			convolution.load( m, in_scale );
		}
		CustomConvolutionShortScale( Environment &env, RawImage &in, const short matrix[5][5], uint32_t in_scale) :
			ITest(env), a(env.ctx, in, CopyData::YES), c(env.ctx, in.width(), in.height(), Format::LUMINANCE16),
			convolution(env.ctx, 5, 5, Format::LUMINANCE16),matrix_size(5)
		{
			RawImage m( 5, 5, Format::LUMINANCE16);
			memcpy(m.data(), (void *)matrix, m.num_pixels() * m.pixel_size());
			convolution.load( m , in_scale );
		}
		void run(ITarget *target)override
		{
			target->convolution(a, c, convolution);
		}
		void validate_outputs() override
		{
			Validate::single_image_output( c, env, env.border_to_exclude_if_undefined( matrix_size / 2) );
		}
	protected:
		Image a, c;
		Convolution convolution;
		uint matrix_size;
};


class CustomConvolutionShort: public CustomConvolutionShortScale
{
	public:
		const std::string name() const override
		{
			std::ostringstream s;
			s << "CustomConvolution" << matrix_size << "x" << matrix_size << "_S16";
			return s.str();
		}
		CustomConvolutionShort( Environment &env, RawImage &in, const short matrix[3][3]) : CustomConvolutionShortScale(env,in,matrix,0) {}
		CustomConvolutionShort( Environment &env, RawImage &in, const short matrix[5][5]) : CustomConvolutionShortScale(env,in,matrix,0) {}
};

class CustomConvolutionShort3x3Sobelx: public CustomConvolutionShort
{
	public:
		const std::string name() const override
		{
			return "CustomConvolution3x3_Sobelx_S16";
		}
		CustomConvolutionShort3x3Sobelx( Environment &env, RawImage &in) : CustomConvolutionShort(env,in,sobel3x3_x) {}
};

class CustomConvolutionShort3x3Laplacian: public CustomConvolutionShort
{
	public:
		const std::string name() const override
		{
			return "CustomConvolution3x3_Laplacian_S16";
		}
		CustomConvolutionShort3x3Laplacian( Environment &env, RawImage &in) : CustomConvolutionShort(env,in,laplacian3x3) {}
};


class CustomConvolutionSeparable3x3: public CustomConvolution
{
	public:
		using CustomConvolution::CustomConvolution;
		const std::string name() const override
		{
			return "Custom Convolution Separable 3x3";
		}
};

class CustomConvolutionSeparable5x5: public CustomConvolution
{
	public:
		using CustomConvolution::CustomConvolution;
		const std::string name() const override
		{
			return "Custom Convolution Separable 5x5";
		}
};

class CustomConvolutionSeparable7x7: public CustomConvolution
{
	public:
		using CustomConvolution::CustomConvolution;
		const std::string name() const override
		{
			return "CustomConvolutionSeparable7x7";
		}
};

class CustomConvolutionSeparable9x9: public CustomConvolution
{
	public:
		using CustomConvolution::CustomConvolution;
		const std::string name() const override
		{
			return "CustomConvolutionSeparable9x9";
		}
};

class CustomConvolution5x5: public CustomConvolution
{
	public:
		using CustomConvolution::CustomConvolution;
		const std::string name() const override
		{
			return "CustomConvolution5x5";
		}
};

class CustomConvolution7x7: public CustomConvolution
{
	public:
		using CustomConvolution::CustomConvolution;
		const std::string name() const override
		{
			return "CustomConvolution7x7";
		}
};

class CustomConvolution5x5Short: public CustomConvolutionShort
{
	public:
		using CustomConvolutionShort::CustomConvolutionShort;
		const std::string name() const override
		{
			return "CustomConvolution5x5_S16";
		}
};

class CustomConvolution7x7Short: public CustomConvolutionShort
{
	public:
		using CustomConvolutionShort::CustomConvolutionShort;
		const std::string name() const override
		{
			return "CustomConvolution7x7_S16";
		}
};


} /* mali_cv */
