#include "mali_cv/Context.h"
#include "mali_cv/Helper.h"
#include "mali_cv/RawImage.h"
#include "mali_cv/Types.h"

#include <vector>
#include <sstream>
#include <memory>

namespace
{
using namespace mali_cv;

template<typename T>
void compute_reference_interleave4x4(RawImage &in, RawImage &ref)
{
	const T* mtx_in   = reinterpret_cast<const T*>(in.data());
	T*       mtx_ref  = reinterpret_cast<T*>(ref.data());
	const int32_t in_rows    = in.height();
	const int32_t in_cols    = in.width();
	const int32_t in_stride  = in_cols;
	const int32_t out_stride = ref.width();

	int32_t y = 0;
	for(; y <= (in_rows - 4); y+=4)
	{
		const T* in_ptr = &mtx_in[y * in_stride];

		for(int32_t x = 0; x < in_cols; x++)
		{
			const T tmp[4] = { in_ptr[x + 0 * in_stride],
			                   in_ptr[x + 1 * in_stride],
			                   in_ptr[x + 2 * in_stride],
			                   in_ptr[x + 3 * in_stride]
			                 };

			T* dst = &mtx_ref[x * 4 + (y / 4) * out_stride];
			memcpy(dst, tmp, sizeof(T) * 4 );
		}
	}

	// Leftover along the Y direction
	const int32_t leftover_y = in_rows - y;

	if(leftover_y != 0)
	{
		const T* in_ptr = &mtx_in[y * in_stride];

		for(int32_t x = 0; x < in_cols; x++)
		{
			T tmp[4] = { 0 };
			for(int32_t k = 0; k < leftover_y; k++)
			{
				tmp[k] = in_ptr[k * in_stride + x];
			}
			T* dst = &mtx_ref[x * 4 + (y / 4) * out_stride];
			memcpy(dst, tmp, sizeof(T) * 4 );
		}
	}
}

template<typename T>
void compute_reference_transpose1xW(RawImage &in, RawImage &out)
{
	const int out_width = static_cast<int>(out.width());
	const int in_width  = static_cast<int>(in.width());
	const int in_height = static_cast<int>(in.height());
	const int W         = 16 / in.element_size();

	const T* in_base_addr  = reinterpret_cast<const T*>(in.data());
	T*       out_base_addr = reinterpret_cast<T*>(out.data());

	int x = 0;
	// If the input width is not multiple of W, we do not compute the last chunk as it will contain garbage
	for(; x <= (in_width - W); x+=W)
	{
		for(int y = 0; y < in_height; y++)
		{
			const T* in_addr  = (in_base_addr + x + y * in_width);
			T*       out_addr = (out_base_addr + y * W + (x / W) * out_width);

			// 128 = 16 * element_size
			memcpy(out_addr, in_addr, 128);
		}
	}
}
}

namespace mali_cv
{
class GEMMInterleave4x4: public ITest
{
	public:
		const std::string name() const override
		{
			return test_name;
		}

		GEMMInterleave4x4(Environment &env, unsigned int cols, unsigned int rows, float rnd_low, float rnd_high, Format fmt) :
			ITest(env),
			in(env.ctx, cols, rows, fmt ),
			out(env.ctx, cols*4, std::ceil(rows/4.f), fmt),
			ref(cols*4, std::ceil(rows/4.f), fmt), test_name(), border_size(rows % 4? 4 : 0)
		{
			std::stringstream tmp_name;
			tmp_name << "GEMMInterleave4x4_M_N_";
			tmp_name << rows;
			tmp_name << "_" << cols;
			test_name = tmp_name.str();

			RawImage tmp_a(cols, rows, fmt);
			tmp_a.random_init(rnd_low, rnd_high);
			in.load(tmp_a);
			switch(tmp_a.element_size())
			{
				case 4:
					compute_reference_interleave4x4<uint32_t>(tmp_a, ref);
					break;
				case 2:
					compute_reference_interleave4x4<uint16_t>(tmp_a, ref);
					break;
				case 1:
					compute_reference_interleave4x4<unsigned char>(tmp_a, ref);
					break;
				default:
					ARM_COMPUTE_ERROR("Element size not supported");
					break;
			}
		}
		void run(ITarget *target) override
		{
			target->interleave_4x4(in,out);
		}
		void validate_outputs() override
		{
			Validate::single_image_output(out, ref, 0.0f, env, border_size);
		}
	protected:
		Image       in, out;
		std::string test_name;
		RawImage    ref;
		int32_t     border_size;
};

class GEMMTranspose1xW : public ITest
{
	public:
		const std::string name() const override
		{
			return test_name;
		}
		GEMMTranspose1xW(Environment &env, unsigned int rows, unsigned int cols, float rnd_low, float rnd_high, Format fmt) :
			ITest(env),
			in(env.ctx, cols, rows, fmt),
			out(env.ctx, rows * (16 / RawImageBase::get_element_size(fmt)), std::ceil(static_cast<float>(cols) / (16 / RawImageBase::get_element_size(fmt))), fmt),
			ref(rows * (16 / RawImageBase::get_element_size(fmt)), std::ceil(static_cast<float>(cols) / (16 / RawImageBase::get_element_size(fmt))), fmt), test_name(),
			border_size(cols % (16 / RawImageBase::get_element_size(fmt)) ? 1 : 0)
		{
			RawImage tmp_a(cols, rows, fmt);
			tmp_a.random_init(rnd_low, rnd_high);
			in.load(tmp_a);

			std::stringstream tmp_name;
			tmp_name << "GEMMTranspose1xW_M_N_";
			tmp_name << 16 / tmp_a.element_size();
			tmp_name << "_" << rows;
			tmp_name << "_" << cols;
			test_name = tmp_name.str();

			switch(tmp_a.element_size())
			{
				case 1:
					compute_reference_transpose1xW<unsigned char>(tmp_a, ref );
					break;
				case 2:
					compute_reference_transpose1xW<uint16_t>(tmp_a, ref );
					break;
				case 4:
					compute_reference_transpose1xW<uint32_t>(tmp_a, ref );
					break;
				default:
					ARM_COMPUTE_ERROR("Not implemented");
					break;
			}
		}
		void run(ITarget *target) override
		{
			target->transpose_1xW(in,out);
		}
		void validate_outputs() override
		{
			/*
				If the width of the input matrix is not multiple of 4, we cannot check if output the values of the last row
				are within the range [lv,hv] because the NEGEMMTranspose1xWKernel will output random values as the padding bytes are not initialized.

				Therefore here we setup the ValidRegion to skip the last row when checking the results.
			*/
			const arm_compute::ValidRegion vr(arm_compute::Coordinates(0,0), arm_compute::TensorShape(ref.width() , (border_size?ref.height()-1:0) ) );
			Validate::single_image_output(env, out, ref, vr, 0.f);
		}
	protected:
		Image        in, out;
		std::string  test_name;
		RawImage     ref;
		const size_t border_size;
};
}
