#include "mali_cv/Array.h"

#include <random>
#include <vector>

namespace
{
mali_cv::TensorInfo calculate_hog_descriptor_tensor_info(mali_cv::HOGInfo hog_info, size_t width, size_t height)
{
	return TensorInfo(hog_info, width, height);
}

inline std::vector<float> random_real(size_t size, int min, int max)
{
	std::vector<float> v(size);

	uint seed = random_seed();
	std::mt19937 rng( seed );
	std::uniform_real_distribution<float> gen( min, max );
	for(size_t i = 0; i < size; ++i)
	{
		v.at(i) = gen(rng);
	}
	return v;
}
}

namespace mali_cv
{
class HOGDescriptor: public ITest
{
	public:
		const std::string name() const override { return "HOGDescriptor";}

		HOGDescriptor( Environment &env, RawImage &in, const Size2D &cell_size, const Size2D &block_size, const Size2D &block_stride, size_t num_bins, mali_cv::HOGNormType normalization_type, float l2_hyst_threshold = 0.2f ) :
		    ITest(env),
		    a(env.ctx, in, CopyData::YES),
		    hog_info(cell_size, block_size, block_size, block_stride, num_bins, normalization_type, l2_hyst_threshold),
		    tensor_info_hog_descriptor(calculate_hog_descriptor_tensor_info(hog_info, in.width(), in.height())),
		    hog(hog_info),
		    b(env.ctx, tensor_info_hog_descriptor.tensor_shape(), DataType::F32, tensor_info_hog_descriptor.num_channels())
		{
		}
		virtual void run(ITarget *target)override
		{
			target->hog_descriptor(a, hog, b);
		}
		virtual void validate_outputs() override
		{
			Validate::single_tensor_output( b, env, 0, 0.1f /* Threshold */ );
		}
	protected:
		Image       a;
		HOGInfo     hog_info;
		TensorInfo  tensor_info_hog_descriptor;
		HOG         hog;
		Tensor      b;
};

class HOGDetector: public ITest
{
	public:
		const std::string name() const override { return "HOGDetector";}

		HOGDetector( Environment &env, unsigned int width, unsigned height, const Size2D &cell_size, const Size2D &block_size, const Size2D &detection_window_size, const Size2D &block_stride, size_t num_bins) :
		    ITest(env),
		    hog_info(cell_size, block_size, detection_window_size, block_stride, num_bins),
		    tensor_info_hog_space(calculate_hog_descriptor_tensor_info(hog_info, width, height)),
		    hog(hog_info),
		    hog_space(env.ctx, tensor_info_hog_space.tensor_shape(), DataType::F32, tensor_info_hog_space.num_channels()),
		    keypoint_objects(env.ctx, MAX_NUM_VALUES),
		    w(width), h(height)
		{
			// Initialize with random values the HOG space
			hog_space.random_fill(-1.0f, +1.0f);

			// Init with random values the HOG descriptor. The values are "unbalanced" to reduce the number of detected objects
			std::vector<float> descriptor = random_real(hog.descriptor_size, -1.0f, 0.3f);
			hog.init_descriptor(descriptor);
		}
		virtual void run(ITarget *target)override
		{
			target->hog_detector(hog_space, hog, keypoint_objects);
		}
		virtual void validate_outputs() override
		{
			Validate::corners_output(keypoint_objects, env, 0, 0.1f, std::make_pair(w, h), 0);
		}
	protected:
		HOGInfo        hog_info;
		TensorInfo     tensor_info_hog_space;
		HOG            hog;
		Tensor         hog_space;
		IKeyPointArray keypoint_objects;
		size_t         w;
		size_t         h;
};

class HOGMultiDetection: public ITest
{
	public:
		const std::string name() const override { return "HOGMultiDetection";}

		HOGMultiDetection( Environment &env, RawImage &in, const Size2D &cell_size, const Size2D &block_size, const Size2D &detection_window_size, const Size2D &block_stride, size_t num_bins, mali_cv::HOGNormType normalization_type, float l2_hyst_threshold = 0.2f ) :
		    ITest(env),
		    a(env.ctx, in, CopyData::YES),
		    hog_info(cell_size, block_size, detection_window_size, block_stride, num_bins, normalization_type, l2_hyst_threshold),
		    hog(hog_info),
		    keypoint_objects(env.ctx, MAX_NUM_VALUES),
		    w(in.width()), h(in.height())
		{
			// Init with random values the HOG descriptor
			std::vector<float> descriptor = random_real(hog.descriptor_size, -1.0f, 1.0f);
			hog.init_descriptor(descriptor);
		}
		virtual void run(ITarget *target)override
		{
			target->hog_multi_detection(a, hog, keypoint_objects);
		}
		virtual void validate_outputs() override
		{
			// NOTE: The threshold used for the strength (0.3f) is different to the other one used for HOGDetector test (0.1f)
			// because the output of HOG descriptor kernel could be different due to different approximation of the result
			Validate::corners_output(keypoint_objects, env, 0, 0.3f, std::make_pair(w, h), 0);
		}
	protected:
		Image          a;
		HOGInfo        hog_info;
		HOG            hog;
		IKeyPointArray keypoint_objects;
		size_t         w;
		size_t         h;
};
}
