#include "mali_cv/Context.h"
#include "mali_cv/Types.h"

#include "arm_compute/core/Utils.h"

#include <sstream>
#include <string>
#include <iostream>

namespace
{
	mali_cv::TensorShape lc_calculate_output_shape(mali_cv::TensorShape in_shape, mali_cv::TensorShape kernel_shape,
	                                            uint conv_stride_x, uint conv_stride_y,
	                                            uint conv_pad_x, uint conv_pad_y, mali_cv::DimensionRoundingType round)
	{
		mali_cv::TensorShape out_shape(in_shape);
		// Calculate output dimensions
		std::pair<unsigned int, unsigned int> scaled_dims = arm_compute::scaled_dimensions(in_shape.x(),
		                                                                              in_shape.y(),
		                                                                              kernel_shape.x(),
		                                                                              kernel_shape.y(),
		                                                                              PadStrideInfo(conv_stride_x, conv_stride_y, conv_pad_x, conv_pad_y, round));
		out_shape.set(0, scaled_dims.first);
		out_shape.set(1, scaled_dims.second);
		out_shape.set(2, kernel_shape[3]);
		return out_shape;
	}
	mali_cv::TensorShape lc_calculate_bias_shape(mali_cv::TensorShape kernel_shape, mali_cv::TensorShape output_shape)
	{
		mali_cv::TensorShape bias_shape;
		bias_shape.set(0, output_shape.z());
		bias_shape.set(1, kernel_shape[4]);
		return bias_shape;
	}
}

namespace mali_cv
{
class LocallyConnectedLayer : public ITest
{
	public:
		const std::string name() const override
		{
			return _name;
		}
		LocallyConnectedLayer(Environment &env, TensorShape shape, TensorShape kernel_shape, DataType dt, uint conv_stride_x, uint conv_stride_y,
		                 uint conv_pad_x, uint conv_pad_y, DimensionRoundingType round)
		: ITest(env),
		  _output_shape(lc_calculate_output_shape(shape, kernel_shape, conv_stride_x, conv_stride_y, conv_pad_x, conv_pad_y, round)),
		  _input(env.ctx, shape, dt),
		  _weights(env.ctx, kernel_shape, dt),
		  _bias(env.ctx, lc_calculate_bias_shape(kernel_shape, _output_shape), dt),
		  _output(env.ctx, _output_shape, dt),
		  _conv_stride_x(conv_stride_x), _conv_stride_y(conv_stride_y),
		  _conv_pad_x(conv_pad_x), _conv_pad_y(conv_pad_y), _round(round),
		  _name()
		{
			// Initialize inputs randomly
			_input.random_fill(-2, 2);
			_weights.random_fill(-2, 2);
			_bias.random_fill(-2, 2);
			_output.random_fill(0, 0);

			// Setup test name
			std::stringstream tmp_name;
			tmp_name << "LocallyConnectedLayer";
			tmp_name << "_stridex_" << conv_stride_x;
			tmp_name << "_stridey_" << conv_stride_y;
			tmp_name << "_padx_" << conv_pad_x;
			tmp_name << "_pady_" << conv_pad_y;
			_name = tmp_name.str();
		}

		virtual void run(ITarget *target) override
		{
			target->locally_connected_layer(_input, _weights, &_bias, _output, PadStrideInfo(_conv_stride_x, _conv_stride_y, _conv_pad_x, _conv_pad_y, _round));
		}
		virtual void validate_outputs() override
		{
			Validate::single_tensor_output( _output, env, 0, 0.01f /* Threshold */ );
		}
	protected:
		TensorShape _output_shape;
		Tensor _input, _weights, _bias, _output;
		uint32_t _conv_stride_x, _conv_stride_y, _conv_pad_x, _conv_pad_y;
		DimensionRoundingType _round;
		std::string _name;
};
}
