#include "mali_cv/Context.h"
#include "mali_cv/Helper.h"
#include "mali_cv/Types.h"
#include "arm_compute/core/Utils.h"

#include <array>
#include <cstdint>
#include <sstream>
#include <string>
#include <random>

namespace mali_cv
{
	class NonLinearFilter: public ITest
	{
		public:
			const std::string name() const override { return _name; }

			NonLinearFilter(Environment &env, RawImage &a, NonLinearFilterFunction function, unsigned int mask_size, MatrixPattern pattern, uint8_t* mask = nullptr)
			: ITest(env), _a(env.ctx, a, CopyData::YES), _c(env.ctx, a.width(), a.height(), Format::U8),
			  _func(function), _pattern(pattern), _mask(mask), _mask_size(mask_size), _name()
			{
				std::stringstream tmp_name;
				tmp_name << "NonLinearFilter_"<< arm_compute::string_from_non_linear_filter_function(function)
					<< "_" << arm_compute::string_from_matrix_pattern(pattern) << "_" << _mask_size << "x" << _mask_size;
				_name = tmp_name.str();
			}
			NonLinearFilter(const NonLinearFilter&) = delete;
			const NonLinearFilter& operator=(const NonLinearFilter&)=delete;
			virtual void run(ITarget *target)override
			{
				target->non_linear_filter(_a, _c, _func, _pattern, _mask_size, _mask);
			}
			virtual void validate_outputs() override
			{
				Validate::single_image_output( _c, env, env.border_to_exclude_if_undefined(_mask_size / 2) );
			}
		protected:
			Image _a, _c;
			NonLinearFilterFunction _func;
			MatrixPattern _pattern;
			uint8_t *_mask;
			unsigned int _mask_size;
			std::string _name;
	};
	class NonLinearFilterBox : public NonLinearFilter
	{
		public:
			NonLinearFilterBox(Environment &env, RawImage &a, NonLinearFilterFunction function, unsigned int mask_size)
			: NonLinearFilter(env, a, function, mask_size, MatrixPattern::BOX, nullptr)
			{}
	};
	class NonLinearFilterCross : public NonLinearFilter
	{
		public:
			NonLinearFilterCross(Environment &env, RawImage &a, NonLinearFilterFunction function, unsigned int mask_size)
			: NonLinearFilter(env, a, function, mask_size, MatrixPattern::CROSS, nullptr)
			{}
	};
	class NonLinearFilterDisk : public NonLinearFilter
	{
		public:
			NonLinearFilterDisk(Environment &env, RawImage &a, NonLinearFilterFunction function, unsigned int mask_size)
			: NonLinearFilter(env, a, function, mask_size, MatrixPattern::DISK, nullptr)
			{}
	};
	class NonLinearFilterOther : public NonLinearFilter
	{
		public:
			NonLinearFilterOther(Environment &env, RawImage &a, NonLinearFilterFunction function, unsigned int mask_size)
			: NonLinearFilter(env, a, function, mask_size, MatrixPattern::OTHER, nullptr), _mask_rand{ {0} }
			{
				std::mt19937 gen(random_seed());
				std::uniform_int_distribution<> dis(0, 1);

				// Fill mask with 0 or 255 at random
				std::generate(_mask_rand.begin(), _mask_rand.end(), [&]()
				{
				    return dis(gen) * 255;
				});

				// Ensure at least one element is 255
				std::uniform_int_distribution<> dis2(0, std::min(mask_size * mask_size - 1, 24U));
				_mask_rand[dis2(gen)] = 255;

				_mask = &_mask_rand[0];
			}
		protected:
			std::array<uint8_t, 25> _mask_rand;
	};
}
