namespace mali_cv
{
class GaussianPyramid: public ITest
{
	public:
		const std::string name() const override
		{
			return "GaussianPyramid";
		}
		GaussianPyramid( Environment &env, RawImage &in, int pyramid_levels, float scale):ITest(env), a(env.ctx, in, CopyData::YES), pyramid(env.ctx, pyramid_levels, scale, in.width(), in.height(), in.format())
		{
		}
		virtual void run(ITarget *target)override
		{
			target->gaussian_pyramid(a, pyramid);
		}
		virtual void validate_outputs() override
		{
			std::vector<ImagePtr> levels;
			pyramid.store( levels );

			for(uint i = 0; i < levels.size(); i++)
			{
				Validate::single_image_output( *levels[i], env, env.border_to_exclude_if_undefined( i ) );
			}
		}
	protected:
		Image a;
		Pyramid pyramid;
};

class LaplacianPyramid: public ITest
{
	public:
		const std::string name() const override
		{
			return "LaplacianPyramid";
		}
		LaplacianPyramid( Environment &env, RawImage &in, int pyramid_levels) :
			ITest(env),
			a(env.ctx, in, CopyData::YES),
			b(env.ctx, in.width()/ pow(2,pyramid_levels-1), in.height()/ pow(2,pyramid_levels-1), Format::LUMINANCE16),
			pyramid(env.ctx, pyramid_levels, 0.5f,  in.width(), in.height(), Format::S16)
		{
			ERROR_ON( in.format() != Format::LUMINANCE );
			ERROR_ON( 0 == pyramid_levels );
			ERROR_ON( 0 == ( in.width() / pow(2,pyramid_levels-1) ) );
			ERROR_ON( 0 == ( in.height()/ pow(2,pyramid_levels-1) ) );
		}
		virtual void run(ITarget *target)override
		{
			target->laplacian_pyramid(a,pyramid,b);
		}
		virtual void validate_outputs() override
		{
			std::vector<ImagePtr> levels;
			pyramid.store( levels );
			for(uint i = 0; i < levels.size(); i++)
			{
				Validate::single_image_output( *levels[i], env, env.border_to_exclude_if_undefined( i + 2 ) );
			}
			Validate::single_image_output( b, env, env.border_to_exclude_if_undefined( levels.size() - 1 + 2 ) );
		}
	protected:
		Image a,b;
		Pyramid pyramid;
};

class LaplacianReconstruct: public LaplacianPyramid
{
	public:
		const std::string name() const override
		{
			return "LaplacianReconstruct";
		}
		LaplacianReconstruct( Environment &env, RawImage &in, int levels) :
			LaplacianPyramid(env,in,levels),c(env.ctx,in.width(),in.height(),in.format()), levels(levels)
		{
		}
		void run(ITarget *target)override
		{
			target->laplacian_pyramid(a,pyramid,b,false);
			target->laplacian_reconstruct(pyramid,b,c);
		}
		void validate_outputs() override
		{
			// Validate Laplacian pyramid
			std::vector<ImagePtr> laplacian_images;
			pyramid.store( laplacian_images );

			for(unsigned int i = 0; i < laplacian_images.size(); i++)
			{
				Validate::single_image_output( *laplacian_images[i], env, env.border_to_exclude_if_undefined( i + 2 ) );
			}

			Validate::single_image_output( b, env, env.border_to_exclude_if_undefined( laplacian_images.size() - 1 + 2 ) );

			// Validate reconstructed image
			Validate::single_image_output( c, env, env.border_to_exclude_if_undefined( (levels - 1 + 2) * pow(2, levels - 1) ) );
		}
	private:
		Image c;
		size_t levels;
};
} /* mali_cv */
