#include "kernels.h"

using namespace arm_compute;

namespace
{
armcomputeMultiImage *to_MultiImage(vx_image img)
{
    if(img == nullptr)
        return nullptr;
    ARM_COMPUTE_ERROR_ON(img->memory.arm_image == nullptr);
    return static_cast<armcomputeMultiImage *>(img->memory.arm_image);
}

armcomputeImage *to_Image(vx_image img)
{
    if(img == nullptr)
        return nullptr;
    auto multi = to_MultiImage(img);
#if defined(USE_ARM_COMPUTE_OPENCL)
    return static_cast<armcomputeImage *>(multi->cl_plane(0));
#else
    return static_cast<armcomputeImage *>(multi->plane(0));
#endif
}

armcomputeLut *to_Lut(vx_array arr)
{
    if(arr == nullptr)
        return nullptr;
    ARM_COMPUTE_ERROR_ON(arr->memory.arm_array == nullptr || arr->memory.is_lut == 0);

    return static_cast<armcomputeLut *>(arr->memory.arm_array);
}

armcomputePyramid *to_Pyramid(vx_pyramid pyr)
{
    if(pyr == nullptr)
        return nullptr;
    ARM_COMPUTE_ERROR_ON(pyr->arm_pyramid == nullptr);
    return static_cast<armcomputePyramid *>(pyr->arm_pyramid);
}

#if defined(USE_ARM_COMPUTE_OPENCL)
void unmap(vx_image img)
{
    if(img != nullptr)
    {
        auto             multi = to_MultiImage(img);
        armcomputeImage *_img  = static_cast<armcomputeImage *>(multi->cl_plane(0));
        _img->unmap(arm_compute::CLScheduler::get().queue());
    }
}

void unmap(armcomputeMultiImage *multi_img)
{
    for(unsigned int i = 0; i < 3; ++i)
    {
        armcomputeImage *_img = static_cast<armcomputeImage *>(multi_img->cl_plane(i));
        if(_img != nullptr && _img->cl_buffer().get() != nullptr)
        {
            _img->unmap(arm_compute::CLScheduler::get().queue());
        }
    }
}

void unmap(vx_distribution dist)
{
    auto *_dist = static_cast<armcomputeDistribution1D *>(dist->memory.arm_distribution);
    ;
    _dist->unmap();
}

void unmap(armcomputeArray<KeyPoint> *img)
{
    if(img != nullptr && img->cl_buffer().get() != nullptr)
    {
        img->unmap(arm_compute::CLScheduler::get().queue());
    }
}

void unmap(armcomputeDistribution1D *dist)
{
    dist->unmap();
}

void unmap(armcomputeArray<Coordinates2D> *img)
{
    if(img != nullptr)
    {
        img->unmap();
    }
}

void unmap(armcomputeLut *lut)
{
    lut->unmap();
}

void unmap(std::unique_ptr<arm_compute::CLTensor> &img)
{
    img->unmap();
}

void unmap(arm_compute::CLTensor *img)
{
    img->unmap();
}

template <typename T, typename... Args>
void unmap(T &&arg, Args &&... args)
{
    unmap(arg);
    unmap(std::forward<Args>(args)...);
}
#else
#define unmap(...) ARM_COMPUTE_NOP(...)
#endif

#if defined(USE_ARM_COMPUTE_OPENCL)
void map(vx_image img)
{
    if(img != nullptr)
    {
        auto             multi = to_MultiImage(img);
        armcomputeImage *_img  = static_cast<armcomputeImage *>(multi->cl_plane(0));
        _img->map(arm_compute::CLScheduler::get().queue());
    }
}

void map(armcomputeMultiImage *multi_img)
{
    for(unsigned int i = 0; i < 3; ++i)
    {
        armcomputeImage *_img = static_cast<armcomputeImage *>(multi_img->cl_plane(i));
        if(_img != nullptr && _img->cl_buffer().get() != nullptr)
        {
            _img->map(arm_compute::CLScheduler::get().queue(), true);
        }
    }
}

void map(armcomputeArray<KeyPoint> *img)
{
    if(img != nullptr)
    {
        img->map(arm_compute::CLScheduler::get().queue(), true);
    }
}

void map(vx_distribution dist)
{
    auto *_dist = static_cast<armcomputeDistribution1D *>(dist->memory.arm_distribution);
    ;
    _dist->map();
}

void map(armcomputeDistribution1D *dist)
{
    dist->map();
}

void map(armcomputeArray<Coordinates2D> *img)
{
    if(img != nullptr)
    {
        img->map();
    }
}

void map(armcomputeLut *lut)
{
    lut->map();
}

void map(std::unique_ptr<arm_compute::CLTensor> &img)
{
    img->map();
}

void map(arm_compute::CLTensor *img)
{
    img->map();
}

template <typename T, typename... Args>
void map(T &&arg, Args &&... args)
{
    map(arg);
    map(std::forward<Args>(args)...);
}
#else
#define map(...) ARM_COMPUTE_NOP(...)
#endif

#if defined(USE_ARM_COMPUTE_OPENCL)
void unmapPyramidLevels(vx_pyramid pyr)
{
    vx_size n_levels = pyr->numLevels;
    for(unsigned int i = 0; i < n_levels; ++i)
    {
        ARM_COMPUTE_ERROR_ON(pyr->levels[i] == nullptr);
        auto img = static_cast<armcomputeImage *>((to_MultiImage(pyr->levels[i]))->cl_plane(0));
        img->unmap(arm_compute::CLScheduler::get().queue());
    }
}
#else
#define unmapPyramidLevels(pyr) ARM_COMPUTE_NOP(pyr)
#endif

#if defined(USE_ARM_COMPUTE_OPENCL)
void mapPyramidLevels(vx_pyramid pyr)
{
    vx_size n_levels = pyr->numLevels;
    for(unsigned int i = 0; i < n_levels; ++i)
    {
        ARM_COMPUTE_ERROR_ON(pyr->levels[i] == nullptr);
        auto img = static_cast<armcomputeImage *>((to_MultiImage(pyr->levels[i]))->cl_plane(0));
        img->map(arm_compute::CLScheduler::get().queue(), true);
    }
}
#else
#define mapPyramidLevels(pyr) ARM_COMPUTE_NOP(pyr)
#endif

armcomputeDistribution1D *to_Distribution1D(vx_distribution dist)
{
    if(dist == nullptr)
        return nullptr;
    ARM_COMPUTE_ERROR_ON(dist->memory.arm_distribution == nullptr);
    armcomputeDistribution1D *_dist = static_cast<armcomputeDistribution1D *>(dist->memory.arm_distribution);
    map(_dist);
    _dist->clear();
    unmap(_dist);
    return _dist;
}

template <class T>
armcomputeArray<T> *to_Array(vx_array arr)
{
    if(arr == nullptr)
        return nullptr;
    ARM_COMPUTE_ERROR_ON(arr->memory.arm_array == nullptr);
    return static_cast<armcomputeArray<T> *>(arr->memory.arm_array);
}

std::vector<std::unique_ptr<armcomputeImage>> to_Remap(vx_remap table)
{
    std::vector<std::unique_ptr<armcomputeImage>> RemapTable;
    RemapTable.clear();

    // Query table info
    vx_uint32 _src_w, _src_h, _dst_w, _dst_h;
    vxQueryRemap(table, VX_REMAP_SOURCE_WIDTH, &_src_w, sizeof(_src_w));
    vxQueryRemap(table, VX_REMAP_SOURCE_HEIGHT, &_src_h, sizeof(_src_h));
    vxQueryRemap(table, VX_REMAP_DESTINATION_WIDTH, &_dst_w, sizeof(_dst_w));
    vxQueryRemap(table, VX_REMAP_DESTINATION_HEIGHT, &_dst_h, sizeof(_dst_h));

    // Allocate maps to be used.
    auto map_x = std::unique_ptr<armcomputeImage>(new armcomputeImage());
    auto map_y = std::unique_ptr<armcomputeImage>(new armcomputeImage());
    map_x->allocator()->init(TensorInfo(_dst_w, _dst_h, Format::F32));
    map_y->allocator()->init(TensorInfo(_dst_w, _dst_h, Format::F32));

    map(map_x, map_y);

    // Copy table content to maps
    for(int32_t y = 0; y < _dst_h; ++y)
    {
        for(int32_t x = 0; x < _dst_w; ++x)
        {
            vx_float32 x0, y0;
            const auto map_x_ptr = reinterpret_cast<float *>(map_x->ptr_to_element(Coordinates(x, y)));
            const auto map_y_ptr = reinterpret_cast<float *>(map_y->ptr_to_element(Coordinates(x, y)));

            // Get Remap point
            vxGetRemapPoint(table, x, y, &x0, &y0);

            // Set to image
            *map_x_ptr = x0;
            *map_y_ptr = y0;
        }
    }

    unmap(map_x, map_y);

    // Return table
    RemapTable.push_back(std::move(map_x));
    RemapTable.push_back(std::move(map_y));

    return RemapTable;
}

armcomputeImage *get_image_from_pyramid(vx_pyramid pyr, const int level)
{
    armcomputeImage *img;
#if defined(USE_ARM_COMPUTE_OPENCL)
    img = static_cast<armcomputeImage *>((to_MultiImage(pyr->levels[level]))->cl_plane(0));
#else
    img = static_cast<armcomputeImage *>((to_MultiImage(pyr->levels[level]))->plane(0));
#endif

    return img;
}

void copy_pyramid_levels(vx_pyramid pyr, bool copy_back = false)
{
    armcomputePyramid *ne_pyr   = to_Pyramid(pyr);
    vx_size            n_levels = pyr->numLevels;
    std::string        dst_name = (copy_back) ? std::string("VX") : std::string("NE");
    for(unsigned int i = 0; i < n_levels; ++i)
    {
        ARM_COMPUTE_ERROR_ON(pyr->levels[i] == nullptr);
        auto img     = get_image_from_pyramid(pyr, i);
        auto img_pyr = ne_pyr->get_pyramid_level(i);

        map(img, img_pyr);

        uint8_t *dst      = (copy_back) ? img->buffer() : img_pyr->buffer();
        uint8_t *src      = (copy_back) ? img_pyr->buffer() : img->buffer();
        size_t   dst_size = (copy_back) ? img->info()->total_size() : img_pyr->info()->total_size();
        size_t   src_size = (copy_back) ? img_pyr->info()->total_size() : img->info()->total_size();
        ARM_COMPUTE_ERROR_ON(nullptr == dst || nullptr == src);
        size_t size_to_copy = std::min(src_size, dst_size);
        memcpy(dst, src, size_to_copy);

        unmap(img, img_pyr);
    }
}

BorderMode convert_to_BorderMode(const vx_border_t *borders)
{
    BorderMode mode = BorderMode::UNDEFINED;
    switch(borders->mode)
    {
        case VX_BORDER_UNDEFINED:
            mode = BorderMode::UNDEFINED;
            break;
        case VX_BORDER_CONSTANT:
            mode = BorderMode::CONSTANT;
            break;
        case VX_BORDER_REPLICATE:
            mode = BorderMode::REPLICATE;
            break;
        default:
            ARM_COMPUTE_ERROR("Unknown border mode");
    }
    return mode;
}

Channel get_scalar_channel(vx_scalar s)
{
    Channel channel = Channel::C0;
    vx_enum v;
    vxCopyScalar(s, &v, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
    switch(v)
    {
        case VX_CHANNEL_0:
            channel = Channel::C0;
            break;
        case VX_CHANNEL_1:
            channel = Channel::C1;
            break;
        case VX_CHANNEL_2:
            channel = Channel::C2;
            break;
        case VX_CHANNEL_3:
            channel = Channel::C3;
            break;
        case VX_CHANNEL_R:
            channel = Channel::R;
            break;
        case VX_CHANNEL_G:
            channel = Channel::G;
            break;
        case VX_CHANNEL_B:
            channel = Channel::B;
            break;
        case VX_CHANNEL_A:
            channel = Channel::A;
            break;
        case VX_CHANNEL_Y:
            channel = Channel::Y;
            break;
        case VX_CHANNEL_U:
            channel = Channel::U;
            break;
        case VX_CHANNEL_V:
            channel = Channel::V;
            break;
        default:
            ARM_COMPUTE_ERROR("Invalid channel");
    }
    return channel;
}

ConvertPolicy get_scalar_convert_policy(vx_scalar s)
{
    ConvertPolicy p = ConvertPolicy::WRAP;

    vx_enum v;
    vxCopyScalar(s, &v, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
    switch(v)
    {
        case VX_CONVERT_POLICY_WRAP:
            p = ConvertPolicy::WRAP;
            break;
        case VX_CONVERT_POLICY_SATURATE:
            p = ConvertPolicy::SATURATE;
            break;
        default:
            ARM_COMPUTE_ERROR("Unknown convert policy");
    }
    return p;
}

RoundingPolicy get_scalar_round_policy(vx_scalar s)
{
    RoundingPolicy p = RoundingPolicy::TO_ZERO;

    vx_enum v;
    vxCopyScalar(s, &v, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
    switch(v)
    {
        case VX_ROUND_POLICY_TO_ZERO:
            p = RoundingPolicy::TO_ZERO;
            break;
        case VX_ROUND_POLICY_TO_NEAREST_EVEN:
            p = RoundingPolicy::TO_NEAREST_EVEN;
            break;
        default:
            ARM_COMPUTE_ERROR("Unknown rounding policy");
    }
    return p;
}

Termination get_scalar_termination_policy(vx_scalar s)
{
    Termination p = Termination::TERM_CRITERIA_ITERATIONS;

    vx_enum v;
    vxCopyScalar(s, &v, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
    switch(v)
    {
        case VX_TERM_CRITERIA_ITERATIONS:
            p = Termination::TERM_CRITERIA_ITERATIONS;
            break;
        case VX_TERM_CRITERIA_EPSILON:
            p = Termination::TERM_CRITERIA_EPSILON;
            break;
        case VX_TERM_CRITERIA_BOTH:
            p = Termination::TERM_CRITERIA_BOTH;
            break;
        default:
            ARM_COMPUTE_ERROR("Unknown termination policy");
    }
    return p;
}

InterpolationPolicy get_scalar_interpolation_policy(vx_scalar s)
{
    InterpolationPolicy p = InterpolationPolicy::NEAREST_NEIGHBOR;

    vx_enum v;
    vxCopyScalar(s, &v, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
    switch(v)
    {
        case VX_INTERPOLATION_NEAREST_NEIGHBOR:
            p = InterpolationPolicy::NEAREST_NEIGHBOR;
            break;
        case VX_INTERPOLATION_AREA:
            p = InterpolationPolicy::AREA;
            break;
        case VX_INTERPOLATION_BILINEAR:
            p = InterpolationPolicy::BILINEAR;
            break;
        default:
            ARM_COMPUTE_ERROR("Unknown interpolation type");
    }
    return p;
}

int32_t get_scalar_norm_policy(vx_scalar s)
{
    int32_t norm_type = 0;
    vx_enum v;
    vxCopyScalar(s, &v, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
    switch(v)
    {
        case VX_NORM_L1:
            norm_type = 1;
            break;
        case VX_NORM_L2:
            norm_type = 2;
            break;
        default:
            ARM_COMPUTE_ERROR("Unknown norm type");
    }
    return norm_type;
}

ThresholdType get_threshold_type(vx_threshold t)
{
    ThresholdType p = ThresholdType::BINARY;

    vx_enum v;
    vxQueryThreshold(t, VX_THRESHOLD_TYPE, &v, sizeof(vx_enum));
    switch(v)
    {
        case VX_THRESHOLD_TYPE_BINARY:
            p = ThresholdType::BINARY;
            break;
        case VX_THRESHOLD_TYPE_RANGE:
            p = ThresholdType::RANGE;
            break;
        default:
            ARM_COMPUTE_ERROR("Unknown threshold type");
    }
    return p;
}

MatrixPattern get_matrix_pattern(vx_enum pattern)
{
    MatrixPattern p = MatrixPattern::OTHER;

    switch(pattern)
    {
        case VX_PATTERN_BOX:
            p = MatrixPattern::BOX;
            break;
        case VX_PATTERN_CROSS:
            p = MatrixPattern::CROSS;
            break;
        case VX_PATTERN_DISK:
            p = MatrixPattern::DISK;
            break;
        default:
            p = MatrixPattern::OTHER;
            break;
    }
    return p;
}

NonLinearFilterFunction get_scalar_nonlinearfunction(vx_scalar function)
{
    NonLinearFilterFunction f = NonLinearFilterFunction::MEDIAN;

    vx_enum v;
    vxCopyScalar(function, &v, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
    switch(v)
    {
        case VX_NONLINEAR_FILTER_MEDIAN:
            f = NonLinearFilterFunction::MEDIAN;
            break;
        case VX_NONLINEAR_FILTER_MIN:
            f = NonLinearFilterFunction::MIN;
            break;
        case VX_NONLINEAR_FILTER_MAX:
            f = NonLinearFilterFunction::MAX;
            break;
        default:
            ARM_COMPUTE_ERROR("Unknown non linear function type");
    }
    return f;
}

void check_scalar_type(vx_scalar scalar, vx_enum expected_type)
{
    vx_enum stype;
    vxQueryScalar(scalar, VX_SCALAR_TYPE, &stype, sizeof(stype));
    ARM_COMPUTE_ERROR_ON(stype != expected_type);
}

template <class T, vx_enum vx_type_enum>
void set_scalar(vx_scalar s, T val)
{
    check_scalar_type(s, vx_type_enum);
    vxCopyScalar(s, &val, VX_WRITE_ONLY, VX_MEMORY_TYPE_HOST);
}

template <class T, vx_enum vx_type_enum>
T get_scalar(vx_scalar s)
{
    check_scalar_type(s, vx_type_enum);
    T v;
    vxCopyScalar(s, &v, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
    return v;
}
}

vx_status vxAbsDiff(vx_image in1, vx_image in2, vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeAbsoluteDifference f;
    unmap(in1, in2, output);
    f.configure(to_Image(in1), to_Image(in2), to_Image(output));
    f.run();
    map(in1, in2, output);

    return VX_SUCCESS;
}

vx_status vxAccumulate(vx_image input, vx_image accum)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeAccumulate f;
    unmap(input, accum);
    f.configure(to_Image(input), to_Image(accum));
    f.run();
    map(input, accum);
    return VX_SUCCESS;
}

vx_status vxAccumulateSquare(vx_image input, vx_scalar scalar, vx_image accum)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    uint32_t _shift = get_scalar<vx_uint32, VX_TYPE_UINT32>(scalar);

    armcomputeAccumulateSquared f;
    unmap(input, accum);
    f.configure(to_Image(input), _shift, to_Image(accum));
    f.run();
    map(input, accum);

    return VX_SUCCESS;
}

vx_status vxAccumulateWeighted(vx_image input, vx_scalar scalar, vx_image accum)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    vx_float32 _scalar = get_scalar<vx_float32, VX_TYPE_FLOAT32>(scalar);

    armcomputeAccumulateWeighted f;
    unmap(input, accum);
    f.configure(to_Image(input), _scalar, to_Image(accum));
    f.run();
    map(input, accum);

    return VX_SUCCESS;
}

vx_status vxAddition(vx_image in0, vx_image in1, vx_scalar policy_param, vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    ConvertPolicy _policy = get_scalar_convert_policy(policy_param);

    armcomputeArithmeticAddition f;
    unmap(in0, in1, output);
    f.configure(to_Image(in0), to_Image(in1), to_Image(output), _policy);
    f.run();
    map(in0, in1, output);

    return VX_SUCCESS;
}

vx_status vxSubtraction(vx_image in0, vx_image in1, vx_scalar policy_param, vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    ConvertPolicy _policy = get_scalar_convert_policy(policy_param);

    armcomputeArithmeticSubtraction f;
    unmap(in0, in1, output);
    f.configure(to_Image(in0), to_Image(in1), to_Image(output), _policy);
    f.run();
    map(in0, in1, output);

    return VX_SUCCESS;
}

vx_status vxAnd(vx_image in0, vx_image in1, vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeBitwiseAnd f;
    unmap(in0, in1, output);
    f.configure(to_Image(in0), to_Image(in1), to_Image(output));
    f.run();
    map(in0, in1, output);

    return VX_SUCCESS;
}
vx_status vxOr(vx_image in0, vx_image in1, vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeBitwiseOr f;
    unmap(in0, in1, output);
    f.configure(to_Image(in0), to_Image(in1), to_Image(output));
    f.run();
    map(in0, in1, output);

    return VX_SUCCESS;
}
vx_status vxXor(vx_image in0, vx_image in1, vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeBitwiseXor f;
    unmap(in0, in1, output);
    f.configure(to_Image(in0), to_Image(in1), to_Image(output));
    f.run();
    map(in0, in1, output);

    return VX_SUCCESS;
}
vx_status vxNot(vx_image input, vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeBitwiseNot f;
    unmap(input, output);
    f.configure(to_Image(input), to_Image(output));
    f.run();
    map(input, output);

    return VX_SUCCESS;
}

vx_status vxBox3x3(vx_image src, vx_image dst, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode       border_mode = convert_to_BorderMode(bordermode);
    armcomputeBox3x3 f;
    unmap(src, dst);
    f.configure(to_Image(src), to_Image(dst), border_mode, bordermode->constant_value.U8);
    f.run();
    map(src, dst);

    return VX_SUCCESS;
}

vx_status vxCannyEdge(vx_image input, vx_threshold hyst, vx_scalar gradient_size, vx_scalar norm_type, vx_image output, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    vx_int32 _low_thresh, _up_thresh;
    vxQueryThreshold(hyst, VX_THRESHOLD_THRESHOLD_LOWER, &_low_thresh, sizeof(_low_thresh));
    vxQueryThreshold(hyst, VX_THRESHOLD_THRESHOLD_UPPER, &_up_thresh, sizeof(_up_thresh));
    vx_int32 _gradient_size = get_scalar<vx_int32, VX_TYPE_INT32>(gradient_size);
    vx_enum  _norm_type     = get_scalar_norm_policy(norm_type);

    BorderMode          border_mode = convert_to_BorderMode(bordermode);
    armcomputeCannyEdge f;
    unmap(input, output);
    f.configure(to_Image(input), to_Image(output), _up_thresh, _low_thresh, _gradient_size, _norm_type, border_mode, bordermode->constant_value.U8);
    f.run();
    map(input, output);

    return VX_SUCCESS;
}

vx_status vxChannelCombine(vx_image inputs[4], vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    armcomputeChannelCombine f;

    //Is the output single plane or not ?
    if(output->memory.nptrs == 1)
    {
        unmap(inputs[0], inputs[1], inputs[2], inputs[3], output);
        f.configure(to_Image(inputs[0]), to_Image(inputs[1]), to_Image(inputs[2]), to_Image(inputs[3]), to_Image(output));
        f.run();
        map(inputs[0], inputs[1], inputs[2], inputs[3], output);
    }
    else
    {
        unmap(inputs[0], inputs[1], inputs[2], output);
        f.configure(to_Image(inputs[0]), to_Image(inputs[1]), to_Image(inputs[2]), to_MultiImage(output));
        f.run();
        map(inputs[0], inputs[1], inputs[2], to_MultiImage(output));
    }

    return VX_SUCCESS;
}

vx_status vxChannelExtract(vx_image src, vx_scalar channel, vx_image dst)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    Channel channel_id = get_scalar_channel(channel);

    armcomputeChannelExtract f;
    if(src->memory.nptrs == 1)
    {
        unmap(src, dst);
        f.configure(to_Image(src), channel_id, to_Image(dst));
        f.run();
        map(src, dst);
    }
    else
    {
        unmap(to_MultiImage(src), dst);
        f.run();
        map(to_MultiImage(src), dst);
    }

    return VX_SUCCESS;
}

vx_status vxConvertColor(vx_image src, vx_image dst)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeColorConvert f;
    if(src->memory.nptrs == 1)
    {
        if(dst->memory.nptrs == 1)
        {
            unmap(src, dst);
            f.configure(to_Image(src), to_Image(dst));
            f.run();
            map(src, dst);
        }
        else
        {
            unmap(src, to_MultiImage(dst));
            f.configure(to_Image(src), to_MultiImage(dst));
            f.run();
            map(src, to_MultiImage(dst));
        }
    }
    else
    {
        if(dst->memory.nptrs == 1)
        {
            unmap(to_MultiImage(src), dst);
            f.configure(to_MultiImage(src), to_Image(dst));
            f.run();
            map(to_MultiImage(src), dst);
        }
        else
        {
            unmap(to_MultiImage(src), to_MultiImage(dst));
            f.configure(to_MultiImage(src), to_MultiImage(dst));
            f.run();
            map(to_MultiImage(src), to_MultiImage(dst));
        }
    }

    return VX_SUCCESS;
}

vx_status vxConvertDepth(vx_image input, vx_image output, vx_scalar spol, vx_scalar sshf)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    ConvertPolicy policy = get_scalar_convert_policy(spol);
    vx_int32      shift  = get_scalar<vx_int32, VX_TYPE_INT32>(sshf);

    armcomputeDepthConvert f;
    unmap(input, output);
    f.configure(to_Image(input), to_Image(output), policy, shift);
    f.run();
    map(input, output);

    return VX_SUCCESS;
}

vx_status vxConvolve(vx_image src, vx_convolution conv, vx_image dst, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    vx_size   _rows, _cols;
    vx_uint32 _scale;
    vxQueryConvolution(conv, VX_CONVOLUTION_ROWS, &_rows, sizeof(_rows));
    vxQueryConvolution(conv, VX_CONVOLUTION_COLUMNS, &_cols, sizeof(_cols));
    vxQueryConvolution(conv, VX_CONVOLUTION_SCALE, &_scale, sizeof(_scale));
    BorderMode border_mode = convert_to_BorderMode(bordermode);

    std::vector<vx_int16> _conv(_rows * _cols);
    vxCopyConvolutionCoefficients(conv, _conv.data(), VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
    std::reverse(_conv.begin(), _conv.end());

    unmap(src, dst);
    if(_rows == _cols)
    {
        if(3 == _rows)
        {
            armcomputeConvolution3x3 f;
            f.configure(to_Image(src), to_Image(dst), _conv.data(), _scale, border_mode, bordermode->constant_value.U8);
            f.run();
        }
        else if(5 == _rows)
        {
            armcomputeConvolution5x5 f;
            f.configure(to_Image(src), to_Image(dst), _conv.data(), _scale, border_mode, bordermode->constant_value.U8);
            f.run();
        }
        else if(7 == _rows)
        {
            armcomputeConvolution7x7 f;
            f.configure(to_Image(src), to_Image(dst), _conv.data(), _scale, border_mode, bordermode->constant_value.U8);
            f.run();
        }
        else if(9 == _rows)
        {
            armcomputeConvolution9x9 f;
            f.configure(to_Image(src), to_Image(dst), _conv.data(), _scale, border_mode, bordermode->constant_value.U8);
            f.run();
        }
        else
        {
            ARM_COMPUTE_ERROR("Unsupported convolution size \n");
        }
    }
    else
    {
        armcomputeConvolutionRectangle f;
        f.configure(to_Image(src), to_Image(dst), _conv.data(), _cols, _rows, _scale, border_mode, bordermode->constant_value.U8);
        f.run();
    }
    map(src, dst);

    return VX_SUCCESS;
}

vx_status vxConvolution3x3(vx_image src, vx_image dst, vx_int16 conv[3][3], const vx_border_t *borders)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeConvolution3x3 f;
    BorderMode               border_mode = convert_to_BorderMode(borders);

    unmap(src, dst);
    f.configure(to_Image(src), to_Image(dst), (const int16_t *)conv, 0, border_mode, borders->constant_value.U8);
    f.run();
    map(src, dst);

    return VX_SUCCESS;
}

vx_status vxDilate3x3(vx_image src, vx_image dst, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode       border_mode = convert_to_BorderMode(bordermode);
    armcomputeDilate f;
    unmap(src, dst);
    f.configure(to_Image(src), to_Image(dst), border_mode, bordermode->constant_value.U8);
    f.run();
    map(src, dst);

    return VX_SUCCESS;
}

vx_status vxEqualizeHist(vx_image src, vx_image dst)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeEqualizeHistogram f;
    unmap(src, dst);
    f.configure(to_Image(src), to_Image(dst));
    f.run();
    map(src, dst);

    return VX_SUCCESS;
}

vx_status vxErode3x3(vx_image src, vx_image dst, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode      border_mode = convert_to_BorderMode(bordermode);
    armcomputeErode f;
    unmap(src, dst);
    f.configure(to_Image(src), to_Image(dst), border_mode, bordermode->constant_value.U8);
    f.run();
    map(src, dst);

    return VX_SUCCESS;
}

vx_status vxFast9Corners(vx_image src, vx_scalar sens, vx_scalar nonm,
                         vx_array points, vx_scalar num_corners, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    vx_float32                 _sens   = get_scalar<vx_float32, VX_TYPE_FLOAT32>(sens);
    vx_bool                    _nonm   = get_scalar<vx_bool, VX_TYPE_BOOL>(nonm);
    armcomputeArray<KeyPoint> *_points = to_Array<KeyPoint>(points);

    BorderMode            border_mode = convert_to_BorderMode(bordermode);
    armcomputeFastCorners f;
    unmap(src, _points);
#if defined(USE_ARM_COMPUTE_OPENCL)
    unsigned int _num_corners = 0;
    f.configure(to_Image(src), _sens, _nonm, _points, &_num_corners, border_mode, bordermode->constant_value.U8);
#else
    f.configure(to_Image(src), _sens, _nonm, _points, border_mode, bordermode->constant_value.U8);
#endif
    f.run();
    map(src, _points);

    // Set number of corners
    if(num_corners)
    {
        set_scalar<size_t, VX_TYPE_SIZE>(num_corners, _points->num_values());
    }

    return VX_SUCCESS;
}

vx_status vxGaussian3x3(vx_image src, vx_image dst, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode            border_mode = convert_to_BorderMode(bordermode);
    armcomputeGaussian3x3 f;
    unmap(src, dst);
    f.configure(to_Image(src), to_Image(dst), border_mode, bordermode->constant_value.U8);
    f.run();
    map(src, dst);

    return VX_SUCCESS;
}

vx_status vxNonLinearFilter(vx_scalar function, vx_image src, vx_matrix mask, vx_image dst, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    vx_size _rows, _cols;
    vx_enum _pattern_enum;
    uint8_t _mask[25];
    vxQueryMatrix(mask, VX_MATRIX_ROWS, &_rows, sizeof(_rows));
    vxQueryMatrix(mask, VX_MATRIX_COLUMNS, &_cols, sizeof(_cols));
    vxQueryMatrix(mask, VX_MATRIX_PATTERN, &_pattern_enum, sizeof(vx_enum));

    ARM_COMPUTE_ERROR_ON_MSG(_rows != _cols, "Mask should be square\n");

    MatrixPattern           _pattern  = get_matrix_pattern(_pattern_enum);
    NonLinearFilterFunction _function = get_scalar_nonlinearfunction(function);
    if(MatrixPattern::OTHER == _pattern)
    {
        vxCopyMatrix(mask, (void *)_mask, VX_WRITE_ONLY, VX_MEMORY_TYPE_HOST);
    }

    BorderMode                border_mode = convert_to_BorderMode(bordermode);
    armcomputeNonLinearFilter f;
    unmap(src, dst);
    f.configure(to_Image(src), to_Image(dst), _function, _rows, _pattern, &_mask[0], border_mode, bordermode->constant_value.U8);
    f.run();
    map(src, dst);

    return VX_SUCCESS;
}

vx_status vxHarrisCorners(vx_image input, vx_scalar strength_thresh, vx_scalar min_distance, vx_scalar sensitivity, vx_scalar gradient_size, vx_scalar block_size,
                          vx_array corners, vx_scalar num_corners, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    vx_float32                 _strength_thresh = get_scalar<vx_float32, VX_TYPE_FLOAT32>(strength_thresh);
    vx_float32                 _min_distance    = get_scalar<vx_float32, VX_TYPE_FLOAT32>(min_distance);
    vx_float32                 _sensitivity     = get_scalar<vx_float32, VX_TYPE_FLOAT32>(sensitivity);
    vx_int32                   _gradient_size   = get_scalar<vx_int32, VX_TYPE_INT32>(gradient_size);
    vx_int32                   _block_size      = get_scalar<vx_int32, VX_TYPE_INT32>(block_size);
    armcomputeArray<KeyPoint> *_corners         = to_Array<KeyPoint>(corners);

    BorderMode              border_mode = convert_to_BorderMode(bordermode);
    armcomputeHarrisCorners f;
    unmap(input, _corners);
    f.configure(to_Image(input), _strength_thresh, _min_distance, _sensitivity, _gradient_size, _block_size,
                _corners, border_mode, bordermode->constant_value.U8);
    f.run();
    map(input, _corners);

    // Set number of corners
    if(num_corners)
    {
        set_scalar<size_t, VX_TYPE_SIZE>(num_corners, _corners->num_values());
    }

    return VX_SUCCESS;
}

vx_status vxHistogram(vx_image src, vx_distribution dist)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    armcomputeHistogram f;
    unmap(src, dist);
    f.configure(to_Image(src), to_Distribution1D(dist));
    f.run();
    map(src, dist);

    return VX_SUCCESS;
}

vx_status vxGaussianPyramid(vx_image input, vx_pyramid gaussian, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode border_mode = convert_to_BorderMode(bordermode);
    vx_float32 _scale;
    vxQueryPyramid(gaussian, VX_PYRAMID_SCALE, &_scale, sizeof(_scale));
    unmapPyramidLevels(gaussian);
    copy_pyramid_levels(gaussian);

    unmap(input);
    if(VX_SCALE_PYRAMID_ORB == _scale)
    {
        armcomputeGaussianPyramidOrb f;
        f.configure(to_Image(input), to_Pyramid(gaussian), border_mode, bordermode->constant_value.U8);
        f.run();
    }
    else
    {
        armcomputeGaussianPyramidHalf f;
        f.configure(to_Image(input), to_Pyramid(gaussian), border_mode, bordermode->constant_value.U8);
        f.run();
    }
    map(input);
    copy_pyramid_levels(gaussian, true);
    mapPyramidLevels(gaussian);

    return VX_SUCCESS;
}

vx_status vxLaplacianPyramid(vx_image input, vx_pyramid laplacian, vx_image output, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode border_mode = convert_to_BorderMode(bordermode);
    unmapPyramidLevels(laplacian);
    copy_pyramid_levels(laplacian);

    armcomputeLaplacianPyramid f;
    unmap(input, output);
    f.configure(to_Image(input), to_Pyramid(laplacian), to_Image(output), border_mode, bordermode->constant_value.U8);
    f.run();
    map(input, output);

    copy_pyramid_levels(laplacian, true);
    mapPyramidLevels(laplacian);

    return VX_SUCCESS;
}

vx_status vxLaplacianReconstruct(vx_pyramid laplacian, vx_image input, vx_image output, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode border_mode = convert_to_BorderMode(bordermode);

    unmapPyramidLevels(laplacian);
    copy_pyramid_levels(laplacian);

    armcomputeLaplacianReconstruct f;
    unmap(input, output);
    f.configure(to_Pyramid(laplacian), to_Image(input), to_Image(output), border_mode, bordermode->constant_value.U8);
    f.run();
    map(input, output);
    mapPyramidLevels(laplacian);

    return VX_SUCCESS;
}

vx_status vxIntegralImage(vx_image src, vx_image dst)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeIntegralImage f;
    unmap(src, dst);
    f.configure(to_Image(src), to_Image(dst));
    f.run();
    map(src, dst);

    return VX_SUCCESS;
}

vx_status vxMagnitude(vx_image grad_x, vx_image grad_y, vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputeMagnitude f;
    unmap(grad_x, grad_y, output);
    f.configure(to_Image(grad_x), to_Image(grad_y), to_Image(output));
    f.run();
    map(grad_x, grad_y, output);

    return VX_SUCCESS;
}

vx_status vxMeanStdDev(vx_image input, vx_scalar mean, vx_scalar stddev)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    float _mean, _stddev;

    armcomputeMeanStdDev f;
    unmap(input);
    f.configure(to_Image(input), &_mean, &_stddev);
    f.run();
    map(input);

    set_scalar<float, VX_TYPE_FLOAT32>(mean, _mean);
    set_scalar<float, VX_TYPE_FLOAT32>(stddev, _stddev);

    return VX_SUCCESS;
}

vx_status vxMedian3x3(vx_image src, vx_image dst, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode          border_mode = convert_to_BorderMode(bordermode);
    armcomputeMedian3x3 f;
    unmap(src, dst);
    f.configure(to_Image(src), to_Image(dst), border_mode, bordermode->constant_value.U8);
    f.run();
    map(src, dst);

    return VX_SUCCESS;
}

vx_status vxMinMaxLoc(vx_image input, vx_scalar minVal, vx_scalar maxVal, vx_array minLoc, vx_array maxLoc, vx_scalar minCount, vx_scalar maxCount)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    vx_enum   _image_type;
    vx_int32  _min, _max;
    vx_uint32 _minCount, _maxCount;

    vxQueryImage(input, VX_IMAGE_FORMAT, &_image_type, sizeof(_image_type));

    armcomputeMinMaxLocation f;
    unmap(input, to_Array<Coordinates2D>(minLoc), to_Array<Coordinates2D>(maxLoc));
    f.configure(to_Image(input), &_min, &_max,
                to_Array<Coordinates2D>(minLoc), to_Array<Coordinates2D>(maxLoc),
                &_minCount, &_maxCount);
    f.run();
    map(input, to_Array<Coordinates2D>(minLoc), to_Array<Coordinates2D>(maxLoc));

    if(VX_DF_IMAGE_U8 == _image_type)
    {
        set_scalar<vx_uint8, VX_TYPE_UINT8>(minVal, _min);
        set_scalar<vx_uint8, VX_TYPE_UINT8>(maxVal, _max);
    }
    else if(VX_DF_IMAGE_S16 == _image_type)
    {
        set_scalar<vx_int16, VX_TYPE_INT16>(minVal, _min);
        set_scalar<vx_int16, VX_TYPE_INT16>(maxVal, _max);
    }
    else
    {
        ARM_COMPUTE_ERROR("Unsupported image type\n");
    }

    if(minCount)
    {
        set_scalar<vx_uint32, VX_TYPE_UINT32>(minCount, _minCount);
    }
    if(maxCount)
    {
        set_scalar<vx_uint32, VX_TYPE_UINT32>(maxCount, _maxCount);
    }

    return VX_SUCCESS;
}

vx_status vxOpticalFlowPyrLK_full(vx_pyramid old_pyramid, vx_pyramid new_pyramid, vx_array prevPts, vx_array estimatedPts, vx_array nextPts, vx_scalar termination, vx_scalar epsilon,
                                  vx_scalar num_iterations, vx_scalar use_initial_estimate, vx_scalar window_dimension, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    Termination _termination          = get_scalar_termination_policy(termination);
    vx_float32  _epsilon              = get_scalar<vx_float32, VX_TYPE_FLOAT32>(epsilon);
    vx_uint32   _num_iterations       = get_scalar<vx_uint32, VX_TYPE_UINT32>(num_iterations);
    vx_bool     _use_initial_estimate = get_scalar<vx_bool, VX_TYPE_BOOL>(use_initial_estimate);
    vx_size     _window_dimension     = get_scalar<vx_size, VX_TYPE_SIZE>(window_dimension);

    BorderMode border_mode = convert_to_BorderMode(bordermode);
    unmapPyramidLevels(old_pyramid);
    unmapPyramidLevels(new_pyramid);
    copy_pyramid_levels(old_pyramid);
    copy_pyramid_levels(new_pyramid);

    armcomputeOpticalFlow f;

    unmap(to_Array<KeyPoint>(prevPts), to_Array<KeyPoint>(nextPts));
    f.configure(to_Pyramid(old_pyramid), to_Pyramid(new_pyramid), to_Array<KeyPoint>(prevPts), to_Array<KeyPoint>(estimatedPts), to_Array<KeyPoint>(nextPts),
                _termination, _epsilon, _num_iterations, _window_dimension, _use_initial_estimate, border_mode, bordermode->constant_value.U8);
    f.run();
    map(to_Array<KeyPoint>(prevPts), to_Array<KeyPoint>(nextPts));

    mapPyramidLevels(old_pyramid);
    mapPyramidLevels(new_pyramid);

    return VX_SUCCESS;
}

vx_status vxPhase(vx_image grad_x, vx_image grad_y, vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    armcomputePhase f;
    unmap(grad_x, grad_y, output);
    f.configure(to_Image(grad_x), to_Image(grad_y), to_Image(output));
    f.run();
    map(grad_x, grad_y, output);

    return VX_SUCCESS;
}

vx_status vxMultiply(vx_image in0, vx_image in1, vx_scalar scale_param, vx_scalar opolicy_param, vx_scalar rpolicy_param, vx_image output)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    vx_float32     _scale   = get_scalar<vx_float32, VX_TYPE_FLOAT32>(scale_param);
    ConvertPolicy  _opolicy = get_scalar_convert_policy(opolicy_param);
    RoundingPolicy _rpolicy = get_scalar_round_policy(rpolicy_param);

    armcomputePixelWiseMultiplication f;
    unmap(in0, in1, output);
    f.configure(to_Image(in0), to_Image(in1), to_Image(output), _scale, _opolicy, _rpolicy);
    f.run();
    map(in0, in1, output);
    return VX_SUCCESS;
}

vx_status vxRemap(vx_image src_image, vx_remap table, vx_scalar policy, vx_image dst_image, const vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode                                    border_mode         = convert_to_BorderMode(bordermode);
    InterpolationPolicy                           _interpolation_type = get_scalar_interpolation_policy(policy);
    std::vector<std::unique_ptr<armcomputeImage>> remap               = to_Remap(table);

    armcomputeRemap f;
    unmap(src_image, dst_image);
    f.configure(to_Image(src_image), remap.at(0).get(), remap.at(1).get(), to_Image(dst_image), _interpolation_type, border_mode, bordermode->constant_value.U8);
    f.run();
    map(src_image, dst_image);

    return VX_SUCCESS;
}

vx_status vxScaleImage(vx_image src_image, vx_image dst_image, vx_scalar stype, vx_border_t *bordermode, vx_float64 *interm, vx_size size)
{
    InterpolationPolicy _interpolation_type = get_scalar_interpolation_policy(stype);
    if(_interpolation_type == InterpolationPolicy::AREA)
    {
        printf("%s:%d %s (NOT IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
        return VX_FAILURE;
    }
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);
    BorderMode      border_mode = convert_to_BorderMode(bordermode);
    armcomputeScale f;
    unmap(src_image, dst_image);
    f.configure(to_Image(src_image), to_Image(dst_image), _interpolation_type, border_mode, bordermode->constant_value.U8);
    f.run();
    map(src_image, dst_image);

    return VX_SUCCESS;
}

vx_status vxSobel3x3(vx_image input, vx_image grad_x, vx_image grad_y, vx_border_t *bordermode)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode         border_mode = convert_to_BorderMode(bordermode);
    armcomputeSobel3x3 f;
    unmap(input, grad_x, grad_y);
    f.configure(to_Image(input), to_Image(grad_x), to_Image(grad_y), border_mode, bordermode->constant_value.U8);
    f.run();
    map(input, grad_x, grad_y);

    return VX_SUCCESS;
}

vx_status vxTableLookup(vx_image src, vx_lut lut, vx_image dst)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    vx_enum _lut_type;
    vxQueryLUT(lut, VX_LUT_TYPE, &_lut_type, sizeof(_lut_type));

    armcomputeTableLookup f;
    unmap(src, to_Lut((vx_array)lut), dst);
    f.configure(to_Image(src), to_Lut((vx_array)lut), to_Image(dst));
    f.run();
    map(src, to_Lut((vx_array)lut), dst);

    return VX_SUCCESS;
}

vx_status vxThreshold(vx_image src_image, vx_threshold threshold, vx_image dst_image)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    vx_int32 _low_thresh, _up_thresh, _thresh, _true_value, _false_value;
    vxQueryThreshold(threshold, VX_THRESHOLD_THRESHOLD_VALUE, &_thresh, sizeof(_thresh));
    vxQueryThreshold(threshold, VX_THRESHOLD_THRESHOLD_LOWER, &_low_thresh, sizeof(_low_thresh));
    vxQueryThreshold(threshold, VX_THRESHOLD_THRESHOLD_UPPER, &_up_thresh, sizeof(_up_thresh));
    vxQueryThreshold(threshold, VX_THRESHOLD_TRUE_VALUE, &_true_value, sizeof(_true_value));
    vxQueryThreshold(threshold, VX_THRESHOLD_FALSE_VALUE, &_false_value, sizeof(_false_value));
    ThresholdType _threshold_type = get_threshold_type(threshold);

    armcomputeThreshold f;
    vx_int32            _final_threshold = (ThresholdType::BINARY == _threshold_type) ? _thresh : _low_thresh;
    unmap(src_image, dst_image);
    f.configure(to_Image(src_image), to_Image(dst_image), _final_threshold, _false_value, _true_value, _threshold_type, _up_thresh);
    f.run();
    map(src_image, dst_image);

    return VX_SUCCESS;
}

vx_status vxWarpAffine(vx_image src_image, vx_matrix matrix, vx_scalar stype, vx_image dst_image, const vx_border_t *borders)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode          border_mode         = convert_to_BorderMode(borders);
    InterpolationPolicy _interpolation_type = get_scalar_interpolation_policy(stype);

    float _matrix[3][2] = { { 0 } };
    vxCopyMatrix(matrix, (void *)&_matrix[0][0], VX_READ_ONLY, VX_MEMORY_TYPE_HOST);

    armcomputeWarpAffine f;
    unmap(src_image, dst_image);
    f.configure(to_Image(src_image), to_Image(dst_image), &_matrix[0][0], _interpolation_type, border_mode, borders->constant_value.U8);
    f.run();
    map(src_image, dst_image);

    return VX_SUCCESS;
}

vx_status vxWarpPerspective(vx_image src_image, vx_matrix matrix, vx_scalar stype, vx_image dst_image, const vx_border_t *borders)
{
    printf("%s:%d %s (IMPLEMENTED)\n", __FILE__, __LINE__, __FUNCTION__);

    BorderMode          border_mode         = convert_to_BorderMode(borders);
    InterpolationPolicy _interpolation_type = get_scalar_interpolation_policy(stype);

    float _matrix[3][3] = { { 0 } };
    vxCopyMatrix(matrix, (void *)&_matrix[0][0], VX_READ_ONLY, VX_MEMORY_TYPE_HOST);

    armcomputeWarpPerspective f;
    unmap(src_image, dst_image);
    f.configure(to_Image(src_image), to_Image(dst_image), &_matrix[0][0], _interpolation_type, border_mode, borders->constant_value.U8);
    f.run();
    map(src_image, dst_image);

    return VX_SUCCESS;
}
