/*
 * Copyright (c) 2012-2016 The Khronos Group Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and/or associated documentation files (the
 * "Materials"), to deal in the Materials without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Materials, and to
 * permit persons to whom the Materials are furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Materials.
 *
 * MODIFICATIONS TO THIS FILE MAY MEAN IT NO LONGER ACCURATELY REFLECTS
 * KHRONOS STANDARDS. THE UNMODIFIED, NORMATIVE VERSIONS OF KHRONOS
 * SPECIFICATIONS AND HEADER INFORMATION ARE LOCATED AT
 *    https://www.khronos.org/registry/
 *
 * THE MATERIALS ARE PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * MATERIALS OR THE USE OR OTHER DEALINGS IN THE MATERIALS.
 */

/*!
 * \file
 * \brief The Absolute Difference Kernel.
 * \author Erik Rainey <erik.rainey@gmail.com>
 */

#include <VX/vx.h>
#include <VX/vx_helper.h>
#include <c_model.h>
#include <vx_internal.h>

static vx_status VX_CALLBACK vxAbsDiffKernel(vx_node node, const vx_reference parameters[], vx_uint32 num)
{
    vx_status status = VX_ERROR_INVALID_PARAMETERS;
    if(num == 3)
    {
        vx_image in1    = (vx_image)parameters[0];
        vx_image in2    = (vx_image)parameters[1];
        vx_image output = (vx_image)parameters[2];
        status          = vxAbsDiff(in1, in2, output);
    }
    return status;
}

static vx_status VX_CALLBACK vxAbsDiffInputValidator(vx_node node, vx_uint32 index)
{
    vx_status status = VX_ERROR_INVALID_PARAMETERS;
    if(index == 0)
    {
        vx_image     input = 0;
        vx_parameter param = vxGetParameterByIndex(node, index);

        vxQueryParameter(param, VX_PARAMETER_REF, &input, sizeof(input));
        if(input)
        {
            vx_df_image format = 0;
            vxQueryImage(input, VX_IMAGE_FORMAT, &format, sizeof(format));
            if(format == VX_DF_IMAGE_U8
               || format == VX_DF_IMAGE_S16
#if defined(EXPERIMENTAL_USE_S16)
               || format == VX_DF_IMAGE_U16
#endif
              )
                status = VX_SUCCESS;
            vxReleaseImage(&input);
        }
        vxReleaseParameter(&param);
    }
    else if(index == 1)
    {
        vx_image     images[2];
        vx_parameter param[2] =
        {
            vxGetParameterByIndex(node, 0),
            vxGetParameterByIndex(node, 1),
        };
        vxQueryParameter(param[0], VX_PARAMETER_REF, &images[0], sizeof(images[0]));
        vxQueryParameter(param[1], VX_PARAMETER_REF, &images[1], sizeof(images[1]));
        if(images[0] && images[1])
        {
            vx_uint32   width[2], height[2];
            vx_df_image format[2];

            vxQueryImage(images[0], VX_IMAGE_WIDTH, &width[0], sizeof(width[0]));
            vxQueryImage(images[1], VX_IMAGE_WIDTH, &width[1], sizeof(width[1]));
            vxQueryImage(images[0], VX_IMAGE_HEIGHT, &height[0], sizeof(height[0]));
            vxQueryImage(images[1], VX_IMAGE_HEIGHT, &height[1], sizeof(height[1]));
            vxQueryImage(images[0], VX_IMAGE_FORMAT, &format[0], sizeof(format[0]));
            vxQueryImage(images[1], VX_IMAGE_FORMAT, &format[1], sizeof(format[1]));
            if(width[0] == width[1] && height[0] == height[1] && format[0] == format[1])
            {
                status = VX_SUCCESS;
            }
            vxReleaseImage(&images[0]);
            vxReleaseImage(&images[1]);
        }
        vxReleaseParameter(&param[0]);
        vxReleaseParameter(&param[1]);
    }
    return status;
}

static vx_status VX_CALLBACK vxAbsDiffOutputValidator(vx_node node, vx_uint32 index, vx_meta_format_t *ptr)
{
    vx_status status = VX_ERROR_INVALID_PARAMETERS;
    if(index == 2)
    {
        vx_parameter param[2] =
        {
            vxGetParameterByIndex(node, 0),
            vxGetParameterByIndex(node, 1),
        };
        if((vxGetStatus((vx_reference)param[0]) == VX_SUCCESS) && (vxGetStatus((vx_reference)param[1]) == VX_SUCCESS))
        {
            vx_image images[2];
            vxQueryParameter(param[0], VX_PARAMETER_REF, &images[0], sizeof(images[0]));
            vxQueryParameter(param[1], VX_PARAMETER_REF, &images[1], sizeof(images[1]));
            if(images[0] && images[1])
            {
                vx_uint32   width[2], height[2];
                vx_df_image format = 0;
                vxQueryImage(images[0], VX_IMAGE_FORMAT, &format, sizeof(format));
                vxQueryImage(images[0], VX_IMAGE_WIDTH, &width[0], sizeof(width[0]));
                vxQueryImage(images[1], VX_IMAGE_WIDTH, &width[1], sizeof(width[1]));
                vxQueryImage(images[0], VX_IMAGE_HEIGHT, &height[0], sizeof(height[0]));
                vxQueryImage(images[1], VX_IMAGE_HEIGHT, &height[1], sizeof(height[1]));
                if(width[0] == width[1] && height[0] == height[1] && (format == VX_DF_IMAGE_U8
                                                                      || format == VX_DF_IMAGE_S16
#if defined(EXPERIMENTAL_USE_S16)
                                                                      || format == VX_DF_IMAGE_U16
#endif
                                                                     ))
                {
                    ptr->type             = VX_TYPE_IMAGE;
                    ptr->dim.image.format = format;
                    ptr->dim.image.width  = width[0];
                    ptr->dim.image.height = height[1];
                    status                = VX_SUCCESS;
                }
                vxReleaseImage(&images[0]);
                vxReleaseImage(&images[1]);
            }
            vxReleaseParameter(&param[0]);
            vxReleaseParameter(&param[1]);
        }
    }
    return status;
}

static vx_param_description_t absdiff_kernel_params[] =
{
    { VX_INPUT, VX_TYPE_IMAGE, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_IMAGE, VX_PARAMETER_STATE_REQUIRED },
    { VX_OUTPUT, VX_TYPE_IMAGE, VX_PARAMETER_STATE_REQUIRED },
};

vx_kernel_description_t absdiff_kernel =
{
    VX_KERNEL_ABSDIFF,
    "org.khronos.openvx.absdiff",
    vxAbsDiffKernel,
    absdiff_kernel_params, dimof(absdiff_kernel_params),
    NULL,
    vxAbsDiffInputValidator,
    vxAbsDiffOutputValidator,
    NULL,
    NULL,
};
