/*
 * Copyright (c) 2012-2016 The Khronos Group Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and/or associated documentation files (the
 * "Materials"), to deal in the Materials without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Materials, and to
 * permit persons to whom the Materials are furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Materials.
 *
 * MODIFICATIONS TO THIS FILE MAY MEAN IT NO LONGER ACCURATELY REFLECTS
 * KHRONOS STANDARDS. THE UNMODIFIED, NORMATIVE VERSIONS OF KHRONOS
 * SPECIFICATIONS AND HEADER INFORMATION ARE LOCATED AT
 *    https://www.khronos.org/registry/
 *
 * THE MATERIALS ARE PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * MATERIALS OR THE USE OR OTHER DEALINGS IN THE MATERIALS.
 */

/*!
 * \file
 * \brief The Filter Kernels.
 * \author Erik Rainey <erik.rainey@gmail.com>
 */

#include <VX/vx.h>
#include <VX/vx_helper.h>
#include <VX/vx_lib_debug.h>
#include <VX/vx_lib_extras.h>
#include <VX/vxu.h>
#include <extra_nodes.h>
#include <vx_internal.h>

static vx_param_description_t harris_kernel_params[] =
{
    { VX_INPUT, VX_TYPE_IMAGE, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED }, // strength_thresh
    { VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED }, // min_distance
    { VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED }, // sensitivity
    { VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED }, // gradient size
    { VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED }, // block size
    { VX_OUTPUT, VX_TYPE_ARRAY, VX_PARAMETER_STATE_REQUIRED },
    { VX_OUTPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_OPTIONAL },
};

static vx_status VX_CALLBACK vxHarrisCornersKernel(vx_node node, const vx_reference parameters[], vx_uint32 num)
{
    vx_status status = VX_FAILURE;

    if(num == dimof(harris_kernel_params))
    {
        vx_border_t bordermode;
        vx_image    src         = (vx_image)parameters[0];
        vx_scalar   str         = (vx_scalar)parameters[1];
        vx_scalar   min         = (vx_scalar)parameters[2];
        vx_scalar   sen         = (vx_scalar)parameters[3];
        vx_scalar   win         = (vx_scalar)parameters[4];
        vx_scalar   blk         = (vx_scalar)parameters[5];
        vx_array    arr         = (vx_array)parameters[6];
        vx_scalar   num_corners = (vx_scalar)parameters[7];
        status                  = vxQueryNode(node, VX_NODE_BORDER, &bordermode, sizeof(bordermode));
        if(status == VX_SUCCESS)
        {
            status = vxHarrisCorners(src, str, min, sen, win, blk, arr, num_corners, &bordermode);
        }
    }

    return status;
}

static vx_status VX_CALLBACK vxHarrisInputValidator(vx_node node, vx_uint32 index)
{
    vx_status status = VX_ERROR_INVALID_PARAMETERS;
    if(index == 0)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_image input = 0;
            status         = vxQueryParameter(param, VX_PARAMETER_REF, &input, sizeof(input));
            if((status == VX_SUCCESS) && (input))
            {
                vx_df_image format = 0;
                status             = vxQueryImage(input, VX_IMAGE_FORMAT, &format, sizeof(format));
                if((status == VX_SUCCESS) && (format == VX_DF_IMAGE_U8))
                {
                    status = VX_SUCCESS;
                }
                vxReleaseImage(&input);
            }
            vxReleaseParameter(&param);
        }
    }
    else if(index == 1)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_scalar strength_thresh = 0;
            status                    = vxQueryParameter(param, VX_PARAMETER_REF, &strength_thresh, sizeof(strength_thresh));
            if((status == VX_SUCCESS) && (strength_thresh))
            {
                vx_enum type = 0;
                vxQueryScalar(strength_thresh, VX_SCALAR_TYPE, &type, sizeof(type));
                if(type == VX_TYPE_FLOAT32)
                {
                    status = VX_SUCCESS;
                }
                else
                {
                    status = VX_ERROR_INVALID_TYPE;
                }
                vxReleaseScalar(&strength_thresh);
            }
            vxReleaseParameter(&param);
        }
    }
    else if(index == 2)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_scalar min_dist = 0;
            status             = vxQueryParameter(param, VX_PARAMETER_REF, &min_dist, sizeof(min_dist));
            if((status == VX_SUCCESS) && (min_dist))
            {
                vx_enum type = 0;
                vxQueryScalar(min_dist, VX_SCALAR_TYPE, &type, sizeof(type));
                if(type == VX_TYPE_FLOAT32)
                {
                    vx_float32 d = 0.0f;
                    status       = vxCopyScalar(min_dist, &d, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
                    if((status == VX_SUCCESS) && (0.0 <= d) && (d <= 30.0))
                    {
                        status = VX_SUCCESS;
                    }
                    else
                    {
                        status = VX_ERROR_INVALID_VALUE;
                    }
                }
                else
                {
                    status = VX_ERROR_INVALID_TYPE;
                }
                vxReleaseScalar(&min_dist);
            }
            vxReleaseParameter(&param);
        }
    }
    else if(index == 3)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_scalar sens = 0;
            status         = vxQueryParameter(param, VX_PARAMETER_REF, &sens, sizeof(sens));
            if((status == VX_SUCCESS) && (sens))
            {
                vx_enum type = 0;
                vxQueryScalar(sens, VX_SCALAR_TYPE, &type, sizeof(type));
                if(type == VX_TYPE_FLOAT32)
                {
                    vx_float32 k = 0.0f;
                    vxCopyScalar(sens, &k, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
                    VX_PRINT(VX_ZONE_INFO, "k = %lf\n", k);
                    if((0.040000f <= k) && (k < 0.150001f))
                    {
                        status = VX_SUCCESS;
                    }
                    else
                    {
                        status = VX_ERROR_INVALID_VALUE;
                    }
                }
                else
                {
                    status = VX_ERROR_INVALID_TYPE;
                }
                vxReleaseScalar(&sens);
            }
            vxReleaseParameter(&param);
        }
    }
    else if(index == 4 || index == 5)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_scalar scalar = 0;
            status           = vxQueryParameter(param, VX_PARAMETER_REF, &scalar, sizeof(scalar));
            if((status == VX_SUCCESS) && (scalar))
            {
                vx_enum type = 0;
                vxQueryScalar(scalar, VX_SCALAR_TYPE, &type, sizeof(type));
                if(type == VX_TYPE_INT32)
                {
                    vx_int32 size = 0;
                    vxCopyScalar(scalar, &size, VX_READ_ONLY, VX_MEMORY_TYPE_HOST);
                    VX_PRINT(VX_ZONE_INFO, "size = %u\n", size);
                    if((size == 3) || (size == 5) || (size == 7))
                    {
                        status = VX_SUCCESS;
                    }
                    else
                    {
                        status = VX_ERROR_INVALID_VALUE;
                    }
                }
                else
                {
                    status = VX_ERROR_INVALID_TYPE;
                }
                vxReleaseScalar(&scalar);
            }
            vxReleaseParameter(&param);
        }
    }
    return status;
}

static vx_status VX_CALLBACK vxHarrisOutputValidator(vx_node node, vx_uint32 index, vx_meta_format_t *ptr)
{
    vx_status status = VX_ERROR_INVALID_PARAMETERS;
    if(index == 6)
    {
        ptr->type                = VX_TYPE_ARRAY;
        ptr->dim.array.item_type = VX_TYPE_KEYPOINT;
        ptr->dim.array.capacity  = 0; /* no defined capacity requirement */
        status                   = VX_SUCCESS;
    }
    else if(index == 7)
    {
        ptr->dim.scalar.type = VX_TYPE_SIZE;
        status               = VX_SUCCESS;
    }
    return status;
}

vx_kernel_description_t harris_kernel =
{
    VX_KERNEL_HARRIS_CORNERS,
    "org.khronos.openvx.harris_corners",
    vxHarrisCornersKernel,
    harris_kernel_params, dimof(harris_kernel_params),
    NULL,
    vxHarrisInputValidator,
    vxHarrisOutputValidator,
    NULL,
    NULL
};
