/*
 * Copyright (c) 2012-2016 The Khronos Group Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and/or associated documentation files (the
 * "Materials"), to deal in the Materials without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Materials, and to
 * permit persons to whom the Materials are furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Materials.
 *
 * MODIFICATIONS TO THIS FILE MAY MEAN IT NO LONGER ACCURATELY REFLECTS
 * KHRONOS STANDARDS. THE UNMODIFIED, NORMATIVE VERSIONS OF KHRONOS
 * SPECIFICATIONS AND HEADER INFORMATION ARE LOCATED AT
 *    https://www.khronos.org/registry/
 *
 * THE MATERIALS ARE PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * MATERIALS OR THE USE OR OTHER DEALINGS IN THE MATERIALS.
 */

/*!
 * \file
 * \brief The PyrLK optical flow kernel.
 * \author Tomer Schwartz
 */

#include <VX/vx.h>
#include <VX/vx_helper.h>

#include <c_model.h>
#include <extra_nodes.h>
#include <vx_internal.h>

typedef struct _vx_keypoint_t_optpyrlk_internal
{
    vx_float32 x;               /*!< \brief The x coordinate. */
    vx_float32 y;               /*!< \brief The y coordinate. */
    vx_float32 strength;        /*!< \brief The strength of the keypoint. */
    vx_float32 scale;           /*!< \brief Unused field reserved for future use. */
    vx_float32 orientation;     /*!< \brief Unused field reserved for future use. */
    vx_int32   tracking_status; /*!< \brief A zero indicates a lost point. */
    vx_float32 error;           /*!< \brief An tracking method specific error. */

} vx_keypoint_t_optpyrlk_internal;

#define INT_ROUND(x, n) (((x) + (1 << ((n)-1))) >> (n))

static vx_param_description_t optpyrlk_kernel_params[] =
{
    { VX_INPUT, VX_TYPE_PYRAMID, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_PYRAMID, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_ARRAY, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_ARRAY, VX_PARAMETER_STATE_REQUIRED },
    { VX_OUTPUT, VX_TYPE_ARRAY, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED },
};

static vx_status VX_CALLBACK vxOpticalFlowPyrLKKernel(vx_node node, const vx_reference *parameters, vx_uint32 num)
{
    vx_status status = VX_FAILURE;

    if(num == dimof(optpyrlk_kernel_params))
    {
        vx_border_t bordermode;
        vx_pyramid  old_pyramid          = (vx_pyramid)parameters[0];
        vx_pyramid  new_pyramid          = (vx_pyramid)parameters[1];
        vx_array    prevPts              = (vx_array)parameters[2];
        vx_array    estimatedPts         = (vx_array)parameters[3];
        vx_array    nextPts              = (vx_array)parameters[4];
        vx_scalar   termination          = (vx_scalar)parameters[5];
        vx_scalar   epsilon              = (vx_scalar)parameters[6];
        vx_scalar   num_iterations       = (vx_scalar)parameters[7];
        vx_scalar   use_initial_estimate = (vx_scalar)parameters[8];
        vx_scalar   window_dimension     = (vx_scalar)parameters[9];
        status                           = vxQueryNode(node, VX_NODE_BORDER, &bordermode, sizeof(bordermode));
        status                           = vxOpticalFlowPyrLK_full(old_pyramid, new_pyramid, prevPts, estimatedPts, nextPts, termination, epsilon, num_iterations, use_initial_estimate, window_dimension,
                                                                   &bordermode);
    }

    return status;
}

static vx_status VX_CALLBACK vxOpticalFlowPyrLKInputValidator(vx_node node, vx_uint32 index)
{
    vx_status status = VX_ERROR_INVALID_PARAMETERS;
    if(index == 0 || index == 1)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_pyramid input = 0;
            vxQueryParameter(param, VX_PARAMETER_REF, &input, sizeof(input));
            if(input)
            {
                vx_size level = 0;
                vxQueryPyramid(input, VX_PYRAMID_LEVELS, &level, sizeof(level));
                if(level != 0)
                {
                    status = VX_SUCCESS;
                }
                vxReleasePyramid(&input);
            }
            vxReleaseParameter(&param);
        }
    }
    else if(index == 2 || index == 3)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_array arr = 0;
            vxQueryParameter(param, VX_PARAMETER_REF, &arr, sizeof(arr));
            if(arr)
            {
                vx_enum item_type = 0;
                vxQueryArray(arr, VX_ARRAY_ITEMTYPE, &item_type, sizeof(item_type));
                if(item_type == VX_TYPE_KEYPOINT)
                {
                    status = VX_SUCCESS;
                }
                vxReleaseArray(&arr);
            }
            vxReleaseParameter(&param);
        }
    }
    else if(index == 5)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_scalar sens = 0;
            status         = vxQueryParameter(param, VX_PARAMETER_REF, &sens, sizeof(sens));
            if((status == VX_SUCCESS) && (sens))
            {
                vx_enum type = 0;
                vxQueryScalar(sens, VX_SCALAR_TYPE, &type, sizeof(type));
                if(type == VX_TYPE_ENUM)
                {
                    status = VX_SUCCESS;
                }
                else
                {
                    status = VX_ERROR_INVALID_TYPE;
                }
                vxReleaseScalar(&sens);
            }
            vxReleaseParameter(&param);
        }
    }
    else if(index == 6)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_scalar sens = 0;
            status         = vxQueryParameter(param, VX_PARAMETER_REF, &sens, sizeof(sens));
            if((status == VX_SUCCESS) && (sens))
            {
                vx_enum type = 0;
                vxQueryScalar(sens, VX_SCALAR_TYPE, &type, sizeof(type));
                if(type == VX_TYPE_FLOAT32)
                {
                    status = VX_SUCCESS;
                }
                else
                {
                    status = VX_ERROR_INVALID_TYPE;
                }
                vxReleaseScalar(&sens);
            }
            vxReleaseParameter(&param);
        }
    }
    else if(index == 7)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_scalar sens = 0;
            status         = vxQueryParameter(param, VX_PARAMETER_REF, &sens, sizeof(sens));
            if((status == VX_SUCCESS) && (sens))
            {
                vx_enum type = 0;
                vxQueryScalar(sens, VX_SCALAR_TYPE, &type, sizeof(type));
                if(type == VX_TYPE_UINT32)
                {
                    status = VX_SUCCESS;
                }
                else
                {
                    status = VX_ERROR_INVALID_TYPE;
                }
                vxReleaseScalar(&sens);
            }
            vxReleaseParameter(&param);
        }
    }
    else if(index == 8)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_scalar sens = 0;
            status         = vxQueryParameter(param, VX_PARAMETER_REF, &sens, sizeof(sens));
            if((status == VX_SUCCESS) && (sens))
            {
                vx_enum type = 0;
                vxQueryScalar(sens, VX_SCALAR_TYPE, &type, sizeof(type));
                if(type == VX_TYPE_BOOL)
                {
                    status = VX_SUCCESS;
                }
                else
                {
                    status = VX_ERROR_INVALID_TYPE;
                }
                vxReleaseScalar(&sens);
            }
            vxReleaseParameter(&param);
        }
    }
    else if(index == 9)
    {
        vx_parameter param = vxGetParameterByIndex(node, index);
        if(vxGetStatus((vx_reference)param) == VX_SUCCESS)
        {
            vx_scalar sens = 0;
            status         = vxQueryParameter(param, VX_PARAMETER_REF, &sens, sizeof(sens));
            if((status == VX_SUCCESS) && (sens))
            {
                vx_enum type = 0;
                vxQueryScalar(sens, VX_SCALAR_TYPE, &type, sizeof(type));
                if(type == VX_TYPE_SIZE)
                {
                    status = VX_SUCCESS;
                }
                else
                {
                    status = VX_ERROR_INVALID_TYPE;
                }
                vxReleaseScalar(&sens);
            }
            vxReleaseParameter(&param);
        }
    }
    return status;
}

static vx_status VX_CALLBACK vxOpticalFlowPyrLKOutputValidator(vx_node node, vx_uint32 index, vx_meta_format_t *ptr)
{
    vx_status status = VX_ERROR_INVALID_PARAMETERS;
    if(index == 4)
    {
        vx_array     arr      = 0;
        vx_size      capacity = 0;
        vx_parameter param    = vxGetParameterByIndex(node, 2);
        vxQueryParameter(param, VX_PARAMETER_REF, &arr, sizeof(arr));
        vxQueryArray(arr, VX_ARRAY_CAPACITY, &capacity, sizeof(capacity));

        ptr->type                = VX_TYPE_ARRAY;
        ptr->dim.array.item_type = VX_TYPE_KEYPOINT;
        ptr->dim.array.capacity  = capacity;

        status = VX_SUCCESS;

        vxReleaseArray(&arr);
        vxReleaseParameter(&param);
    }
    return status;
}

#ifdef __cplusplus
extern "C"
#endif
vx_kernel_description_t optpyrlk_kernel =
{
    VX_KERNEL_OPTICAL_FLOW_PYR_LK,
    "org.khronos.openvx.optical_flow_pyr_lk",
    vxOpticalFlowPyrLKKernel,
    optpyrlk_kernel_params, dimof(optpyrlk_kernel_params),
    NULL,
    vxOpticalFlowPyrLKInputValidator,
    vxOpticalFlowPyrLKOutputValidator,
    NULL,
    NULL,
};
