/*
 * Copyright (c) 2012-2016 The Khronos Group Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and/or associated documentation files (the
 * "Materials"), to deal in the Materials without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Materials, and to
 * permit persons to whom the Materials are furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Materials.
 *
 * MODIFICATIONS TO THIS FILE MAY MEAN IT NO LONGER ACCURATELY REFLECTS
 * KHRONOS STANDARDS. THE UNMODIFIED, NORMATIVE VERSIONS OF KHRONOS
 * SPECIFICATIONS AND HEADER INFORMATION ARE LOCATED AT
 *    https://www.khronos.org/registry/
 *
 * THE MATERIALS ARE PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * MATERIALS OR THE USE OR OTHER DEALINGS IN THE MATERIALS.
 */

/*!
 * \file
 * \brief The Image Pyramid Kernel
 * \author Erik Rainey <erik.rainey@gmail.com>
 */

#include <VX/vx.h>
#include <VX/vx_helper.h>
#include <VX/vx_lib_debug.h>
#include <VX/vx_lib_extras.h>

#include <extra_nodes.h>
#include <math.h>
#include <vx_internal.h>

static vx_param_description_t gaussian_pyramid_kernel_params[] =
{
    { VX_INPUT, VX_TYPE_IMAGE, VX_PARAMETER_STATE_REQUIRED },
    { VX_OUTPUT, VX_TYPE_PYRAMID, VX_PARAMETER_STATE_REQUIRED },
};

static vx_status VX_CALLBACK vxGaussianPyramidKernel(vx_node node, const vx_reference parameters[], vx_uint32 num)
{
    vx_status status = VX_FAILURE;

    if(num == dimof(gaussian_pyramid_kernel_params))
    {
        vx_border_t bordermode;
        vx_image    input    = (vx_image)parameters[0];
        vx_pyramid  gaussian = (vx_pyramid)parameters[1];
        status               = vxQueryNode(node, VX_NODE_BORDER, &bordermode, sizeof(bordermode));
        status               = vxGaussianPyramid(input, gaussian, &bordermode);
    }

    return status;
}

static vx_status VX_CALLBACK vxGaussianPyramidInputValidator(vx_node node, vx_uint32 index)
{
    vx_status status = VX_ERROR_INVALID_PARAMETERS;
    if(index == 0)
    {
        vx_image     input = 0;
        vx_parameter param = vxGetParameterByIndex(node, index);

        vxQueryParameter(param, VX_PARAMETER_REF, &input, sizeof(input));
        if(input)
        {
            vx_df_image format = 0;
            vxQueryImage(input, VX_IMAGE_FORMAT, &format, sizeof(format));
            if(format == VX_DF_IMAGE_U8)
            {
                status = VX_SUCCESS;
            }
            vxReleaseImage(&input);
        }
        vxReleaseParameter(&param);
    }
    return status;
}

static vx_status VX_CALLBACK vxGaussianPyramidOutputValidator(vx_node node, vx_uint32 index, vx_meta_format_t *ptr)
{
    vx_status status = VX_ERROR_INVALID_PARAMETERS;
    if(index == 1)
    {
        vx_image     src       = 0;
        vx_parameter src_param = vxGetParameterByIndex(node, 0);
        vx_parameter dst_param = vxGetParameterByIndex(node, index);

        vxQueryParameter(src_param, VX_PARAMETER_REF, &src, sizeof(src));
        if(src)
        {
            vx_pyramid dst = 0;
            vxQueryParameter(dst_param, VX_PARAMETER_REF, &dst, sizeof(dst));

            if(dst)
            {
                vx_uint32   width = 0, height = 0;
                vx_df_image format;
                vx_size     num_levels;
                vx_float32  scale;

                vxQueryImage(src, VX_IMAGE_WIDTH, &width, sizeof(width));
                vxQueryImage(src, VX_IMAGE_HEIGHT, &height, sizeof(height));
                vxQueryImage(src, VX_IMAGE_FORMAT, &format, sizeof(format));
                vxQueryPyramid(dst, VX_PYRAMID_LEVELS, &num_levels, sizeof(num_levels));
                vxQueryPyramid(dst, VX_PYRAMID_SCALE, &scale, sizeof(scale));

                /* fill in the meta data with the attributes so that the checker will pass */
                ptr->type               = VX_TYPE_PYRAMID;
                ptr->dim.pyramid.width  = width;
                ptr->dim.pyramid.height = height;
                ptr->dim.pyramid.format = format;
                ptr->dim.pyramid.levels = num_levels;
                ptr->dim.pyramid.scale  = scale;
                status                  = VX_SUCCESS;
                vxReleasePyramid(&dst);
            }
            vxReleaseImage(&src);
        }
        vxReleaseParameter(&dst_param);
        vxReleaseParameter(&src_param);
    }
    return status;
}

vx_kernel_description_t gaussian_pyramid_kernel =
{
    VX_KERNEL_GAUSSIAN_PYRAMID,
    "org.khronos.openvx.gaussian_pyramid",
    vxGaussianPyramidKernel,
    gaussian_pyramid_kernel_params, dimof(gaussian_pyramid_kernel_params),
    NULL,
    vxGaussianPyramidInputValidator,
    vxGaussianPyramidOutputValidator,
    NULL,
    NULL
};

/* --------------------------------------------------------------------------*/
/* Laplacian Pyramid */

static vx_param_description_t laplacian_pyramid_kernel_params[] =
{
    { VX_INPUT, VX_TYPE_IMAGE, VX_PARAMETER_STATE_REQUIRED },
    { VX_OUTPUT, VX_TYPE_PYRAMID, VX_PARAMETER_STATE_REQUIRED },
    { VX_OUTPUT, VX_TYPE_IMAGE, VX_PARAMETER_STATE_REQUIRED },
};

static vx_status VX_CALLBACK vxLaplacianPyramidKernel(vx_node node, const vx_reference parameters[], vx_uint32 num)
{
    vx_status status = VX_FAILURE;

    if(num == dimof(laplacian_pyramid_kernel_params))
    {
        vx_border_t bordermode;
        vx_image    input     = (vx_image)parameters[0];
        vx_pyramid  laplacian = (vx_pyramid)parameters[1];
        vx_image    output    = (vx_image)parameters[2];
        status                = vxQueryNode(node, VX_NODE_BORDER, &bordermode, sizeof(bordermode));
        status                = vxLaplacianPyramid(input, laplacian, output, &bordermode);
    }

    return status;
}

static vx_status VX_CALLBACK vxLaplacianPyramidInputValidator(vx_node node, vx_uint32 index)
{
    vx_status status = VX_SUCCESS;
    if(index == 0)
    {
        vx_image     input = 0;
        vx_parameter param = vxGetParameterByIndex(node, index);

        status |= vxQueryParameter(param, VX_PARAMETER_REF, &input, sizeof(input));
        if(status == VX_SUCCESS && input)
        {
            vx_df_image format = 0;
            status |= vxQueryImage(input, VX_IMAGE_FORMAT, &format, sizeof(format));
            if(status == VX_SUCCESS)
            {
                if(format != VX_DF_IMAGE_U8)
                    status = VX_ERROR_INVALID_PARAMETERS;
            }

            status |= vxReleaseImage(&input);
        }

        status |= vxReleaseParameter(&param);
    }

    return status;
}

static vx_status VX_CALLBACK vxLaplacianPyramidOutputValidator(vx_node node, vx_uint32 index, vx_meta_format_t *ptr)
{
    vx_status status = VX_SUCCESS;
    if(index == 1)
    {
        vx_image     input     = 0;
        vx_parameter src_param = vxGetParameterByIndex(node, 0);
        vx_parameter dst_param = vxGetParameterByIndex(node, index);

        status |= vxQueryParameter(src_param, VX_PARAMETER_REF, &input, sizeof(input));
        if(status == VX_SUCCESS && input)
        {
            vx_uint32   width     = 0;
            vx_uint32   height    = 0;
            vx_df_image format    = 0;
            vx_pyramid  laplacian = 0;

            status |= vxQueryImage(input, VX_IMAGE_WIDTH, &width, sizeof(width));
            status |= vxQueryImage(input, VX_IMAGE_HEIGHT, &height, sizeof(height));
            status |= vxQueryImage(input, VX_IMAGE_FORMAT, &format, sizeof(format));

            status |= vxQueryParameter(dst_param, VX_PARAMETER_REF, &laplacian, sizeof(laplacian));
            if(status == VX_SUCCESS && laplacian)
            {
                vx_uint32   pyr_width  = 0;
                vx_uint32   pyr_height = 0;
                vx_df_image pyr_format = 0;
                vx_float32  pyr_scale  = 0;
                vx_size     pyr_levels = 0;

                status |= vxQueryPyramid(laplacian, VX_PYRAMID_WIDTH, &pyr_width, sizeof(pyr_width));
                status |= vxQueryPyramid(laplacian, VX_PYRAMID_HEIGHT, &pyr_height, sizeof(pyr_height));
                status |= vxQueryPyramid(laplacian, VX_PYRAMID_FORMAT, &pyr_format, sizeof(pyr_format));
                status |= vxQueryPyramid(laplacian, VX_PYRAMID_SCALE, &pyr_scale, sizeof(pyr_scale));
                status |= vxQueryPyramid(laplacian, VX_PYRAMID_LEVELS, &pyr_levels, sizeof(pyr_levels));

                if(status == VX_SUCCESS)
                {
                    if(pyr_width == width && pyr_height == height && pyr_format == VX_DF_IMAGE_S16 && pyr_scale == VX_SCALE_PYRAMID_HALF)
                    {
                        /* fill in the meta data with the attributes so that the checker will pass */
                        ptr->type               = VX_TYPE_PYRAMID;
                        ptr->dim.pyramid.width  = pyr_width;
                        ptr->dim.pyramid.height = pyr_height;
                        ptr->dim.pyramid.format = pyr_format;
                        ptr->dim.pyramid.scale  = pyr_scale;
                        ptr->dim.pyramid.levels = pyr_levels;
                    }
                    else
                        status = VX_ERROR_INVALID_PARAMETERS;
                }

                status |= vxReleasePyramid(&laplacian);
            }

            status |= vxReleaseImage(&input);
        }

        status |= vxReleaseParameter(&dst_param);
        status |= vxReleaseParameter(&src_param);
    }

    if(index == 2)
    {
        vx_pyramid   laplacian = 0;
        vx_parameter src_param = vxGetParameterByIndex(node, 1);
        vx_parameter dst_param = vxGetParameterByIndex(node, index);

        status |= vxQueryParameter(src_param, VX_PARAMETER_REF, &laplacian, sizeof(laplacian));
        if(status == VX_SUCCESS && laplacian)
        {
            vx_uint32   width      = 0;
            vx_uint32   height     = 0;
            vx_df_image format     = 0;
            vx_size     levels     = 0;
            vx_image    last_level = 0;
            vx_image    output     = 0;

            status |= vxQueryPyramid(laplacian, VX_PYRAMID_LEVELS, &levels, sizeof(levels));

            last_level = vxGetPyramidLevel(laplacian, (vx_uint32)(levels - 1));

            status |= vxQueryImage(last_level, VX_IMAGE_WIDTH, &width, sizeof(width));
            status |= vxQueryImage(last_level, VX_IMAGE_HEIGHT, &height, sizeof(height));
            status |= vxQueryImage(last_level, VX_IMAGE_FORMAT, &format, sizeof(format));

            status |= vxQueryParameter(dst_param, VX_PARAMETER_REF, &output, sizeof(output));
            if(status == VX_SUCCESS && output)
            {
                vx_uint32   dst_width  = 0;
                vx_uint32   dst_height = 0;
                vx_df_image dst_format = 0;

                status |= vxQueryImage(output, VX_IMAGE_WIDTH, &dst_width, sizeof(dst_width));
                status |= vxQueryImage(output, VX_IMAGE_HEIGHT, &dst_height, sizeof(dst_height));
                status |= vxQueryImage(output, VX_IMAGE_FORMAT, &dst_format, sizeof(dst_format));

                if(status == VX_SUCCESS)
                {
                    if(dst_width == width && dst_height == height && dst_format == VX_DF_IMAGE_S16)
                    {
                        /* fill in the meta data with the attributes so that the checker will pass */
                        ptr->type             = VX_TYPE_IMAGE;
                        ptr->dim.image.width  = dst_width;
                        ptr->dim.image.height = dst_height;
                        ptr->dim.image.format = dst_format;
                    }
                    else
                        status = VX_ERROR_INVALID_PARAMETERS;
                }

                status |= vxReleaseImage(&output);
            }

            status |= vxReleaseImage(&last_level);
            status |= vxReleasePyramid(&laplacian);
        }

        status |= vxReleaseParameter(&dst_param);
        status |= vxReleaseParameter(&src_param);
    }

    return status;
}

vx_kernel_description_t laplacian_pyramid_kernel =
{
    VX_KERNEL_LAPLACIAN_PYRAMID,
    "org.khronos.openvx.laplacian_pyramid",
    vxLaplacianPyramidKernel,
    laplacian_pyramid_kernel_params, dimof(laplacian_pyramid_kernel_params),
    NULL,
    vxLaplacianPyramidInputValidator,
    vxLaplacianPyramidOutputValidator,
    NULL,
    NULL
};

/* Laplacian Reconstruct */

#define VX_SCALE_PYRAMID_DOUBLE (2.0f)

static vx_param_description_t laplacian_reconstruct_kernel_params[] =
{
    { VX_INPUT, VX_TYPE_PYRAMID, VX_PARAMETER_STATE_REQUIRED },
    { VX_INPUT, VX_TYPE_IMAGE, VX_PARAMETER_STATE_REQUIRED },
    { VX_OUTPUT, VX_TYPE_IMAGE, VX_PARAMETER_STATE_REQUIRED },
};

static vx_status VX_CALLBACK vxLaplacianReconstructKernel(vx_node node, const vx_reference parameters[], vx_uint32 num)
{
    vx_status status = VX_FAILURE;

    if(num == dimof(laplacian_reconstruct_kernel_params))
    {
        vx_border_t bordermode;
        vx_pyramid  laplacian = (vx_pyramid)parameters[0];
        vx_image    input     = (vx_image)parameters[1];
        vx_image    output    = (vx_image)parameters[2];
        status                = vxQueryNode(node, VX_NODE_BORDER, &bordermode, sizeof(bordermode));
        status                = vxLaplacianReconstruct(laplacian, input, output, &bordermode);
    }

    return status;
}

static vx_status VX_CALLBACK vxLaplacianReconstructInputValidator(vx_node node, vx_uint32 index)
{
    vx_status status = VX_SUCCESS;
    if(index == 0)
    {
        vx_pyramid   laplacian = 0;
        vx_parameter param     = vxGetParameterByIndex(node, index);

        status |= vxQueryParameter(param, VX_PARAMETER_REF, &laplacian, sizeof(laplacian));
        if(status == VX_SUCCESS && laplacian)
        {
            vx_df_image pyr_format = 0;
            vx_float32  pyr_scale  = 0;

            status |= vxQueryPyramid(laplacian, VX_PYRAMID_FORMAT, &pyr_format, sizeof(pyr_format));
            status |= vxQueryPyramid(laplacian, VX_PYRAMID_SCALE, &pyr_scale, sizeof(pyr_scale));

            if(status == VX_SUCCESS)
            {
                if((pyr_format != VX_DF_IMAGE_S16) || (pyr_scale != VX_SCALE_PYRAMID_HALF))
                    status = VX_ERROR_INVALID_PARAMETERS;
            }

            status |= vxReleasePyramid(&laplacian);
        }

        status |= vxReleaseParameter(&param);
    }

    if(index == 1)
    {
        vx_pyramid laplacian = 0;
        vx_image   input     = 0;

        vx_parameter param0 = vxGetParameterByIndex(node, 0);
        vx_parameter param1 = vxGetParameterByIndex(node, index);

        status |= vxQueryParameter(param0, VX_PARAMETER_REF, &laplacian, sizeof(laplacian));
        status |= vxQueryParameter(param1, VX_PARAMETER_REF, &input, sizeof(input));
        if(status == VX_SUCCESS && laplacian && input)
        {
            vx_size     pyr_levels            = 0;
            vx_uint32   pyr_last_level_width  = 0;
            vx_uint32   pyr_last_level_height = 0;
            vx_image    pyr_last_level_img    = 0;
            vx_uint32   lowest_res_width      = 0;
            vx_uint32   lowest_res_height     = 0;
            vx_df_image lowest_res_format     = 0;

            status |= vxQueryPyramid(laplacian, VX_PYRAMID_LEVELS, &pyr_levels, sizeof(pyr_levels));

            pyr_last_level_img = vxGetPyramidLevel(laplacian, (vx_uint32)(pyr_levels - 1));

            status |= vxQueryImage(pyr_last_level_img, VX_IMAGE_WIDTH, &pyr_last_level_width, sizeof(pyr_last_level_width));
            status |= vxQueryImage(pyr_last_level_img, VX_IMAGE_HEIGHT, &pyr_last_level_height, sizeof(pyr_last_level_height));

            status |= vxQueryImage(input, VX_IMAGE_WIDTH, &lowest_res_width, sizeof(lowest_res_width));
            status |= vxQueryImage(input, VX_IMAGE_HEIGHT, &lowest_res_height, sizeof(lowest_res_height));
            status |= vxQueryImage(input, VX_IMAGE_FORMAT, &lowest_res_format, sizeof(lowest_res_format));

            if(status == VX_SUCCESS)
            {
                if((pyr_last_level_width != lowest_res_width) || (pyr_last_level_height != lowest_res_height) || (lowest_res_format != VX_DF_IMAGE_S16))
                {
                    status = VX_ERROR_INVALID_PARAMETERS;
                }
            }

            status |= vxReleasePyramid(&laplacian);
            status |= vxReleaseImage(&input);
            status |= vxReleaseImage(&pyr_last_level_img);
        }

        status |= vxReleaseParameter(&param0);
        status |= vxReleaseParameter(&param1);
    }

    return status;
}

static vx_status VX_CALLBACK vxLaplacianReconstructOutputValidator(vx_node node, vx_uint32 index, vx_meta_format_t *ptr)
{
    vx_status status = VX_SUCCESS;
    if(index == 2)
    {
        vx_pyramid   laplacian = 0;
        vx_parameter src_param = vxGetParameterByIndex(node, 0);
        vx_parameter dst_param = vxGetParameterByIndex(node, index);

        status |= vxQueryParameter(src_param, VX_PARAMETER_REF, &laplacian, sizeof(laplacian));
        if(status == VX_SUCCESS && laplacian)
        {
            vx_uint32 pyr_width  = 0;
            vx_uint32 pyr_height = 0;
            vx_image  output     = 0;

            status |= vxQueryPyramid(laplacian, VX_PYRAMID_WIDTH, &pyr_width, sizeof(pyr_width));
            status |= vxQueryPyramid(laplacian, VX_PYRAMID_HEIGHT, &pyr_height, sizeof(pyr_height));

            status |= vxQueryParameter(dst_param, VX_PARAMETER_REF, &output, sizeof(output));

            if(status == VX_SUCCESS && output)
            {
                vx_uint32   dst_width  = 0;
                vx_uint32   dst_height = 0;
                vx_df_image dst_format = 0;

                status |= vxQueryImage(output, VX_IMAGE_WIDTH, &dst_width, sizeof(dst_width));
                status |= vxQueryImage(output, VX_IMAGE_HEIGHT, &dst_height, sizeof(dst_height));
                status |= vxQueryImage(output, VX_IMAGE_FORMAT, &dst_format, sizeof(dst_format));

                if(status == VX_SUCCESS)
                {
                    if(dst_width == pyr_width && dst_height == pyr_height && dst_format == VX_DF_IMAGE_U8)
                    {
                        /* fill in the meta data with the attributes so that the checker will pass */
                        ptr->type             = VX_TYPE_IMAGE;
                        ptr->dim.image.width  = dst_width;
                        ptr->dim.image.height = dst_height;
                        ptr->dim.image.format = dst_format;
                    }
                    else
                        status = VX_ERROR_INVALID_PARAMETERS;
                }

                status |= vxReleaseImage(&output);
            }

            status |= vxReleasePyramid(&laplacian);
        }

        status |= vxReleaseParameter(&dst_param);
        status |= vxReleaseParameter(&src_param);
    }

    return status;
}

vx_kernel_description_t laplacian_reconstruct_kernel =
{
    VX_KERNEL_LAPLACIAN_RECONSTRUCT,
    "org.khronos.openvx.laplacian_reconstruct",
    vxLaplacianReconstructKernel,
    laplacian_reconstruct_kernel_params, dimof(laplacian_reconstruct_kernel_params),
    NULL,
    vxLaplacianReconstructInputValidator,
    vxLaplacianReconstructOutputValidator,
    NULL,
    NULL
};
