/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#ifndef ARM_COMPUTE_FIXED_POINT_INT32_H
#define ARM_COMPUTE_FIXED_POINT_INT32_H

// TODO These functions where implemented to be used in softmax-uint8 kernel and therefore processes only vectors of length 16.
// But they can be managed to process arbitrary vector length using VEC_DATA_TYPE(int, size) definition to be more reusable.

/** For each element of input vector, the corresponding bits of the result item are set
 * if the input item is zero.
 */
inline int16 fp_mask_if_zero(int16 a)
{
    const int16 zero = 0;
    const int16 one  = ~0;
    return select(zero, one, a == 0);
}

/** For each element of input vector, the corresponding bits of the result item are set
 * if the input item is non-zero.
 */
inline int16 fp_mask_if_non_zero(int16 a)
{
    const int16 zero = 0;
    const int16 one  = ~0;
    return select(zero, one, a != 0);
}

/** Each bit of the result is set to the corresponding bit of either then_val or
 * else_val depending on whether the corresponding bit of if_mask is set.
 * Equivalent to the VBSL instruction in ARM NEON.
 */
inline int16 fp_select_using_mask(int16 if_mask, int16 then_val, int16 else_val)
{
    return (if_mask & then_val) ^ (~if_mask & else_val);
}

/** Correctly rounded to nearest division by a power of two (POT).
 * Also known as a rounding arithmetic right shift.
 *
 * @param[in] x        Argument.
 * @param[in] exponent POT, must be positive number.
 *
 * @return Arithmetic right shift.
 */
inline int16 fp_rounding_divide_by_POT(int16 x, int exponent)
{
    int16       mask      = (1 << exponent) - 1;
    const int16 zero      = 0;
    const int16 one       = 1;
    int16       threshold = (mask >> 1) + select(zero, one, x < 0);
    return (x >> exponent) + select(zero, one, (x & mask) > threshold);
}

/** Calculates the product of a integer value by a power of two (POT), with either a positive exponent
 * (equivalent to an arithmetic left shift, saturating) or a negative exponent
 * (equivalent to an arithmetic right shift, rounding to nearest).
 *
 * @param[in] x        Argument.
 * @param[in] exponent POT, can be positive or negative number.
 *
 * @return Arithmetic left or right shift.
 */
inline int16 fp_saturating_rounding_mult_by_POT(int16 x, int exponent)
{
    if(exponent < 0)
        return fp_rounding_divide_by_POT(x, -exponent);

    const int16 min           = INT_MIN;
    const int16 max           = INT_MAX;
    int         threshold     = ((1 << (31 - exponent)) - 1);
    int16       positive_mask = fp_mask_if_non_zero(x > threshold);
    int16       negative_mask = fp_mask_if_non_zero(x < -threshold);
    int16       result        = x << exponent;
    result                    = fp_select_using_mask(positive_mask, max, result);
    result                    = fp_select_using_mask(negative_mask, min, result);
    return result;
}

/** Calculates (a+b)/2, rounded to the nearest integer.
 * Equivalent to VRHADD in the ARM NEON instruction set.
 */
inline int16 fp_rounding_half_sum(int16 a, int16 b)
{
    long16       a64       = convert_long16(a);
    long16       b64       = convert_long16(b);
    long16       sum       = a64 + b64;
    const long16 one       = 1;
    const long16 minus_one = -1;
    long16       sign      = select(minus_one, one, sum >= 0);
    return convert_int16((sum + sign) / 2);
}

/** Product of two numbers, interpreting them as fixed-point values in the interval [-1, 1),
 * rounding to the nearest value, and saturating -1 * -1 to the maximum value.
 * This is equivalent to the VQRDMULH instruction in ARM NEON.
 */
inline int16 saturating_rounding_doubling_high_mul(int16 a, int16 b)
{
    int16  overflow     = a == b && a == INT_MIN;
    long16 a_64         = convert_long16(a);
    long16 b_64         = convert_long16(b);
    long16 ab_64        = a_64 * b_64;
    long16 mask1        = 1 << 30;
    long16 mask2        = 1 - (1 << 30);
    long16 nudge        = select(mask2, mask1, ab_64 >= 0);
    long16 mask         = 1ll << 31;
    int16  ab_x2_high32 = convert_int16((ab_64 + nudge) / mask);
    return select(ab_x2_high32, INT_MAX, overflow);
}

/** Fixed-point multiplication.
 *
 * @param[in] a Argument 1 in fixed-point format Q(a).
 * @param[in] b Argument 2 in fixed-point format Q(b).
 *
 * @return Result in fixed-point format Q(a+b).
 */
inline int16 fp_mult(int16 a, int16 b)
{
    return saturating_rounding_doubling_high_mul(a, b);
}

/** Calculates \f$ exp(x) \f$ for x in [-1/4, 0).
 *
 * @param[in] a Argument in fixed-point format Q0.
 *
 * @return Result in fixed-point format Q0.
 */
inline int16 fp_exp_on_interval_between_negative_one_quarter_and_0_excl(int16 a)
{
    const int16 constant_term                            = 1895147668;
    const int16 constant_1_over_3                        = 715827883;
    const int   kFractionalBits                          = 31;
    int16       x                                        = a + (1 << (kFractionalBits - 3));
    int16       x2                                       = fp_mult(x, x);
    int16       x3                                       = fp_mult(x2, x);
    int16       x4                                       = fp_mult(x2, x2);
    int16       x4_over_4                                = fp_rounding_divide_by_POT(x4, 2);
    int16       x4_over_24_plus_x3_over_6_plus_x2        = fp_mult((x4_over_4 + x3), constant_1_over_3) + x2;
    int16       x4_over_24_plus_x3_over_6_plus_x2_over_2 = fp_rounding_divide_by_POT(x4_over_24_plus_x3_over_6_plus_x2, 1);
    return constant_term + fp_mult(constant_term, x + x4_over_24_plus_x3_over_6_plus_x2_over_2);
}

/** Calculates \f$ exp(x) \f$ for x < 0.
 *
 * @param[in] a            Argument in fixed-point format Q(kIntegerBits).
 * @param[in] kIntegerBits Number of integer bit in argument.
 *
 * @return Result in fixed-point format Q0.
 */
inline int16 fp_exp_on_negative_values(int16 a, int kIntegerBits)
{
    const int kFractionalBits                        = 31 - kIntegerBits;
    int16     kOneQuarter                            = 1 << (kFractionalBits - 2);
    int16     mask                                   = kOneQuarter - 1;
    int16     a_mod_quarter_minus_one_quarter        = (a & mask) - kOneQuarter;
    int16     a_mod_quarter_minus_one_quarter_scaled = a_mod_quarter_minus_one_quarter << kIntegerBits;
    int16     result                                 = fp_exp_on_interval_between_negative_one_quarter_and_0_excl(a_mod_quarter_minus_one_quarter_scaled);
    int16     remainder                              = a_mod_quarter_minus_one_quarter - a;

#define GEMMLOWP_EXP_BARREL_SHIFTER(Exponent, FixedPointMultiplier)                        \
    if(kIntegerBits > Exponent)                                                            \
    {                                                                                      \
        const int kShiftAmount = kIntegerBits > Exponent ? kFractionalBits + Exponent : 0; \
        result                 = fp_select_using_mask(                                     \
                                                                                           fp_mask_if_non_zero(remainder & (1 << kShiftAmount)),                          \
                                                                                           fp_mult(result, FixedPointMultiplier), result);                                \
    }
    GEMMLOWP_EXP_BARREL_SHIFTER(-2, 1672461947);
    GEMMLOWP_EXP_BARREL_SHIFTER(-1, 1302514674);
    GEMMLOWP_EXP_BARREL_SHIFTER(+0, 790015084);
    GEMMLOWP_EXP_BARREL_SHIFTER(+1, 290630308);
    GEMMLOWP_EXP_BARREL_SHIFTER(+2, 39332535);
    GEMMLOWP_EXP_BARREL_SHIFTER(+3, 720401);
    GEMMLOWP_EXP_BARREL_SHIFTER(+4, 242);
#undef GEMMLOWP_EXP_BARREL_SHIFTER

    if(kIntegerBits > 5)
    {
        const int16 clamp = -(1 << (kFractionalBits + 5));
        result            = fp_select_using_mask(fp_mask_if_non_zero(a < clamp), 0, result);
    }

    const int16 Q0_one = INT_MAX;
    return fp_select_using_mask(fp_mask_if_zero(a), Q0_one, result);
}

/** Calculates \f$ 1 / (1 + x) \f$ for x in (0, 1).
 *
 * @param[in] a Argument in fixed-point format Q0.
 *
 * @return Result in fixed-point format Q0.
 */
inline int16 fp_one_over_one_plus_x_for_x_in_0_1(int16 a)
{
    const int16 Q0_one            = INT_MAX;
    const int16 Q2_one            = 1 << (31 - 2);
    int16       half_denominator  = fp_rounding_half_sum(a, Q0_one);
    const int16 Q2_48_over_17     = 1515870810;
    const int16 Q2_neg_32_over_17 = -1010580540;
    int16       x                 = Q2_48_over_17 + fp_mult(half_denominator, Q2_neg_32_over_17);
    for(int i = 0; i < 3; i++)
    {
        int16 half_denominator_times_x           = fp_mult(half_denominator, x);
        int16 one_minus_half_denominator_times_x = Q2_one - half_denominator_times_x;
        int16 tmp                                = fp_mult(x, one_minus_half_denominator_times_x);
        x                                        = x + fp_saturating_rounding_mult_by_POT(tmp, 2);
    }
    return fp_saturating_rounding_mult_by_POT(x, 1);
}

/** Considering the integer value as fixed-point, change the number of integer bits and update value accordingly.
 *
 * @param[in] value          Value to be rescaled.
 * @param[in] srcIntegerBits Old number of integer bits.
 * @param[in] dstIntegerBits New number of integer bits.
 *
 * @return Rescaled value.
 */
inline int16 fp_rescale(int16 value, int srcIntegerBits, int dstIntegerBits)
{
    int exponent = srcIntegerBits - dstIntegerBits;
    return fp_saturating_rounding_mult_by_POT(value, exponent);
}

#endif // ARM_COMPUTE_FIXED_POINT_INT32_H
