/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#ifndef __ARM_COMPUTE_TEST_VXACCESSOR_H__
#define __ARM_COMPUTE_TEST_VXACCESSOR_H__

#include "arm_compute/core/TensorShape.h"
#include "arm_compute/core/Utils.h"
#include "tests/IAccessor.h"

#include <VX/vx.h>

namespace arm_compute
{
namespace test
{
/** Accessor implementation for vx_image objects. */
class VXAccessor : public IAccessor
{
public:
    /** Create an accessor for the given @p image.
     *
     * @param[in, out] image To be accessed image.
     *
     * @note The memory is mapped by the constructor.
     */
    VXAccessor(vx_image &image);

    VXAccessor(const VXAccessor &) = delete;
    VXAccessor &operator=(const VXAccessor &) = delete;
    VXAccessor(VXAccessor &&)                 = default;
    VXAccessor &operator=(VXAccessor &&) = default;

    /** Destructor that unmaps the memory. */
    ~VXAccessor();

    TensorShape      shape() const override;
    size_t           element_size() const override;
    size_t           size() const override;
    Format           format() const override;
    DataType         data_type() const override;
    int              num_channels() const override;
    int              num_elements() const override;
    PaddingSize      padding() const override;
    int              fixed_point_position() const override;
    QuantizationInfo quantization_info() const override;
    const void *operator()(const Coordinates &coord) const override;
    void *operator()(const Coordinates &coord) override;

private:
    vx_rectangle_t get_image_size(const vx_image &image) const;

    vx_image                  &_image;
    void                      *_mapping{ nullptr };
    vx_imagepatch_addressing_t _addr{ 0, 0, 0, 0, 0, 0, 0, 0 };
    vx_map_id                  _map_id{ 0 };
};

inline VXAccessor::VXAccessor(vx_image &image)
    : _image{ image }
{
    vx_rectangle_t      rect  = get_image_size(image);
    constexpr vx_uint32 plane = 0;

    vxMapImagePatch(_image, &rect, plane, &_map_id, &_addr, &_mapping, VX_READ_AND_WRITE, VX_MEMORY_TYPE_NONE, VX_NOGAP_X);
}

inline VXAccessor::~VXAccessor()
{
    vxUnmapImagePatch(_image, _map_id);
}

inline TensorShape VXAccessor::shape() const
{
    vx_rectangle_t rect = get_image_size(_image);

    return TensorShape(rect.end_x, rect.end_y);
}

inline size_t VXAccessor::element_size() const
{
    return element_size_from_data_type(data_type());
}

inline size_t VXAccessor::size() const
{
    vx_size size = 0;

    vxQueryImage(_image, VX_IMAGE_SIZE, &size, sizeof(vx_size));

    return size;
}

inline Format VXAccessor::format() const
{
    vx_df_image format = 0;

    vxQueryImage(_image, VX_IMAGE_FORMAT, &format, sizeof(vx_df_image));

    switch(format)
    {
        case VX_DF_IMAGE_VIRT:
            return Format::UNKNOWN;
        case VX_DF_IMAGE_RGB:
            return Format::RGB888;
        case VX_DF_IMAGE_RGBX:
            return Format::RGBA8888;
        case VX_DF_IMAGE_NV12:
            return Format::NV12;
        case VX_DF_IMAGE_NV21:
            return Format::NV21;
        case VX_DF_IMAGE_UYVY:
            return Format::UYVY422;
        case VX_DF_IMAGE_YUYV:
            return Format::YUYV422;
        case VX_DF_IMAGE_IYUV:
            return Format::IYUV;
        case VX_DF_IMAGE_YUV4:
            return Format::YUV444;
        case VX_DF_IMAGE_U8:
            return Format::U8;
        case VX_DF_IMAGE_U16:
            return Format::U16;
        case VX_DF_IMAGE_S16:
            return Format::S16;
        case VX_DF_IMAGE_U32:
            return Format::U32;
        case VX_DF_IMAGE_S32:
            return Format::S32;
        default:
            ARM_COMPUTE_ERROR("Unsupported image format!");
    }
}

inline DataType VXAccessor::data_type() const
{
    return data_type_from_format(format());
}

inline int VXAccessor::num_channels() const
{
    return num_channels_from_format(format());
}

inline int VXAccessor::num_elements() const
{
    return shape().total_size();
}

inline PaddingSize VXAccessor::padding() const
{
    return PaddingSize(0);
}

inline int VXAccessor::fixed_point_position() const
{
    return 0;
}

inline QuantizationInfo VXAccessor::quantization_info() const
{
    return QuantizationInfo{};
}

inline const void *VXAccessor::operator()(const Coordinates &coord) const
{
    return vxFormatImagePatchAddress2d(_mapping, coord.x(), coord.y(), &_addr);
}

inline void *VXAccessor::operator()(const Coordinates &coord)
{
    return vxFormatImagePatchAddress2d(_mapping, coord.x(), coord.y(), &_addr);
}

inline vx_rectangle_t VXAccessor::get_image_size(const vx_image &image) const
{
    vx_rectangle_t rect{ 0, 0, 0, 0 };

    vxQueryImage(image, VX_IMAGE_WIDTH, &rect.end_x, sizeof(vx_uint32));
    vxQueryImage(image, VX_IMAGE_HEIGHT, &rect.end_y, sizeof(vx_uint32));

    return rect;
}
} // namespace test
} // namespace arm_compute
#endif /* __ARM_COMPUTE_TEST_VXACCESSOR_H__ */
