/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#ifndef __ARM_COMPUTE_TEST_VX_VXLUTACCESSOR_H__
#define __ARM_COMPUTE_TEST_VX_VXLUTACCESSOR_H__

#include "ILutAccessor.h"

#include <VX/vx.h>

namespace arm_compute
{
namespace test
{
/** Accessor implementation for vx_lut objects. */
template <typename T>
class VXLutAccessor : public ILutAccessor<T>
{
public:
    /** Create an accessor for the given @p vx_lut.
     */
    VXLutAccessor(vx_lut &lut)
        : _lut{ lut }
    {
        void *_lut_ptr;
        vxMapLUT(_lut, &_map_id, &_lut_ptr, VX_READ_AND_WRITE, VX_MEMORY_TYPE_NONE, 0);
        _lut_tmp = static_cast<T *>(_lut_ptr);
    }

    ~VXLutAccessor()
    {
        vxUnmapLUT(_lut, _map_id);
    }

    VXLutAccessor(const VXLutAccessor &) = delete;
    VXLutAccessor &operator=(const VXLutAccessor &) = delete;
    VXLutAccessor(VXLutAccessor &&)                 = default;
    VXLutAccessor &operator=(VXLutAccessor &&) = default;

    int num_elements() const override
    {
        vx_size count = 0;
        vxQueryLUT(_lut, VX_LUT_COUNT, &count, sizeof(count));
        return count;
    }

    const T &operator[](T input_value) const override
    {
        vx_uint32 index_offset = 0;
        vxQueryLUT(_lut, VX_LUT_OFFSET, &index_offset, sizeof(index_offset));
        int32_t real_index = static_cast<int32_t>(input_value) + index_offset;

        if(0 <= real_index && real_index < num_elements())
        {
            return _lut_tmp[real_index];
        }
        ARM_COMPUTE_ERROR("Error index not in range.");
    }

    T &operator[](T input_value) override
    {
        vx_uint32 index_offset = 0;
        vxQueryLUT(_lut, VX_LUT_OFFSET, &index_offset, sizeof(index_offset));
        int32_t real_index = static_cast<int32_t>(input_value) + index_offset;

        if(0 <= real_index && real_index < num_elements())
        {
            return _lut_tmp[real_index];
        }
        ARM_COMPUTE_ERROR("Error index not in range.");
    }

private:
    vx_lut   &_lut;
    T        *_lut_tmp = nullptr;
    vx_map_id _map_id  = 0;
};
} // namespace test
} // namespace arm_compute
#endif /* __ARM_COMPUTE_TEST_VX_VXLUTACCESSOR_H__ */
