/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "arm_compute/core/TensorShape.h"
#include "arm_compute/core/Types.h"
#include "arm_compute/runtime/NEON/functions/NEGEMM.h"
#include "arm_compute/runtime/Tensor.h"
#include "arm_compute/runtime/TensorAllocator.h"
#include "tests/AssetsLibrary.h"
#include "tests/Globals.h"
#include "tests/NEON/Accessor.h"
#include "tests/Utils.h"
#include "tests/datasets/MatrixMultiplyGEMMDataset.h"
#include "tests/framework/Fixture.h"
#include "tests/framework/Macros.h"
#include "tests/framework/datasets/Datasets.h"
#include "utils/TypePrinter.h"

#include "cblas.h"

namespace arm_compute
{
namespace test
{
/** Fixture that can be used for BLAS */
template <typename TensorType>
class BLASGEMMFixture : public framework::Fixture
{
public:
    template <typename...>
    void setup(TensorShape shape_a, TensorShape shape_b, TensorShape shape_c, TensorShape shape_dst, float alpha, float beta)
    {
        // Create tensors
        _a   = create_tensor<TensorType>(shape_a, DataType::F32, 1);
        _b   = create_tensor<TensorType>(shape_b, DataType::F32, 1);
        _c   = create_tensor<TensorType>(shape_c, DataType::F32, 1);
        _dst = create_tensor<TensorType>(shape_dst, DataType::F32, 1);
        // Allocate tensors
        _a.allocator()->allocate();
        _b.allocator()->allocate();
        _c.allocator()->allocate();
        _dst.allocator()->allocate();
        // Fill tensors
        fill(Accessor(_a), 0);
        fill(Accessor(_b), 1);
        fill(Accessor(_c), 2);
        _alpha     = alpha;
        _beta      = beta;
        _shape_a   = shape_a;
        _shape_b   = shape_b;
        _shape_c   = shape_c;
        _shape_dst = shape_dst;
    }

    void run()
    {
        // Compute GEMM function
        auto ptr_a = reinterpret_cast<const float *>(_a.buffer());
        auto ptr_b = reinterpret_cast<const float *>(_b.buffer());
        auto ptr_c = reinterpret_cast<float *>(_c.buffer());
        cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, _shape_a.y(), _shape_b.x(), _shape_a.x(), _alpha, ptr_a, _shape_a.x(), ptr_b, _shape_b.x(), _beta, ptr_c, _shape_c.x());
    }

    void teardown()
    {
        _a.allocator()->free();
        _b.allocator()->free();
        _c.allocator()->free();
        _dst.allocator()->free();
    }

private:
    template <typename U>
    void fill(U &&tensor, int i)
    {
        switch(tensor.data_type())
        {
            case DataType::F32:
            {
                std::uniform_real_distribution<> distribution(-1.0f, 1.0f);
                library->fill(tensor, distribution, i);
                break;
            }
            default:
            {
                ARM_COMPUTE_ERROR("Not implemented");
                library->fill_tensor_uniform(tensor, i);
                break;
            }
        }
    }

    TensorType  _a{};
    TensorType  _b{};
    TensorType  _c{};
    TensorType  _dst{};
    TensorShape _shape_a{};
    TensorShape _shape_b{};
    TensorShape _shape_c{};
    TensorShape _shape_dst{};
    float       _alpha{};
    float       _beta{};
};

TEST_SUITE(BLAS)
REGISTER_FIXTURE_DATA_TEST_CASE(MatrixMultiplyGEMM, BLASGEMMFixture<Tensor>, framework::DatasetMode::ALL, datasets::MatrixMultiplyGEMMDataset());
TEST_SUITE_END()
} // namespace test
} // namespace arm_compute
