/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "arm_compute/core/Types.h"
#include "arm_compute/runtime/Tensor.h"
#include "arm_compute/runtime/TensorAllocator.h"
#include "tests/NEON/Accessor.h"
#include "tests/datasets/SmallGEMMDataset.h"
#include "tests/framework/Asserts.h"
#include "tests/framework/Macros.h"
#include "tests/framework/datasets/Datasets.h"
#include "tests/validation/Validation.h"

#include "arm_compute/core/TensorShape.h"
#include "tests/AssetsLibrary.h"
#include "tests/Globals.h"
#include "tests/framework/Fixture.h"
#include "tests/validation/CPP/GEMM.h"
#include "tests/validation/Helpers.h"

#include "cblas.h"
#include <random>

namespace arm_compute
{
namespace test
{
namespace validation
{
template <typename TensorType, typename AccessorType, typename T>
class BLASGEMMValidationFixture : public framework::Fixture
{
public:
    template <typename...>
    void setup(TensorShape shape_a, TensorShape shape_b, TensorShape shape_c, TensorShape output_shape, float alpha, float beta)
    {
        _target    = compute_target(shape_a, shape_b, shape_c, output_shape, alpha, beta);
        _reference = compute_reference(shape_a, shape_b, shape_c, output_shape, alpha, beta);
    }

protected:
    template <typename U>
    void fill(U &&tensor, int i)
    {
        switch(tensor.data_type())
        {
            case DataType::F32:
            {
                std::uniform_real_distribution<> distribution(-1.0f, 1.0f);
                library->fill(tensor, distribution, i);
                break;
            }
            default:
            {
                ARM_COMPUTE_ERROR("Not supported");
                break;
            }
        }
    }

    TensorType compute_target(const TensorShape &shape_a, const TensorShape &shape_b, const TensorShape &shape_c, const TensorShape &output_shape, float alpha, float beta)
    {
        // Create tensors
        TensorType a   = create_tensor<TensorType>(shape_a, DataType::F32, 1);
        TensorType b   = create_tensor<TensorType>(shape_b, DataType::F32, 1);
        TensorType c   = create_tensor<TensorType>(shape_c, DataType::F32, 1);
        TensorType dst = create_tensor<TensorType>(output_shape, DataType::F32, 1);
        // Allocate tensors
        a.allocator()->allocate();
        b.allocator()->allocate();
        c.allocator()->allocate();
        dst.allocator()->allocate();
        // Fill tensors
        fill(AccessorType(a), 0);
        fill(AccessorType(b), 1);
        fill(AccessorType(c), 2);
        // Compute GEMM function
        auto ptr_a = reinterpret_cast<const T *>(a.buffer());
        auto ptr_b = reinterpret_cast<const T *>(b.buffer());
        auto ptr_c = reinterpret_cast<T *>(c.buffer());
        cblas_sgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, shape_a.y(), shape_b.x(), shape_a.x(), alpha, ptr_a, shape_a.x(), ptr_b, shape_b.x(), beta, ptr_c, shape_c.x());
        dst.copy_from(c);
        return dst;
    }

    SimpleTensor<T> compute_reference(const TensorShape &shape_a, const TensorShape &shape_b, const TensorShape &shape_c, const TensorShape &output_shape, float alpha, float beta)
    {
        ARM_COMPUTE_UNUSED(output_shape);
        // Create reference
        SimpleTensor<T> a{ shape_a, DataType::F32, 1 };
        SimpleTensor<T> b{ shape_b, DataType::F32, 1 };
        SimpleTensor<T> c{ shape_c, DataType::F32, 1 };
        // Fill reference
        fill(a, 0);
        fill(b, 1);
        fill(c, 2);
        return reference::gemm<T>(a, b, c, alpha, beta);
    }

    TensorType      _target{};
    SimpleTensor<T> _reference{};
};

namespace
{
constexpr AbsoluteTolerance<float> tolerance_f(0.001f); /**< Tolerance value for comparing reference's output against implementation's output for floating point data types */

/** CNN data types */
const auto CNNDataTypes = framework::dataset::make("DataType",
{
    DataType::F32,
});
} // namespace

TEST_SUITE(BLAS)
TEST_SUITE(GEMM)

template <typename T>
using BLASGEMMFixture = BLASGEMMValidationFixture<Tensor, Accessor, T>;

TEST_SUITE(Float)
TEST_SUITE(FP32)
FIXTURE_DATA_TEST_CASE(RunSmall, BLASGEMMFixture<float>, framework::DatasetMode::PRECOMMIT, datasets::SmallGEMMDataset())
{
    // Validate output
    validate(Accessor(_target), _reference, tolerance_f);
}
TEST_SUITE_END()
TEST_SUITE_END()

TEST_SUITE_END()
TEST_SUITE_END()
} // namespace validation
} // namespace test
} // namespace arm_compute
