/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "GEMMLowp.h"

#include "arm_compute/core/Types.h"

namespace arm_compute
{
namespace test
{
namespace validation
{
namespace reference
{
template <typename T>
SimpleTensor<int32_t> gemmlowp_matrix_multiply_core(const SimpleTensor<T> &a, const SimpleTensor<T> &b, int32_t a_offset, int32_t b_offset)
{
    TensorShape shape(b.shape()[0], a.shape()[1]);

    SimpleTensor<int32_t> c(shape, DataType::S32);

    const int K       = a.shape().x();
    const int b_width = b.shape().x();
    const int rows    = c.shape().y(); //M
    const int cols    = c.shape().x(); //N

    std::vector<int32_t> acc;
    acc.resize(cols);

    for(int i = 0; i < rows; ++i)
    {
        for(int j = 0; j < cols; ++j)
        {
            acc[j] = 0;
        }
        for(int k = 0; k < K; ++k)
        {
            const int32_t tmp_a = a_offset + static_cast<int32_t>(a[k + i * K]);
            for(int j = 0; j < b_width; ++j)
            {
                const int32_t tmp_b       = b_offset + static_cast<int32_t>(b[j + k * b_width]);
                const int32_t mult_as_int = tmp_a * tmp_b;
                acc[j] += mult_as_int;
            }
        }
        for(int j = 0; j < cols; ++j)
        {
            c[j + i * cols] = acc[j];
        }
    }

    return c;
}

// used to validate assembly kernels which don't know anything about offsets
SimpleTensor<int32_t> gemmlowp(const SimpleTensor<int8_t> &a, const SimpleTensor<int8_t> &b)
{
    return gemmlowp_matrix_multiply_core(a, b, 0, 0);
}

template <typename T>
SimpleTensor<uint8_t> gemmlowp_quantize_down_int32_to_uint8_scale(const SimpleTensor<T> &in, int32_t result_offset, int32_t result_mult_int, int32_t result_shift)
{
    SimpleTensor<uint8_t> dst(in.shape(), DataType::QASYMM8);

    for(int i = 0; i < in.num_elements(); ++i)
    {
        const int32_t result = ((in[i] + result_offset) * result_mult_int) >> result_shift;
        dst[i]               = static_cast<uint8_t>(std::max(0, std::min(255, result)));
    }

    return dst;
}

template SimpleTensor<int32_t> gemmlowp_matrix_multiply_core(const SimpleTensor<uint8_t> &a, const SimpleTensor<uint8_t> &b, int32_t a_offset, int32_t b_offset);
template SimpleTensor<uint8_t> gemmlowp_quantize_down_int32_to_uint8_scale(const SimpleTensor<int32_t> &a, int32_t result_offset, int32_t result_mult_int, int32_t result_shift);
} // namespace reference
} // namespace validation
} // namespace test
} // namespace arm_compute
