/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "arm_compute/core/Types.h"
#include "tests/AssetsLibrary.h"
#include "tests/Globals.h"
#include "tests/VX/VXAccessor.h"
#include "tests/datasets/BorderModeDataset.h"
#include "tests/datasets/ShapeDatasets.h"
#include "tests/framework/Fixture.h"
#include "tests/framework/Macros.h"
#include "tests/framework/datasets/Datasets.h"
#include "tests/validation/CPP/HarrisCornerDetector.h"
#include "tests/validation/Helpers.h"
#include "tests/validation/VX/Utils.h"
#include "tests/validation/Validation.h"

#include <VX/vx.h>

namespace arm_compute
{
namespace test
{
namespace validation
{
namespace
{
class VXHarrisCornersFixture : public framework::Fixture
{
public:
    template <typename...>
    void setup(TensorShape shape, int gradient_size, int block_size, BorderMode border_mode)
    {
        _context = vxCreateContext();
        CHECK_VX(vxGetStatus(reinterpret_cast<vx_reference>(_context)));

        HarrisCornersParameters params = harris_corners_parameters();

        _target    = compute_target(shape, gradient_size, block_size, border_mode, params);
        _reference = compute_reference(shape, gradient_size, block_size, border_mode, params);
    }

protected:
    template <typename U>
    void fill(U &&tensor)
    {
        library->fill_tensor_uniform(tensor, 0);
    }

    std::vector<KeyPoint> compute_target(const TensorShape &shape, int gradient_size, int block_size, BorderMode border_mode, const HarrisCornersParameters &params)
    {
        // Create graph
        CHECK_VX_TYPE(vx_graph, graph, vxCreateGraph(_context));

        // Create image
        CHECK_VX_TYPE(vx_image, src, vxCreateImage(_context, shape.x(), shape.y(), VX_DF_IMAGE_U8));

        // Fill image
        fill(VXAccessor(src));

        // Create input VX variables
        CHECK_VX_TYPE(vx_scalar, vx_threshold, vxCreateScalar(_context, VX_TYPE_FLOAT32, &params.threshold));
        CHECK_VX_TYPE(vx_scalar, vx_min_dist, vxCreateScalar(_context, VX_TYPE_FLOAT32, &params.min_dist));
        CHECK_VX_TYPE(vx_scalar, vx_sensitivity, vxCreateScalar(_context, VX_TYPE_FLOAT32, &params.sensitivity));
        CHECK_VX_TYPE(vx_scalar, vx_num_corners, vxCreateScalar(_context, VX_TYPE_SIZE, nullptr));
        vx_border_t vx_border = get_vx_border(border_mode, params.constant_border_value);

        CHECK_VX_TYPE(vx_array, vx_corners, vxCreateArray(_context, VX_TYPE_KEYPOINT, shape.total_size()));
        CHECK_VX_TYPE(vx_node, harris, vxHarrisCornersNode(graph, src, vx_threshold, vx_min_dist, vx_sensitivity, gradient_size, block_size, vx_corners, vx_num_corners));
        CHECK_VX(vxSetNodeAttribute(harris, VX_NODE_BORDER, &vx_border, sizeof(vx_border)));

        // Compute function
        CHECK_VX(vxProcessGraph(graph));

        size_t num_corners = 0;
        CHECK_VX(vxCopyScalar(vx_num_corners, &num_corners, VX_READ_ONLY, VX_MEMORY_TYPE_HOST));

        vx_size   stride = 0;
        void     *base   = nullptr;
        vx_map_id map_id = 0;
        CHECK_VX(vxMapArrayRange(vx_corners, 0, num_corners, &map_id, &stride, &base, VX_READ_ONLY, VX_MEMORY_TYPE_HOST, 0));

        std::vector<KeyPoint> corners;
        corners.reserve(num_corners);

        for(size_t i = 0; i < num_corners; ++i)
        {
            KeyPoint corner;

            corner.x               = vxArrayItem(vx_keypoint_t, base, i, stride).x;               //NOLINT
            corner.y               = vxArrayItem(vx_keypoint_t, base, i, stride).y;               //NOLINT
            corner.strength        = vxArrayItem(vx_keypoint_t, base, i, stride).strength;        //NOLINT
            corner.scale           = vxArrayItem(vx_keypoint_t, base, i, stride).scale;           //NOLINT
            corner.orientation     = vxArrayItem(vx_keypoint_t, base, i, stride).orientation;     //NOLINT
            corner.tracking_status = vxArrayItem(vx_keypoint_t, base, i, stride).tracking_status; //NOLINT
            corner.error           = vxArrayItem(vx_keypoint_t, base, i, stride).error;           //NOLINT

            corners.emplace_back(corner);
        }

        CHECK_VX(vxUnmapArrayRange(vx_corners, map_id));

        // Release objects
        CHECK_VX(vxReleaseScalar(&vx_threshold));
        CHECK_VX(vxReleaseScalar(&vx_min_dist));
        CHECK_VX(vxReleaseScalar(&vx_sensitivity));
        CHECK_VX(vxReleaseScalar(&vx_num_corners));
        CHECK_VX(vxReleaseImage(&src));
        CHECK_VX(vxReleaseGraph(&graph));

        return corners;
    }

    std::vector<KeyPoint> compute_reference(const TensorShape &shape, int gradient_size, int block_size, BorderMode border_mode, const HarrisCornersParameters &params)
    {
        // Create reference
        SimpleTensor<uint8_t> src{ shape, Format::U8 };

        // Fill reference
        fill(src);

        return reference::harris_corner_detector<uint8_t>(src, params.threshold, params.min_dist, params.sensitivity, gradient_size, block_size, border_mode, params.constant_border_value);
    }

    vx_context            _context{};
    std::vector<KeyPoint> _target{};
    std::vector<KeyPoint> _reference{};
};

const auto data = combine(framework::dataset::make("GradientSize", { 3, 5, 7 }), combine(framework::dataset::make("BlockSize", { 3, 5, 7 }), datasets::BorderModes()));
} // namespace

TEST_SUITE(VX)
TEST_SUITE(HarrisCorners)

//FIXME(COMPMID-542): Mismatches and segfault (?)
DISABLED_FIXTURE_DATA_TEST_CASE(RunSmall, VXHarrisCornersFixture, framework::DatasetMode::PRECOMMIT, combine(datasets::Small2DShapes(), data))
{
    // Validate output
    validate_keypoints(_target.begin(), _target.end(), _reference.begin(), _reference.end());
}
//FIXME(COMPMID-542): Mismatches and segfault (?)
DISABLED_FIXTURE_DATA_TEST_CASE(RunLarge, VXHarrisCornersFixture, framework::DatasetMode::NIGHTLY, combine(datasets::Large2DShapes(), data))
{
    // Validate output
    validate_keypoints(_target.begin(), _target.end(), _reference.begin(), _reference.end());
}

TEST_SUITE_END()
TEST_SUITE_END()
} // namespace validation
} // namespace test
} // namespace arm_compute
