/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "arm_compute/core/Types.h"
#include "tests/PaddingCalculator.h"
#include "tests/Utils.h"
#include "tests/VX/VXAccessor.h"
#include "tests/datasets/ShapeDatasets.h"
#include "tests/framework/Asserts.h"
#include "tests/framework/Exceptions.h"
#include "tests/framework/Macros.h"
#include "tests/framework/datasets/Datasets.h"
#include "tests/validation/VX/Utils.h"
#include "tests/validation/Validation.h"
#include "tests/validation/fixtures/HistogramFixture.h"
#include "utils/TypePrinter.h"

#include <VX/vx.h>

namespace arm_compute
{
namespace test
{
namespace validation
{
namespace
{
void validate_histogram(const std::vector<uint32_t> &tensor, const SimpleTensor<uint32_t> &reference)
{
    RelativeTolerance<uint32_t> tolerance_value(0);
    int64_t                     num_mismatches = 0;
    int64_t                     num_elements   = 0;

    int         target_size = tensor.size();
    TensorShape target_shape(static_cast<size_t>(target_size), static_cast<size_t>(1));
    ARM_COMPUTE_EXPECT(compare_dimensions(target_shape, reference.shape()), framework::LogLevel::ERRORS);

    const int min_elements = std::min(target_size, reference.num_elements());

    // Iterate over all elements
    for(int element_idx = 0; element_idx < min_elements; ++element_idx)
    {
        const Coordinates id = index2coord(reference.shape(), element_idx);

        const uint32_t &target_value    = tensor[element_idx];
        const uint32_t &reference_value = reinterpret_cast<const uint32_t *>(reference(id))[0];

        if(!compare<RelativeTolerance<uint32_t>>(target_value, reference_value, tolerance_value))
        {
            ARM_COMPUTE_TEST_INFO("id = " << id);
            ARM_COMPUTE_TEST_INFO("channel = " << 0);
            ARM_COMPUTE_TEST_INFO("target = " << std::setprecision(5) << framework::make_printable(target_value));
            ARM_COMPUTE_TEST_INFO("reference = " << std::setprecision(5) << framework::make_printable(reference_value));
            framework::ARM_COMPUTE_PRINT_INFO();

            ++num_mismatches;
        }

        ++num_elements;
    }

    if(num_elements > 0)
    {
        const int32_t tolerance_number          = 0;
        const int64_t absolute_tolerance_number = tolerance_number * num_elements;
        const float   percent_mismatches        = static_cast<float>(num_mismatches) / num_elements * 100.f;

        ARM_COMPUTE_TEST_INFO(num_mismatches << " values (" << std::fixed << std::setprecision(2) << percent_mismatches
                              << "%) mismatched (maximum tolerated " << std::setprecision(2) << tolerance_number << "%)");
        ARM_COMPUTE_EXPECT(num_mismatches <= absolute_tolerance_number, framework::LogLevel::ERRORS);
    }
}

class VXHistogramFixture : public framework::Fixture
{
public:
    template <typename...>
    void setup(TensorShape shape)
    {
        std::mt19937                          gen(library->seed());
        std::uniform_int_distribution<size_t> distribution_size_t(1, 10);
        _num_bins = distribution_size_t(gen);
        std::uniform_int_distribution<int32_t>  distribution_int32_t(0, 125);
        const size_t                            offset = distribution_int32_t(gen);
        std::uniform_int_distribution<uint32_t> distribution_uint32_t(1, 255 - offset);
        const size_t                            range = distribution_uint32_t(gen);

        _context = vxCreateContext();
        CHECK_VX(vxGetStatus(reinterpret_cast<vx_reference>(_context)));

        _target    = compute_target(shape, offset, range);
        _reference = compute_reference(shape, offset, range);
    }

protected:
    template <typename U>
    void fill(U &&tensor)
    {
        library->fill_tensor_uniform(tensor, 0);
    }

    std::vector<uint32_t> compute_target(const TensorShape &shape, int32_t offset, uint32_t range)
    {
        // Create graph
        CHECK_VX_TYPE(vx_graph, graph, vxCreateGraph(_context));

        // Create images
        CHECK_VX_TYPE(vx_image, src, vxCreateImage(_context, shape.x(), shape.y(), VX_DF_IMAGE_U8));

        // Fill images
        fill(VXAccessor(src));

        // Create distribution
        CHECK_VX_TYPE(vx_distribution, dst_distribution, vxCreateDistribution(_context, _num_bins, offset, range));

        // Create and configure node
        CHECK_VX_TYPE(vx_node, hist_node, vxHistogramNode(graph, src, dst_distribution));

        // Compute function
        CHECK_VX(vxVerifyGraph(graph));
        CHECK_VX(vxProcessGraph(graph));

        // Create the vector and copy the histogram
        std::vector<uint32_t> dst;
        dst.resize(_num_bins);
        CHECK_VX(vxCopyDistribution(dst_distribution, dst.data(), VX_READ_ONLY, VX_MEMORY_TYPE_HOST));

        // Release objects
        CHECK_VX(vxReleaseImage(&src));
        CHECK_VX(vxReleaseGraph(&graph));

        return dst;
    }

    SimpleTensor<uint32_t> compute_reference(const TensorShape &shape, int32_t offset, uint32_t range)
    {
        // Create reference
        SimpleTensor<uint8_t> src{ shape, Format::U8 };

        // Fill reference
        fill(src);

        return reference::histogram<uint8_t>(src, _num_bins, offset, range);
    }

    size_t                 _num_bins{};
    vx_context             _context{};
    std::vector<uint32_t>  _target{};
    SimpleTensor<uint32_t> _reference{};
};
} // namespace

TEST_SUITE(VX)
TEST_SUITE(Histogram)

FIXTURE_DATA_TEST_CASE(RunSmall, VXHistogramFixture, framework::DatasetMode::PRECOMMIT, datasets::Small2DShapes())
{
    // Validate output
    validate_histogram(_target, _reference);
}
FIXTURE_DATA_TEST_CASE(RunLarge, VXHistogramFixture, framework::DatasetMode::NIGHTLY, datasets::Large2DShapes())
{
    // Validate output
    validate_histogram(_target, _reference);
}

TEST_SUITE_END()
TEST_SUITE_END()
} // namespace validation
} // namespace test
} // namespace arm_compute
