/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "arm_compute/core/Types.h"
#include "tests/PaddingCalculator.h"
#include "tests/VX/VXAccessor.h"
#include "tests/datasets/BorderModeDataset.h"
#include "tests/datasets/ShapeDatasets.h"
#include "tests/framework/Asserts.h"
#include "tests/framework/Macros.h"
#include "tests/framework/datasets/Datasets.h"
#include "tests/validation/VX/Utils.h"
#include "tests/validation/Validation.h"
#include "tests/validation/fixtures/Median3x3Fixture.h"

#include <VX/vx.h>

namespace arm_compute
{
namespace test
{
namespace validation
{
namespace
{
namespace
{
constexpr unsigned int filter_size = 3;              /* Size of the kernel/filter in number of elements. */
constexpr BorderSize   border_size(filter_size / 2); /* Border size of the kernel/filter around its central element. */
} // namespace

class VXMedian3x3Fixture : public framework::Fixture
{
public:
    template <typename...>
    void setup(TensorShape shape, BorderMode border_mode)
    {
        std::mt19937                           gen(library->seed());
        std::uniform_int_distribution<uint8_t> distribution(0, 255);
        const uint8_t                          constant_border_value = distribution(gen);
        _border_mode                                                 = border_mode;

        _context = vxCreateContext();
        CHECK_VX(vxGetStatus(reinterpret_cast<vx_reference>(_context)));

        _target    = compute_target(shape, border_mode, constant_border_value);
        _reference = compute_reference(shape, border_mode, constant_border_value);
    }

protected:
    template <typename U>
    void fill(U &&tensor)
    {
        library->fill_tensor_uniform(tensor, 0);
    }

    vx_image compute_target(const TensorShape &shape, BorderMode border_mode, uint8_t constant_border_value)
    {
        vx_border_t vxmode = { VX_BORDER_UNDEFINED, { { constant_border_value } } };

        switch(border_mode)
        {
            case BorderMode::UNDEFINED:
                vxmode.mode = VX_BORDER_UNDEFINED;
                break;
            case BorderMode::CONSTANT:
                vxmode.mode              = VX_BORDER_CONSTANT;
                vxmode.constant_value.U8 = constant_border_value; // NOLINT
                break;
            case BorderMode::REPLICATE:
                vxmode.mode = VX_BORDER_REPLICATE;
                break;
            default:
                ARM_COMPUTE_ERROR("NOT SUPPORTED!");
        }

        // Create graph
        CHECK_VX_TYPE(vx_graph, graph, vxCreateGraph(_context));

        // Create images
        CHECK_VX_TYPE(vx_image, src, vxCreateImage(_context, shape.x(), shape.y(), VX_DF_IMAGE_U8));
        CHECK_VX_TYPE(vx_image, dst, vxCreateImage(_context, shape.x(), shape.y(), VX_DF_IMAGE_U8));

        // Fill images
        fill(VXAccessor(src));

        // Create and configure node
        CHECK_VX_TYPE(vx_node, median_node, vxMedian3x3Node(graph, src, dst));
        CHECK_VX(vxSetNodeAttribute(median_node, VX_NODE_BORDER, &vxmode, sizeof(vxmode)));

        // Compute function
        CHECK_VX(vxVerifyGraph(graph));
        CHECK_VX(vxProcessGraph(graph));

        // Release objects
        CHECK_VX(vxReleaseImage(&src));
        CHECK_VX(vxReleaseGraph(&graph));

        return dst;
    }

    SimpleTensor<uint8_t> compute_reference(const TensorShape &shape, BorderMode border_mode, uint8_t constant_border_value)
    {
        // Create reference
        SimpleTensor<uint8_t> src{ shape, Format::U8 };

        // Fill reference
        fill(src);

        return reference::median3x3<uint8_t>(src, border_mode, constant_border_value);
    }

    BorderMode            _border_mode{};
    vx_context            _context{};
    vx_image              _target{};
    SimpleTensor<uint8_t> _reference{};
};
} // namespace

TEST_SUITE(VX)
TEST_SUITE(Median3x3)

FIXTURE_DATA_TEST_CASE(RunSmall, VXMedian3x3Fixture, framework::DatasetMode::PRECOMMIT, combine(datasets::Small2DShapes(),
                                                                                                datasets::BorderModes()))
{
    // Validate output
    validate(VXAccessor(_target), _reference, shape_to_valid_region(_reference.shape(), (_border_mode == BorderMode::UNDEFINED), border_size));

    //Release the target
    CHECK_VX(vxReleaseImage(&_target));
}
FIXTURE_DATA_TEST_CASE(RunLarge, VXMedian3x3Fixture, framework::DatasetMode::NIGHTLY, combine(datasets::Large2DShapes(),
                                                                                              datasets::BorderModes()))
{
    // Validate output
    validate(VXAccessor(_target), _reference, shape_to_valid_region(_reference.shape(), (_border_mode == BorderMode::UNDEFINED), border_size));

    //Release the target
    CHECK_VX(vxReleaseImage(&_target));
}

TEST_SUITE_END()
TEST_SUITE_END()
} // namespace validation
} // namespace test
} // namespace arm_compute
