/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "arm_compute/core/Helpers.h"
#include "arm_compute/core/Types.h"
#include "arm_compute/runtime/Tensor.h"
#include "tests/PaddingCalculator.h"
#include "tests/VX/VXAccessor.h"
#include "tests/datasets/BorderModeDataset.h"
#include "tests/datasets/ShapeDatasets.h"
#include "tests/framework/Asserts.h"
#include "tests/framework/Macros.h"
#include "tests/framework/datasets/Datasets.h"
#include "tests/validation/Helpers.h"
#include "tests/validation/VX/Utils.h"
#include "tests/validation/Validation.h"
#include "tests/validation/fixtures/ScaleFixture.h"

#include <VX/vx.h>

namespace arm_compute
{
namespace test
{
namespace validation
{
namespace
{
constexpr AbsoluteTolerance<uint8_t> tolerance(1);
constexpr float                      tolerance_num = 0.01f;

class VXScaleFixture : public framework::Fixture
{
public:
    template <typename...>
    void setup(TensorShape shape, InterpolationPolicy policy, BorderMode border_mode)
    {
        _shape       = shape;
        _policy      = policy;
        _border_mode = border_mode;

        std::mt19937                           generator(library->seed());
        std::uniform_real_distribution<float>  distribution_float(0.25, 4);
        const float                            scale_x = distribution_float(generator);
        const float                            scale_y = distribution_float(generator);
        std::uniform_int_distribution<uint8_t> distribution_u8(0, 255);
        uint8_t                                constant_border_value = distribution_u8(generator);

        _context = vxCreateContext();
        CHECK_VX(vxGetStatus(reinterpret_cast<vx_reference>(_context)));

        _target    = compute_target(shape, scale_x, scale_y, policy, border_mode, constant_border_value);
        _reference = compute_reference(shape, scale_x, scale_y, policy, border_mode, constant_border_value);
    }

protected:
    template <typename U>
    void fill(U &&tensor)
    {
        library->fill_tensor_uniform(tensor, 0);
    }

    vx_image compute_target(const TensorShape &shape, const float scale_x, const float scale_y,
                            InterpolationPolicy policy, BorderMode border_mode, uint8_t constant_border_value)
    {
        vx_enum     vxpolicy = VX_INTERPOLATION_NEAREST_NEIGHBOR;
        vx_border_t vxmode   = { VX_BORDER_UNDEFINED, { { constant_border_value } } };

        switch(policy)
        {
            case InterpolationPolicy::NEAREST_NEIGHBOR:
                vxpolicy = VX_INTERPOLATION_NEAREST_NEIGHBOR;
                break;
            case InterpolationPolicy::BILINEAR:
                vxpolicy = VX_INTERPOLATION_BILINEAR;
                break;
            case InterpolationPolicy::AREA:
                vxpolicy = VX_INTERPOLATION_AREA;
                break;
            default:
                ARM_COMPUTE_ERROR("NOT SUPPORTED!");
        }
        switch(border_mode)
        {
            case BorderMode::UNDEFINED:
                vxmode.mode = VX_BORDER_UNDEFINED;
                break;
            case BorderMode::CONSTANT:
                vxmode.mode              = VX_BORDER_CONSTANT;
                vxmode.constant_value.U8 = constant_border_value; // NOLINT
                break;
            case BorderMode::REPLICATE:
                vxmode.mode = VX_BORDER_REPLICATE;
                break;
            default:
                ARM_COMPUTE_ERROR("NOT SUPPORTED!");
        }
        // Create graph
        CHECK_VX_TYPE(vx_graph, graph, vxCreateGraph(_context));

        // Create images
        CHECK_VX_TYPE(vx_image, src, vxCreateImage(_context, shape.x(), shape.y(), VX_DF_IMAGE_U8));

        TensorShape shape_scaled(shape);
        shape_scaled.set(0, shape.x() * scale_x);
        shape_scaled.set(1, shape.y() * scale_y);
        CHECK_VX_TYPE(vx_image, dst, vxCreateImage(_context, shape_scaled[0], shape_scaled[1], VX_DF_IMAGE_U8));

        // Fill images
        fill(VXAccessor(src));

        // Create and configure node
        CHECK_VX_TYPE(vx_node, scale_node, vxScaleImageNode(graph, src, dst, vxpolicy));
        CHECK_VX(vxSetNodeAttribute(scale_node, VX_NODE_BORDER, &vxmode, sizeof(vxmode)));

        // Compute function
        CHECK_VX(vxVerifyGraph(graph));
        CHECK_VX(vxProcessGraph(graph));

        // Release objects
        CHECK_VX(vxReleaseImage(&src));
        CHECK_VX(vxReleaseGraph(&graph));

        return dst;
    }

    SimpleTensor<uint8_t> compute_reference(const TensorShape &shape, const float scale_x, const float scale_y,
                                            InterpolationPolicy policy, BorderMode border_mode, uint8_t constant_border_value)
    {
        // Create reference
        SimpleTensor<uint8_t> src{ shape, Format::U8 };

        // Fill reference
        fill(src);

        return reference::scale<uint8_t>(src, scale_x, scale_y, policy, border_mode, constant_border_value);
    }

    vx_context            _context{};
    vx_image              _target{};
    SimpleTensor<uint8_t> _reference{};
    TensorShape           _shape{};
    InterpolationPolicy   _policy{};
    BorderMode            _border_mode{};
};
} // namespace

TEST_SUITE(VX)
TEST_SUITE(Scale)

DISABLED_FIXTURE_DATA_TEST_CASE(RunSmall, VXScaleFixture, framework::DatasetMode::PRECOMMIT, combine(combine(datasets::Small2DShapes(), framework::dataset::make("InterpolationPolicy",
{ InterpolationPolicy::NEAREST_NEIGHBOR, InterpolationPolicy::BILINEAR })),
datasets::BorderModes()))
{
    //Create valid region
    TensorInfo        src_info(_shape, Format::U8);
    const ValidRegion valid_region = calculate_valid_region_scale(src_info, _reference.shape(), _policy, BorderSize(1), (_border_mode == BorderMode::UNDEFINED));

    // Validate output
    validate(VXAccessor(_target), _reference, valid_region, tolerance);

    //Release the target
    CHECK_VX(vxReleaseImage(&_target));
}
FIXTURE_DATA_TEST_CASE(RunLarge, VXScaleFixture, framework::DatasetMode::NIGHTLY, combine(combine(datasets::Large2DShapes(), framework::dataset::make("InterpolationPolicy",
{ InterpolationPolicy::NEAREST_NEIGHBOR, InterpolationPolicy::BILINEAR })),
datasets::BorderModes()))
{
    //Create valid region
    TensorInfo        src_info(_shape, Format::U8);
    const ValidRegion valid_region = calculate_valid_region_scale(src_info, _reference.shape(), _policy, BorderSize(1), (_border_mode == BorderMode::UNDEFINED));

    // Validate output
    validate(VXAccessor(_target), _reference, valid_region, tolerance, tolerance_num);

    //Release the target
    CHECK_VX(vxReleaseImage(&_target));
}

TEST_SUITE_END()
TEST_SUITE_END()
} // namespace validation
} // namespace test
} // namespace arm_compute
