/*
 * Copyright (c) 2017 ARM Limited.
 *
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "arm_compute/core/Helpers.h"
#include "arm_compute/core/Types.h"
#include "arm_compute/runtime/Tensor.h"
#include "tests/PaddingCalculator.h"
#include "tests/VX/VXAccessor.h"
#include "tests/datasets/ShapeDatasets.h"
#include "tests/framework/Asserts.h"
#include "tests/framework/Macros.h"
#include "tests/framework/datasets/Datasets.h"
#include "tests/validation/Helpers.h"
#include "tests/validation/VX/Utils.h"
#include "tests/validation/Validation.h"
#include "tests/validation/fixtures/WarpAffineFixture.h"

#include <VX/vx.h>

namespace arm_compute
{
namespace test
{
namespace validation
{
namespace
{
/** Tolerance */
constexpr AbsoluteTolerance<uint8_t> tolerance(1);

class VXWarpAffineFixture : public framework::Fixture
{
public:
    template <typename...>
    void setup(TensorShape shape, InterpolationPolicy policy, BorderMode border_mode)
    {
        // Generate a random constant value if border_mode is constant
        std::mt19937                           generator(library->seed());
        std::uniform_int_distribution<uint8_t> distribution_u8(0, 255);
        uint8_t                                constant_border_value = distribution_u8(generator);

        // Create the matrix
        std::array<float, 6> matrix{ {} };
        fill_warp_matrix<6>(matrix);

        _context = vxCreateContext();
        CHECK_VX(vxGetStatus(reinterpret_cast<vx_reference>(_context)));

        _target    = compute_target(shape, matrix.data(), policy, border_mode, constant_border_value);
        _reference = compute_reference(shape, matrix.data(), policy, border_mode, constant_border_value);
    }

protected:
    template <typename U>
    void fill(U &&tensor)
    {
        library->fill_tensor_uniform(tensor, 0);
    }

    vx_image compute_target(const TensorShape &shape, const float *matrix, InterpolationPolicy policy, BorderMode border_mode, uint8_t constant_border_value)
    {
        vx_enum     vxpolicy = VX_INTERPOLATION_NEAREST_NEIGHBOR;
        vx_border_t vxmode   = { VX_BORDER_UNDEFINED, { { 0 } } };

        switch(policy)
        {
            case InterpolationPolicy::NEAREST_NEIGHBOR:
                vxpolicy = VX_INTERPOLATION_NEAREST_NEIGHBOR;
                break;
            case InterpolationPolicy::BILINEAR:
                vxpolicy = VX_INTERPOLATION_BILINEAR;
                break;
            case InterpolationPolicy::AREA:
                vxpolicy = VX_INTERPOLATION_AREA;
                break;
            default:
                ARM_COMPUTE_ERROR("NOT SUPPORTED!");
        }
        switch(border_mode)
        {
            case BorderMode::UNDEFINED:
                vxmode.mode = VX_BORDER_UNDEFINED;
                break;
            case BorderMode::CONSTANT:
                vxmode.mode              = VX_BORDER_CONSTANT;
                vxmode.constant_value.U8 = constant_border_value; // NOLINT
                break;
            case BorderMode::REPLICATE:
                vxmode.mode = VX_BORDER_REPLICATE;
                break;
            default:
                ARM_COMPUTE_ERROR("NOT SUPPORTED!");
        }

        CHECK_VX_TYPE(vx_matrix, vxmatrix, vxCreateMatrix(_context, VX_TYPE_FLOAT32, 2, 3));
        ARM_COMPUTE_ERROR_ON((matrix == nullptr));
        CHECK_VX(vxCopyMatrix(vxmatrix, const_cast<float *>(matrix), VX_WRITE_ONLY, VX_MEMORY_TYPE_HOST)); // NOLINT

        // Create graph
        CHECK_VX_TYPE(vx_graph, graph, vxCreateGraph(_context));

        // Create images
        CHECK_VX_TYPE(vx_image, src, vxCreateImage(_context, shape.x(), shape.y(), VX_DF_IMAGE_U8));
        CHECK_VX_TYPE(vx_image, dst, vxCreateImage(_context, shape.x(), shape.y(), VX_DF_IMAGE_U8));

        // Fill images
        fill(VXAccessor(src));

        // Create and configure node
        CHECK_VX_TYPE(vx_node, warp_node, vxWarpAffineNode(graph, src, vxmatrix, vxpolicy, dst));
        CHECK_VX(vxSetNodeAttribute(warp_node, VX_NODE_BORDER, &vxmode, sizeof(vxmode)));

        // Compute function
        CHECK_VX(vxVerifyGraph(graph));
        CHECK_VX(vxProcessGraph(graph));

        // Release objects
        CHECK_VX(vxReleaseImage(&src));
        CHECK_VX(vxReleaseGraph(&graph));

        return dst;
    }

    SimpleTensor<uint8_t> compute_reference(const TensorShape &shape, const float *matrix, InterpolationPolicy policy, BorderMode border_mode, uint8_t constant_border_value)
    {
        // Create reference
        SimpleTensor<uint8_t> src{ shape, Format::U8 };

        // Create the valid mask Tensor
        _valid_mask = SimpleTensor<uint8_t>(shape, Format::U8);

        // Fill reference
        fill(src);

        return reference::warp_affine<uint8_t>(src, _valid_mask, matrix, policy, border_mode, constant_border_value);
    }

    vx_context            _context{};
    vx_image              _target{};
    SimpleTensor<uint8_t> _reference{};
    SimpleTensor<uint8_t> _valid_mask{};
};
} // namespace

TEST_SUITE(VX)
TEST_SUITE(WarpAffine)

FIXTURE_DATA_TEST_CASE(RunSmall, VXWarpAffineFixture, framework::DatasetMode::PRECOMMIT, combine(combine(datasets::Small2DShapes(),
                                                                                                         framework::dataset::make("InterpolationPolicy", { InterpolationPolicy::NEAREST_NEIGHBOR, InterpolationPolicy::BILINEAR })),
                                                                                                 framework::dataset::make("BorderMode", { BorderMode::UNDEFINED, BorderMode::CONSTANT })))
{
    // Validate output
    validate(VXAccessor(_target), _reference, _valid_mask, tolerance, 0.01f);

    //Release the target
    CHECK_VX(vxReleaseImage(&_target));
}
FIXTURE_DATA_TEST_CASE(RunLarge, VXWarpAffineFixture, framework::DatasetMode::NIGHTLY, combine(combine(datasets::Large2DShapes(),
                                                                                                       framework::dataset::make("InterpolationPolicy", { InterpolationPolicy::NEAREST_NEIGHBOR, InterpolationPolicy::BILINEAR })),
                                                                                               framework::dataset::make("BorderMode", { BorderMode::UNDEFINED, BorderMode::CONSTANT })))
{
    // Validate output
    validate(VXAccessor(_target), _reference, _valid_mask, tolerance, 0.01f);

    //Release the target
    CHECK_VX(vxReleaseImage(&_target));
}

TEST_SUITE_END()
TEST_SUITE_END()
} // namespace validation
} // namespace test
} // namespace arm_compute
